/**
 * 
 */
package sim.application;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.Arrays;

import sim.application.util.SJFrameColorClassifier;
import sim.application.util.SJFrameImage;
import sim.application.util.SJFrameSIMLearning;
import sim.exception.SConstructorException;
import sim.exception.SNoImplementationException;
import sim.graphics.SChart;
import sim.graphics.STexture;
import sim.math.SBufferedMatrix;
import sim.nn.SNNData;
import sim.nn.SNNUtil;
import sim.nn.layer.SFullConnectedLayer;
import sim.nn.network.SFullConnectedNetwork;
import sim.nn.network.SNetwork;
import sim.nn.reader.SNeuralClassifiedImageReader;
import sim.nn.reader.SNeuralDataReader;
import sim.nn.trainer.STrainer;
import sim.nn.wdbc.SWDBCData;
import sim.util.SKeyboard;
import sim.util.SLog;
import sim.util.SManyFilesFoundException;

/**
 * L'application <b>SIMDeepLearning</b> permet la ralisation d'un apprentissage  l'aide du concept de rseau de neurones.
 * 
 * @author Simon Vzina
 * @since 2018-06-18
 * @version 2022-01-07
 */
public class SIMLearning {

  /**
   * La constante <b>CONSOLE_MODE</b> dtermine si l'application sera excut en mode console.
   */
  private static final boolean CONSOLE_MODE = false;
  
  /**
   * La mthode lanant l'application de SIMLearning.
   * 
   * @param args
   * @throws IOException 
   * @throws FileNotFoundException 
   * @throws SManyFilesFoundException 
   */
  public static void main(String[] args) throws SManyFilesFoundException, FileNotFoundException, IOException
  {
    if(CONSOLE_MODE)
      mainConsole();
    else
      mainApplication();
  }

  /**
   * Mthode pour lancer l'application de SIMLearning en mode console.
   * 
   * @throws IOException 
   * @throws FileNotFoundException 
   * @throws SManyFilesFoundException 
   */
  private static void mainConsole() throws SManyFilesFoundException, FileNotFoundException, IOException
  {
    boolean repeat;
    
    do{
      
      repeat = false;
      
      SLog.logWriteLine("Veuillez choisir le type de donnes en analyse :");
      SLog.logWriteLine("(1) Prlaboratoire - Partie 1 (activation)");
      SLog.logWriteLine("(2) Table de vrit ET (AND)");
      SLog.logWriteLine("(3) Breast Cancer Wisconsin (Diagnostic) Data Set");
      SLog.logWriteLine("(4) MNIST Data Set");
      SLog.logWriteLine("(5) Prlaboratoire - Partie 2 (propagation de l'erreur)");
      SLog.logWriteLine("(6) Apprentissage de la table de vrit ET (AND)");
      SLog.logWriteLine("(7) Apprentissage Breast Cancer Wisconsin (Diagnostic) Data Set");
      SLog.logWriteLine("(8) Apprentissage MNIST Data Set");
      SLog.logWriteLine("(9) Apprentissage Classification Couleur");
      SLog.logWriteLine();
      
      int choix = SKeyboard.readInt("Effectuez votre choix : ");
      
      switch(choix)
      {
        case 1 : prelaboratoirePartie1(); break;  
        case 2 : and(); break;
        case 3 : wdbc(); break;
        case 4 : mnist(); break;
        case 5 : prelaboratoirePartie2(); break;  
        case 6 : learningAnd(); break;
        case 7 : learningWDBC(); break;
        case 8 : learningMNIST(); break;
        case 9 : learningColorClassifier(); break;
        default : repeat = true;
      }
      
    }while(repeat);   
  }
  
  /**
   * Mthode pour lancer l'application SIMLearning en version graphique.
   */
  private static void mainApplication()
  {
    SJFrameSIMLearning frame = new SJFrameSIMLearning();
    frame.setVisible(true);
  }
  
  /**
   * Mthode pour raliser des calculs en lien avec l'activation d'un rseau.
   */
  public static void prelaboratoirePartie1()
  {
    SLog.logWriteLine("Prelaboratoire partie 1 - Activation d'un rseau");
    SLog.logWriteLine("------------------------------------------------");
    
    double[] l0 = { 1.0, 4.0 };
    double[] l1 = { 3.0, 2.0 };
    double[] l2 = { -2.0, -3.0 };
    
    SBufferedMatrix W = new SBufferedMatrix(l0, l1, l2);
    SLog.logWriteLine("Matrice des poids W :");
    SLog.logWriteLine(W.toString());
    SLog.logWriteLine();
    
    double[] b = { 0.8, 0.5, -0.2 };
    SLog.logWriteLine("Vecteur des biais B :");
    SLog.logWriteLine(Arrays.toString(b));
    SLog.logWriteLine();
    
    double[] x = { -0.4, 0.6 };
    SLog.logWriteLine("Vecteur x des donnes :");
    SLog.logWriteLine(Arrays.toString(x));
    SLog.logWriteLine();
    
    SFullConnectedLayer layer = new SFullConnectedLayer(W, b, "tanh");
    SLog.logWriteLine("Description de la couche de neurones : " + layer);
    
    // Effectuer la propagation des donnes de la couche (l'activation).
    double[] a = layer.forwardPropagation(x);
    
    SLog.logWriteLine("Rsultat du prlaboratoire, partie 1 (valuer l'activation a) :");
    SLog.logWriteLine("a = " + Arrays.toString(a));
  }
  
  /**
   * Mthode pour raliser des calculs en lien avec la propagation de l'erreur d'un rseau.
   */
  public static void prelaboratoirePartie2()
  {
    SLog.logWriteLine("Prelaboratoire partie 2 - Propagation de l'erreur d'un rseau");
    SLog.logWriteLine("-------------------------------------------------------------");
    
    double[] l0 = { 1.0, 4.0 };
    double[] l1 = { 3.0, 2.0 };
    double[] l2 = { -2.0, -3.0 };
    
    SBufferedMatrix W = new SBufferedMatrix(l0, l1, l2);
    SLog.logWriteLine("Matrice des poids W :");
    SLog.logWriteLine(W.toString());
    SLog.logWriteLine();
    
    double[] b = { 0.8, 0.5, -0.2 };
    SLog.logWriteLine("Vecteur des biais B :");
    SLog.logWriteLine(Arrays.toString(b));
    SLog.logWriteLine();
    
    double[] x = { -0.4, 0.6 };
    SLog.logWriteLine("Vecteur x des donnes :");
    SLog.logWriteLine(Arrays.toString(x));
    SLog.logWriteLine();
    
    SFullConnectedLayer layer = new SFullConnectedLayer(W, b, "tanh");
    SLog.logWriteLine("Description de la couche de neurones : " + layer);
    
    // Effectuer la propagation des donnes de la couche (l'activation).
    double[] a = layer.forwardPropagation(x);
    
    SLog.logWriteLine("Activation de la couche de neurone :");
    SLog.logWriteLine("a = " + Arrays.toString(a));
    SLog.logWriteLine();
    
    double[] delta = { 0.2, 0.1, -0.1 };
    SLog.logWriteLine("Vecteur des erreurs delta :");
    SLog.logWriteLine(Arrays.toString(delta));
    SLog.logWriteLine();
    
    // Effectuer la propagation des erreurs par la couche.
    double[] delta_out = layer.backPropagation(delta);
    
    SLog.logWriteLine("Rsultat du prlaboratoire, partie 2 (valuer la propagation de l'erreur) :");
    SLog.logWriteLine("error = " + Arrays.toString(delta_out));
  }
  
  /**
   * Mthode pour faire l'analyse par rseau de neurones de la table de vrit AND.
   */
  public static void and()
  {
    final String ERROR_FUNCTION_NAME = "energy";  // nom de la fonction d'erreur utilise (ncessaire  la construction d'un entraneur.
    int NB = 5;                                   // nombre de rseau non calibr affich.    
    
    try{
     
      SLog.logWriteLine("**********************************************");
      SLog.logWriteLine("Analyse des donnes : Table de vrit ET (AND)");
      SLog.logWriteLine("**********************************************");
      
      //--------------------
      // RSEAU NON ENTRAN
      //--------------------
      SLog.logWriteLine("----------------------------------------------------------");
      SLog.logWriteLine("SITUATION 1 : Rseau sans entranement (rseau alatoire).");
      SLog.logWriteLine("----------------------------------------------------------");
      SLog.logWriteLine();
       
      // Affichage des caractristiques du rseau.
      SFullConnectedNetwork network = buildANDNetwork();
            
      SLog.logWriteLine("Description du rseau : ");
      SLog.logWriteLine(network.getDescription());
      SLog.logWriteLine();
            
      // Affichage de la classification de plusieurs rseaux non entrans.
      SLog.logWriteLine("Taux de classification de " + NB + " rseaux ayant des paramtres diffrents :");
      STrainer trainer = null;
      
      for(int i = 0; i < NB; i++)
      {
        // Construction du rseau de neurones.
        network = buildANDNetwork();
        trainer = buildANDTrainer(network, ERROR_FUNCTION_NAME);
        trainer.computeErrorsTestingData();
        SLog.logWriteLine(trainer.getStatistics());
      }
      
      SLog.logWriteLine();
      
      // Affichage de la table sans entranement.
      SLog.logWriteLine("Affichage de la table de vrit calcul par le dernier rseau non entran :");
      SLog.logWriteLine();
      showAllExampleConsole(trainer);
      
      //----------------
      // RSEAU ENTRAN
      //----------------
      SLog.logWriteLine("--------------------------------------------------------");
      SLog.logWriteLine("SITUATION 2 : Rseau avec entranement (rseau calibr).");
      SLog.logWriteLine("--------------------------------------------------------");
      SLog.logWriteLine();
      
      // Construction du rseau de neurones.
      network = buildANDTrainedNetwork();
      
      // Affichage des caractristiques du rseau.
      SLog.logWriteLine("Description du rseau : ");
      SLog.logWriteLine(network.getDescription());
      SLog.logWriteLine();
      
      trainer = buildANDTrainer(network, ERROR_FUNCTION_NAME);
      trainer.computeErrorsTestingData();
      
      // Affichage des rsultats de la classification.
      SLog.logWriteLine("Taux de classification du rseau entran :");
      SLog.logWriteLine(trainer.getStatistics());    
      SLog.logWriteLine();
    
      // Affichage de la table avec entranement.
      SLog.logWriteLine("Affichage de la table de vrit calcule par le rseau pralablement entran :");
      SLog.logWriteLine();
      showAllExampleConsole(trainer);
      
      SLog.logWriteLine();
      
    }catch(SNoImplementationException e){
      SLog.logWriteLine("Puisque l'activation d'un rseau n'a pas t implmente, la table de vrit AND ne peut pas tre analyse.");
      SLog.logException(e);
      SLog.logWriteLine();
    }
  }
  
  /**
   * Mthode pour raliser l'entranement d'un rseau  l'analyse de la table de vrit ET (AND).
   */
  public static void learningAnd()
  {
    //******************************************************
    // LABORATOIRE : L'apprentissage par rseau de neurones.
    //
    // VALEURS PAR DFAUT : EPOCH = 5, LEARNING_RATE = 0.1
    //******************************************************
    final int EPOCH = 5;
    final double LEARNING_RATE = 0.1;
    
    final String ERROR_FUNCTION_NAME = "energy";    // nom de la fonction d'erreur utilise.
    
    try{
      
      SLog.logWriteLine("**********************************************");
      SLog.logWriteLine("Analyse des donnes : Table de vrit ET (AND)");
      SLog.logWriteLine("**********************************************");
            
      // Construction du rseau de neurones.
      SFullConnectedNetwork network = buildANDNetwork();
      
      SLog.logWriteLine("Description du rseau :");
      SLog.logWriteLine(network.getDescription());
      SLog.logWriteLine();
      
      // Construction de l'entraneur.
      STrainer trainer = buildANDTrainer(network, ERROR_FUNCTION_NAME);
      trainer.setLearningRate(LEARNING_RATE);
      trainer.computeErrorsTestingData();
      
      // Affichage des rsultats de la classification.
      SLog.logWriteLine("Taux de classification avant l'entranement :");
      SLog.logWriteLine();
      SLog.logWriteLine(trainer.getStatistics());   
      SLog.logWriteLine();
         
      SLog.logWriteLine("Entranement (1 epoque  la fois) : Taux d'apprentissage = " + LEARNING_RATE);
      SLog.logWriteLine();
      
      // Faire l'entranement.
      for(int i = 0; i < EPOCH; i++)
      {
        trainer.trainAll();                 // entranement sur tous les vecteurs.
        trainer.computeErrorsTestingData();
        
        SLog.logWriteLine("Cycle #" + (i+1) + " : " + trainer.getStatistics());  
      }
      
      // Affichage de tous les vecteurs d'entranement.
      SLog.logWriteLine("Affichage de la table de vrit calcule par le rseau aprs entranement :");
      SLog.logWriteLine();
      showAllExampleConsole(trainer);
      
      SLog.logWriteLine();
    }catch(SNoImplementationException e){
      SLog.logWriteLine("Puisque l'activation/l'apprentissage d'un rseau n'a pas t implmente, la table de vrit ET (AND) ne peut pas tre analyse.");
      SLog.logException(e);
      SLog.logWriteLine();
    } catch(SConstructorException e) {
      SLog.logWriteLine("Message - Une erreur est survenue.");
      SLog.logException(e);
      SLog.logWriteLine();
    }
  }
    
  /**
   * Mthode pour faire l'analyse par rseau de neurones de la base de donnes WBDC.
   */
  public static void wdbc()
  {
    final String ERROR_FUNCTION_NAME = "energy";    // nom de la fonction d'erreur (ncessaire  la construction d'un entraneur)
    final int NB = 4;                                // nombre de rseau non entran
    final int EXAMPLE = 5;                          // nombre d'exemple affich
    
    try{
           
      SLog.logWriteLine("*******************************************************************");
      SLog.logWriteLine("Analyse des donnes : Breast Cancer Wisconsin (Diagnostic) Data Set");
      SLog.logWriteLine("*******************************************************************");
      
      // Lecture de la base de donnes.
      SLog.logWriteLine("Lecture des donnes WDBC.");
      SWDBCData data_train = new SWDBCData("breast-cancer-wisconsin.txt");
      SLog.logWriteLine();
          
      //--------------------
      // RSEAU NON ENTRAN
      //--------------------
      SLog.logWriteLine("----------------------------------------------------------");
      SLog.logWriteLine("SITUATION 1 : Rseau sans entranement (rseau alatoire).");
      SLog.logWriteLine("----------------------------------------------------------");
      SLog.logWriteLine();
      
      // Affichage des caractristiques du rseau.
      SFullConnectedNetwork network = buildWDBCNetwork();
      
      SLog.logWriteLine("Description du rseau : ");
      SLog.logWriteLine(network.getDescription());
      SLog.logWriteLine();
      
      // Affichage de la classification de plusieurs rseaux non entrans.
      SLog.logWriteLine("Taux de classification de " + NB + " rseaux ayant des paramtres diffrents :");
      STrainer trainer = null;
      
      // Affichage de la classification de plusieurs autres rseaux non entran.
      for(int i = 0; i < NB; i++)
      {
        // Construction du rseau de neurones.
        network = buildWDBCNetwork();
        trainer = buildWDBCTrainer(network, ERROR_FUNCTION_NAME, data_train);
        trainer.computeErrorsTestingData();
        SLog.logWriteLine(trainer.getStatistics());           
      }
      
      SLog.logWriteLine();
            
      // Affichage de quelques activations avec rseau non entrain.
      SLog.logWriteLine("Affichage de quelques activations calcules par le dernier rseau non entran :");
      showRandomExampleConsole(trainer, EXAMPLE);
      SLog.logWriteLine();
      
      //----------------
      // RSEAU ENTRAN
      //----------------
      SLog.logWriteLine("--------------------------------------------------------");
      SLog.logWriteLine("SITUATION 2 : Rseau avec entranement (rseau calibr).");
      SLog.logWriteLine("--------------------------------------------------------");
      SLog.logWriteLine();
      
      // Construction du rseau de neurones.
      network = buildWDBCTrainedNetwork();
      
      // Construction de l'entraneur.
      trainer = buildWDBCTrainer(network, ERROR_FUNCTION_NAME, data_train);
      trainer.computeErrorsTestingData();      
      
      // Affichage des rsultats de la classification.
      SLog.logWriteLine("Taux de classification du rseaux entran :");   
      SLog.logWriteLine(trainer.getStatistics());
      SLog.logWriteLine();
    
      // Affichage de quelques activations avec rseau non entrain.
      SLog.logWriteLine("Affichage de quelques activations calcules par le rseau entran :");
      showRandomExampleConsole(trainer, EXAMPLE);
      SLog.logWriteLine();
      
    }catch(SNoImplementationException e){
      SLog.logWriteLine("Puisque l'activation d'un rseau n'a pas t implment, les donnes WBDC ne peuvent pas tre analyses.");
      SLog.logException(e);
      SLog.logWriteLine();
    }catch(SConstructorException e){
      SLog.logWrite("Message - Une erreur est survenue.");
      SLog.logException(e);
      SLog.logWriteLine();
    }
  }
  
  /**
   * Mthode pour faire l'apprentissage de la classification de la collection des donnes WBDC.
   */
  public static void learningWDBC()
  {
    //******************************************************
    // LABORATOIRE : L'apprentissage par rseau de neurones.
    //
    // VALEURS PAR DFAUT : CYCLE = 50, LEARNING_RATE = 0.1
    //******************************************************
    final int CYCLE = 50;
    final double LEARNING_RATE = 0.1;
    
    final int EXAMPLE = 5;                        // nombre d'examples illustrs
    final String ERROR_FUNCTION_NAME = "energy";  // nom de la fonction d'erreur 
    
    try{
    
      SLog.logWriteLine("*********************************************************************");
      SLog.logWriteLine("* Analyse des donnes : Breast Cancer Wisconsin (Diagnostic) Data Set");
      SLog.logWriteLine("*********************************************************************");
      
      // Lecture de la base de donnes.
      SLog.logWriteLine("Lecture des donnes WDBC.");
      SWDBCData data_train = new SWDBCData("breast-cancer-wisconsin.txt");
      SLog.logWriteLine();
      
      // Construction du rseau de neurones.
      SFullConnectedNetwork network = buildWDBCNetwork();
      
      // Description du rseau.
      SLog.logWriteLine("Description du rseau :");
      SLog.logWriteLine(network.getDescription());
      SLog.logWriteLine();
      
      // Construction de l'entraneur.
      STrainer trainer = buildWDBCTrainer(network, ERROR_FUNCTION_NAME, data_train);
      trainer.setLearningRate(LEARNING_RATE);
      trainer.computeErrorsTestingData();
      
      // Affichage des rsultats de la classification.
      SLog.logWriteLine("Taux de classification du rseaux (avant l'entranement) :");   
      SLog.logWriteLine(trainer.getStatistics());
      SLog.logWriteLine();
      
      SLog.logWriteLine("Entranement (1 vecteur  la fois) : Taux d'apprentissage = " + LEARNING_RATE);
      SLog.logWriteLine();
      
      // Faire l'entranement.
      for(int i = 0; i < CYCLE; i++)
      {
        trainer.train();
        trainer.computeErrorsTestingData();
        SLog.logWriteLine("Cycle #" + (i+1) + " : " + trainer.getStatistics());  
      }
         
      // Affichage de quelques exemples.
      showRandomExampleConsole(trainer, EXAMPLE);
      
      SLog.logWriteLine();
    }catch(SNoImplementationException e){
      SLog.logWriteLine("Puisque l'activation/l'apprentissage d'un rseau n'a pas t implment, les donnes WBDC ne peuvent pas tre analyses.");
      SLog.logException(e);      
      SLog.logWriteLine();
    }catch(SConstructorException e) {
      SLog.logWriteLine("Une erreur est survenue lors de la lecture des donnes.");
      SLog.logException(e);
      SLog.logWriteLine();
    }
  }
  
  /**
   * Mthode pour faire l'analyse par rseau de neurones de la base de donnes MNIST.
   */
  public static void mnist()
  {
    final String ERROR_FUNCTION_NAME = "energy";              // nom de la fonction d'erreur (ncessaire  la construction d'un entraneur)
    final String FILE_NAME = "mnist_Network.txt";             // nom du fichier contenant la description du rseau entran.
    final int EXAMPLE = 10;                                   // nombre d'exemple affich.
    int NB = 5;                                               // nombre de rseau non entran utilis.  
    
    try{
            
      SLog.logWriteLine("************************************");
      SLog.logWriteLine("Analyse des donnes : MNIST Data Set");
      SLog.logWriteLine("************************************");
           
      // Lecture de la base de donnes.
      SNeuralDataReader reader = new SNeuralClassifiedImageReader("mnist_png", false);
      SLog.logWriteLine("Lecture des donnes de test de MNIST.");
      reader.readAllTestingData();
      SLog.logWriteLine();
      
      //--------------------
      // RSEAU NON ENTRAN
      //--------------------
      SLog.logWriteLine("SITUATION 1 : Rseau sans entranement (rseau alatoire).");
      SLog.logWriteLine("----------------------------------------------------------");
      SLog.logWriteLine();
      
      // Affichage des caractristiques du rseau.
      SNetwork network = buildMNISTNetwork();
      
      SLog.logWriteLine("Description du rseau : ");
      SLog.logWriteLine(network.getDescription());
      SLog.logWriteLine();
           
      // Construction de l'entraineur pour valider les donnes de test.
      STrainer trainer = new STrainer(network, ERROR_FUNCTION_NAME);
      trainer.addTesting(reader.getAllTestingData());
      
      // Affichage de la classification de plusieurs rseaux non entrans.
      SLog.logWriteLine("Taux de classification de " + NB + " rseaux ayant des paramtres diffrents :");
            
      for(int i = 0; i < NB; i++)
      {
        network = buildMNISTNetwork();
        trainer = new STrainer(network, ERROR_FUNCTION_NAME);
        trainer.addTesting(reader.getAllTestingData());
        trainer.computeErrorsTestingData();
        SLog.logWriteLine(trainer.getStatistics());         
      }
      
      SLog.logWriteLine();
     
      //----------------
      // RSEAU ENTRAN
      //----------------
      SLog.logWriteLine("SITUATION 2 : Rseau avec entranement (rseau calibr).");
      SLog.logWriteLine("--------------------------------------------------------");
      SLog.logWriteLine();
      
      // Construction du rseau de neurones.
      SLog.logWriteLine("Lecture du fichier : " + FILE_NAME);
      network = buildMNISTTrainedNetwork(FILE_NAME);
      
      // Construction de l'entraneur.      
      trainer = new STrainer(network, ERROR_FUNCTION_NAME);
      trainer.addTesting(reader.getAllTestingData());
      trainer.computeErrorsTestingData();      
      
      // Affichage des rsultats de la classification.
      SLog.logWriteLine("Taux de classification du rseaux entran :");   
      SLog.logWriteLine(trainer.getStatistics());
      SLog.logWriteLine();
    
      // Affichage de quelques activations avec rseau non entrain.
      SLog.logWriteLine("Affichage de quelques activations calcules par le rseau entran :");
      showRandowExampleFrame(trainer, EXAMPLE);
      SLog.logWriteLine();
            
    }catch(SNoImplementationException e){
      SLog.logWriteLine("Puisque l'activation d'un rseau n'a pas t implment, les donnes MNIST ne peuvent pas tre analyses.");
      SLog.logException(e);
      SLog.logWriteLine();
    }catch(SConstructorException e) {
      SLog.logWriteLine("Message - Une erreur est survenue  la lecture.");
      SLog.logException(e);
      SLog.logWriteLine();
    }
    
  }
  
  /**
   * Mthode pour faire l'apprentissage de la classification de la collection des donnes MNIST.
   */
  public static void learningMNIST()
  {
    //************************************************************************
    // LABORATOIRE : L'apprentissage par rseau de neurones.
    //
    // VALEURS PAR DFAUT : CYCLE = 100, TRAINING = 100 et LEARNING_RATE = 0.1 
    //************************************************************************
    final int CYCLE = 100;
    final int TRAINING = 100;  
    final double LEARNING_RATE = 0.1;
    
    final String ERROR_FUNCTION_NAME = "energy";  // nom de la fonction d'erreur (ncessaire  la construction d'un entraneur)
    final int EXAMPLE = 10;                       // Nombre d'examples illustrs.
    
    try
    {
      SLog.logWriteLine("************************************");
      SLog.logWriteLine("Analyse des donnes : MNIST Data Set");
      SLog.logWriteLine("************************************");
      
      // Lecture de la base de donnes.
      SNeuralDataReader reader = new SNeuralClassifiedImageReader("mnist_png", false);
      SLog.logWriteLine("Lecture des donnes d'entrainement de MNIST.");
      reader.readAllTrainingData();
      SLog.logWriteLine("Lecture des donnes de test de MNIST.");
      reader.readAllTestingData();
      SLog.logWriteLine();
      
      // Construction du rseau de neurones.
      SFullConnectedNetwork network = buildMNISTNetwork();
      
      SLog.logWriteLine("Description du rseau :");
      SLog.logWriteLine(network.getDescription());
      SLog.logWriteLine();
      
      // Construction de l'entraneur.
      STrainer trainer = new STrainer(network, ERROR_FUNCTION_NAME);
      trainer.setLearningRate(LEARNING_RATE);
      
      trainer.addTraining(reader.getAllTrainingData());
      trainer.addTesting(reader.getAllTestingData());
      trainer.computeErrorsTestingData();
      
      // Affichage des rsultats de la classification.
      SLog.logWriteLine("Taux de classification du rseaux entran :");   
      SLog.logWriteLine(trainer.getStatistics());
      SLog.logWriteLine();
      
      // Faire l'entranement.
      SLog.logWriteLine("Entranement (1 vecteur  la fois, affichage  tous les " + TRAINING + " vecteurs) : Taux d'apprentissage = " + LEARNING_RATE);
      SLog.logWriteLine();
      
      // Affichage du frame de statistique.
      SJFrameImage frame = null;
      
      // Itrer sur tous les cycle.
      for(int i = 0; i < CYCLE; i++)
      {
        // Raliser plusieurs entranement.
        for(int n = 0; n < TRAINING; n++)
          trainer.train();
                  
        // valuer et afficher les statistiques.
        trainer.computeErrorsTestingData();
        SLog.logWriteLine("Cycle #" + (i+1) + " : " + trainer.getStatistics());     
        
        // Affichage de l'volution de l'erreur du rseau.
        if(trainer.getStatisticFunction().getPoints().size() == 2)
        {
          frame = new SJFrameImage("Statistique - volution de l'erreur en fonction du nombre d'entrainement", new SChart(trainer.getStatisticFunction().createRealFunction(1000), 0.0, 2.0, 500).buildBufferedImage());
          frame.setVisible(true);
        }
        else
          if(frame != null)
          {        
            SChart c = new SChart(trainer.getStatisticFunction().createRealFunction(1000), 0.0, 2.0, 500);
            frame.setBufferedImage(c.buildBufferedImage());
            frame.repaint();
          }
        
        try {
          Thread.sleep(5);
        } catch (InterruptedException e) {
         throw new RuntimeException("Une erreur inattendue est survenue.", e);
        }
        
      }
      SLog.logWriteLine();
      
      // Affichage de quelques vecteurs aprs entranement.
      SLog.logWriteLine("Affichage de l'activation de quelques vecteurs :");
      showRandowExampleFrame(trainer, EXAMPLE);
      
      SLog.logWriteLine();
    }catch(SNoImplementationException e){
      SLog.logWriteLine("Puisque l'activation/l'apprentissage d'un rseau n'a pas t implment, les donnes MNIST ne peuvent pas tre analyses.");
      SLog.logException(e);
      SLog.logWriteLine();
    }catch(SConstructorException e) {
      SLog.logWriteLine("Une erreur est survenue lors de la lecture des donnes.");
      SLog.logException(e);
      SLog.logWriteLine();
    }
    
  }
  
  /**
   * Mthode qui lance l'application graphique de l'apprentissage de la classification des couleurs.
   */
  public static void learningColorClassifier()
  {
    SJFrameColorClassifier frame = new SJFrameColorClassifier();
    frame.setVisible(true);
  }
  
  /**
   * Mthode permettant de construire un rseau de neurones pour les donnes AND.
   * 
   * @return Le rseau de neurones.
   */
  private static SFullConnectedNetwork buildANDNetwork()
  {
    return new SFullConnectedNetwork(2,1);
  }
  
  /**
   * Mthode permettant de construire un rseau de neurones pour les donnes AND.
   * 
   * @return Le rseau de neurones.
   */
  private static SFullConnectedNetwork buildWDBCNetwork()
  {
    // Prendre la version  9 paramtres.
    return new SFullConnectedNetwork(9,1);
  }
  
  /**
   * Mthode permettant de construire un rseau de neurones.
   * 
   * @return La construction d'un rseau de neurones.
   */
  private static SFullConnectedNetwork buildMNISTNetwork()
  {
    // Au besoin, vous pourrez changer le nom des fonctions d'activation
    // pour sigmoid -> tanh
    
    SFullConnectedLayer l1 = new SFullConnectedLayer(784, 16, "sigmoid");
    SFullConnectedLayer l2 = new SFullConnectedLayer(16, 16, "sigmoid");
    SFullConnectedLayer l3 = new SFullConnectedLayer(16, 10, "sigmoid");
    
    return new SFullConnectedNetwork(l1, l2, l3);
  }
  
  /**
   * Mthode permettant de construire un rseau de neurones pour les donnes AND dj entran.
   * 
   * @return Le rseau de neurones.
   */
  private static SFullConnectedNetwork buildANDTrainedNetwork()
  {
    // Paramtre du rseau entran.
    double[] w0 = { 1.1579168099386101 , 1.0902882438737216 };
    double[] b = { -1.898585250637241 };
    
    // Construction du rseau.
    double[][] tab = { w0 };
    SBufferedMatrix w = new SBufferedMatrix(tab);
    SFullConnectedLayer l1 = new SFullConnectedLayer(w, b, "sigmoid");
    SFullConnectedNetwork net = new SFullConnectedNetwork(l1);
    
    return net;
  }
  
  /**
   * Mthode permettant de construire un rseau de neurones pour les donnes WBDC dj entran.
   * 
   * @return Le rseau de neurones.
   */
  private static SFullConnectedNetwork buildWDBCTrainedNetwork()
  {
    // Paramtre du rseau entran.
    double[] w0 = { 3.737545405284475 , 3.030348599930828 , 2.400833527257423 , 1.7003540855005037 , 0.3056863839432416 , 4.118271532220821 , 1.6554129358022023 , 2.0135285024820724 , 1.8871721610048746 };
    double[] b = { -7.2068548377404 };
    
    // Construction du rseau.
    double[][] tab = { w0 };
    SBufferedMatrix w = new SBufferedMatrix(tab);
    SFullConnectedLayer l1 = new SFullConnectedLayer(w, b, "sigmoid");
    SFullConnectedNetwork net = new SFullConnectedNetwork(l1);
    
    return net;
  }
  
  /**
   * Mthode permettant de construire un rseau de neurones pour les donnes MNIST dj entran.
   * 
   * @param file_name Le nom du fichier du rseau de neurones.
   * @return Le rseau de neurones.
   */
  private static SNetwork buildMNISTTrainedNetwork(String file_name)
  {
    return SNNUtil.readNetwork(file_name);
  }
  
  /**
   * Mthode pour construire un entraneur de table de vrit ET.
   * 
   * @param network Le rseau affect  l'entraneur.
   * @return L'entraneur.
   */
  private static STrainer buildANDTrainer(SFullConnectedNetwork network, String error_function_name)
  {
    // Vecteur d'entranement de la fonction x1 ET x2.  (AND)
    double[][] training_data = {  
        { 0.0 , 0.0 },
        { 0.0, 1.0 },
        { 1.0, 0.0 },
        { 1.0, 1.0 } };
        
    // Rsultat de la fonction x1 == x2 associe au vecteur d'entre.
    double[][] result = {  
        { 0.0 },
        { 0.0 },
        { 0.0 },
        { 1.0 } };
    
    // Construire l'entraneur.
    STrainer trainer = new STrainer(network, error_function_name);
    
    // Remplir l'entraneur avec des donnes de validation.
    for(int i = 0; i< training_data.length; i++)
    {
      SNNData data = new SNNData(training_data[i], training_data[i], result[i]);
      
      trainer.addTraining(data);
      trainer.addTesting(data);
    }
    
    return trainer;
  }
  
  /**
   * Mthode pour faire la construction d'un entraneur de rseau de neurones  la base de donnes WBDC.
   * 
   * @param network Le rseau  entraner.
   * @param data_train Les donnes de l'entranement.
   * @return L'entraneur.
   */
  private static STrainer buildWDBCTrainer(SFullConnectedNetwork network, String error_function_name, SWDBCData data_train)
  {
    STrainer trainer = new STrainer(network, error_function_name);
    trainer.addTraining(data_train.getDataList());
    trainer.addTesting(data_train.getDataList());
    
    return trainer;
  }
    
  /**
   * Mthode pour afficher quelques vecteurs de validation choisis alatoirement par un entraneur.
   * L'affichage sera ralis dans un JFrame.
   * 
   * @param trainer L'entraneur.
   * @param EXAMPLE Le nombre de vecteurs affichs.
   */
  private static void showRandowExampleFrame(STrainer trainer, final int EXAMPLE)
  {
    
    SNetwork network = trainer.getNetwork();
    
    // Itrer sur un nombre d'examen  traiter.
    for(int e = 0; e < EXAMPLE; e++)
    {
      // Obtenir un vecteur de test alatoirement.
      SNNData random_data = trainer.getRandomTestingData();
            
      // Affectation de l'activation du rseau avec la donne.
      random_data.setActivation(network.forwardPropagation(random_data.getVector()));
                  
      // Affichage console.
      SLog.logWriteLine("Vecteur #" + (e+1) + ", fichier '" + ((STexture)random_data.getObj()).getFileName() + "'");
      SLog.logWriteLine("Rsultat du Network :" + Arrays.toString(random_data.getActivation()));
      SLog.logWriteLine("Rsultat attendu : " + random_data.getExpectedClassification());
      SLog.logWriteLine();
      
      // Affichage de l'image.
      showFrameMNIST(random_data, random_data.getActivationClassification(), random_data.getExpectedClassification());
    }
  }
  
  /**
   * Mthode pour afficher quelques vecteurs de validation choisis alatoirement par un entraneur.
   * L'affichage sera ralis dans une console.
   * 
   * @param trainer L'entraneur.
   * @param EXAMPLE Le nombre de vecteurs affichs.
   */
  private static void showRandomExampleConsole(STrainer trainer, final int EXAMPLE)
  {
    SFullConnectedNetwork network = (SFullConnectedNetwork) trainer.getNetwork();
    
    // Affichage de quelques vecteurs dtermin alatoirement.
    for(int e = 0; e < EXAMPLE; e++)
    {
      SNNData random_data = trainer.getRandomTrainingData();
      
      double[] out = network.forwardPropagation(random_data.getVector());
      SLog.logWriteLine("Vecteur #" + (e+1));
      SLog.logWriteLine("Expression du vecteur : " + Arrays.toString(random_data.getVector()));
      SLog.logWriteLine("Rsultat du Network :" + Arrays.toString(out));
      SLog.logWriteLine("Rsultat attendu : " + Arrays.toString(random_data.getExpected()));
      SLog.logWriteLine();
    }
  }
  
  /**
   * Mthode pour afficher tous les vecteurs de validation d'un entraneur.
   * L'affichage sera ralis dans une console.
   * 
   * @param trainer L'entraneur.
   */
  private static void showAllExampleConsole(STrainer trainer)
  {
    SFullConnectedNetwork network = (SFullConnectedNetwork) trainer.getNetwork();
    
    // Affichage de tous les vecteurs.
    for(int e = 0; e < trainer.getTestingSize(); e++)
    {
      SNNData data = trainer.getTestingList().get(e);
      
      double[] out = network.forwardPropagation(data.getVector());
      SLog.logWriteLine("Vecteur #" + (e+1));
      SLog.logWriteLine("Expression du vecteur : " + Arrays.toString(data.getVector()));
      SLog.logWriteLine("Rsultat du Network :" + Arrays.toString(out));
      SLog.logWriteLine("Rsultat attendu : " + Arrays.toString(data.getExpected()));
      SLog.logWriteLine();
    }
  }
  
  /**
   * Mthode pour afficher une image de la base de donne MNIST dans un JFrame. 
   * 
   * @param data La donne  afficher.
   * @param calculated Le rsultat calcul par le rseau de neurones.
   * @param expected Le rsultat attendu.
   */
  private static void showFrameMNIST(SNNData data, String calculated, String expected)
  {
    STexture texture = (STexture)data.getObj();
    String frame_name = "Texture '" + texture.getFileName() + "' : Calculated --> " + calculated + " ; Expected --> " + expected;
    
    // Affichage d'une image, avec interpolation des couleurs (pour harmoniser l'agrandissement)
    SJFrameImage frame = new SJFrameImage(frame_name, texture.getBufferedImage(), true);
    frame.setVisible(true);
  }
    
}
