/**
 * 
 */
package sim.math;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import sim.exception.SConstructorException;
import sim.exception.SNoImplementationException;
import sim.util.SArrays;

/**
 * La classe <b>SArrayTensor</b> reprsente un tenseur dont les donnes internes sont organises dans un tableau.
 * 
 * @author Simon Vezina
 * @since 2019-12-28
 * @version 2022-09-13 (Version labo : Apprentissage par rseau de neurones)
 */
public class SArrayTensor {

  /**
   * Le variable <b>array</b> reprsente le tableau contenant l'ensemble des donnes de ce tenseur.
   */
  protected double[] array;
  
  /**
   * La variable <b>dimension</b> reprsente la taille de chaque dimension du tenseur.
   */
  private final int[] dimension;
  
  /**
   * La variable <b>offset</b> reprsente le dcalage  effectuer dans l'indexage du tableau <u>array</u> pour avoir accs aux donnes en changement de dimension.
   */
  private final int[] offset;
  
  /**
   * Constructeur d'un tenseur  une dimension  partir d'un tableau de valeurs.
   * 
   * @param array Le tableau de valeurs.
   */
  public SArrayTensor(double[] array) 
  {
    this(array, array.length);
  }

  /**
   * Constructeur d'un tenseur vide avec dimension variable.
   * 
   * @param dimension Les dimension du tenseur.
   */
  public SArrayTensor(int... dimension)
  {
    this(new double[computeArrayLength(dimension)], dimension);
  }
  
  /**
   * Constructeur d'un tenseur  partir d'un tableau de donnes et des dimensions du tenseur.
   * Il est important de prciser que dans ce constructure, le tableau de donne ne sera pas copie
   * mais la rfrence au tableau sera inserr dans le tenseur.
   * 
   * @param array Les donnes du tenseur.
   * @param dimension Les dimensions du tenseur.
   * @throws SConstructorException Si les dimensions du teneur ne concordent pas avec le nombre de donnes dans le tableau.
   */
  public SArrayTensor(double[] array, int... dimension) throws SConstructorException
  {
    this(array, true, dimension);
  }
  
  /**
   * Constructeur d'un tenseur avec l'option de copier ou non les donnes.
   * 
   * @param array
   * @param with_copy
   * @param dimension
   * @throws SConstructorException
   */
  private SArrayTensor(double[] array, boolean with_copy, int... dimension) throws SConstructorException
  {
    // Validation des dimensions.
    if(dimension.length < 1)
      throw new SConstructorException("Erreur SArrayTensor 001 : Les dimensions du tenseur n'ont pas t dfinies.");
    
    int size = dimension[0];
    
    for(int i = 1; i < dimension.length; i++)
      size *= dimension[i];
    
    if(size != array.length)
      throw new SConstructorException("Erreur SArrayTensor 002 : Le produit des dimensions '" + size + "' ne concorde pas avec le nombre d'lments du tableau tait '" + array.length + "'.");
    
    // Faire l'affectation des donnes.
    if(with_copy)
    {
      // Avec copie de toutes les donnes
      this.array = array.clone();
    }
    else
    {
      // Avec rfrence sur le tableau.
      this.array = array;
    }
    
    // Construction du tableau des dimensions.
    this.dimension = new int[dimension.length];
    
    for(int i = 0; i < this.dimension.length; i++)
      this.dimension[i] = dimension[i];
        
    // Construction du tableau des offsets.
    this.offset = new int[dimension.length];
    
    this.offset[this.offset.length -1] = 1;
    
    for(int i = this.offset.length -2; i > -1; i--)
      this.offset[i] = dimension[i+1] * this.offset[i+1];
  }
  
  /**
   * Mthode pour construire un tenseur avec une rfrence sur les donnes et non une copie.
   * 
   * @param array
   * @param dimension
   * @return
   */
  public static SArrayTensor buildWithReference(double[] array, int... dimension)
  {
    return new SArrayTensor(array, false, dimension);
  }
  
  /**
   * Mthode pour construire un tenseur d'ordre 3  partir d'une list de matrice.
   *    
   * @param list La liste des matrices.
   * @return Le tenseur d'ordre 3 regroupant les matrices.
   * @throws IllegalArgumentException Si les matrices ne sont pas tous de mme taille.
   */
  public static SArrayTensor buildWithMatrixList(List<SBufferedMatrix> list) throws IllegalArgumentException
  {
    if(list.size() < 1)
      throw new IllegalArgumentException("Erreur SArrayTensor 003 : La liste de matrice est vide.");
    
    // Vrifier que toutes les matrices sont de mme dimension.
    int nb_line = list.get(0).getNbLine();
    int nb_column = list.get(0).getNbColumn();
    
    for(SBufferedMatrix m : list)
      if(m.getNbLine() != nb_line || m.getNbColumn() != nb_column)
        throw new IllegalArgumentException("Erreur SArrayTensor 004 : Une matrice n'est pas conforme avec les dimensions de la 1re tant (ligne = " + nb_line + "; colonne = " + nb_column +").");
    
    // Construction du tenseur.
    SArrayTensor t = new SArrayTensor(list.size(), nb_line, nb_column);
    
    // Remplir le tenseur.
    for(int d = 0; d < list.size(); d++)
      for(int i = 0; i < nb_line; i++)
        for(int j = 0; j < nb_column; j++)
          t.set(list.get(d).get(i, j), d, i, j);
    
    // Retourner le tenseur.
    return t;
  }
  
  @Override
  public SArrayTensor clone()
  {
    return new SArrayTensor(this.array, this.dimension);
  }
  
  /**
   * Mthode pour dterminer si deux tenseurs ont les mme dimensions.
   * 
   * @param tensor Le tenseur  comparer
   * @return <b>true</b> si les deux tenseurs ont la mme dimension et <b>false</b> sinon.
   */
  public boolean asSameDimension(SArrayTensor tensor)
  {
    return Arrays.equals(this.dimension, tensor.dimension);
  }
  
  /**
   * Mthode pour obtenir les indices dans le tenseur  partir d'un index dans le tableau du tenseur.
   * 
   * @param array_index L'indice de dans le tableau des donnes du tenseur.
   * @return Les indices du tenseur.
   */
  public int[] getTensorIndex(int array_index) throws ArrayIndexOutOfBoundsException
  {
    if(array_index < 0 || array_index > (this.array.length-1))
      throw new ArrayIndexOutOfBoundsException("Erreur SArrayTensor 005 : L'indice " + array_index + " est  l'extrieur de l'interval [0," + (this.array.length-1) + "].");
    
    // Le tableau des indices.
    int[] index = new int[this.offset.length];
    
    // Convertir l'indice du tableau en indice dans le tenseur.
    for(int i = 0; i < index.length; i++)
    {
      index[i] = array_index / this.offset[i];
      array_index -= index[i]*this.offset[i];
    }
    
    return index;
  }
  
  /**
   * Mthode pour obtenir l'ordre du tenseur.
   * 
   * @return L'ordre du tenseur.
   */
  public int getOrder()
  {
    return this.dimension.length;
  }
  
  /**
   * Mthode pour obtenir une valeur du tenseur  partir des indices du tenseur.
   * 
   * @param index Les indice [i][j][k]... du tenseur.
   * @return La valeur du tenseur en indice demand.
   * @throws IllegalArgumentException 
   * @throws IndexOutOfBoundsException
   */
  public double get(int... index) throws IllegalArgumentException, IndexOutOfBoundsException
  {    
    int location = computeIndex(index);
      
    if(location == -1)
      throw new IndexOutOfBoundsException("Erreur SArrayTensor 003 : Les index index = " + Arrays.toString(index) + " ne correspondent pas aux dimensions " + Arrays.toString(this.dimension) + ".");
      
    return this.array[location];
  }
  
  /**
   * 
   * @param array_index
   * @return
   * @throws ArrayIndexOutOfBoundsException
   */
  public double getWithArrayIndex(int array_index) throws ArrayIndexOutOfBoundsException
  {
    return get(getTensorIndex(array_index));
  }
  
  /**
   * Mthode pour obtenir une valeur dans le tenseur.
   * Si la coordonne dans le tenseur est hors limite, la valeur 0.0 sera retourne.
   * 
   * @param index Les indices dans le tenseur.
   * @return La valeur en indices dans le tenseur.
   * 
   * @throws IllegalArgumentException
   */
  public double getZeroIfOutside(int... index) throws IllegalArgumentException
  {
    int location = computeIndex(index);
    
    if(location != -1)
      return this.array[location];
    else
      return 0.0;      
  }
   
  /**
   * Mthode pour obtenir le tableau des donnes du tenseur.
   * Il est important de prciser que ce tableau correspond aux donnes internes et non une copie.
   * Toute modification  ce tableau va engendrer des modifications au tenseur.
   * 
   * @return Le tableau des donnes dcrivant le tenseur.
   */
  public double[] getAccesArray()
  {
    return this.array;
  }
  
  /**
   * Mthode pour obtenir les dimension du tenseur.
   * 
   * @return Les dimensions du tenseur.
   */
  public int[] getAccesDimension()
  {
    return this.dimension;
  }
  
  /**
   * ...
   * 
   * @param value
   * @param index
   * @throws IllegalArgumentException
   * @throws IndexOutOfBoundsException
   */
  public void set(double value, int... index) throws IllegalArgumentException, IndexOutOfBoundsException
  {
    this.array[computeIndex(index)] = value;
  }
    
  /**
   * Mthode pour dfinir un nouveau tableau de donnes au tenseur.
   * 
   * @param array Le nouveau tableau des donnes.
   * @throws IllegalArgumentException Si la taille du tableau ne concorde pas avec les dimensions du tenseur.
   */
  public void setArray(double[] array) throws IllegalArgumentException
  {
    if(array.length != this.array.length)
      throw new IllegalArgumentException("Erreur SArrayTensor 004 : Le nouveau tableau possde une taille de " + array.length + " ce qui ne correspond pas  l'ancienne taille tant de " + this.array.length + ".");
    
    this.array = array;
  }
  
  /**
   * Mthode pour faire l'ajout d'un tenseur au tenseur courant. Le tenseur courant sera modifi.
   * 
   * @param tensor Le tenseur qui sera ajout.
   * @return Le tenseur courant.
   * @throws IllegalArgumentException Si le tenseur  ajouter n'est pas de mme dimension que celui courant.
   */
  public SArrayTensor addWith(SArrayTensor tensor) throws IllegalArgumentException
  {
    // Validation du nombre d'lment dans les tenseurs.
    if(this.array.length != tensor.array.length)
      throw new IllegalArgumentException("Erreur SArrayTensor 005: Le tenseur courant contient " + this.array.length + " lments et le tenseur  ajouter en contient " + tensor.array.length + " lments ce qui n'est pas identique.");
    
    // Validation des dimensions des tenseurs.
    for(int i = 0; i < this.dimension.length; i++)
      if(this.dimension[i] != tensor.dimension[i])
        throw new IllegalArgumentException("Erreur SArrayTensor 006: Le tenseur en dimension i = " + i + " courant contient " + this.dimension[i] + " lments et le tenseur  ajouter en dimension i = " + i + " en contient " + tensor.dimension[i] + " lments ce qui n'est pas identique.");
      
    // Ajouter les lments du nouveau tenseur.
    for(int i = 0; i < this.array.length; i++)
      this.array[i] += tensor.array[i];
    
    return this;
  }
  
  /**
   * Mthode pour raliser la multiplication du tenseur courant avec une constante. Le tenseur courant sera modifi.
   *  
   * @param value La valeur constante.
   * @return Le tenseur courant.
   */
  public SArrayTensor multiplyWith(double value)
  {
    for(int i = 0; i < this.array.length; i++)
      this.array[i] *= value;
    
    return this;
  }
  
  /**
   * Mthode pour raliser la multiplication d'un tenseur composante par composante avec un autre tenseur de mme ordre.
   * Le tenseur courant sera modifi.
   * 
   * @param tensor Le tenseur  multiplier.
   * @return Le tenseur courant.
   */
  public SArrayTensor multiplyWith(SArrayTensor tensor)
  {
    // Validation du nombre d'lment dans les tenseurs.
    if(this.array.length != tensor.array.length)
      throw new IllegalArgumentException("Erreur SArrayTensor 007: Le tenseur courant contient " + this.array.length + " lments et le tenseur  ajouter en contient " + tensor.array.length + " lments ce qui n'est pas identique.");
    
    // Validation des dimensions des tenseurs.
    for(int i = 0; i < this.dimension.length; i++)
      if(this.dimension[i] != tensor.dimension[i])
        throw new IllegalArgumentException("Erreur SArrayTensor 008: Le tenseur en dimension i = " + i + " courant contient " + this.dimension[i] + " lments et le tenseur  ajouter en dimension i = " + i + " en contient " + tensor.dimension[i] + " lments ce qui n'est pas identique.");
      
    // Ajouter les lments du nouveau tenseur.
    for(int i = 0; i < this.array.length; i++)
      this.array[i] *= tensor.array[i];
    
    return this;
  }
  
  /**
   * Mthode qui effectue le produit de convolution avec un autre tenseur.
   * L'ordre du tenseur courant doit tre suprieur  l'ordre du tenseur  multiplier.
   * Le nombre d'indice utilis dans le produit est trs important.
   * Pour l'instant, c'est seulement la multiplication d'un ordre 3 avec un ordre 3  2 indices qui sera support.
   *  
   * @param tensor Le tenseur  multiplier en colol
   * @param index
   * @return
   */
  public double convolutionProduct(SArrayTensor tensor, int... index) throws IllegalArgumentException
  {
    if(tensor.getOrder() > this.getOrder())
      throw new IllegalArgumentException("Erreur SArrayTensor 008 : L'ordre du tenseur courant est d'ordre " + this.getOrder() + " et le tenseur en produit de convolution est d'ordre " + tensor.getOrder() + " ce qui n'est pas possible pour un produire de convolution.");
    
    if(index.length > tensor.getOrder())
      throw new IllegalArgumentException("Erreur SArrayTensor 007 : L'ordre du tenseur  mutiplier est " + tensor.getOrder() + ", mais le nombre d'indice dans le produit est gal  " + index.length + " ce qui n'est pas conforme pour le produit de convolution.");
    
    // Cas particulier du produit de convolution d'ordre 3 avec ordre 2 (il y aura une sommation sur le 1er indice).
    if(this.getOrder() == 3 && tensor.getOrder() == 3 && index.length == 2)
      return convolutionMultiplicationOrder3With2Index(tensor, index);
    
    throw new SNoImplementationException();
  }
  
  /**
   * Mthode pour mettre  zro l'ensemble des lments du tenseur.
   */
  public void clear()
  {
    for(int i = 0; i < this.array.length; i++)
      this.array[i] = 0.0;
  }
  
  /**
   * Mthode pour convertir un tenseur d'ordre 2 en matrice.
   * 
   * @return La matrice correspondante au tenseur.
   * @throws SInvalidTensorSizeException Si le tenseur n'est pas d'ordre 2.
   */
  public SBufferedMatrix toSBufferedMatrix() throws SInvalidTensorSizeException
  {
    // Validation de l'ordre du tenseur.
    if(getOrder() != 2)
      throw new SInvalidTensorSizeException("Erreur SArrayTensor 007 : Le tenseur est d'ordre " + getOrder() + " ce qui n'est pas d'odre 2 pour une matrice.");
    
    // La liste des lignes.
    List<double[]> line_list = new ArrayList<double[]>();
    
    // Obtenir la liste des lignes.
    for(int i = 0; i < this.dimension[0]; i++)
      line_list.add(Arrays.copyOfRange(this.array, i*this.offset[0], (i+1)*this.offset[0]));
      
    // Construire la matrice.
    return new SBufferedMatrix(line_list);
  }
  
  /**
   * <p>
   * Mthode pour convertir un tenseur en liste de tenseur d'ordre infrieur.
   * </p>
   * <p>
   * Par exemple, un tenseur 4x3x2 donnera une liste de 4 tenseurs 3x2.
   * 
   * @return La liste des tenseurs d'ordre infrieur.
   * @throws SInvalidTensorSizeException Si le tenseur est d'ordre 1.
   */
  public List<SArrayTensor> toListLowerOrderTensor() throws SInvalidTensorSizeException
  {
    if(getOrder() == 1)
      throw new SInvalidTensorSizeException("Erreur SArrayTensor 008 : Puisque le tenseur est d'ordre 1, nous n'acceptons pas de produire une liste de nombre, car a ne correspondra pas  la dfinition d'un tenseur.");
    
    List<SArrayTensor> list = new ArrayList<SArrayTensor>();
    
    // Dfinir les nouvelles dimension (les mmes sauf qu'on exclut la premire).
    int[] new_dimension = Arrays.copyOfRange(this.dimension, 1, this.dimension.length);
    
    for(int i = 0 ; i < this.dimension[0]; i++)
    {
      double[] values = Arrays.copyOfRange(this.array, i*this.offset[0], (i+1)*this.offset[0]);
      list.add(SArrayTensor.buildWithReference(values, new_dimension));
    }
      
    return list;
  }
  
  /**
   * ...
   * 
   * @param tensor
   * @param index
   * @return
   * @throws IllegalArgumentException
   */
  private double convolutionMultiplicationOrder3With2Index(SArrayTensor tensor, int... index) throws IllegalArgumentException
  {
    throw new SNoImplementationException();
    
  }
  
  /**
   * Mthode pour obtenir l'indice dans le tableau de la coordonne de tenseur.
   * Si les coordonnes ne sont pas situ dans les dimensions, l'indice -1 sera retourne.
   * 
   * @param index
   * @return
   * @throws IllegalArgumentException
   */
  private int computeIndex(int... index) throws IllegalArgumentException
  {
    // Vrification que le nombre d'index concorde avec le nombre de dimension du tenseur.
    if(index.length != this.dimension.length)
      throw new IllegalArgumentException("Erreur SArrayTensor 004 : Le nombre d'index " + index.length + " ne concorde pas avec le nombre de dimension du tenseur tant de " + this.dimension.length + ".");
  
    // Vrification du choix des index. S'il y en a un qui ne concorde pas, la position -1 sera retourn.
    for(int i = 0; i < this.dimension.length; i++)
      if(index[i] < 0 || index[i] >= this.dimension[i])
        return -1;
    
    // valuation de l'index.
    switch(index.length)
    {
      case 1 : return index[0];
      
      case 2 : return index[0]*this.offset[0] + index[1];
      
      case 3 : return index[0]*this.offset[0] + index[1]*this.offset[1] + index[2];
      
      case 4 : return index[0]*this.offset[0] + index[1]*this.offset[1] + index[2]*this.offset[2] + index[3];
    
      // Calcul plus lent de l'index pour un tenseur  arbitraire.
      default :
        
      int location = 0;
        
      for(int i = 0; i < this.dimension.length; i++)
        location += index[i]*this.offset[i];
      
      return location;
    }
  }
  
  /**
   * Mthode pour dterminer le nombre d'lments dans le tenseur  partir d'un tableau de dimension.
   * 
   * @param dimension Le tableau des dimensions.
   * @return Le nombre d'lments dans le tenseur.
   */
  private static int computeArrayLength(int[] dimension)
  {
    int length = 1;
    
    for(int i = 0; i < dimension.length; i++)
      length *= dimension[i];
    
    return length;
  }
  
  @Override
  public int hashCode() 
  {
    final int prime = 31;
    int result = 1;
    result = prime * result + Arrays.hashCode(array);
    result = prime * result + Arrays.hashCode(dimension);
    result = prime * result + Arrays.hashCode(offset);
    return result;
  }

  @Override
  public boolean equals(Object obj) 
  {
    if (this == obj)
      return true;
    
    if (obj == null)
      return false;
    
    if (getClass() != obj.getClass())
      return false;
    
    SArrayTensor other = (SArrayTensor) obj;
    
    if (!SArrays.nearlyEquals(array, other.array))
      return false;
    
    if (!Arrays.equals(dimension, other.dimension))
      return false;
    
    if (!Arrays.equals(offset, other.offset))
      return false;
    
    return true;
  }

  @Override
  public String toString()
  {
    return "[array=" + Arrays.toString(array) + ", dimension=" + Arrays.toString(dimension) + "]";
  } 
  
}
