/**
 * 
 */
package sim.math;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import sim.exception.SNoImplementationException;
import sim.util.SArrays;

/**
 * La classe <b>SBufferedMatrix</b> un buffer de matrice.
 * Le contenu de cette matrice sera transforme selon les oprations qui lui seront appliques.
 * Il est alors inportant de copier l'tat de la matrice si l'on dsire garder son contenu d'origine
 * ou bien utiliser la classe <b>SMatrix</b> qui est immuable. 
 * 
 * @author Simon Vezina
 * @since 2017-03-11
 * @version 2022-09-13 (Version labo : Apprentissage par rseau de neurones)
 */
public class SBufferedMatrix {

  /**
   * La constante <b>ZERO_DIMENSION_MATRIX</b> reprsente une matrice sans dimension. 
   * Ainsi, cette matrice possde 0 ligne et 0 colonne.
   */
  public final static SBufferedMatrix ZERO_DIMENSION_MATRIX = new SBufferedMatrix(0,0);
  
  /**
   * La variable <b>data</b> reprsente l'information contenu dans la matrice.
   * La premire rfrence sera la ligne, la deuxime sera la colonne.
   */
  private final double[][] data;
 
  /**
   * La variable <b>nb_line</b> reprsente le nombre de lignes contenus dans la matrice. 
   */
  private final int nb_line;
  
  /**
   * La variable <b>nb_column</b> reprsente le nombre de colonnes contenus dans la matrice. 
   */
  private final int nb_column;
  
  //----------------
  // CONSTRUCTEUR //
  //----------------
  
  /**
   * Constructeur d'un buffer de matrice vide ayant un nombre de lignes et de colonnes dtermines.
   * 
   * @param nb_line Le nombre de ligne dans la matrice.
   * @param nb_column Le nombre de colonne dans la matrice.
   * @throws SInvalidMatrixSizeException Si la matrice est dfinie avec un nombre de ligne ou de colonne invalide.
   */
  public SBufferedMatrix(int nb_line, int nb_column) throws SInvalidMatrixSizeException
  {
    if(nb_line < 0)
      throw new SInvalidMatrixSizeException("Erreur SBufferedMatrix 001 : La matrice contient un nombre ngatif de ligne.");
    
    if(nb_column < 0)
      throw new SInvalidMatrixSizeException("Erreur SBufferedMatrix 002 : La matrice contient un nombre ngatif de colonne.");
    
    this.nb_line = nb_line;
    this.nb_column = nb_column;
    
    // Faire l'allocation de mmoire  la dimension des lignes.
    data = memoryAllocation(nb_line, nb_column);
  }
  
  /**
   * Constructeur d'un buffer de matrice. Les informations passes dans le constructeur seront copies.
   * Ainsi, une modification des donnes par des oprations mathmatiques apportes  l'objet <b>SBufferedMatrix</b>
   * n'aura pas d'effet sur les objets utilises lors de l'appel de ce constructeur.
   * 
   * @param matrix Les informations  indroduire dans le buffer de matrice.
   * @throws SInvalidMatrixSizeException Si le nombre de colonnes n'est pas unique pour l'ensemble des lignes de la matrice.
   * @throws SInvalidMatrixSizeException Si la matrice ne contient pas de ligne.
   */
  public SBufferedMatrix(double[]... matrix) throws SInvalidMatrixSizeException
  {
    // tablir le nombre de colonnes et le nombre de lignes officielles de la matrice.
    this.nb_line = matrix.length;
    this.nb_column = matrix[0].length;
    
    // Vrification du nombre de colonne dans chaque ligne
    for(int l = 0; l < nb_line; l++)
      if(matrix[l].length != nb_column)
        throw new SInvalidMatrixSizeException("Erreur SBufferedMatrix 004 : La ligne " + l + " contient " + matrix[l].length + " colonnes ce qui n'est pas en accord avec la dfinition de la matrice tant de " + nb_column + " colonnes.");
  
    // Faire l'allocation de mmoire  la dimension des lignes.
    data = memoryAllocation(nb_line, nb_column);
       
    // Effectuer la copie des donnes des matrices.
    copyData(matrix, data, nb_line, nb_column);
  }

  /**
   * Constructeur d'un buffer de matrice. Les informations passes dans le constructeur seront copies.
   * Ainsi, une modification des donnes par des oprations mathmatiques apportes  l'objet <b>SBufferedMatrix</b>
   * n'aura pas d'effet sur les objets utilises lors de l'appel de ce constructeur.
   * 
   * @param matrix Les informations  indroduire dans le buffer de matrice.
   * @throws SInvalidMatrixSizeException Si le nombre de colonnes n'est pas unique pour l'ensemble des lignes de la matrice.
   * @throws SInvalidMatrixSizeException Si la matrice ne contient pas de ligne.
   */
  public SBufferedMatrix(List<double[]> eq_list) throws SInvalidMatrixSizeException
  {
    // tablir le nombre de colonnes et le nombre de lignes officielles de la matrice.
    if(eq_list.size() == 0)
    {
      this.nb_line = 0;
      this.nb_column = 0;
    }
    else
    {
      this.nb_line = eq_list.size();
      this.nb_column = eq_list.get(0).length;
    }
    
    // Vrification du nombre de colonne dans chaque ligne
    for(int l = 0; l < nb_line; l++)
      if(eq_list.get(l).length != nb_column)
        throw new SInvalidMatrixSizeException("Erreur SBufferedMatrix 006 : La ligne " + l + " contient " + eq_list.get(l).length + " colonnes ce qui n'est pas en accord avec la dfinition de la matrice tant de " + nb_column + " colonnes.");
  
    // Faire l'allocation de mmoire  la dimension des lignes.
    data = memoryAllocation(nb_line, nb_column);
    
    // Effectuer la copie de l'information.
    copyData(eq_list, data, nb_line, nb_column);
  }
  
  //------------
  // MTHODES //
  //------------
  
  @Override
  public SBufferedMatrix clone()
  {
    // Retourner la matrice vide.
    if(this.getNbLine() == 0 && this.getNbColumn() == 0)
      return ZERO_DIMENSION_MATRIX;
    else
      return new SBufferedMatrix(this.data);
  }
  
  /**
   * Mthode pour obtenir l'ensemble des donnes de la matrice.
   * 
   * @return Les donnes de la matrice.
   */
  public double[][] getData()
  {
    return data;
  }
  
  /**
   * Mthode pour obtenir le nombre de lignes de la matrice.
   * 
   * @return Le nombre de lignes de la matrice.
   */
  public int getNbLine()
  {
    return this.nb_line;
  }
  
  /**
   * Mthode pour obtenir le nombre de colonne de la matrice.
   * 
   * @return Le nombre de colonnes de la matrice.
   */
  public int getNbColumn()
  {
    return this.nb_column;
  }
  
  /**
   * Mthode pour obtenir la valeur d'une matrice. L'indexage se fait par une rfrence  la ligne ainsi qu' une colonne.
   * 
   * @param l Le numro de la ligne de la matrice.
   * @param c Le numro de la colonne de la matrice.
   * @return La valeur de la matrice  la ligne et  la colonne exige.
   * @throws ArrayIndexOutOfBoundsException Si le numro de ligne ou de colonne n'est pas admissible dans la matrice.
   */
  public double get(int l, int c) throws ArrayIndexOutOfBoundsException
  {
    return this.data[l][c];
  }
  
  /**
   * Mthode pour obtenir accs  une ligne de la matrice.
   * Effectuer une opration mathmatique sur le contenu de ce tableau influencera le contenu de la matrice.
   * 
   * @param l La ligne  obtenir
   * @return Une ligne de la matrice.
   * @throws ArrayIndexOutOfBoundsException Si le numro de ligne n'est pas admissible dans la matrice.
   */
  public double[] getLine(int l) throws ArrayIndexOutOfBoundsException
  {
    return data[l];
  }
  
  /**
   * Mthode pour obtenir un accs en rfrence sur toutes les lignes de la matrice.
   * 
   * @return La liste des lignes de la matrice.
   */
  public List<double[]> getLineList()
  {
    List<double[]> list = new ArrayList<double[]>();
    
    for(int i = 0; i < getNbLine(); i++)
      list.add(getLine(i));
    
    return list;
  }
  
  /**
   * Mthode pour obtenir le nombre de pivots disponibles dans la matrice.
   * 
   * @return Le nombre de pivot de la matrice.
   */
  public int getNbPivot()
  {
    throw new SNoImplementationException("La mthode n'est pas implmente.");
  }
  
  /**
   * Mthode pour modifier le contenu d'une cellule de la matrice.
   * 
   * @param l Le numro de la ligne de la matrice.
   * @param c Le numro de la colonne de la matrice.
   * @param value La valeur  changer dans la matrice.
   * @throws ArrayIndexOutOfBoundsException Si le numro de ligne ou de colonne n'est pas admissible dans la matrice.
   */
  public void set(int l, int c, double value) throws ArrayIndexOutOfBoundsException
  {
    data[l][c] = value;
  }
  
  /**
   * Mthode pour affecter une nouvelle ligne  la matrice.
   * 
   * @param l L'indice de la ligne change par la nouvelle ligne.
   * @param line La nouvelle ligne de la matrice.
   * @throws SInvalidMatrixSizeException Si la nouvelle ligne introduite dans la matrice n'a pas le bon nombre de colonne.
   * @throws ArrayIndexOutOfBoundsException Si le numro de ligne n'est pas admissible dans la matrice.
   */
  public void setLine(int l, double[] line) throws SInvalidMatrixSizeException, ArrayIndexOutOfBoundsException
  {
    // Vrifier que la nouvelle ligne introduite contient le bon nombre de colonne
    if(line.length != nb_column)
      throw new SInvalidMatrixSizeException("Erreur SBufferedMatrix 007 : La nouvelle ligne affecte  la matrice contient " + line.length + " colonnes ce qui n'est pas en accord avec la dfinition de la matrice tant de " + nb_column + " colonnes.");
    
    data[l] = line;
  }
  
  /**
   * Mthode pour faire la permutation d'une ligne avec une autre dans la matrice.
   * 
   * @param l1 La 1er ligne  permutter.
   * @param l2 La 2e ligne  permutter.
   * @throws ArrayIndexOutOfBoundsException Si le numro de ligne n'est pas admissible dans la matrice.
   */
  public void swapLine(int l1, int l2) throws ArrayIndexOutOfBoundsException
  {
    throw new SNoImplementationException("La mthode n'est pas implmente.");
    
  }
  
  /**
   * Mthode pour mettre  zro tous les termes de la matrice.
   */
  public void clear()
  {
    for(int i = 0; i < nb_line; i++)
      for(int j = 0; j < nb_column; j++)
        data[i][j] = 0.0;
  }
  
  /**
   * Mthode pour ajouter une valeur  une position spcifier dans la matrice.
   * 
   * @param l L'index de la ligne.
   * @param c L'index de la colonne.
   * @param value La valeur ajoute.
   */
  public void add(int l, int c, double value)
  {
    double v = get(l, c) + value;
    set(l, c, v);
  }
  
  /**
   * Mthode pour faire l'ajout d'une matrice  la matrice courante. La matrice courante sera modifie.
   * 
   * @param m La matrice  ajouter.
   * @return La rfrence sur la matrice ayant subit l'addition.
   * @throws IndexOutOfBoundsException Si la matrice  ajouter n'est pas de mme dimension que celle courante.
   */
  public SBufferedMatrix addWith(SBufferedMatrix m) throws IndexOutOfBoundsException
  {
    for(int i = 0; i < this.getNbLine(); i++)
      SArrays.add(this.getLine(i), m.getLine(i), this.getLine(i));
    
    return this;
  }
  
  /**
   * Mthode pour faire l'addition de deux lignes d'une matrice et d'introduire le rsultat en remplaant une ligne existante par celle calcule.
   * 
   * @param A Multiplication applique  la ligne 1 avant l'addition.
   * @param l1 L'index de la ligne 1.
   * @param B Multiplicateur applique  la ligne 2 avant l'addition.
   * @param l2 L'index de la ligne 2.
   * @param l_result L'indexe de la ligne o sera introduit le rsultat de l'addition des deux lignes. Cette indexe <u>peut tre gal</u>  <b>l1</b> ou <b>l2</b>.
   * @throws ArrayIndexOutOfBoundsException Si le numro de ligne n'est pas admissible dans la matrice.
   */
  public void addLines(double A, int l1, double B, int l2, int l_result) throws ArrayIndexOutOfBoundsException
  {
    throw new SNoImplementationException("La mthode n'est pas implmente.");
    
  }
  
  /**
   * Mthode pour multiplier la matrice par une constante. 
   * Ceci correspond  multiplier l'ensemble des composants de la matrice par la constante.
   * 
   * @param v La constante.
   * @return Une rfrence sur la matrice ayant subit la multiplication par la constante.
   */
  public SBufferedMatrix multiplyWith(double v)
  {
    for(int i = 0; i < this.getNbLine(); i++)
      SArrays.multiply(this.getLine(i), v, this.getLine(i));
    
    return this;
  }
  
  /**
   * <p>
   * Mthode pour effectuer le produit de Hadamard correspondant au produit des composantes Aij*Bij 
   * pour deux matrice A et B.
   * </p>
   * <p>
   * Rfrence : https://fr.wikipedia.org/wiki/Produit_matriciel_de_Hadamard
   * </p>
   * 
   * @param m La matrice  multiplier.
   * @return La rfrence sur la matrice ayant subit le produit de Hadamard.
   */
  public SBufferedMatrix hadamardProduct(SBufferedMatrix m) throws IllegalArgumentException
  {
    // Validation des dimensions de la matrice.
    if(this.nb_column != m.getNbColumn())
      throw new IllegalArgumentException("Erreur SBufferedMatrix 008 : La matrice m possde " + m.getNbColumn() + " colonnes ce qui n'est pas identique  " + this.nb_column);
    
    if(this.nb_line != m.getNbLine())
      throw new IllegalArgumentException("Erreur SBufferedMatrix 009 : La matrice m possde " + m.getNbLine() + " lignes ce qui n'est pas identique  " + this.nb_line);
    
    // Effectuer le produit des lments des deux matrices.
    for(int i = 0; i < this.getNbLine(); i++)
      SArrays.multiply(this.getLine(i), m.getLine(i), this.getLine(i));
    
    return this;
  }
  
  /**
   * <p>
   * ...
   * </p>
   * <p>
   * Rfrence : https://fr.wikipedia.org/wiki/Produit_de_Kronecker
   * </p>
   * 
   * @param m
   */
  public SBufferedMatrix kroneckerProduct(SBufferedMatrix matrix)
  {
    // Soient A une matrice de taille m x n et B une matrice de taille p x q. 
    // Leur produit tensoriel est la matrice A (x) B de taille mp par nq, 
    // dfinie par blocs successifs de taille p x q, le bloc d'indice i,j valant a[i,j]*B
    SBufferedMatrix result = new SBufferedMatrix(this.getNbLine()*matrix.getNbLine(), this.getNbColumn()*matrix.getNbColumn());
    
    // Soit this est une matrice m x n.
    // Soit m est une matrice p x q.
    
    for(int m = 0; m < this.getNbLine(); m++)
      for(int n = 0; n < this.getNbColumn(); n++)
        for(int p = 0; p < matrix.getNbLine(); p++)
          for(int q = 0; q < matrix.getNbColumn(); q++)
            result.set(m*matrix.getNbLine() + p, n*matrix.getNbColumn() + q, this.get(m, n)*matrix.get(p, q));
    
    return result;
  }
  
  /**
   * Mthode pour faire la multiplication d'une ligne par une scalaire.
   * Le rsultat de l'opration sera affect sur la ligne courante. 
   * 
   * @param A Le scalaire en multiplication avec la ligne.
   * @param l L'index de la ligne qui va subir la multiplication par un scalaire.
   * @throws ArrayIndexOutOfBoundsException Si le numro de ligne n'est pas admissible dans la matrice.
   */
  public void multiplyLine(double A, int l) throws ArrayIndexOutOfBoundsException
  {
    throw new SNoImplementationException("La mthode n'est pas implmente.");
    
  }
   
  
  /**
   * Mthode pour effectuer la rduction de Gauss-Jordan  la matrice.
   * Le rsultat sera interrompu si la matrice est singulire.
   * 
   * @throws SSingularMatrixException Si durant la rduction de Gauss-Jordan, on dtermine que la matrice est singulire. La matrice sera alors laisse dans un <u>tat invalide</u> de rduction Gauss-Jordan puisque l'opration aura t interrompue.
   */
  public void gaussJordanReduction() throws SSingularMatrixException
  {
    throw new SNoImplementationException("La mthode n'est pas implmente.");
    
  }
    
  /**
   * Mthode qui effectue la rduction de Gauss-Jordan sur une colonne <u>lanant uniquement</u> la valeur 1  la position du pivot.
   * 
   * @param pivot La position du pivot dans la rduction de type Gauss-Jordan.
   * @throws ArrayIndexOutOfBoundsException Si le numro du pivot n'est pas admissible dans la matrice.
   */
  public void gaussJordanColumnReduction(int pivot) throws ArrayIndexOutOfBoundsException
  {
    throw new SNoImplementationException("La mthode n'est pas implmente.");
    
  }
  
  /**
   * Mthode pour <b>augmenter</b> la matrice.
   * Cette opration consiste  inverser le signe de la dernire colonne ce qui
   * transforme les lignes de la matrice de la forme
   * <ul>A*x1 + B*x2 + C*x3 + ... + Z = 0</ul>
   * dans la forme
   * <ul>A*x1 + B*x2 + C*x3 + ... = -Z</ul>
   * 
   * @return La matrice augmente (comparativement  sa version prcdente).
   */
  public void toAugmented()
  {
    // Faire l'itration des lignes et changer le signe de la donne de la dernire colonne.    
    for(int l = 0; l < nb_line; l++)
      data[l][nb_column-1] = -1*data[l][nb_column-1];
  }
  
  /**
   * Mthode pour obtenir l'indice de la ligne ayant la plus grande valeur (en valeur absolue)  la position du pivot.
   * La recherche va s'effectuer uniquement chez les lignes dont l'indice est suprieur  l'indice du pivot (donc ligne infrieur).
   * 
   * @param pivot L'indice du pivot.
   * @return L'indice de la ligne ayant la plus grande valeur (en valeur absolue)  la position du pivot.
   * @throws ArrayIndexOutOfBoundsException Si le numro du pivot n'est pas admissible dans la matrice.
   */
  public int findLineWithMaxValueForPivot(int pivot) throws ArrayIndexOutOfBoundsException
  {
    throw new SNoImplementationException("La mthode n'est pas implmente.");
    
  }
    
  //----------------------
  // MTHODES STATIQUES //
  //----------------------
  
  /**
   * Mthode pour gnrer une matrice identit d'une dimension quelconque.
   * 
   * @param size La dimension de la matrice identit.
   * @return Une matrice identit.
   */
  public static SBufferedMatrix identity(int size)
  {
    SBufferedMatrix buffer = new SBufferedMatrix(size, size);
    
    for(int i = 0; i < buffer.getNbLine(); i++)
      buffer.set(i, i, 1.0);
    
    return buffer;
  }
  
  /**
   * Mthode pour construire un vecteur  partir d'un tableau de valeur.
   * Il est important de prciser qu'un vecteur correspond  une matrice  n lignes et 1 colonne.
   * 
   * @param v Les valeurs du vecteur.
   * @return Le vecteur (matrice  une colonne) associ aux donnes.
   */
  public static SBufferedMatrix columnVector(double[] v)
  {
    // Construire une matrice  une colonne.
    SBufferedMatrix buffer = new SBufferedMatrix(v.length, 1);
    
    // Remplir la matrice  une colonne.
    for(int i = 0; i < v.length; i++)
      buffer.set(i, 0, v[i]);
    
    return buffer;
  }
  
  //------------------------
  // MTHODES UTILITAIRES //
  //------------------------
  
  /**
   * Mthode pour faire l'allocation de la mmoire  la matrice.
   * 
   * @param nb_line Le nombre de ligne de la matrice.
   * @param nb_column Le nombre de colonne de la matrice.
   * @return Le tableau  deux dimensions correspondant  l'allocation de la mmoire de la matrice.
   */
  private double[][] memoryAllocation(int nb_line, int nb_column)
  {
    // Faire l'allocation de mmoire  la dimension des lignes.
    double[][] m = new double[nb_line][];
    
    // Faire l'allocation de mmoire  la dimension des colonnes.
    for(int l = 0; l < nb_line; l++)
      m[l] = new double[nb_column];
    
    return m;
  }
  
  /**
   * Mthode pour effectuer la copy des donnes d'une matrice  une autre.
   * 
   * @param data La matrice  copier.
   * @param copy La matrice copie.
   * @param nb_line Le nombre de ligne des matrices.
   * @param nb_column Le nombre de colonne des matrices.
   */
  private void copyData(double[][] data, double[][] copy, int nb_line, int nb_column)
  {
    // Effectuer la copie de l'information.
    for(int l = 0; l < nb_line; l++)
      for(int c = 0; c < nb_column; c++)
        copy[l][c] = data[l][c];
  }
  
  /**
   * Mthode pour effectuer la copy des donnes d'une matrice  une autre.
   * 
   * @param data La liste de tableau correspondant  la matrice  copier.
   * @param copy La matrice copie.
   * @param nb_line Le nombre de ligne des matrices.
   * @param nb_column Le nombre de colonne des matrices.
   */
  private void copyData(List<double[]> data, double[][] copy, int nb_line, int nb_column)
  {
    // Effectuer la copie de l'information.
    for(int l = 0; l < nb_line; l++)
      for(int c = 0; c < nb_column; c++)
        copy[l][c] = data.get(l)[c];
  }
     
  //----------------------
  // MTHODE OVERLOADED //
  //----------------------
  
  @Override
  public String toString()
  {
    String expression = "\n"; // dbuter avec un saut de ligne
    
    for(int l = 0; l < getNbLine(); l++)
      for(int c = 0; c < getNbColumn(); c++)
      {
        if(c == 0)
          expression = expression.concat("|  ");
        
        expression = expression.concat(Double.toString(get(l, c)) + "  ");
                
        if(c == getNbColumn() - 1)
          expression = expression.concat("|\n");  // avec saut de ligne
      }
    
    return "SBufferedMatrix [data = " + expression + "]";
  }

  @Override
  public int hashCode()
  {
    final int prime = 31;
    int result = 1;
    result = prime * result + Arrays.hashCode(data);
    result = prime * result + nb_column;
    result = prime * result + nb_line;
    return result;
  }

  @Override
  public boolean equals(Object obj) 
  {
    if (this == obj)
      return true;
    
    if (obj == null)
      return false;
    
    if (getClass() != obj.getClass())
      return false;
    
    SBufferedMatrix other = (SBufferedMatrix) obj;
    
    if (nb_column != other.nb_column)
      return false;
    
    if (nb_line != other.nb_line)
      return false;
    
    // Comparaison des valeurs  l'intrieur de la matrice.
    // Nous allons itrer sur l'ensemble des lignes et vrifier si les tableaux ( 1D) des lignes sont "nearlyEquals".
    //  cet tape, nous avons la garantie que les deux SBufferedMatrix  comparer ont les mme dimensions.
    for(int l = 0; l < nb_line; l++)
      if(!SArrays.nearlyEquals(getLine(l), other.getLine(l)))
        return false;
    
    return true;
  }
  
}

