/**
 * 
 */
package sim.math;

/**
 * La classe <b>SDiscreteFunction</b> est une classe utilitaire permettant d'effectuer des oprations mathmatiques sur des fonctions discrtes.
 * 
 * @author Simon Vzina
 * @since 2018-03-29
 * @version 2020-01-31
 */
public class SDiscreteFunction {

  /**
   * Mthode pour obtenir l'intervalle dx entre deux lments conscutif de la fonction discrte f(x).
   * 
   * @param fx Le tableau des valeurs de la fonction f(x).
   * @param x_min La valeur minimale du domaine.
   * @param x_max La valeur maximale du domaine.
   * @return La distance dx entre deux lments discret du domaine de la fonction f(x).
   * @throws IllegalArgumentException Si le tableau de valeurs est de taille infrieure  2.
   */
  public static double dx(double[] fx, double x_min, double x_max) throws IllegalArgumentException
  {
    if(fx.length < 2)
      throw new IllegalArgumentException("Erreur SDiscreteFunction 001 : La fonction fx possde moins de 2 lments ce qui ne permet pas de dfinir un paramtre dx.");
    
    //-----------------------------------------------
    // VERSION LABORATOIRE : Les fonctions discrtes.
    //-----------------------------------------------
    //throw new SNoImplementationException("La mthode n'a pas t implmente.");
    
    return (x_max - x_min) / (fx.length-1);
  }
  
  /**
   * Mthode pour obtenir l'indice i associ  une coordonne x d'une fonction discrte f(x).
   * Lorsque x = x_min, i = 0. Lorsque x = x_max, i = length-1.
   * 
   * @param x La valeur x  convertir.
   * @param fx Le tableau reprsentant la fonction.
   * @param x_min Le dbut du domaine.
   * @param x_max La fin du domaine.
   * @return L'indice dans le tableau de la fonction.
   * @throws IllegalArgumentException Si x n'est pas dans l'intervalle x_min et x_max.
   */
  public static int xToIndex(double x, double[] fx, double x_min, double x_max) throws IllegalArgumentException
  {    
    if(x < x_min || x > x_max)
      throw new IllegalArgumentException("Erreur SDiscreteFunction 001 : Le paramtre x = " + x + " n'est pas dans l'intervalle x_min = " + x_min + " et x_max = " + x_max + ".");
    
    return (int)Math.round((x - x_min) / dx(fx, x_min, x_max));   
  }
  
  /**
   * Mthode pour obtenir une coordonne x associ  un indice i d'une fonction discrte f(x).
   * Lorsque i = 0, x = x_min. Lorsque i = length-1, x = x_max.
   * 
   * @param i L'indice dans le tableau de la fonction  convertir.
   * @param fx Le tableau reprsentant la fonction.
   * @param x_min Le dbut du domaine.
   * @param x_max La fin du domaine.
   * @return La coordonne x.
   * @throws IllegalArgumentException Si l'indice i n'est pas accessible dans le tableau.
   */
  public static double indexToX(int i, double[] fx, double x_min, double x_max) throws IllegalArgumentException
  {    
    if(i < 0 || i >= fx.length)
      throw new IllegalArgumentException("Erreur SDiscreteFunction 001 : Le paramtre i = " + i + " n'est pas dans l'intervalle du tableau de taille " + fx.length + ".");
    
    return x_min + i*dx(fx, x_min, x_max);
  }
  
  /**
   * Mthode pour valuer une fonction discrte f(x)  une coordonne x.
   * 
   * @param x La coordonne x.
   * @param fx Le tableau reprsentant la fonction.
   * @param x_min Le dbut du domaine.
   * @param x_max La fin du domaine.
   * @return La valeur de la fonction f(x) en coordonne x.
   * @throws IllegalArgumentException Si x n'est pas dans l'intervalle de x_min et x_max.
   */
  public static double f(double x, double[] fx, double x_min, double x_max) throws IllegalArgumentException
  {
    return fx[xToIndex(x, fx, x_min, x_max)];
  }
  
  /**
   * Mthode pour valuer la drive f'(x) d'une fonction discrte f(x)  une coordonne x.
   * 
   * @param x La coordonne x.
   * @param fx Le tableau reprsentant la fonction.
   * @param x_min Le dbut du domaine.
   * @param x_max La fin du domaine.
   * @return La drive de la fonction f(x) en coordonne x.
   * @throws IllegalArgumentException Si x n'est pas dans l'intervalle de x_min et x_max.
   */
  public static double dfdx(double x, double[] fx, double x_min, double x_max) throws IllegalArgumentException
  {
    return dfdx(xToIndex(x, fx, x_min, x_max), fx, dx(fx, x_min, x_max));
  }
  
  /**
   * Mthode pour valuer la drive f'(x) d'une fonction discrte f(x)  un indice i.
   * 
   * @param i L'indice dans le tableau.
   * @param fx Le tableau reprsentant la fonction.
   * @param dx La distance entre deux lments conscutifs du tableau.
   * @return La drive de la fonction en indice i.
   * @throws IndexOutOfBoundsException Si l'indice i n'est pas accessible.
   */
  public static double dfdx(int i, double[] fx, double dx) throws IndexOutOfBoundsException
  {
    //-----------------------------------------------
    // VERSION LABORATOIRE : Les fonctions discrtes.
    //-----------------------------------------------
    //throw new SNoImplementationException("La mthode n'a pas t implmente.");
    
    // Cas gnrale : La drive  5 points.
    if(i > 1 && i < fx.length -2)
      return fivePointsDerivate(i, fx, dx);
    
    // Cas particulier 1 : Le permier lment du tableau.
    if(i == 0)
      return rightDerivate(i, fx, dx);
    
    // Cas particulier 2 : Le dernier lment du tableau.
    if(i == fx.length -1)
      return leftDerivate(i, fx, dx);
    
    // Cas particuler 3 : La drive  trois points.
    return middleDerivate(i, fx, dx);
  }
  
  /**
   * Mthode qui value la drive f'(x) d'une fonction discrte f(x)  un indice i selon la mthode de la <u>drive  gauche</u>.
   * 
   * @param i L'indice dans le tableau.
   * @param fx Le tableau reprsentant la fonction.
   * @param dx La distance entre deux lments conscutifs du tableau.
   * @return La drive de la fonction en indice i.
   * @throws IndexOutOfBoundsException Si l'indice i n'est pas accessible.
   */
  public static double leftDerivate(int i, double[] fx, double dx) throws IndexOutOfBoundsException
  {
    //-----------------------------------------------
    // VERSION LABORATOIRE : Les fonctions discrtes.
    //-----------------------------------------------
    //throw new SNoImplementationException("La mthode n'a pas t implmente.");
    
    return (fx[i] - fx[i-1]) / dx;
  }
  
  /**
   * Mthode qui value la drive f'(x) d'une fonction discrte f(x)  un indice i selon la mthode de la <u>drive  droite</u>.
   * 
   * @param i L'indice dans le tableau.
   * @param fx Le tableau reprsentant la fonction.
   * @param dx La distance entre deux lments conscutifs du tableau.
   * @return La drive de la fonction en indice i.
   * @throws IndexOutOfBoundsException Si l'indice i n'est pas accessible.
   */
  public static double rightDerivate(int i, double[] fx, double dx) throws IndexOutOfBoundsException
  {
    //-----------------------------------------------
    // VERSION LABORATOIRE : Les fonctions discrtes.
    //-----------------------------------------------
    //throw new SNoImplementationException("La mthode n'a pas t implmente.");
    
    return (fx[i+1] - fx[i]) / dx;
  }
  
  /**
   * Mthode qui value la drive f'(x) d'une fonction discrte f(x)  un indice i selon la mthode de la <u>drive au centre</u>.
   * 
   * @param i L'indice dans le tableau.
   * @param fx Le tableau reprsentant la fonction.
   * @param dx La distance entre deux lments conscutifs du tableau.
   * @return La drive de la fonction en indice i.
   * @throws IndexOutOfBoundsException Si l'indice i n'est pas accessible.
   */
  public static  double middleDerivate(int i, double[] fx, double dx) throws IndexOutOfBoundsException
  {
    //-----------------------------------------------
    // VERSION LABORATOIRE : Les fonctions discrtes.
    //-----------------------------------------------
    //throw new SNoImplementationException("La mthode n'a pas t implmente.");
    
    return (fx[i+1] - fx[i-1]) / (2*dx);
  }
  
  /**
   * Mthode qui value la drive f'(x) d'une fonction discrte f(x)  un indice i selon la mthode de la <u>drive  cinq points</u>.
   * 
   * Rfrence : https://en.wikipedia.org/wiki/Five-point_stencil
   * 
   * @param i L'indice dans le tableau.
   * @param fx Le tableau reprsentant la fonction.
   * @param dx La distance entre deux lments conscutifs du tableau.
   * @return La drive de la fonction en indice i.
   * @throws IndexOutOfBoundsException Si l'indice i n'est pas accessible.
   */
  public static double fivePointsDerivate(int i, double[] fx, double dx) throws IndexOutOfBoundsException
  {
    //-----------------------------------------------
    // VERSION LABORATOIRE : Les fonctions discrtes.
    //-----------------------------------------------
    //throw new SNoImplementationException("La mthode n'a pas t implmente.");
    
    return (-1*fx[i+2] + 8*fx[i+1] - 8*fx[i-1] + fx[i-2]) / (12*dx);
  }
  
  /**
   * Mthode qui value la drive discrte f'(x) d'une fonction f(x).
   * 
   * @param fx Le tableau reprsentant la fonction.
   * @param x_min Le dbut du domaine.
   * @param x_max La fin du domaine.
   * @return Un tableau correspondant  la drive de chaque lment du tableau de la fonction prcdente.
   */
  public static double[] derivate(double[] fx, double dx)
  {
    //-----------------------------------------------
    // VERSION LABORATOIRE : Les fonctions discrtes.
    //-----------------------------------------------
    //throw new SNoImplementationException("La mthode n'a pas t implmente.");
    
    double[] function = new double[fx.length];
       
    for(int i = 0; i < fx.length; i++)
      function[i] = dfdx(i, fx, dx);
    
    return function;
  }
  
  /**
   * Mthode pour valuer la primitive F(x) d'une fonction discrte f(x)  une coordonne x.
   * 
   * @param x
   * @param fx
   * @param C La constante d'intgration de la fonction F(x) qui est value  la coordonne x_min.
   * @param x_min
   * @param x_max
   * @return
   * @throws IllegalArgumentException Si x n'est pas dans l'intervalle de x_min et x_max.
   */
  public static double Fx(double x, double[] fx, double C, double x_min, double x_max) throws IllegalArgumentException
  {    
    int i = xToIndex(x, fx, x_min, x_max);
    double dx = dx(fx, x_min, x_max);
    
    return Fx(0, i, fx, C, dx);    
  }
  
  /**
   * Mthode pour valuer la primitive F(x) d'une fonction discrte f(x) entre un indice i et j.
   * 
   * @param i
   * @param j
   * @param fx
   * @param C
   * @param dx
   * @return
   * @throws IndexOutOfBoundsException
   */
  public static double Fx(int i, int j, double[] fx, double C, double dx) throws IndexOutOfBoundsException, IllegalArgumentException
  {
    if(i > j)
      throw new IllegalArgumentException("Erreur SDiscreteFunction 005 : L'indice i = " + i + " est plus grand que l'argument j = " + j + ".");
   
    //-----------------------------------------------
    // VERSION LABORATOIRE : Les fonctions discrtes.
    //-----------------------------------------------
    //throw new SNoImplementationException("La mthode n'a pas t implmente.");
    
    if(i == j)
      return C;
    else
      return middleIntegrate(i, j, fx, C, dx);
   
  }
    
  /**
   * ...
   * 
   * @param i
   * @param j
   * @param fx
   * @param C
   * @param dx
   * @return
   * @throws IndexOutOfBoundsException
   */
  public static double leftIntegrate(int i, int j, double[] fx, double C, double dx) throws IndexOutOfBoundsException
  {
    //-----------------------------------------------
    // VERSION LABORATOIRE : Les fonctions discrtes.
    //-----------------------------------------------
    //throw new SNoImplementationException("La mthode n'a pas t implmente.");
    
    // Affectation de la condition initiale.
    double sum = C;
    
    // valuation de l'intgrale par la mthode du trapze.
    for(int index = i; index < j; index++)
      sum += fx[index] * dx;
    
    return sum;
  }
  
  /**
   * ...
   * 
   * @param i
   * @param j
   * @param fx
   * @param C
   * @param dx
   * @return
   * @throws IndexOutOfBoundsException
   */
  public static double rightIntegrate(int i, int j, double[] fx, double C, double dx) throws IndexOutOfBoundsException
  {
    //-----------------------------------------------
    // VERSION LABORATOIRE : Les fonctions discrtes.
    //-----------------------------------------------
    //throw new SNoImplementationException("La mthode n'a pas t implmente.");
    
    // Affectation de la condition initiale.
    double sum = C;
    
    // valuation de l'intgrale par la mthode du trapze.
    for(int index = i; index < j; index++)
      sum += fx[index+1] * dx;
    
    return sum;
  }
  
  /**
   * ...
   * 
   * @param i
   * @param j
   * @param fx
   * @param C
   * @param dx
   * @return
   * @throws IndexOutOfBoundsException
   */
  public static double middleIntegrate(int i, int j, double[] fx, double C, double dx) throws IndexOutOfBoundsException
  {
    //-----------------------------------------------
    // VERSION LABORATOIRE : Les fonctions discrtes.
    //-----------------------------------------------
    //throw new SNoImplementationException("La mthode n'a pas t implmente.");
    
    // Affectation de la condition initiale.
    double sum = C;
    
    // valuation de l'intgrale par la mthode du trapze.
    for(int index = i; index < j; index++)
      sum += (fx[index] + fx[index+1]) * dx / 2.0;
         
    return sum;
  }
  
  /**
   * Mthode pour valuer l'intgrale F(x)  partir de la fonction f(x) en indice i.
   *  
   * @param i L'indice de la fonction f(x).
   * @param f La fonction f(x)
   * @param F Les valeurs de la fonction F(x) dj values qui serviront de conditions initiales (constante C) dans l'valuation de la fonction pour d'autre indice i.
   * @param dx L'intervalle de domaine entre deux indices conscutifs dans le tableau reprsentatant f(x).
   * @return L'intgrale en indice i.
   * @throws IndexOutOfBoundsException S'il y a eu une erreur dans l'accs dans le tableau.
   * 
   * 
   * 
   */
  public static double Fx(int i, double[] f, double[] F, double dx) throws IndexOutOfBoundsException 
  {
    //-----------------------------------------------
    // VERSION LABORATOIRE : Les fonctions discrtes.
    //-----------------------------------------------
    //throw new SNoImplementationException("La mthode n'a pas t implmente.");
     
    // Choisir le bon algorithme en fonction de l'indice.
    switch(i)
    {
      case 0 : return F[0];
      
      case 1 : return middleRuleIntegrate(i, f, F[i-1], dx);
      
      case 2 : return simpsonRuleIntegrate(i, f, F[i-2], dx);
      
      case 3 : return simpsonAdvRuleIntegrate(i, f, F[i-3], dx);
      
      default : return booleRuleIntegrate(i, f, F[i-4], dx);
    }
  }
  
  /**
   * ...
   * 
   * @param i
   * @param f
   * @param F
   * @param dx
   * @return
   * @throws IndexOutOfBoundsException
   */
  public static double leftRuleIntegrate(int i, double[] f, double C, double dx) throws IndexOutOfBoundsException
  {
    //-----------------------------------------------
    // VERSION LABORATOIRE : Les fonctions discrtes.
    //-----------------------------------------------
    //throw new SNoImplementationException("La mthode n'a pas t implmente.");
    
    return f[i-1] * dx + C;
  }
  
  /**
   * 
   * @param i
   * @param f
   * @param F
   * @param dx
   * @return
   * @throws IndexOutOfBoundsException
   */
  public static double rightRuleIntegrate(int i, double[] f, double C, double dx) throws IndexOutOfBoundsException
  {
    //-----------------------------------------------
    // VERSION LABORATOIRE : Les fonctions discrtes.
    //-----------------------------------------------
    //throw new SNoImplementationException("La mthode n'a pas t implmente.");
    
    return f[i] * dx + C;
  }
  
  /**
   * 
   * @param i
   * @param f
   * @param F
   * @param dx
   * @return
   * @throws IndexOutOfBoundsException
   */
  public static double middleRuleIntegrate(int i, double[] f, double C, double dx) throws IndexOutOfBoundsException
  {
    //-----------------------------------------------
    // VERSION LABORATOIRE : Les fonctions discrtes.
    //-----------------------------------------------
    //throw new SNoImplementationException("La mthode n'a pas t implmente.");
    
    return (f[i-1] + f[i]) * dx / 2.0 + C;
  }
  
  /**
   * Mthode qui effectue le calcul de l'aire sous la courbe d'une fonction dans l'intervalle [i-2, i]
   * selon la mthode de Simpson 1/3.
   * 
   * Rfrence : https://fr.wikipedia.org/wiki/M%C3%A9thode_de_Simpson
   * 
   * @param i
   * @param f
   * @param F
   * @param dx
   * @return
   * @throws IndexOutOfBoundsException
   */
  public static double simpsonRuleIntegrate(int i, double[] f, double C, double dx) throws IndexOutOfBoundsException
  {
    //-----------------------------------------------
    // VERSION LABORATOIRE : Les fonctions discrtes.
    //-----------------------------------------------
    //throw new SNoImplementationException("La mthode n'a pas t implmente.");
    
    return (f[i-2] + 4*f[i-1] + f[i]) * dx / 3.0 + C;
  }
  
  /**
   * Mthode qui effectue le calcul de l'aire sous la courbe d'une fonction dans l'intervalle [i-3, i]
   * selon la mthode de Simpson 3/8.
   * 
   * Rfrence : http://mathworld.wolfram.com/Simpsons38Rule.html
   * 
   * @param i
   * @param f
   * @param F
   * @param dx
   * @return
   * @throws IndexOutOfBoundsException
   */
  public static double simpsonAdvRuleIntegrate(int i, double[] f, double C, double dx) throws IndexOutOfBoundsException
  {
    //-----------------------------------------------
    // VERSION LABORATOIRE : Les fonctions discrtes.
    //-----------------------------------------------
    //throw new SNoImplementationException("La mthode n'a pas t implmente.");
    
    return (f[i-3] + 3*f[i-2] + 3*f[i-1] + f[i]) * (3*dx) / 8.0 + C;
  }
  
  /**
   * Mthode qui effectue le calcul de l'aire sous la courbe d'une fonction dans l'intervalle [i-4, i]
   * selon la mthode de Boole.
   * 
   * Rfrence : https://en.wikipedia.org/wiki/Boole%27s_rule
   * 
   * @param i
   * @param f
   * @param F
   * @param dx
   * @return
   * @throws IndexOutOfBoundsException
   */
  public static double booleRuleIntegrate(int i, double[] f, double C, double dx) throws IndexOutOfBoundsException
  {
    //-----------------------------------------------
    // VERSION LABORATOIRE : Les fonctions discrtes.
    //-----------------------------------------------
    //throw new SNoImplementationException("La mthode n'a pas t implmente.");
    return (7*f[i-4] + 32*f[i-3] + 12*f[i-2] + 32*f[i-1] + 7*f[i]) * (4*dx) / 90.0 + C;
  }
  
  /**
   * Mthode pour valuer la prmitive F(x) d'une fonction discrte f(x).
   * 
   * @param f
   * @param C
   * @param x_min
   * @param x_max
   * @return
   */
  public static double[] integrate(double[] f, double C, double x_min, double x_max)
  {
    //-----------------------------------------------
    // VERSION LABORATOIRE : Les fonctions discrtes.
    //-----------------------------------------------
    //throw new SNoImplementationException("La mthode n'a pas t implmente.");
    
    double[] F = new double[f.length];
    double dx = dx(f, x_min, x_max);
    
    // Affectation de la condition intiale.
    F[0] = C;
    
    // valuer la primitive pour tous les indices.
    for(int i = 1; i < f.length; i++)
      F[i] = Fx(i, f, F, dx);
    
    return F;
  }
  
}
