/**
 * 
 */
package sim.nn;

/**
 * <p>
 * L'interface <b>SActivable</b> reprsente une structure dans un rseau de neurone qui est <u>activable</u> comme par exemple :
 * <ul>- Un rseau de neurones</ul>
 * <ul>- Une couche de neurones</ul>
 * <ul>- Une fonction neuronale</ul>
 * </p>
 * <p>
 * De plus, une structure activable (propagation des donnes vers l'avant) pourra galement propager une erreur vers l'arrire (<i>back-propagation</i>).
 * </p>
 * 
 * @author Simon Vezina
 * @since 2022-01-03
 * @version 2022-01-09
 */
public interface SActivable {
  
  /**
   * <p>
   * Mthode pour faire la propagation des donnes vers l'avant dans un rseau de neurones. 
   * </p>
   * <p>
   * Cette action porte galement le nom <u>d'activation</u>.
   * </p>
   * 
   * @param inputs Le vecteur d'entre de la couche.
   * @return L'espace mmoire o est enregistr le rsultat de la propagation des donnes vers l'avant.
   * @throws IllegalArgumentException Si le paramtre ne possde pas le bon nombre de neurones.
   */
  public double[] forwardPropagation(double[] inputs) throws IllegalArgumentException;
  
  /**
   * Mthode pour faire la propagation de l'erreur vers l'arrire dans un rseau de neurones.
   * 
   * @param delta L'erreur qui sera propage  l'intrieur de la couche.
   * @return L'erreur propage par la couche.
   * @throws IllegalArgumentException Si le paramtre ne possde pas le bon nombre de neurones.
   */
  public double[] backPropagation(double[] delta) throws IllegalArgumentException;
  
  /**
   * Mthode pour dterminer si la structure (ex: rseau, couche ou fonction) contient des paramtres (ex : poids, biais, scale, shift).
   * 
   * @return <b>true</b> si la fonction possde des paramtres et <b>false</b> sinon.
   */
  public default boolean asParameter()
  {
    return this.getNumberOfParameter() != 0;
  }
  
  /**
   * Mthode pour obtenir le nombre de neurones en entre dans la couche.
   * 
   * @return Le nombre de neurones d'entre.
   */
  public int getInputsSize();
  
  /**
   * Mthode pour obtenir le nombre de neurones de sortie de la couche.
   * 
   * @return Le nombre de neurones de sortie.
   */
  public int getOutputsSize();
  
  /**
   * Mthode pour obtenir le niveau de profondeur  l'entre de la couche.
   * 
   * @return Le niveau de profondeur  l'entre de la couche.
   */
  public default int getInputsDepth()
  {
    return 1;
  }
    
  /**
   * Mthode pour obtenir le niveau de profondeur  la sortie de la couche.
   * 
   * @return Le niveau de profondeur  la sortie de la couche.
   */
  public default int getOutputsDepth()
  {
    return 1;
  }
    
  /**
   * Mthode pour obtenir les dimensions de la couche de neurones.
   * Si la couche possde une profondeur de 1, l'criture sera sous la forme "In x Out".
   * Si la couche possde plusieurs dimensions, l'crituire sera sous la forme "In (depth In) x Out (depth Out)".
   * 
   * @return Les dimensions de la couche de neurones.
   */
  public default String getDimension()
  {
    return getInputsSize() + " x " + getOutputsSize();
  }
       
  /**
   * Mthode pour obtenir le nombre de paramtres associ  l'objet.
   * 
   * @return Le nombre de paramtre dans la couche du rseau.
   */
  public default int getNumberOfParameter()
  {
    return 0;
  }
    
}
