/**
 * 
 */
package sim.nn;

import java.util.ArrayList;
import java.util.List;

/**
 * <p>
 * L'interface <b>SBatchActivable</b> reprsente une structure dans un rseau de neurone pouvant 
 * raliser l'activation par regroupement (<i>batch</i>). 
 * </p>
 * <p>
 * Cette activation conciste  activer
 * un rseau avec plusieurs donnes simultanment influenant ainsi l'activation de l'enseble
 * des donnes tout au long du passage dans le rseau.
 * </p>
 * 
 * @author Simon Vezina
 * @since 2022-01-03
 * @version 2022-01-05
 */
public interface SBatchActivable extends SActivable {
  
  @Override
  public default double[] forwardPropagation(double[] inputs) throws IllegalArgumentException
  {
    List<double[]> list = new ArrayList<double[]>();
    list.add(inputs);
    
    return this.batchForwardPropagation(1, list).get(0);
  }
  
  @Override
  public default double[] backPropagation(double[] delta) throws IllegalArgumentException
  {
    List<double[]> list = new ArrayList<double[]>();
    list.add(delta);
    
    return this.batchBackPropagation(1, list).get(0);
  }
  
  /**
   * Mthode pour raliser la propagation des donnes par <i>batch</i> vers l'avant dans un rseau de neurones.
   * Cette action porte galement de <i>batch-activation</i>.
   * 
   * @param b La taille de la <i>batch</i>.
   * @param batch_inputs La liste des vecteurs d'entre correspondant  la <i>batch</i>.
   * @return La liste des espaces mmoires o seront enregistres la propagation des donnes vers l'avant.
   * @throws IllegalArgumentException Si un argument de la mthode n'est pas adquat.
   */
  public List<double[]> batchForwardPropagation(int b, List<double[]> batch_inputs) throws IllegalArgumentException;
  
  /**
   * Mthode pour raliser la progation de l'erreur par <i>batch</i> vers l'arrire dans un rseau de neurones.
   * 
   * @param b La taille de la <i>batch</i>.
   * @param batch_delta La liste des erreurs  propager correspondant  la <i>batch</i>.
   * @return La liste des espaces mmoires o seront enregistres la propagations des erreur ver l'arrire.
   * @throws IllegalArgumentException Si un argument de la mthode n'est pas adquat.
   */
  public List<double[]> batchBackPropagation(int b, List<double[]> batch_delta) throws IllegalArgumentException;
    
  /**
   * Mthode pour obtenir la taille de la <i>batch</i> associ  la dernire activation.
   * 
   * @return La taille de la <i>batch</i> de la dernire activation.
   */
  public int getBatchSize();
    
  /**
   * <p>
   * Mthode pour dfinir la taille de la <i>batch</i> qui sera utilis lors de la prochaine activation.
   * </p>
   * <p>
   * De plus, si la taille de la <b>batch</b> est plus imposante que la taille maximale supporte,
   * il y aura allocation de mmoire pour permettre l'excution de la prochaine activation.
   * Ainsi, un appel  la mthode
   * <ul><b>public boolean</b> upgradeMaximumBatchSize(<b>int</b> size);</ul>
   * sera ncessaire dans l'implmentation de cette mthode.
   * <p>
   * 
   * @param b La taille de la batch.
   * @throws IllegalArgumentException Si b est infrieur  1.
   */
  public void setBatchSize(int b) throws IllegalArgumentException;
    
  /**
   * <p>
   * Mthode permettant d'valuer la taille maximale de la <i>batch</i> pouvant tre trait.
   * </p>
   * <p>
   * Si une <i>batch</i> de taille suprieure doit tre traite, alors une allocation de mmoire devra tre ralise afin d'y enregistrer les rsultats.
   * Cette allocation pourra tre ralise pralablement ou dynamiquement lorsque l'espace mmoire sera ncessaire.
   * </p>
   * 
   * @return La taille maximum de la <i>batch</i> pouvant tre traite.
   */
  public int getMaximumBatchSize();
    
  /**
   * <p>
   * Mthode pour augmenter la taille maximale de la <i>batch</i> pouvant tre traiter.
   * </p>
   * <p>
   * Cette mthode augmentera uniquement la taille maximale. Si un nombre infrieur  la taille maximale
   * est propose, il n'y aura pas de modification de la taille maximale.
   * </p>
   * 
   * @param size La nouvelle taille maximale de la <i>batch</i>.
   * @return <b>true</p> si la taille a t change  la hausse et <b>false</b> sinon.
   */
  public boolean upgradeMaximumBatchSize(int size);
      
}
