/**
 * 
 */
package sim.nn;

import java.awt.Color;
import java.awt.image.BufferedImage;

import sim.graphics.SColor;
import sim.math.SMath;
import sim.util.SArrays;
import sim.util.SStringUtil;

/**
 * La classe <b>SDataVectorializer</b> est une classe utilitaire permettant 
 * de transformer des donnes de rseau de neurones en vecteurs pouvant tre activs dans un rseau de neurones.  
 * 
 * @author Simon Vezina
 * @since 2022-03-06
 * @version 2022-03-06
 */
public class SDataVectorializer {

  /**
   * La constante <b>LOWER_CASE_VECTOR_SIZE</b> reprsente la dimension pour un vecteur de caractre contenant uniquement des caractres minuscules. 
   */
  private final static int LOWER_CASE_VECTOR_DIMENSION = 27;
  
  /**
   * La constante <b>DEFAULT_BLACK_VALUE</b> reprsente la valeur numrique <u>de rfrence</u> attribuable  la couleur noire tant {@value}.
   */
  private final static double DEFAULT_BLACK_VALUE = 0.0;
  
  /**
   * La constante <b>DEFAULT_WHITE_VALUE</b> reprsente la valeur numrique <u>de rfrence</u> attribuable  la couleur blanche tant {@value}.
   */
  private final static double DEFAULT_WHITE_VALUE = 1.0;
    
  /**
   * La variable <b>black_value</b> reprsente la valeur numrique accorde  une couleur noire.
   */
  private double black_value;
  
  /**
   * La variable <b>white_value</b> reprsente la valeur numrique accorde  une couleur blanche.
   */
  protected double white_value;
  
  /**
   * ...
   */
  public SDataVectorializer()
  {
    this.black_value = DEFAULT_BLACK_VALUE;
    this.white_value = DEFAULT_WHITE_VALUE;
  }
  
  /**
   * <p>
   * Mthode pour transformer une chane de caractres en vecteur pour rseau de neurones.
   * </p>
   * <p>
   * La chane de caractre sera transform en matrice de taille n x 27 o n sera le nombre de caractres de la chane
   * et 27 correspond au nombre de dimensions associes au type de cararactre interprtable dans la chane. Dans ces 27 cases,
   * ils seront tous gales  "0" sauf pour la case correspondant en association au caractre qui sera gal  "1".
   * Tous les caractres n'tant pas des lettres seront remplaces par le caractre " ".
   * <p> 
   * 
   * @param string La chane de caractres.
   * @return Le vecteur.
   */
  public double[] stringTo27CaractersVector(String string)
  {
    // Dfinition de l'espacer mmoire pour le vecteur.
    double[] vector = new double[LOWER_CASE_VECTOR_DIMENSION*string.length()];
    
    // Mettre les caractres en minuscule.
    string = string.toLowerCase();
    
    // Retirer les caractres non lettre.
    string = SStringUtil.replaceAllNonAlphaCaracter(string, ' ');
    
    // Itrer sur l'enemble des caractres.
    int position;
    
    for(int i = 0; i < string.length(); i++)
    {
      char c = string.charAt(i);
      
      // Identification de la colonne o sera inscrit la valeur 1.0.
      if(c !=  ' ')
        position = (int)(string.charAt(i) - 'a');
      else
        position = LOWER_CASE_VECTOR_DIMENSION-1;
      
      vector[i*LOWER_CASE_VECTOR_DIMENSION + position] = 1.0;
    }
    
    return vector;
  }

  /**
   * <p>
   * Mthode pour gnrer un vecteur  partir d'une image en nuance de gris. 
   * Si l'image est en couleur, il y aura une conversion de la couleur en nuance de gris.
   * Les donnes sont prsents dans l'ordre en itrant ligne par ligne.
   * </p>
   * 
   * @see https://stackoverflow.com/questions/9131678/convert-a-rgb-image-to-grayscale-image-reducing-the-memory-in-java
   * @param image L'image noire et blanche.
   * @return Le vecteur associ  l'image.
   */
  public double[] greyImageToVector(BufferedImage image)
  {
    // Allocation de la mmoire pour le vecteur.
    double[] array = new double[image.getHeight()*image.getWidth()];
    
    int index = 0;
    
    // Itration sur les pixels de l'image (i = x, j = y).
    for(int y = 0; y < image.getHeight(); y++)          
      for(int x = 0; x < image.getWidth(); x++)         
      {
        // Rfrence : 
        int rgb = image.getRGB(x, y);
        int r = (rgb >> 16) & 0xFF;
        int g = (rgb >> 8) & 0xFF;
        int b = rgb & 0xFF;
        
        int greyLevel = (r + g + b) / 3;
        int grey = (greyLevel << 16) + (greyLevel << 8) + greyLevel;
        
        // Construire une couleur dans l'espace des couleurs (valeur entre 0.0 et 1.0).
        SColor color = new SColor(new Color(grey));
        
        // Interpolation entre le paramtre de couleur noire et blanc en fonction de la couleur red.
        array[index] = SMath.linearInterpolation(this.black_value, this.white_value, color.getRed());
                
        // Changer le positionnement dans le vecteur.
        index++;
      }
    
    return array;
  }
 
  /**
   * <p>
   * Mthode pour gnrer un vecteur  partir d'une image en couleur.
   * Les donnes sont prsents dans l'ordre en itrant ligne par ligne et une couleur  la fois.
   * </p>
   * @param image L'image en couleur.
   * @return Le vecteur associ  l'image.
   */
  public double[] coloredImageToVector(BufferedImage image)
  {
    // Taille de l'image.
    final int size = image.getHeight()*image.getWidth();
    
    double[] array_red = new double[size];
    double[] array_green = new double[size];
    double[] array_blue = new double[size];
    
    int index = 0;
    
    // Itrer sur l'ensemble des couleurs de l'image
    for(int y = 0; y < image.getHeight(); y++)          // Iterer sur la ligne.
      for(int x = 0; x < image.getWidth(); x++)         // Iterer sur la colonne.
      {
        // La couleur en (i,j)
        Color color = new Color(image.getRGB(x, y));
        
        // Extraire la couleur par canaux 0 .. 1.
        float[] compArray = new float[4];
        color.getColorComponents(compArray);
        
        // Recadrer la valeur entre les extremums 
        array_red[index] = SMath.linearInterpolation(this.black_value, this.white_value, (double)compArray[0]);
        array_green[index] = SMath.linearInterpolation(this.black_value, this.white_value, (double)compArray[1]);
        array_blue[index] = SMath.linearInterpolation(this.black_value, this.white_value, (double)compArray[2]);
               
        // Changer le positionnement dans le vecteur.
        index++;
      }
    
    // Retourner un vecteur o chaque canaux de couleur est plac dans l'ordre rouge, vert, bleu.
    return SArrays.merge(array_red, array_green, array_blue);
  }
  
}
