/**
 * 
 */
package sim.nn;

import sim.util.SArrays;

/**
 * La classe <b>SNNData</p> reprsente une donne d'entranement pour une rseau de neurone.
 * 
 * @author Simon Vzina
 * @since 2018-03-04
 * @version 2021-12-17
 */
public class SNNData implements Comparable<SNNData> {

  /**
   * La constante <b>DEFAULT_ERROR</b> est l'erreur par dfaut d'une donne non active qui sera {@value}.
   */
  private final static double DEFAULT_ERROR = Double.POSITIVE_INFINITY;
  
  /**
   * La constante <b>DEFAULT_BOOLEAN_CLASSIFIER</b> reprsente les deux expression utilises pour classifier une donne boolenne.
   */
  private final static String[] DEFAULT_BOOLEAN_CLASSIFIER = { "FALSE", "TRUE" };
  
  /**
   * La constante <b>DEFAULT_CLASSIFIER</b> reprsente le nom par dfaut de la classification attendue de la donne tant {@value}.
   */
  private final static String DEFAULT_CLASSIFIER = "classe indetermine";
  
  /**
   * La variable <b>obj</b> reprsente l'objet de l'entranement.
   */
  private final Object obj;
  
  /**
   * La variable <b>vector</b> reprsente le vecteur dcrivant l'objet en entranement.
   */
  private final double[] vector;
  
  /**
   * La variable <b>expected</b> reprsente la valeur attendue  obtenir par le rseau de neurone pour l'objet d'entranement.
   */
  private final double[] expected;
  
  /**
   * La variable <b>all_classification<b> reprsente l'ensemble des valeurs admissibles de classification de la donne. 
   */
  private final String[] all_classification;
  
  /**
   * La variable <b>activation</b> reprsente l'tat d'activation de la donne par un rseau.
   */
  private final double[] activation;
  
  /**
   * La variable <b>error</b> reprsente l'erreur associe  la donne dans le rseau de neurone.
   */
  private double error;
  
  //----------------
  // CONSTRUCTEUR //
  //----------------
  
  /**
   * Constructeur d'une donne de rseau de neurones.
   * 
   * @param obj L'objet  traiter.
   * @param vector Le vecteur reprsentant la donne.
   * @param expected L'activation attendue associe  la donne.
   */
  public SNNData(Object obj, double[] vector, double[] expected) 
  {
    this.obj = obj;
    this.vector = vector;
            
    this.expected = expected;
    this.activation = new double[expected.length];
    
    this.error = DEFAULT_ERROR;
    
    this.all_classification = new String[expected.length];
    
    for(int i = 0; i < this.all_classification.length; i++)
      this.all_classification[i] = DEFAULT_CLASSIFIER;
  }
  
  /**
   * Constructeur d'une donne de rseau de neurones avec descriptions de l'ensemble des classifications admissibles.
   * 
   * @param obj L'objet  traiter.
   * @param vector Le vecteur reprsentant la donne.
   * @param expected L'activation attendue associe  la donne.
   * @param all_classification La liste des mots reprsentant chaque classfication admissible.
   */
  public SNNData(Object obj, double[] vector, double[] expected, String... all_classification)
  {
    this.obj = obj;
    this.vector = vector;
    
    this.expected = expected;
    this.activation = new double[expected.length];
    
    this.error = DEFAULT_ERROR;
    this.all_classification = all_classification;
  } 
  
  //------------
  // MTHODES //
  //------------
  
  /**
   * Mthode pour obtenir le nombre de neurones correspondant  la donne.
   * 
   * @return Le nombre de neurones correspondant  la donne.
   */
  public int getDataSize()
  {
    return this.vector.length;
  }
  
  /**
   * Mthode pour obtenir le nombre de neurones correspondant  la valeur attendue de la donne.
   * 
   * @return Le nombre de neurones de la valeur attendue.
   */
  public int getExpectedSize()
  {
    return this.expected.length;
  }
    
  /**
   * Mthode pour obtenir l'objet reprsentant la donne.
   * 
   * @return L'objet de la donne.
   */
  public Object getObj()
  {
    return this.obj;
  }

  /**
   * Mthode pour obtenir le vecteur reprsentant la donne.
   * 
   * @return Le vecteur de la donne.
   * @throws UnsupportedOperationException Si le vecteur est null.
   */
  public double[] getVector() throws UnsupportedOperationException
  {
    if(this.vector == null)
      throw new UnsupportedOperationException("Erreur SNNData 001 : La donne propose l'usage d'un vecteur  plusieurs profondeurs.");
    
    return this.vector;
  }
 
  /**
   * Mthode pour obtenir la valeur attendue associe au vecteur de la donne.
   * 
   * @return La valeur attendue.
   */
  public double[] getExpected()
  {
    return this.expected;
  }

  /**
   * Mthode pour obtenir l'activation de la donne gnr par un rseau.
   * 
   * @return L'activatio de la donne.
   */
  public double[] getActivation()
  {
    return this.activation;
  }
  /**
   * Mthode pour obtenir l'erreur value entre la donne et un rseau de neurone.
   * Si la donne n'a pas pralablement t valu par un rseau de neurone, le rsultat sera l'infini.
   * 
   * @return L'erreur de la donne.
   */
  public double getError()
  {
    return this.error;
  }
  
  /**
   * Mthode pour obtenir l'tat de la classification de la donne par le rseau.
   * Si la donne est bien classifie, alors le neurone avec la plus grande valeur sera associ  la bonne classe.
   * Si tel est le cas, la classification a t russi. Autrement, la donne est toujours mal classifie.
   * 
   * @return <b>true</b> si la donne est bien classifie par le rseau et <b>false</b> sinon.
   */
  public boolean isClassified()
  {
    // Classification  1 neurone.
    if(this.activation.length == 1)
      return Math.round(this.activation[0]) == this.expected[0];
    
    // Classification  N neurones.
    // Il suffit de vrifier si le numro de neurone ayant la plus grande valeur correspond au numro du neurone attendu.
    return SArrays.findIndexOfMax(this.activation) == SArrays.findIndexOfMax(this.expected);   
  }
  
  /**
   * Mthode pour obtenir la valeur de la classification de la donne par le rseau (pas ncessairement la bonne).
   * La valeur correspond au "nom" de la classe obtenu lors de l'activation parmi l'ensemble disponible.
   * 
   * @return Le nom de la classe dtermin par l'activation de la donne dans le rseau.
   */
  public String getActivationClassification()
  {
    // Nom de la classe dtermin par une activation  1 neurone.
    if(this.activation.length == 1)
      return DEFAULT_BOOLEAN_CLASSIFIER[Math.round((int)this.activation[0])];
        
    // Nom de la classe dtermine par l'activation de la donne.
    return this.all_classification[SArrays.findIndexOfMax(this.activation)];
  }
  
  /**
   * Mthode pour obtenir la valeur de la classification attendue de la donne.
   * 
   * @return Le nom de la classe attendu de la donne.
   */
  public String getExpectedClassification()
  {
    // Nom de la classe dtermin par une activation  1 neurone.
    if(this.activation.length == 1)
      return DEFAULT_BOOLEAN_CLASSIFIER[Math.round((int)this.expected[0])];
        
    // Nom de la classe dtermin par une activation  2 neurones.
    return this.all_classification[SArrays.findIndexOfMax(this.expected)];
  }
  
  /**
   * Mthode pour dfinir l'erreur gnre par l'activation d'une donne par un rseau de neurones.
   * 
   * @param error L'erreur.
   */
  public void setError(double error)
  {
    this.error = error;
  }
  
  /**
   * Mthode pour dfinir l'activation de la donne gnr par un rseau. 
   * La donnes passe en paramtre sera <u>copie</u> dans la structure de la donne.
   * 
   * @param activation L'activation de la donne par un rseau.
   */
  public void setActivation(double[] activation)
  {
    if(activation.length != this.activation.length)
      throw new IllegalArgumentException("Erreur SNNData 002 : L'activation  attribuer  la donne n'a pas la bonne taille (" + activation.length + ").");
    
    SArrays.copy(activation, this.activation);
  }
    
  /**
   * Mthode pour dfinir l'activation et l'erreur gnres par un rseau lors de l'usage de la donne.
   * 
   * @param activation L'activation du rseau avec la donne.
   * @param errors L'erreur du rseau avec l'activation de la donne.
   */
  public void setActivationAndErrors(double[] activation, double[] errors)
  {
    this.setActivation(activation);
    this.setError(SArrays.sumElementsArray(errors));
  }
  
  @Override
  public String toString() 
  {
    return this.obj.toString();
  }

  @Override
  public int compareTo(SNNData o)
  {
    return Double.compare(this.error, o.error); 
  }
  
}
