package sim.nn;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import sim.nn.layer.SBatchConvolutionLayer;
import sim.nn.layer.SBatchDropoutLayer;
import sim.nn.layer.SBatchFullConnectedLayer;
import sim.nn.layer.SBatchLayer;
import sim.nn.layer.SBatchPoolingLayer;
import sim.nn.layer.SConvolutionLayer;
import sim.nn.layer.SDropoutLayer;
import sim.nn.layer.SFullConnectedLayer;
import sim.nn.layer.SLayer;
import sim.nn.layer.SPoolingLayer;
import sim.nn.network.SBatchCustomNetwork;
import sim.nn.network.SBatchFullConnectedNetwork;
import sim.nn.network.SCustomNetwork;
import sim.nn.network.SFullConnectedNetwork;
import sim.nn.network.SNetwork;
import sim.readwrite.object.SMatrixRW;
import sim.readwrite.object.nn.SBatchConvolutionLayerRW;
import sim.readwrite.object.nn.SBatchFullConnectedLayerRW;
import sim.readwrite.object.nn.SConvolutionLayerRW;
import sim.readwrite.object.nn.SDropoutLayerRW;
import sim.readwrite.object.nn.SFullConnectedLayerRW;
import sim.readwrite.object.nn.SLayerRW;
import sim.readwrite.object.nn.SNeuralNetworkRW;
import sim.readwrite.object.nn.SPoolingLayerRW;
import sim.util.SLog;
import sim.util.SReadingException;

/**
 * La classe <b>SNNUtil</b> contient des fonctionnalits pratique lors de l'usage de rseau de neurones.
 * 
 * @author Simon Vzina
 * @since 2018-05-11
 * @version 2022-03-04
 */
public class SNNUtil {

  /**
   * Mthode pour faire la lecture d'un fichier afin d'y construire un rseau.
   * 
   * @param file_name Le nom du fichier (avec extension).
   * @return Le rseau construire  partir des informations du fichier.
   * @throws IllegalArgumentException
   */
  public static SNetwork readNetwork(String file_name) throws IllegalArgumentException
  {
    // Faire la lecture du rseau de neurones.
    SNeuralNetworkRW net = new SNeuralNetworkRW(file_name);
    
    try {
      net.read();
    } catch (SReadingException e) {
      SLog.logException(e);
      throw new RuntimeException();
    } 
    
    List<SLayerRW> layer_list = net.getLayerList();
    
    // Validation s'il y a des couches dans le rseau.
    if(layer_list.size() == 0)
      throw new IllegalArgumentException("Erreur SNNUtil 001 : Le fichier '" + file_name + "' reprsente un rseau sans couche.");
    
    // Choisir le type de rseau qui sera construit.
    if(net.isBatchNetwork())
      //return buildBatchFullConnectedNetwork(layer_list);
      return buildBatchNetwork(layer_list);
    else
      return buildNetwork(layer_list);
  }
      
  /**
   * ...
   * 
   * @param layer_list
   * @return
   */
  private static SNetwork buildNetwork(List<SLayerRW> layer_list)
  {
    List<SLayer> list = new ArrayList<SLayer>();
    
    for(SLayerRW layerRW : layer_list)
    {
      // Type de couche non valide.
      if(layerRW instanceof SBatchFullConnectedLayerRW)
        throw new IllegalArgumentException("Erreur SNNUtil 003 : Une couche de type 'SBatchFullConnectedLayerRW' n'est pas support dans un rseau de type 'Custom'.");
      
      // Type de couche valide.
      if(layerRW instanceof SFullConnectedLayerRW)
      {
        /*
        SFullConnectedLayerRW full_connected_layer = (SFullConnectedLayerRW) layerRW;
        
        SFullConnectedLayer layer = new SFullConnectedLayer(
            new SBufferedMatrix(full_connected_layer.getWeights().getMatrix()), 
            full_connected_layer.getBias(), 
            new SBufferedMatrix(full_connected_layer.getActivationParameters().getMatrix()), 
            full_connected_layer.getAggregationFunctionName(), 
            full_connected_layer.getActivationFunctionName()
            );
        
        list.add(layer);
        */
        list.add(((SFullConnectedLayerRW) layerRW).toFullConnectedLayer());
        
      }
      else
        if(layerRW instanceof SConvolutionLayerRW)
        {
          /*
          SConvolutionLayerRW convolution_layerRW = (SConvolutionLayerRW) layerRW;
            
          // Construire les filtres.
          List<SFiltersRW> l = convolution_layerRW.getFilters();
          SArrayTensor[] array = new SArrayTensor[l.size()];
            
          for(int i = 0; i < l.size(); i++)
          {
            SFiltersRW filtersRW = l.get(i);
              
            List<SMatrixRW> matrixRW_list = filtersRW.getMatrix();
            List<SBufferedMatrix> matrix_list = new ArrayList<SBufferedMatrix>();
              
            for(SMatrixRW m : matrixRW_list)
              matrix_list.add(new SBufferedMatrix(m.getMatrix()));
              
            array[i] = SArrayTensor.buildWithMatrixList(matrix_list);
          }
                      
          SFiltersWeights W = new SFiltersWeights(array);
            
          SConvolutionLayer layer = new SConvolutionLayer(
              W, 
              new SBias(convolution_layerRW.getBias()), 
              new SActivationParameters(new SBufferedMatrix(convolution_layerRW.getActivationParameters().getMatrix())),
              convolution_layerRW.getInputSize(), 
              convolution_layerRW.getPadding(), 
              convolution_layerRW.getStride(), 
              convolution_layerRW.getActivationFunctionName()
          );
          
          list.add(layer);
          */
          list.add(((SConvolutionLayerRW) layerRW).toConvolutionLayer());
          
        }
        else
          if(layerRW instanceof SDropoutLayerRW)
          {
            SDropoutLayerRW dropout_layerRW = (SDropoutLayerRW) layerRW;
              
            list.add(new SDropoutLayer(dropout_layerRW.getInputSize(), dropout_layerRW.getDropout()));
          }
          else
            if(layerRW instanceof SPoolingLayerRW)
            {
              SPoolingLayerRW pooling_layerRW = (SPoolingLayerRW) layerRW;
              
              list.add(new SPoolingLayer(pooling_layerRW.getInputSize(), pooling_layerRW.getInputDepth(), pooling_layerRW.getFilterOffset(), pooling_layerRW.getPoolingFunctionName()));
            }
            else
              throw new IllegalArgumentException("Erreur SNNUtil 004 : Une couche n'est pas reconnu pour tre introduite dans un rseau custom.");
    }
    
    return new SCustomNetwork(list);
    
  }
  
  /**
   * ...
   * 
   * @param layer_list
   * @return
   */
  private static SNetwork buildBatchNetwork(List<SLayerRW> layer_list)
  {
    List<SBatchLayer> list = new ArrayList<SBatchLayer>();
    
    for(SLayerRW layerRW : layer_list)
    {
      // Type de couche valide.
      if(layerRW instanceof SBatchFullConnectedLayerRW)
      {
        /*
        SBatchFullConnectedLayerRW layer = (SBatchFullConnectedLayerRW) layerRW;
        
        SBatchFullConnectedLayer l = new SBatchFullConnectedLayer(
            new SBufferedMatrix(layer.getWeights().getMatrix()), 
            layer.getBias(), 
            new SBufferedMatrix(layer.getActivationParameters().getMatrix()),
            layer.getScale(), 
            layer.getShift(), 
            layer.getAggregationFunctionName(), 
            layer.getActivationFunctionName()
            );
        
        list.add(l);
        */
        
        list.add(((SBatchFullConnectedLayerRW) layerRW).toBatchFullConnectedLayer());
        
      }
      else
        if(layerRW instanceof SBatchConvolutionLayerRW)
        {
          /*
          SBatchConvolutionLayerRW convolution_layerRW = (SBatchConvolutionLayerRW) layerRW;
            
          // Construire les filtres.
          List<SFiltersRW> l = convolution_layerRW.getFilters();
          SArrayTensor[] array = new SArrayTensor[l.size()];
            
          for(int i = 0; i < l.size(); i++)
          {
            SFiltersRW filtersRW = l.get(i);
              
            List<SMatrixRW> matrixRW_list = filtersRW.getMatrix();
            List<SBufferedMatrix> matrix_list = new ArrayList<SBufferedMatrix>();
              
            for(SMatrixRW m : matrixRW_list)
              matrix_list.add(new SBufferedMatrix(m.getMatrix()));
              
            array[i] = SArrayTensor.buildWithMatrixList(matrix_list);
          }
                      
          SFiltersWeights W = new SFiltersWeights(array);
           
          // Ajouter la couche  la liste.
          SBatchConvolutionLayer layer = new SBatchConvolutionLayer(
              W, 
              new SBias(convolution_layerRW.getBias()), 
              new SActivationParameters(new SBufferedMatrix(convolution_layerRW.getActivationParameters().getMatrix())),
              new SScale(convolution_layerRW.getScale()), 
              new SShift(convolution_layerRW.getShift()), 
              convolution_layerRW.getInputSize(), 
              convolution_layerRW.getPadding(), 
              convolution_layerRW.getStride(), 
              convolution_layerRW.getActivationFunctionName()
              );
          
          list.add(layer);
          */
          list.add(((SBatchConvolutionLayerRW) layerRW).toBatchConvolutionLayer());
          
        }
        else
          if(layerRW instanceof SDropoutLayerRW)
          {
            SDropoutLayerRW dropout_layerRW = (SDropoutLayerRW) layerRW;
              
            list.add(new SBatchDropoutLayer(dropout_layerRW.getInputSize(), dropout_layerRW.getDropout()));
          }
          else
            if(layerRW instanceof SPoolingLayerRW)
            {
              SPoolingLayerRW pooling_layerRW = (SPoolingLayerRW) layerRW;
              
              list.add(new SBatchPoolingLayer(pooling_layerRW.getInputSize(), pooling_layerRW.getInputDepth(), pooling_layerRW.getFilterOffset(), pooling_layerRW.getPoolingFunctionName()));
            }            
            else
              throw new IllegalArgumentException("Erreur SNNUtil 004 : Une couche n'est pas reconnu pour tre introduite dans un rseau custom.");
    }
    
    return new SBatchCustomNetwork(list);
    
  }
  
  /**
   * ...
   * 
   * @param network_file_name
   * @param network
   * @throws IOException
   */
  public static void writeNetwork(String network_file_name, SNetwork network) throws IOException
  {
    if(network instanceof SFullConnectedNetwork)
      writeFullConnectedNetwork(network_file_name, (SFullConnectedNetwork) network);
    else
      if(network instanceof SBatchFullConnectedNetwork)
        writeBatchFullConnectedNetwork(network_file_name, (SBatchFullConnectedNetwork) network);
      else
        if(network instanceof SCustomNetwork)
          writeCustomNetwork(network_file_name, (SCustomNetwork) network);
        else
          if(network instanceof SBatchCustomNetwork)
            writeBatchCustomNetwork(network_file_name, (SBatchCustomNetwork) network);
          else
            throw new IllegalArgumentException("Erreur SNNUtil 003 : Le type de rseau ne peu par tre crit, car il n'est pas support.");
  }
  
  /**
   * Mthode pour faire l'criture du rseau de neurones dans un fichier texte.
   * 
   * @param network_file_name Le nom du fichier  crire.
   * @param network Le rseau de neurones.
   * @throws IOException
   */
  private static void writeFullConnectedNetwork(String network_file_name, SFullConnectedNetwork network) throws IOException
  {
    SLayer[] layers = network.getLayers();
    
    List<SLayerRW> layer_list = new ArrayList<SLayerRW>();
    
    // Construire l'ensemble des couche de rseau en fonction RW.
    for(int i = 0; i < network.getNbLayers(); i++)
    {
      // Puisqu'un SFullConnectedNetwork ne possde qu'une seule sorte de couche,
      // nous pouvons forcer le casting de la sorte.
      SFullConnectedLayer layer = (SFullConnectedLayer) layers[i];
      
      String str = network_file_name.substring(0, network_file_name.length()-4);
      
      SFullConnectedLayerRW l = new SFullConnectedLayerRW(str + "_" + i + "_FULL_CONNECTED_LAYER" + ".txt", 
          layer.getAggregationFunctionName(), 
          layer.getActivationFunctionName(), 
          new SMatrixRW(layer.getWeights().clone().getMatrix().getLineList()), 
          layer.getBias().clone().getVector(),
          new SMatrixRW(layer.getActivationParameters().clone().getMatrix().getLineList())
        );
      
      layer_list.add(l);
    }
    
    SNeuralNetworkRW net = new SNeuralNetworkRW(network_file_name, false, layer_list);
    net.write();
  }
  
  /**
   * ...
   * 
   * @param network_file_name
   * @param network
   * @throws IOException
   */
  private static void writeBatchFullConnectedNetwork(String network_file_name, SBatchFullConnectedNetwork network) throws IOException
  {
    SLayer[] layers = network.getLayers();
    
    List<SLayerRW> layer_list = new ArrayList<SLayerRW>();
    
    // Construire l'ensemble des couche de rseau en fonction RW.
    for(int i = 0; i < network.getNbLayers(); i++)
    {
      // Puisqu'un SFullConnectedNetwork ne possde qu'une seule sorte de couche,
      // nous pouvons forcer le casting de la sorte.
      SBatchFullConnectedLayer layer = (SBatchFullConnectedLayer) layers[i];
      
      String str = network_file_name.substring(0, network_file_name.length()-4);
      SBatchFullConnectedLayerRW l = new SBatchFullConnectedLayerRW(str + "_" + i + "BATCH_FULL_CONNECTED_LAYER" + ".txt", 
          layer.getAggregationFunctionName(), 
          layer.getActivationFunctionName(), 
          new SMatrixRW(layer.getWeights().clone().getMatrix().getLineList()), 
          layer.getBias().clone().getVector(), 
          new SMatrixRW(layer.getActivationParameters().clone().getMatrix().getLineList()),
          layer.getScale().clone().getVector(), 
          layer.getShift().clone().getVector()
          );
      layer_list.add(l);
    }
    
    SNeuralNetworkRW net = new SNeuralNetworkRW(network_file_name, true, layer_list);
    net.write();
  }
  
  /**
   * ...
   * 
   * @param network_file_name
   * @param network
   * @throws IOException
   */
  private static void writeCustomNetwork(String network_file_name, SCustomNetwork network) throws IOException
  {
    SLayer[] layers = network.getLayers();
    
    List<SLayerRW> layer_list = new ArrayList<SLayerRW>();
    
    String str = network_file_name.substring(0, network_file_name.length()-4);
    
    // Construire l'ensemble des couche de rseau en fonction RW.
    for(int i = 0; i < network.getNbLayers(); i++)
    {
      SLayer layer = layers[i];
      
      if(layer instanceof SFullConnectedLayer)
      {
        SFullConnectedLayer l = (SFullConnectedLayer) layer;
        
        SFullConnectedLayerRW l_RW = new SFullConnectedLayerRW(str + "_" + i + "_FULL_CONNECTED_LAYER" + ".txt", 
            l.getAggregationFunctionName(),
            l.getActivationFunctionName(), 
            new SMatrixRW(l.getWeights().clone().getMatrix().getLineList()), 
            l.getBias().clone().getVector(),
            new SMatrixRW(l.getActivationParameters().clone().getMatrix().getLineList())
            ); 
            
        layer_list.add(l_RW);
      }
      else
        if(layer instanceof SConvolutionLayer)
        {
          SConvolutionLayer l = (SConvolutionLayer) layer;
          SConvolutionLayerRW l_RW = new SConvolutionLayerRW(str + "_" + i + "_CONVOLUTION_LAYER" + ".txt", 
              l.getActivationFunctionName(), 
              l.getFiltersWeights().clone(), 
              l.getBias().clone().getVector(), 
              new SMatrixRW(l.getActivationParameters().clone().getMatrix().getLineList()),
              l.getInputsSize(),
              l.getStride(),
              l.getPadding()
              );
          
          layer_list.add(l_RW);
        }
        else
          if(layer instanceof SDropoutLayer)
          {
            SDropoutLayer l = (SDropoutLayer) layer;
            layer_list.add(new SDropoutLayerRW(str + "_" + i + "_DROPOUT_LAYER" + ".txt", l.getInputsSize(), l.getDropout()));        
          }
          else  
            if(layer instanceof SPoolingLayer)
            {
              SPoolingLayer l = (SPoolingLayer) layer;
              layer_list.add(new SPoolingLayerRW(str + "_" + i + "_POOLING_LAYER" + ".txt", l.getInputsSize(), l.getInputsDepth(), l.getFilterOffset(), l.getPoolingFunctionName()));        
            }
            else
              throw new UnsupportedOperationException("Erreur SNNUtil 004 : Type de couche non supporte.");
    }
    
    SNeuralNetworkRW net = new SNeuralNetworkRW(network_file_name, false, layer_list);
    net.write();
  }
  
  /**
   * ...
   * 
   * @param network_file_name
   * @param network
   * @throws IOException
   */
  private static void writeBatchCustomNetwork(String network_file_name, SBatchCustomNetwork network) throws IOException
  {
    SLayer[] layers = network.getLayers();
    
    List<SLayerRW> layer_list = new ArrayList<SLayerRW>();
    
    String str = network_file_name.substring(0, network_file_name.length()-4);
    
    // Construire l'ensemble des couche de rseau en fonction RW.
    for(int i = 0; i < network.getNbLayers(); i++)
    {
      SLayer layer = layers[i];
      
      if(layer instanceof SBatchFullConnectedLayer)
      {
        SBatchFullConnectedLayer l = (SBatchFullConnectedLayer) layer;
        
        SBatchFullConnectedLayerRW l_RW = new SBatchFullConnectedLayerRW(str + "_" + i + "_BATCH_FULL_CONNECTED_LAYER"+ ".txt", 
            l.getAggregationFunctionName(), 
            l.getActivationFunctionName(), 
            new SMatrixRW(l.getWeights().clone().getMatrix().getLineList()), 
            l.getBias().clone().getVector(), 
            new SMatrixRW(l.getActivationParameters().clone().getMatrix().getLineList()), 
            l.getScale().clone().getVector(),
            l.getShift().clone().getVector()
            );
        
        layer_list.add(l_RW);
      }
      else
        if(layer instanceof SBatchConvolutionLayer)
        {
          SBatchConvolutionLayer l = (SBatchConvolutionLayer) layer;
          
          SBatchConvolutionLayerRW l_RW = new SBatchConvolutionLayerRW(str + "_" + i + "_BATCH_CONVOLUTION_LAYER" + ".txt", 
              l.getActivationFunctionName(), 
              l.getFiltersWeights().clone(), 
              l.getBias().clone().getVector(), 
              new SMatrixRW(l.getActivationParameters().clone().getMatrix().getLineList()),
              l.getScale().clone().getVector(),
              l.getShift().clone().getVector(),
              l.getInputsSize(), 
              l.getStride(), 
              l.getPadding()
              );        
          
          layer_list.add(l_RW);
        }
        else
          if(layer instanceof SBatchDropoutLayer)
          {
            SBatchDropoutLayer l = (SBatchDropoutLayer) layer;
            layer_list.add(new SDropoutLayerRW(str + "_" + i + "_BATCH_DROPOUT_LAYER" + ".txt", l.getInputsSize(), l.getDropout()));        
          }
          else  
            if(layer instanceof SBatchPoolingLayer)
            {
              SBatchPoolingLayer l = (SBatchPoolingLayer) layer;
              layer_list.add(new SPoolingLayerRW(str + "_" + i + "_BATCH_POOLING_LAYER" + ".txt", l.getInputsSize(), l.getInputsDepth(), l.getFilterOffset(), l.getPoolingFunctionName()));        
            }
            else
              throw new UnsupportedOperationException("Erreur SNNUtil 004 : Type de couche non supporte.");
    }
    
    SNeuralNetworkRW net = new SNeuralNetworkRW(network_file_name, true, layer_list);
    net.write();
  }
  
}
