/**
 * 
 */
package sim.nn.function;

import java.util.List;

import sim.exception.SNaNValueException;

/**
 * La classe abstrait <b>SAbstractBatchFunction</b> reprsente une fonction de type batch permettant d'activer plusieurs rseaux 
 * en parallle dont la valeur de chacun d'influencera mutuellement selon la nature de la fonction.
 * 
 * @author Simon Vezina
 * @since 2018-10-09
 * @version 2022-01-05
 */
public abstract class SAbstractBatchFunction extends SMemoryBatchFunction  {
  
  /**
   * Constructeur d'une fonction de type <i>batch</i> o le nombre de neurones en entre et en sortie sont identiques.
   * 
   * @param neurons Le nombre de neurones.
   */
  public SAbstractBatchFunction(int neurons)
  {
    super(neurons);
  }
  
  /**
   * Constructeur d'une fonction de type <i>batch</i> dont le nombre de neurones en entre et en sortie sont diffrents.
   * 
   * @param inputs Le nombre de neurones en entre.
   * @param outputs Le nombre de neurones en sortie.
   */
  public SAbstractBatchFunction(int inputs, int outputs)
  {
    super(inputs, outputs); 
  }  
          
  @Override
  public List<double[]> batchForwardPropagation(int b, List<double[]> batch_inputs) throws IllegalArgumentException
  {
    // Vrification de la taille de la batch b  oprer sur le vecteur d'entre.
    if(b > batch_inputs.size())
      throw new IllegalArgumentException("Erreur SAbstractBatchFunction 001 : La taile de la batch a excuter (b=" + b + ") ne peut pas tre plus grande et la batch du vecteur d'entre (m=" + batch_inputs.size() + ").");
    
    // Dfinir la taille de la batch pour la fonction et allocation de mmoire si ncessaire.
    this.setBatchSize(b);
    
    // Conserver l'accs aux donnes d'entre.
    this.batch_forward_inputs = batch_inputs;
    
    // Effectuer l'activation pour l'ensemble de la batch et pour l'ensemble des neurones.
    return forwardPass();
  }
  
  @Override
  public List<double[]> batchBackPropagation(int b, List<double[]> batch_delta) throws IllegalArgumentException
  {
    if(b > batch_delta.size())
      throw new IllegalArgumentException("Erreur SAbstractBatchFunction 002 : Le parametre b (b=" + b + ") ne peut pas tre suprieur  la batch du vecteur (m=" + batch_delta.size() + ").");
  
    if(b != this.getBatchSize())
      throw new IllegalArgumentException("Erreur SAbstractBatchFunction 003 : Le paramtre b n'est pas gal  la taille de la batch en calcul lors de l'activation.");
    
    this.batch_backward_inputs = batch_delta;
    
    return backwardPass();
  }
  
  /**
   * Mthode qui effectue le calcul de la fonction <u>vers l'avant</u> pour le neurone u de la batch b.
   * 
   * @param b L'indice de la batch.
   * @param u L'indice du neurone.
   * @return Le rsultat de la fonction <u>vers l'avant</u> pour le neurone u.
   */
  abstract protected double forward(int b, int u);
  
  /**
   * Mthode qui effectue le calcul de la fonction <u>vers l'arrire</u> correspondant  la propagation de l'erreur de la fonction pour le neurone v de la batch b.
   * 
   * @param b L'indice de la batch.
   * @param v L'indice du neurone.
   * @return Le rsultat de la fonction <u>vers l'arrire</u> correspondant  la propagation de l'erreur au neurone v.
   */
  abstract protected double backward(int b, int v);
  
  /**
   * ...
   * 
   * @return
   * @throws SNaNValueException
   */
  protected List<double[]> forwardPass() throws SNaNValueException
  {
    // Itrer sur l'ensemble des batchs.
    for(int b = 0; b < this.getBatchSize(); b++)
    {
      // Obtenir un vecteur de sortie de la batch batch b.
        double[] output = batch_forward_outputs.get(b);
    
        // Itrer sur l'ensemble des neurones.
        for(int u = 0; u < output.length; u++)
        {
          double value =  forward(b, u);
      
          if(Double.isNaN(value))
            throw new SNaNValueException("Erreur SAbstractBatchFunction 004 : Le neurone u = " + u + " et de batch b = " + b + " est gal  NaN.");
      
          output[u] = value;
        }
    }
       
    return this.batch_forward_outputs;
  }
  
  /**
   * ...
   * 
   * @return
   * @throws SNaNValueException
   */
  protected List<double[]> backwardPass() throws SNaNValueException
  {
    // Itrer sur l'ensemble des batchs.
    for(int b = 0; b < this.getBatchSize(); b++)
    {
      // Obtenir un vecteur de sortie avec profondeur de batch b.
      double[] backward_output = this.batch_backward_outputs.get(b);
    
      for(int v = 0; v < backward_output.length; v++)
      {
        double value =  backward(b, v);
      
          if(Double.isNaN(value))
            throw new SNaNValueException("Erreur SAbstractBatchFunction 005 : Le neurone v = " + v + " et de batch b = " + b + " est gal  NaN.");
          
          backward_output[v] = value;
      }
    }
        
    return this.batch_backward_outputs;
  }
  
  @Override
  public String toString()
  {
    return getFunctionName();
  }

  @Override
  public boolean equals(Object obj) {
    if (this == obj)
      return true;
    if (obj == null)
      return false;
    if (getClass() != obj.getClass())
      return false;
    
    SAbstractBatchFunction other = (SAbstractBatchFunction) obj;
    if (batch_backward_inputs == null) {
      if (other.batch_backward_inputs != null)
        return false;
    } else if (!batch_backward_inputs.equals(other.batch_backward_inputs))
      return false;
    if (batch_backward_outputs == null) {
      if (other.batch_backward_outputs != null)
        return false;
    } else if (!batch_backward_outputs.equals(other.batch_backward_outputs))
      return false;
    if (batch_forward_inputs == null) {
      if (other.batch_forward_inputs != null)
        return false;
    } else if (!batch_forward_inputs.equals(other.batch_forward_inputs))
      return false;
    if (batch_forward_outputs == null) {
      if (other.batch_forward_outputs != null)
        return false;
    } else if (!batch_forward_outputs.equals(other.batch_forward_outputs))
      return false;
    if (batch_size != other.batch_size)
      return false;
    return true;
  }
  
}
