/**
 * 
 */
package sim.nn.function;

import sim.exception.SConstructorException;
import sim.exception.SNaNValueException;
import sim.math.SArrayTensor;

/**
 * La classe abstraite <b>SAbstractFilterFunction</b> reprsente une fonction avec filtre.
 * Cette fonction permet d'accder aux donnes des neurones par un accs en tenseur.
 * 
 * @author Simon Vezina
 * @since 2019-02-03
 * @version 2021-10-18
 */
public abstract class SAbstractFilterFunction extends SMemoryNeuralFunction {

  /**
   * Le paramtre <b>inputOffset</b> reprsente le nombre de ligne (l) et le nombre de colonne (c) de la pseudo-image correspondant  la <u>donne d'entre</u>  2 dimensions.
   * Ainsi, la dimension est [l][c] o l = c = inputOffset.
   */
  protected final int inputOffset;
  
  /**
   * Le paramtre <b>outputOffset</b> reprsente le nombre de ligne (m) et le nombre de colonne (n) de la pseudo-image correspondant  la <u>donne d'entre</u>  2 dimensions.
   * Ainsi, la dimension est [m][n] o m = n = inputOffset.
   */
  protected final int outputOffset;
  
  /**
   * La variable <b>filterOffset</b> reprsente la taille du filtre  appliquer sur la pseudo-image. Par exemple, si filterOffset == 3, les filtres seront des 3x3. 
   * Dans la notation, nous utiliserons la dimension [i][j] o i = j = filterOffset pour accder aux composants des filtres.
   * 
   */
  protected final int filterOffset;
  
  /**
   * La variable <b>padding</b> reprsente l'excdant de bordure utilis dans le produit de convolution.
   */
  protected final int padding;
  
  /**
   * La variable <b>stride</b> reprsente le dcalage du filtre  chaque itration du produit de convolution.
   */
  protected final int stride;
  
  /**
   * La variable <b>input_depth</b> reprsente le niveau de profondeur de l'entre.
   */
  protected final int input_depth;
  
  /**
   * La variable <b>output_depth</b> reprsente le niveau de profondeur de la sortie.
   */
  protected final int output_depth;
  
  /**
   * La variable <b>inputDepthOffset</b> reprsente le offset dans le tableau des neurones pour passer d'une profondeur  une autre pour l'entre.
   */
  protected final int inputDepthOffset;
  
  /**
   * La variable <b>outputDepthOffset</b> reprsente le offset dans le tableau des neurones pour passer d'une profondeur  une autre pour la sortie.
   */
  protected final int outputDepthOffset;
  
  /**
   * ...
   */
  protected final SArrayTensor tensor_forward_input;
  
  /**
   * ...
   */
  protected final SArrayTensor tensor_forward_output;
  
  /**
   * ...
   */
  protected final SArrayTensor tensor_backward_input;
  
  /**
   * ...
   */
  protected final SArrayTensor tensor_backward_output;
  
  /**
   * Constructeur d'une fonction de filtrage.
   * 
   * @param inputs
   * @param inputs_depth
   * @param outputs_depth
   * @param filter_offset
   * @param padding
   * @param stride
   * @throws SConstructorException Si la construction de la fonction est inadquate.
   */
  public SAbstractFilterFunction(int inputs, int inputs_depth, int outputs_depth, int filter_offset, int padding, int stride) throws SConstructorException
  {
    // Pour faire l'allocation de la mmoire, nous devons dterminer le nombre de neurones en sortie.
    super(inputs, SNeuralMath.computeOutputsSizeForSquareInputsDepth(inputs, inputs_depth, outputs_depth, filter_offset, padding, stride));
    
    if(stride < 1)
      throw new SConstructorException("Erreur SAbstractFilterFunction 001 : Le stride doit tre suprieur  1.");
    
    //if(padding < 0)
    if(padding != 0)  
      throw new SConstructorException("Erreur SAbstractFilterFunction 002 : POUR L'INSTANT, LE PADDING N'ES PAS SUPPORT. Le padding pour tre suprieur  0.");
     
    this.input_depth = inputs_depth;
    this.output_depth = outputs_depth;
    this.filterOffset = filter_offset;
    this.stride = stride;
    this.padding = padding;
    
    // Dfinir le offset de l'entre.
    try {
      
      this.inputOffset = SNeuralMath.computeInputsOffsetForSquareInputsDepth(inputs, inputs_depth);
    
    }catch(IllegalArgumentException e){
      throw new SConstructorException("Erreur SAbstractFilterFunction 003 : Une erreur est survenue lors de la construction de l'objet.", e);
    }
    
    // Dfinir le offset de la sortie.
    this.outputOffset = SNeuralMath.computeOutputsOffsetForSquareInputsDepth(inputs, inputs_depth, filter_offset, padding, stride);
    
    this.inputDepthOffset = inputs / inputs_depth;
    
    if(this.inputDepthOffset*inputs_depth != inputs)
      throw new SConstructorException("Erreur SAbstractFilterFunction 004 : Le nombre d'entre input = " + inputs + " n'est pas divisible entirement par la profondeur d'entre input_depth = " + inputs_depth + ".");
    
    this.outputDepthOffset = this.getOutputsSize() / this.output_depth;
    
    // Test de suret.
    if(this.outputDepthOffset*this.output_depth != this.getOutputsSize())
      throw new SConstructorException("Erreur SAbstractFilterFunction 005 : Le nombre de sortie output = " + this.getOutputsSize() + " n'est pas divisible entirement par la profondeur de sortie output_depth = " + this.output_depth + ".");
  
    //---------------------------
    // Construction des tenseurs.
    //---------------------------
    
    // Constructeur du tenseur d'entre de l'activation.
    // Le array de donne sera inject lors de l'activation.
    this.tensor_forward_input = new SArrayTensor(this.input_depth, this.inputOffset, this.inputOffset);
    
    // Construction du tenseur de sortie d'activation.
    // Nous allons injecter le array des rsultats ds maintenant.
    this.tensor_forward_output = new SArrayTensor(this.output_depth, this.outputOffset, this.outputOffset);
    this.tensor_forward_output.setArray(this.forward_outputs);
    
    // Constructeur du tenseur d'entre de la propagation des erreurs.
    // Le array de donne sera inject lors de la propagation des erreurs.
    this.tensor_backward_input = new SArrayTensor(this.output_depth, this.outputOffset, this.outputOffset);
    
    // Construction du tenseur de sortie de la propagation des erreurs.
    // Nous allons injecter le array des rsultats ds maintenant.
    this.tensor_backward_output = new SArrayTensor(this.input_depth, this.inputOffset, this.inputOffset);
    this.tensor_backward_output.setArray(this.backward_outputs);
  }

  @Override
  public int getInputsDepth()
  {
    return this.input_depth;
  }

  @Override
  public int getOutputsDepth()
  {
    return this.output_depth;
  }
  
  /**
   * Mthode pour obtenir la dimension de largeur de l'entre de forme carr.
   * 
   * @return
   */
  public int getInputOffset()
  {
    return this.inputOffset;
  }
  
  /**
   * Mthode pour obtenir la dimension de largeur de la sortie de forme carr.
   * 
   * @return
   */
  public int getOutputOffset()
  {
    return this.outputOffset;
  }
  
  /**
   * Mthode pour obtenir la taille du filtre de filtrage.
   * 
   * @return La taille du filtre.
   */
  public int getFilterOffset()
  {
    return this.filterOffset;
  }
  
  /**
   * Mthode pour obtenir le nombre de case d'largissement de l'espace (padding).
   * 
   * @return Le padding.
   */
  public int getPadding()
  {
    return this.padding;
  }
  
  /**
   * Mthode pour obtenir le nombre de case de glissement de la fonction (stride).
   * 
   * @return Le stride.
   */
  public int getStride()
  {
    return this.stride;
  }
  
  /**
   * ...
   * 
   * @return
   */
  protected double[] forwardPass()
  {
    // Filtrage avec un dcalage du filtre de 1 case.
    if(this.stride == 1)
    {
      // Raliser le filtre tous les filtres (f) de la couche pour toutes les ligne (m) et colonne (n) du tenseur de sortie.
      for(int f = 0; f < this.output_depth; f++)
        for(int m = 0; m < this.outputOffset; m++)
          for(int n = 0; n < this.outputOffset; n++)  
            this.tensor_forward_output.set(filtering(f, m, n), f, m, n);
    }
    else
    {
      // Raliser le produit de convolution pour tous les filtres d de la couche.
      for(int f = 0; f < this.output_depth; f++)
        for(int m = 0; m < this.outputOffset; m++)
          for(int n = 0; n < this.outputOffset; n++)
            this.tensor_forward_output.set( filtering(f, m*stride, n*stride), f, m, n);
    }
    
    // Retourner les champs de sortie modifi par les calculs de tenseur.
    return this.forward_outputs;
  }
  
  /**
   * ...
   * 
   * @return
   * @throws SNaNValueException
   */
  abstract protected double[] backwardPass() throws SNaNValueException;
  
  /**
   * Mthode qui ralise le filtrage du tenseur d'entre dans le but
   * de calculer la valeur de sortie [f][m][n] du tenseur de sortie.
   * 
   * @param f La profondeur de la sortie.
   * @param m La ligne de la sortie.
   * @param n La colonne de la sortie.
   * @return La valeur de sortie [f][m][n] de la fonction de filtrage.
   */
  protected abstract double filtering(int f, int m, int n);
  
  @Override
  public double[] forwardPropagation(double[] inputs) throws IllegalArgumentException
  {
    // Validation du nombre de neurones en entre.
    if(inputs.length != this.getInputsSize())
      throw new IllegalArgumentException("Erreur SAbstractFilterFunction 002 : L'entre n'est pas admissible, car la taille de inputs = " + inputs.length + " n'est pas gal  " + this.getInputsSize());
    
    // Conserver l'accs des entre.
    this.forward_inputs = inputs;
    this.tensor_forward_input.setArray(inputs);
    
    // Effectuer l'activation de la fonction pour l'ensemble des neurones.
    return forwardPass();
  }
  
  @Override
  public double[] backPropagation(double[] delta) throws IllegalArgumentException
  {
    // Validation de la taille de l'entre en propagation de l'erreur.
    if(delta.length != this.getOutputsSize())
      throw new IllegalArgumentException("Erreur SAbstractFilterFunction 001 : La propagation de l'erreur n'est pas admissible, car la taille de delta = " + delta.length + " n'est pas gal  " + this.getOutputsSize() + ".");
    
    // Affectation de l'erreur d'entre dans la fonction.
    this.backward_inputs = delta;
    this.tensor_backward_input.setArray(delta);
    
    // Effectuer la propagation de l'erreur  l'ensemble des neurones.
    return backwardPass();
  }
 
  @Override
  public String getDimension()
  {
    return this.getInputsSize() + " x " + this.getOutputsSize() + " --> (" + this.getInputsDepth() + " x " + this.getInputOffset() + " x " + this.getInputOffset() + ") x (" + this.getOutputsDepth() + " x " + this.getOutputOffset() + " x " + this.getOutputOffset() + ")";
  }
  
  @Override
  public String toString()
  {
    return getFunctionName() + " (" + this.filterOffset + " x " + filterOffset + ")";
  }
  
}
