/**
 * 
 */
package sim.nn.function;

import sim.exception.SNaNValueException;
import sim.exception.SNoImplementationException;

/**
 * La classe abstraite <b>SAbstractNeuralFunction</b> reprsente une fonction neuronale
 * o les mthodes d'activation et de propagation de l'erreur pour un neurone de la fonction resteront  tre dfinies.
 * 
 * @author Simon Vzina
 * @since 2019-02-22
 * @version 2022-09-14 (Version labo : Apprentissage par rseau de neurones)
 */
public abstract class SAbstractNeuralFunction extends SMemoryNeuralFunction {

  /**
   * Constructeur d'une fonction neuronale dont le nombre de neurones en entre et en sortie sont identiques.
   * 
   * @param neurons Le nombre de neurones en entre et en sortie.
   */
  public SAbstractNeuralFunction(int neurons)
  {
    super(neurons);
  }

  /**
   * Constructeur d'une fonction neuronale dont le nombre de neurones en entre et en sortie sont diffrents.
   * 
   * @param inputs_size Le nombre de neurones en entre.
   * @param outputs_size Le nombre de neurones en sortie.
   */
  public SAbstractNeuralFunction(int inputs_size, int outputs_size)
  {
    super(inputs_size, outputs_size);
  }
  
  @Override
  public double[] forwardPropagation(double[] inputs) throws IllegalArgumentException
  {
    // Validation du nombre d'entre dans la fonction.
    if(inputs.length != this.getInputsSize())
      throw new IllegalArgumentException("Erreur SAbstractNeuralFunction 001 : L'entre n'est pas admissible, car la taille de inputs = " + inputs.length + " n'est pas gal  " + this.getInputsSize());
    
    // Conserver l'accs aux donne d'entre.
    this.forward_inputs = inputs;
    
    // Effectuer l'activation sur l'ensemble des neurones.
    return forwardPass();
  }
  
  @Override
  public double[] backPropagation(double[] delta) throws IllegalArgumentException
  {
    // Validation de la taille de l'entre en propagation de l'erreur.
    if(delta.length != this.getOutputsSize())
      throw new IllegalArgumentException("Erreur SAbstractNeuralFunction 002 : La propagation de l'erreur n'est pas admissible, car la taille de delta = " + delta.length + " n'est pas gal  " + this.getOutputsSize() + ".");
    
    // Conserver l'accs aux donne d'entre de la propagation de l'erreur.
    this.backward_inputs = delta;
    
    // Effectuer la propagation de l'erreur sur l'ensemble des neurones.
    return backwardPass();
  }
  
  /**
   * Mthode qui effectue le calcul de la fonction <u>vers l'avant</u> pour le neurone u.
   * Ce mode de calcul est utilis lors du mode de l'activation du rseau.
   * 
   * @param u L'indice du neurone en sortie qui sera calcul.
   * @return Le rsultat de la fonction <u>vers l'avant</u> pour le neurone u.
   */
  abstract protected double forward(int u);
  
  /**
   * Mthode qui effectue le calcul de la fonction <u>vers l'arrire</u> pour le neurone v.
   * Ce mode de calcul correspond  la propagation de l'erreur de la fonction 
   * et il est utilis de la propagation de l'erreur du rseau.
   * 
   * @param v L'indice du neurone en sortie vers laquelle l'erreur sera propage.
   * @return Le rsultat de la fonction <u>vers l'arrire</u> correspondant  la propagation de l'erreur.
   */
  abstract protected double backward(int v);
  
  /**
   * Mthode qui excute la fonction en mode propagation des donnes (vers l'avant).
   * 
   * @return L'excution vers l'avant de la fonction.
   * @throws SNaNValueException Si une valeur NaN a t calcule.
   */
  protected double[] forwardPass() throws SNaNValueException
  {
    throw new SNoImplementationException();
    
  }
  
  /**
   * Mthode qui excute de la fonction en mode propagation des erreurs (vers l'arrire).
   * 
   * @return L'excution vers l'arrire de la fonction.
   * @throws SNaNValueException Si une valeur NaN a t calcule.
   */
  protected double[] backwardPass() throws SNaNValueException
  {
    throw new SNoImplementationException();
    
  }
  
}
