/**
 * 
 */
package sim.nn.function;

import java.util.ArrayList;
import java.util.List;

import sim.nn.SBatchActivable;

/**
 * <p>
 * La classe abstraite <b>SMemoryBatchFunction</b> reprsente une fonction de type <i>batch</i>
 * o l'allocation de la mmoire pour enregistrer les rsultats de l'activation et la propagation des erreurs sera ralis.
 * </p>
 * 
 * @author Simon Vezina
 * @since 2022-01-04
 * @version 2022-01-05
 */
public abstract class SMemoryBatchFunction implements SBatchActivable, SNeuralFunction {

  /**
   * La constante <b>DEFAULT_BATCH_SIZE</b> reprsente l'espace mmoire par dfaut qui sera alou pour la taille maximale de la <i>batch</i> tant gal  {@value}.
   */
  protected final static int DEFAULT_BATCH_SIZE = 4;
   
  /**
   * La variable <b>batch_size</b> reprsente la taille de la batch en calcul dans la fonction.
   * Ce rsultat peut tre infrieur  la taille maximale de la batch.
   */
  protected int batch_size;
  
  /**
   * La variable <b>batch_forward_input</b> reprsente le vecteur d'entre de la fonction dans le sens <u>vers l'avant</u>.  
   */
  protected List<double[]> batch_forward_inputs;
    
  /**
   * La variable <b>batch_forward_output</b> reprsente le vecteur de sortie de la fonction dans le sens <u>vers l'avant</u>.  
   */
  final protected List<double[]> batch_forward_outputs;
  
  /**
   * La variable <b>batch_backward_input</b> reprsente le vecteur d'entre de la fonction dans le sens <u>vers l'arrire</u>.  
   */
  protected List<double[]> batch_backward_inputs;
  
  /**
   * La variable <b>backward_output</b> reprsente le vecteur de sortie de la fonction dans le sens <u>vers l'arrire</u>.  
   */
  final protected List<double[]> batch_backward_outputs;
  
  /**
   * Constructeur d'une fonction de type <i>batch</i> o le nombre de neurones en entre et en sortie sont identiques.
   * 
   * @param neurons Le nombre de neurones.
   */
  public SMemoryBatchFunction(int neurons)
  {
    this(neurons, neurons);
  }
  
  /**
   * Constructeur d'une fonction de type <i>batch</i> o le nombre de neurones en entre et en sortie sont diffrents.
   * 
   * @param inputs Le nombre de neurones en entre.
   * @param outputs Le nombre de neurones en sortie. 
   */
  public SMemoryBatchFunction(int inputs, int outputs) 
  {
    this.batch_size = DEFAULT_BATCH_SIZE;
   
    // Les entres en activation et en propagation d'erreur.
    this.batch_forward_inputs = null;
    this.batch_backward_inputs = null;
    
    // Allocation de la mmoire pour le passage vers l'avant.
    this.batch_forward_outputs = new ArrayList<double[]>();
    
    // Faire l'allocation de la mmoire pour la batch.
    for(int b = 0; b < this.getBatchSize(); b++)
    {
      double[] l = new double[outputs];
      this.batch_forward_outputs.add(l);
    }
    
    // Allocation de la mmoire pour le passage vers l'arrire.
    this.batch_backward_outputs = new ArrayList<double[]>();
    
    // Faire l'allocation de la mmoire pour la batch.
    for(int b = 0; b < this.getBatchSize(); b++)
    {
      double[] l = new double[inputs];
      this.batch_backward_outputs.add(l);
    }
    
  }

  @Override
  public int getInputsSize()
  {
    return this.batch_backward_outputs.get(0).length;
  }
  
  @Override
  public int getOutputsSize()
  {
    return this.batch_forward_outputs.get(0).length;
  }
  
  @Override
  public int getBatchSize()
  {
    return this.batch_size;
  }
  
  @Override
  public int getMaximumBatchSize()
  {
    return this.batch_forward_outputs.size();
  }
  
  @Override
  public void setBatchSize(int b) throws IllegalArgumentException
  {
    if(b < 1)
      throw new IllegalArgumentException("Erreur SMemoryBatchFunction 001 : La taille de la batch ne peut pas tre infreure  1.");
    
    // Dfinir la taille de la batch.
    this.batch_size = b;
    
    // Rallocation de mmoire si la nouvelle taille de la batch est suprieur  l'espace alloue.
    upgradeMaximumBatchSize(this.batch_size);
  }
  
  @Override
  public boolean upgradeMaximumBatchSize(int size)
  {
    // Critre  satisfaire pour augmenter la taille maximale de la batch.
    if(size > this.getMaximumBatchSize())
    {
      // valuer l'ampleur de la mise  jour.
      int upgrade = size - this.batch_forward_outputs.size();
      
      // Augmenter l'espace mmoire uniquement si ncessaire
      if(upgrade > 0)
        for(int i = 0; i < upgrade; i++)
        {
          // Construire le nouveau lment de la batch maximale pour la sortie.
          this.batch_forward_outputs.add(new double[this.getOutputsSize()]);
          
          // Constuire le nouveau lment de la batch maximale pour l'entre.
          this.batch_backward_outputs.add(new double[this.getInputsSize()]);
        }
      
      return true;
    }
    else
      return false;
  }
  
  @Override
  public String toString()
  {
    return getFunctionName();
  }
  
}
