/**
 * 
 */
package sim.nn.function.activation;

import sim.exception.SNoImplementationException;
import sim.nn.function.SAbstractNeuralFunction;
import sim.nn.function.SComputeErrorFunctionGradientActivationParameters;
import sim.nn.parameter.SActivationParameters;

/**
 * La classe <b>SActivationFunction</b> permet de gnrer des fonctions d'activation dans un rseau de neurones en intelligence artificielle.
 * 
 * @author Simon Vzina
 * @since 2018-01-25
 * @version 2022-09-14 (Version labo : Apprentissage par rseau de neurones)
 */
public abstract class SActivationFunction extends SAbstractNeuralFunction implements SComputeErrorFunctionGradientActivationParameters {

  /**
   * <p>
   * La variable <b>K</b> reprsente les paramtres de la fonction d'activation.
   * </p>
   * <p>
   * Il est  not que les fonction d'activation de base (comme sigmode) n'utilise pas de paramtre.
   * Ainsi, ce paramtre sera vide.
   * </p>
   */
  protected SActivationParameters K;
  
  //----------------
  // CONSTRUCTEUR //
  //----------------
  
  /**
   * Constructeur d'une fonction d'activation.
   * 
   * @param neurons Le nombre de neurones dans la fonction d'activation.
   */
  public SActivationFunction(int neurons)
  {
    super(neurons);
    
    // Construction des paramtres de la fonction, s'il y en a.
    // Autrement, un paramtre vide sera construit.
    this.K = SActivationFunctionBuilder.buildParameters(this.getFunctionName(), neurons);
  }

  //------------
  // MTHODES //
  //------------
  
  /**
   * <p>
   * Mthode pour obtenir le nombre de type de paramtre dans la fonction d'activation.
   * </p>
   * <p>
   * Traditionnellement, les fonctions d'activation n'ont pas de paramtre donc aucun type.
   * Cependant, il y a des fonctions avec 1 ou plusieurs type de paramtre qui sont applicable
   * pour l'ensemble des neurones de la fonction.
   * </p>
   * 
   * @return Le nombre de type de paramtre dans la fonction d'activation.
   */
  public int getNbParameterType()
  {
    return 0;
  }
  
  @Override
  public int getNumberOfParameter()
  {
    return this.K.getNumberOfParameter();
  }
    
  /**
   * Mthode qui prcise si la fonction d'activation ncessite dans son calcul l'ensemble des activations de la couche de neurones.
   *  
   * @return <b>true</b> si la fonction d'activation est connecte sur l'ensemble des neurones et <b>false</b> sinon.
   */
  public boolean isFullActivationFunction()
  {
    return false;
  }
      
  /**
   * Mthode qui prcise si la fonction d'activation possde des paramtres.
   * 
   * @return <b>true</b> si la fonction d'activation contient des paramtres et <b>false</b> sinon.
   */
  public boolean isActivationFunctionWithParameters()
  {
    return false;
  }
    
  /**
   * Mthode pour faire l'activation des neurones.
   * 
   * @param Z L'agrgation des neurones.
   * @return L'espace mmoire o est enregistr l'activation de la fonction.
   * @throws IllegalArgumentException Si les paramtres n'ont pas le bon nombre de neurones.
   */
  public double[] activation(double[] Z) throws IllegalArgumentException
  {
    return this.forwardPropagation(Z);
  }
  
  /**
   * Mthode pour faire l'activation des neurones avec des paramtres d'activation.
   * 
   * @param Z L'agrgation des neurones.
   * @param K Les paramtres d'activation.
   * @return L'espace mmoire o est enregistr l'activation de la fonction.
   * @throws IllegalArgumentException Si les paramtres n'ont pas le bon nombre de neurones.
   */
  public double[] activation(double[] Z, SActivationParameters K) throws IllegalArgumentException
  {
    // Validation du paramtre d'activation.
    if(K.getNbParameterType() != this.getNbParameterType())
      throw new IllegalArgumentException("Erreur SActivationFunction 001 : L'activation n'est pas admissible, car le paramtre K contient " + K.getNbParameterType() + " type de paramtre mais la fonction " + this.getFunctionName() + " doit en contenir " + this.getNbParameterType() + ".");
    
    // Validation du nombre de paramtres d'activation par type.
    if(K.getNbParameterType() != 0)
      if(K.getNbNeurons() != this.getInputsSize())
        throw new IllegalArgumentException("Erreur SActivationFunction 002 : L'activation n'est pas admissible, car la taille de K = " + K.getNbNeurons() + " n'est pas gal  " + this.getInputsSize());
    
    // Conserver l'accs au paramtres.
    this.K = K;
    
    // Effectuer l'activation de la fonction.
    return forwardPropagation(Z);
  }
  
  @Override
  protected double backward(int u)
  {
    throw new SNoImplementationException();
    
  }
  
  /**
   * Mthode qui effectue le calcul de la drive de l'activation A par rapport  l'agrgation Z du neurone u.
   * 
   * @param u L'indice du neurone de l'activation A et agrgation Z.
   * @return La drive de l'activation par rapport  l'agrgation.
   */
  public abstract double dAdZ(int u);
   
  /**
   * Mthode qui effectue le calcul de la drive de l'activation A du neurone u par rapport  l'agrgation Z du neurone k.
   * 
   * @param u L'indice du neurone de l'activation A.
   * @param k L'indice du neurone de l'agrgation Z.
   * @return La drive de l'activation par rapport  l'agrgation.
   */
  public double dAdZ(int u, int k)
  {
    if(u != k)
      return 0.0;
    else
      return dAdZ(u);
  }
     
  /**
   * Mthode pour obtenir la drive partielle de la fonction d'activation par rapport au paramtre d'activation K_i_u de type i pour neurone u.
   * 
   * @param i
   * @param u
   * @return
   * @throws IllegalArgumentException
   */
  public double dAdK(int i, int u) throws IllegalArgumentException
  {
    throw new IllegalArgumentException("Erreur SActivationFunction 005 : Cette fonction ne contient pas de paramtre, donc cette drive n'est pas admissible."); 
  }
  
  @Override
  public double dCdK(int i, int u) throws IllegalArgumentException
  {
    throw new SNoImplementationException();
    
  }
  
}
