/**
 * 
 */
package sim.nn.function.activation;

import sim.nn.function.SNeuralFunction;
import sim.nn.parameter.SActivationParameters;
import sim.util.SStringUtil;

/**
 * La classe <b>SActivationFunctionBuilder</b> reprsente un constructeur de fonction d'activation.
 * 
 * @author Simon Vzina
 * @since 2018-03-17
 * @version 2021-12-31
 */
public class SActivationFunctionBuilder {
  
//-------------
  // CONSTANTE //
  //-------------
  
  /**
   * <p>
   * La constante <b>ACTIVATION_FUNCTIONS</b> reprsente la liste des fonctions d'activations pouvant tre reconnues.
   * </p>
   * <p>
   * On y retrouve la liste suivante :
   * <ul>Sigmoid</ul>
   * <ul>ReLU</ul>
   * <ul>Softmax</ul>
   * <ul>Linear</ul>
   * <ul>Tanh</ul>
   * <ul>Gaussian</ul>
   * <ul>Softsign</ul>
   * <ul>Swish</ul>
   * </p>
   */
  public final static String[] ACTIVATION_FUNCTIONS = { 
      SNeuralFunction.SIGMOID_FUNCTION,
      SNeuralFunction.RELU_FUNCTION,
      SNeuralFunction.SOFTMAX_FUNCTION,
      SNeuralFunction.LINEAR_FUNCTION,
      SNeuralFunction.TANH_FUNCTION,
      SNeuralFunction.GAUSSIAN_FUNCTION,
      SNeuralFunction.SOFTSIGN_FUNCTION,
      SNeuralFunction.SWISH_FUNCTION
  };
   
  public final static int SIGMOID_CODE = 0;
  public final static int RELU_CODE = 1;
  public final static int SOFTMAX_CODE = 2;
  public final static int LINEAR_CODE = 3;
  public final static int TANH_CODE = 4;
  public final static int GAUSSIAN_CODE = 5;
  public final static int SOFTSIGN_CODE = 6;
  public final static int SWISH_CODE = 7;
  
  /**
   * Mthode effectuant la construction de la fonction d'activation.
   * 
   * @param function_name Le nom de la fonction d'activation.
   * @param nb_neurons Le nombre de neurones dans la fonction d'activation.
   * @return La fonction d'activation.
   * @throws IllegalArgumentException Si le nom de la fonction d'activation n'est pas reconnu.
   */
  public static SActivationFunction build(String function_name, int nb_neurons) throws IllegalArgumentException
  {
    // Validation du nombre de neurones.
    if(nb_neurons < 1)
      throw new IllegalArgumentException("Erreur SActivationFunctionBuilder 001 : Une fonction d'activation doit contenir un minimum de 1 neurones.");
        
    // Dtermination de la fonction d'activation.
    if(function_name.equalsIgnoreCase(ACTIVATION_FUNCTIONS[SIGMOID_CODE]))
      return new SSigmoidFunction(nb_neurons);
    
    if(function_name.equalsIgnoreCase(ACTIVATION_FUNCTIONS[RELU_CODE]))
      return new SReluFunction(nb_neurons);
    
    if(function_name.equalsIgnoreCase(ACTIVATION_FUNCTIONS[SOFTMAX_CODE]))
      return new SSoftmaxFunction(nb_neurons);
    
    if(function_name.equalsIgnoreCase(ACTIVATION_FUNCTIONS[LINEAR_CODE]))
      return new SLinearFunction(nb_neurons);
    
    if(function_name.equalsIgnoreCase(ACTIVATION_FUNCTIONS[TANH_CODE]))
      return new STanhFunction(nb_neurons);
    
    if(function_name.equalsIgnoreCase(ACTIVATION_FUNCTIONS[GAUSSIAN_CODE]))
      return new SGaussianFunction(nb_neurons);
    
    if(function_name.equalsIgnoreCase(ACTIVATION_FUNCTIONS[SOFTSIGN_CODE]))
      return new SSoftsignFunction(nb_neurons);
    
    if(function_name.equalsIgnoreCase(ACTIVATION_FUNCTIONS[SWISH_CODE]))
      return new SSwishFunction(nb_neurons);
    
    // Erreur, car il n'y a pas eu reconnaissance du nom de la fonction.
    throw new IllegalArgumentException("Erreur SActivationFunctionBuilder 002 : La fonction '" + function_name + "' n'est pas supporte. Voici la liste disponible : " + SStringUtil.join(ACTIVATION_FUNCTIONS, ", "));
  }
  
  /**
   * Mthode pour construire le paramtre d'activation en fonction du type de fonction.
   * 
   * @param function_name Le nom de la fonction d'activation.
   * @param neurons Le nombre de neurones que la fonction d'activation devra activer.
   * @return Le paramtre d'activation.
   */
  public static SActivationParameters buildParameters(String function_name, int neurons) throws IllegalArgumentException
  {
    // Validation du nombre de neurones.
    if(neurons < 1)
      throw new IllegalArgumentException("Erreur SActivationFunctionBuilder 003 : Un paramtre de fonction d'activation doit contenir un minimum de 1 neurones.");
     
    SActivationParameters K;
    
    // Cration des paramtres.
    switch(function_name)
    {
      // Fonction d'activation sans paramtre.
      case SNeuralFunction.SIGMOID_FUNCTION :
      case SNeuralFunction.RELU_FUNCTION :
      case SNeuralFunction.SOFTMAX_FUNCTION :
      case SNeuralFunction.LINEAR_FUNCTION :
      case SNeuralFunction.TANH_FUNCTION :
      case SNeuralFunction.GAUSSIAN_FUNCTION :
      case SNeuralFunction.SOFTSIGN_FUNCTION : 
        K = SActivationParameters.EMPTY_ACTIVATION_PARAMETERS;
        break;
        
      // Fonction d'activation  1 paramtre.
      case SNeuralFunction.SWISH_FUNCTION : 
        K = new SActivationParameters(1, neurons);
        
        // Mettre tous les paramtres  1 pour avoir la fonction f(z) = z * sigmod(z).
        for(int u = 0; u < K.getNbNeurons(); u++)
          K.set(0, u, 1);
        
        break;
        
      default : throw new IllegalArgumentException("Erreur SActivationFunctionBuilder 004 : La fonction '" + function_name + "' n'est pas supporte. Voici la liste disponible : " + SStringUtil.join(ACTIVATION_FUNCTIONS, ", "));
    }
   
    return K;
  }
  
}
