/**
 * 
 */
package sim.nn.function.activation;

import java.util.ArrayList;
import java.util.List;

import sim.exception.SConstructorException;
import sim.exception.SNoImplementationException;
import sim.nn.function.SMemoryBatchFunction;
import sim.nn.function.SComputeErrorFunctionGradientActivationParameters;
import sim.nn.parameter.SActivationParameters;

/**
 * La classe <b>SBatchActivationFunction</b> reprsente une fonction d'activation pouvant tre excut par batch.
 * 
 * @author Simon Vezina
 * @since 2018-10-24
 * @version 2022-09-13 (Version labo : Apprentissage par rseau de neurones)
 */
public class SBatchActivationFunction extends SMemoryBatchFunction implements SComputeErrorFunctionGradientActivationParameters {

  /**
   * La variable <b>activation_function_list</b> reprsente la liste des fonctions d'activation qui seront utilises pour l'ensemble des lments de la batch.
   */
  private final List<SActivationFunction> activation_function_list;
    
  /**
   * <p>
   * La variable <b>K</b> reprsente les paramtres de la fonction d'activation.
   * </p>
   * <p>
   * Il est  not que les fonction d'activation de base (comme sigmode) n'utilise pas de paramtre.
   * Ainsi, ce paramtre sera vide.
   * </p>
   */
  protected SActivationParameters K;
  
  //---------------
  // CONSTRUTEUR //
  //---------------
  
  /**
   * ...
   * 
   * @param neurons
   * @param function_name
   * @throws SConstructorException
   */
  public SBatchActivationFunction(int neurons, String function_name) throws SConstructorException 
  {
    // La fonction d'activation a autant de neurone en entre qu'en sortie.
    super(neurons, neurons);
    
    this.activation_function_list = new ArrayList<SActivationFunction>();
        
    try {
        
      // Chargement de l'espace mmoire pour les batch (l'espace des donnes tant ralis dans SAbstractBatchFunction).
      for(int i = 0; i < this.getMaximumBatchSize(); i++)
        this.activation_function_list.add(SActivationFunctionBuilder.build(function_name, neurons));
    }catch(IllegalArgumentException e) {
      throw new SConstructorException("Erreur SBatchActivationException 001 : La fonction d'activation par batch n'a pas pu tre construite", e);
    }
    
    // Construction des paramtres de la fonction, s'il y en a.
    // Autrement, un paramtre vide sera construit.
    this.K = SActivationFunctionBuilder.buildParameters(this.getFunctionName(), neurons);
  }

  //------------
  // MTHODES //
  //------------
    
  /**
   * Mthode qui prcise si la fonction d'activation possde des paramtres.
   * 
   * @return <b>true</b> si la fonction d'activation contient des paramtres et <b>false</b> sinon.
   */
  public boolean isActivationFunctionWithParameters()
  {
    return this.activation_function_list.get(0).isActivationFunctionWithParameters();
  }
  
  @Override
  public String getFunctionName()
  {
    return this.activation_function_list.get(0).getFunctionName();
  }
   
  /**
   * <p>
   * Mthode pour obtenir le nombre de type de paramtre dans la fonction d'activation.
   * </p>
   * <p>
   * Traditionnellement, les fonctions d'activation n'ont pas de paramtre donc aucun type.
   * Cependant, il y a des fonctions avec 1 ou plusieurs type de paramtre qui sont applicable
   * pour l'ensemble des neurones de la fonction.
   * </p>
   * 
   * @return Le nombre de type de paramtre dans la fonction d'activation.
   */
  public int getNbParameterType()
  {
    return this.activation_function_list.get(0).getNbParameterType();
  }
  
  /**
   * ...
   * 
   * @param b
   * @param batch_inputs
   * @return
   * @throws IllegalArgumentException
   */
  public List<double[]> batchActivation(int b, List<double[]> batch_inputs) throws IllegalArgumentException
  {
    return batchForwardPropagation(b, batch_inputs);
  }
  
  /**
   * ...
   * 
   * @param b
   * @param batch_inputs
   * @param K
   * @return
   * @throws IllegalArgumentException
   */
  public List<double[]> batchActivation(int b, List<double[]> batch_inputs, SActivationParameters K) throws IllegalArgumentException
  {
    // Validation du paramtre d'activation.
    if(K.getNbParameterType() != this.getNbParameterType())
      throw new IllegalArgumentException("Erreur SBatchActivationFunction 001 : L'activation n'est pas admissible, car le paramtre K contient " + K.getNbParameterType() + " type de paramtre mais la fonction " + this.getFunctionName() + " doit en contenir " + this.getNbParameterType() + ".");
    
    // Validation du nombre de paramtres d'activation par type.
    if(K.getNbParameterType() != 0)
      if(K.getNbNeurons() != this.getInputsSize())
        throw new IllegalArgumentException("Erreur SBatchActivationFunction 002 : L'activation n'est pas admissible, car la taille de K = " + K.getNbNeurons() + " n'est pas gal  " + this.getInputsSize());
    
    // Affectation du nouveau paramter.
    this.K = K;
    
    // Effectuer l'activation de la fonction.
    return batchForwardPropagation(b, batch_inputs);
  }
  
  @Override
  public List<double[]> batchForwardPropagation(int b, List<double[]> batch_inputs) throws IllegalArgumentException 
  {
    // Vrification de la taille de la batch b  oprer sur le vecteur d'entre.
    if(b > batch_inputs.size())
      throw new IllegalArgumentException("Erreur SBatchNormalizationFunction 002 : La taile de la batch a excuter ne peut pas tre plus grande et la batch du vecteur d'entre.");
    
    // Dfinir la taille de la batch pour la fonction et allocation de mmoire si ncessaire.
    this.setBatchSize(b);
    
    // Conserver l'accs aux inputs.
    this.batch_forward_inputs = batch_inputs;
        
    // Effectuer le type d'activation en fonction du type de fonction d'activation.
    if(isActivationFunctionWithParameters())
    {
      // Excuter la fonction sur la batch et enregister la rfrence de la solution dans la batch de retour.
      for(int i = 0; i < getBatchSize(); i++)
        this.batch_forward_outputs.set(i, this.activation_function_list.get(i).activation(batch_inputs.get(i), K));
    }
    else
    {
      // Excuter la fonction sur la batch et enregister la rfrence de la solution dans la batch de retour.
      for(int i = 0; i < getBatchSize(); i++)
        this.batch_forward_outputs.set(i, this.activation_function_list.get(i).activation(batch_inputs.get(i)));
    }
    
    return this.batch_forward_outputs;
  }
    
  @Override
  public List<double[]> batchBackPropagation(int b, List<double[]> batch_delta) throws IllegalArgumentException
  {
    // Validation de la taille de la batch  calculer.
    if(b != this.getBatchSize())
      throw new IllegalArgumentException("Erreur SBatchActivationFunction 002 : Le paramtre b = " + b + " n'est pas gal  la taille de la batch en calcul lors de l'activation (" + this.batch_size + ").");
   
    // Excuter la fonction sur la batch et enregister la rfrence de la solution dans la batch de retour.
    for(int i = 0; i < getBatchSize(); i++)
      this.batch_backward_outputs.set(i, this.activation_function_list.get(i).backPropagation(batch_delta.get(i)));
    
    return this.batch_backward_outputs;
  }
  
  @Override
  public boolean upgradeMaximumBatchSize(int size)
  {
    // Mise  jour des taille de la batch pour les donnes.
    if(super.upgradeMaximumBatchSize(size))
    {
      // valuer l'ampleur de la mise  jour.
      int upgrade = size - this.activation_function_list.size();
      
      // Augmenter l'espace mmoire uniquement si ncessaire
      if(upgrade > 0)
        for(int i = 0; i < upgrade; i++)
          this.activation_function_list.add(SActivationFunctionBuilder.build(this.getFunctionName(), this.getOutputsSize()));
      
      return true;
    }
    else
      return false;
  }
  
  @Override
  public String toString()
  {
    return getFunctionName();
  }

  @Override
  public boolean asParameter() 
  {
    return this.activation_function_list.get(0).asParameter();
  }

  @Override
  public double dCdK(int i, int u) throws IllegalArgumentException
  {
    throw new SNoImplementationException();
    
  }
  
}
