/**
 * 
 */
package sim.nn.function.aggregation;

import sim.exception.SNoImplementationException;
import sim.nn.function.SAbstractNeuralFunction;
import sim.nn.function.SComputeErrorFunctionGradientBias;
import sim.nn.function.SComputeErrorFunctionGradientWeights;
import sim.nn.parameter.SBias;
import sim.nn.parameter.SWeights;

/**
 * La classe <b>SAggregationFunction</i> reprsente une fonction d'agrgation de la <u>somme pondre</u>.
 * 
 * @author Simon Vezina
 * @since 2018-08-05
 * @version 2022-09-14 (Version labo : Apprentissage par rseau de neurones)
 */
public abstract class SAggregationFunction extends SAbstractNeuralFunction 
  implements SComputeErrorFunctionGradientWeights, SComputeErrorFunctionGradientBias {

  //-------------
  // VARIABLES //
  //-------------
  
  /**
   * La variable <b>W</b> reprsente les poids de l'agrgation.
   */
  protected SWeights W;
  
  /**
   * La variable <b>B</b> reprsente les biais de l'agrgation.
   */
  protected SBias B;
  
  //----------------
  // CONSTRUCTEUR //
  //----------------
  
  /**
   * Constructeur d'une fonction d'agrgation.
   * 
   * @param input_size Le nombre de neurones en entre.
   * @param output_size Le nombre de neurones en sortie.
   */
  protected SAggregationFunction(int input_size, int output_size)
  {
    super(input_size, output_size);
    
    // Construction des paramtres.
    this.W = new SWeights(input_size, output_size, getOrder());
    this.B = new SBias(output_size);
  }

  //------------
  // MTHODES //
  //------------
  
  /**
   * Mthode pour obtenir l'ordre de la fonction d'agrgation.
   * 
   * @return L'ordre de la fonction d'agrgation.
   */
  public abstract int getOrder();
  
  @Override
  public boolean asParameter()
  {
    return true;
  }
  
  /**
   * Mthode ralisant l'agrgation de l'entre (<i>input</i>) avec les poids (<i>weights</i>) et biais (<i>bias</i>) d'une couche de neurones.
   * 
   * @param inputs L'entre (habituellement, l'activation de la couche prcdente).
   * @param weights Les poids.
   * @param bias Les biais.
   * @return L'agrgation des neurones.
   * @throws IllegalArgumentException Si les paramtres n'ont pas le bon nombre de neurones.
   */
   public double[] aggregation(double[] inputs, SWeights weights, SBias bias) throws IllegalArgumentException
   {
     // Validation de la taille des poids.
     if(weights.getInputSize() != this.getInputsSize() || weights.getOutputSize() != this.getOutputsSize())
       throw new IllegalArgumentException("Erreur SAggregationFunction 001 : L'agrgation n'est pas admissible, car les poids n'ont pas la bonne dimension.");
         
     // Validation de l'ordre des poids.
     if(weights.getOrder() != this.getOrder())
       throw new IllegalArgumentException("Erreur SAggregationFunction 002 : L'agrgation n'est pas admissible, car les poids de la fonction " + this.getFunctionName() + " ne sont d'ordre adquat.");
     
     // Valisation de la taille des biais.
     if(bias.getNeurons() != this.getOutputsSize())
       throw new IllegalArgumentException("Erreur SAggregationFunction 003 : L'agrgation n'est pas admissible, car les biais n'ont pas la bonne dimension.");
     
     // Affectation des paramtres ncessaires au fonctionnement de la fonction d'agrgation.
     this.W = weights;
     this.B = bias;  
      
     // Effectuer l'agrgation sur l'ensemble des neurones de la fonction.
     return this.forwardPropagation(inputs);
   }
  
  /**
   * <p>
   * Mthode pour obtenir la drive de l'agrgation Z_u par rapport au poids W_uv.
   * </p>
   * <p>
   * Il est important de prciser l'interprtation de deux notations :
   * <ul>- Matrice : d(Z_u) / d(W_uv) o u est l'indice de ligne et v est l'indice de colonne.</ul>
   * <ul>- Tenseur : d(Z_u) / d(W_vu) o v est l'indice d'entre et u est l'indice de sortie.</ul>
   * </p> 
   * 
   * @param u L'indice de sortie (tenseur) ou de ligne (matrice). C'est l'indice Z_u.
   * @param v L'indice d'entre (tenseur) ou colonne (matrice). C'est le 2e indice de W_uv.
   * @param 0 L'ordre du poids.
   * @return La drive de Z par rapport  W.
   * @throws IllegalArgumentException Si l'ordre o n'est pas admissible pour la fonction.
   */
  abstract public double dZdW(int u, int v, int o) throws IllegalArgumentException;
     
  /**
   * Mthode pour obtenir la drive de l'agrgation Z_u par rapport  l'entre A_v (l'activation de la couche prcdente).
   * 
   * @param u L'indice de sortie de Z_u.
   * @param v L'indice d'entre de A_v.
   * @return La drive de Z_u par rapport  A_v.
   */
  abstract public double dZdA(int u, int v);
    
  /**
   * Mthode pour obtenir la drive de l'agrgation Z par rapport au biais B.
   * 
   * @param u L'indice de sortie de Z et B.
   * @return La drive de Z par rapport  B.
   */
  abstract  public double dZdB(int u);  
      
  @Override
  public double dCdB(int u)
  {
    throw new SNoImplementationException();
    
  }
  
  
  @Override
  protected double backward(int v)
  {
    throw new SNoImplementationException();
    
  }
  
  @Override
  public boolean equals(Object obj)
  {
    if (this == obj)
      return true;
    if (!super.equals(obj))
      return false;
    if (getClass() != obj.getClass())
      return false;
    
    SAggregationFunction other = (SAggregationFunction) obj;
    if (B == null) {
      if (other.B != null)
        return false;
    } else if (!B.equals(other.B))
      return false;
    if (W == null) {
      if (other.W != null)
        return false;
    } else if (!W.equals(other.W))
      return false;
    
    return true;
  }
  
  
}
