/**
 * 
 */
package sim.nn.function.aggregation;

import sim.math.SBufferedMatrix;
import sim.nn.function.SNeuralFunction;
import sim.util.SStringUtil;

/**
 * La classe <b>SAggregationFunctionBuilder</b> reprsente un constructeur de fonction d'agrgation.
 * 
 * @author Simon Vezina
 * @since 2018-10-01
 * @version 2022-01-08
 */
public class SAggregationFunctionBuilder {

  /**
   * <p>
   * La constante <b>AGGREGATION_FUNCTIONS</b> reprsente la liste des fonctions d'agrgation pouvant tre reconnues.
   * </p>
   * <p>
   * On y retrouve la liste suivante :
   * <ul>weighted_sum</ul>
   * <ul>quadric_sum</ul>
   * </p>
   */
  public final static String[] AGGREGATION_FUNCTIONS = { SNeuralFunction.WEIGHTED_SUM_FUNCTION, SNeuralFunction.QUADRIC_SUM_FUNCTION };
  
  public final static int WEIGHTED_SUM_CODE = 0;
  public final static int QUADRIC_SUM_CODE = 1;
  
  /**
   * Mthode effectuant la construction de la fonction d'agrgation.
   * 
   * @param function_name Le nom de la fonction d'agrgation.
   * @param input Le nombre de neurones dans la fonction d'agrgation en entre.
   * @param output Le nombre de neurones dans la fonction d'agrgation en sortie.
   * @return La fonction d'agrgation.
   * @throws IllegalArgumentException Si le nom de la fonction d'agrgation n'est pas reconnu.
   */
  public static SAggregationFunction build(String function_name, int input, int output) throws IllegalArgumentException
  {
    // Dtermination de la fonction d'activation.
    if(function_name.equalsIgnoreCase(AGGREGATION_FUNCTIONS[WEIGHTED_SUM_CODE]))
      return new SWeightedSumFunction(input, output);
    
    if(function_name.equalsIgnoreCase(AGGREGATION_FUNCTIONS[QUADRIC_SUM_CODE]))
      return new SQuadricSumFunction(input, output);
    
    // Erreur, car il n'y a pas eu reconnaissance du nom de la fonction.
    throw new IllegalArgumentException("Erreur SAggregationFunctionBuilder 001 : La fonction '" + function_name + "' n'est pas supporte. Voici la liste disponible : " + SStringUtil.join(AGGREGATION_FUNCTIONS, ", "));
  }  

  /**
   * Mthode effectuant la construction de la fonction d'agrgation  partir d'une matrice contenant les dimension des poids de la fonction.
   * 
   * @param function_name Le nom de la fonction.
   * @param matrix La matrice des poids (accs au dimension uniquement).
   * @return La fonction d'agrgation.
   * @throws IllegalArgumentException Si le nom de la fonction d'agrgation n'est pas reconnu.
   */
  public static SAggregationFunction build(String function_name, SBufferedMatrix matrix) throws IllegalArgumentException
  {
    // Dtermination de la fonction d'activation.
    if(function_name.equalsIgnoreCase(AGGREGATION_FUNCTIONS[WEIGHTED_SUM_CODE]))
      return new SWeightedSumFunction(matrix.getNbColumn(), matrix.getNbLine());
    
    if(function_name.equalsIgnoreCase(AGGREGATION_FUNCTIONS[QUADRIC_SUM_CODE]))
    {
      // Calcul du nombre d'entre
      int inputs = matrix.getNbColumn() / SQuadricSumFunction.ORDER;
      
      // Validation du nombre de colonne.
      if(inputs*SQuadricSumFunction.ORDER != matrix.getNbColumn())
        throw new IllegalArgumentException("Erreur SAggregationFunctionBuilder 002 : La fonction '" + function_name + "' ne peut pas tre construite, car le nombre de colonne " + matrix.getNbColumn() + " n'est pas compatible avec l'ordre requis de la fonction tant gal  " + SQuadricSumFunction.ORDER + ".");
      
      return new SQuadricSumFunction(inputs, matrix.getNbLine());
    }
    
    // Erreur, car il n'y a pas eu reconnaissance du nom de la fonction.
    throw new IllegalArgumentException("Erreur SAggregationFunctionBuilder 003 : La fonction '" + function_name + "' n'est pas supporte. Voici la liste disponible : " + SStringUtil.join(AGGREGATION_FUNCTIONS, ", "));
  } 
  
}
