/**
 * 
 */
package sim.nn.function.aggregation;

import java.util.ArrayList;
import java.util.List;

import sim.exception.SConstructorException;
import sim.exception.SNoImplementationException;
import sim.nn.function.SMemoryBatchFunction;
import sim.nn.function.SComputeErrorFunctionGradientBias;
import sim.nn.function.SComputeErrorFunctionGradientWeights;
import sim.nn.parameter.SBias;
import sim.nn.parameter.SWeights;

/**
 * La classe <b>SBatchAggregationFunction</b> reprsente une fonction d'aggrgation avec profondeur pouvant tre value en mode batch-activation.
 * 
 * @author Simon Vezina
 * @since 2018-10-24
 * @version 2022-09-13 (Version labo : Apprentissage par rseau de neurones)
 */
public class SBatchAggregationFunction extends SMemoryBatchFunction 
  implements SComputeErrorFunctionGradientWeights, SComputeErrorFunctionGradientBias {

  /**
   * La variable <b>aggregation_function_list</b> reprsente la liste des fonctions d'agrgation.
   */
  private final List<SAggregationFunction> aggregation_function_list;
    
  /**
   * ...
   */
  protected SWeights W;
  
  /**
   * ...
   */
  protected SBias B;
  
  /**
   * ...
   * 
   * @param inputs_size
   * @param outputs_size
   * @param function_name
   * @throws SConstructorException
   */
  public SBatchAggregationFunction(int inputs_size, int outputs_size, String function_name) throws SConstructorException
  {
    super(inputs_size, outputs_size);
        
    // Construction de la liste des fonctions d'agrgation.
    this.aggregation_function_list = new ArrayList<SAggregationFunction>();
       
    try {
      
      for(int i = 0; i < this.getMaximumBatchSize(); i++)
        this.aggregation_function_list.add(SAggregationFunctionBuilder.build(function_name, inputs_size, outputs_size));
    
    }catch(IllegalArgumentException e) {
      throw new SConstructorException("Erreur SBatchAggregationFunction 001 : La fonction d'agrgation de batch n'a pas pu tre construite", e);
    }
    
    // Construction des paramtres (aprs les fonctions pour avoir accs  l'ordre des poids).
    this.W = new SWeights(inputs_size, outputs_size, getOrder());
    this.B = new SBias(outputs_size);  
  }
  
  @Override
  public String getFunctionName()
  {
    return this.aggregation_function_list.get(0).getFunctionName();
  }
     
  @Override
  public boolean asParameter() 
  {
    return this.aggregation_function_list.get(0).asParameter();
  }
    
  /**
   * Mthode pour obtenir l'ordre de la fonction d'agrgation.
   * 
   * @return L'ordre de la fonction d'agrgation.
   */
  public int getOrder()
  {
    return this.aggregation_function_list.get(0).getOrder();
  }
  
  @Override
  public double dCdW(int v, int u, int o) throws IllegalArgumentException
  {
    throw new SNoImplementationException();
    
  }
  
  @Override
  public double dCdB(int u)
  {
    throw new SNoImplementationException();
    
  }
    
  /**
   * ...
   * 
   * @param b
   * @param batch_inputs
   * @param weights
   * @param bias
   * @return
   * @throws IllegalArgumentException
   */
  public List<double[]> batchAggregation(int b, List<double[]> batch_inputs, SWeights weights, SBias bias) throws IllegalArgumentException
  {
    // Validation de la taille des poids.
    if(weights.getInputSize() != this.getInputsSize() || weights.getOutputSize() != this.getOutputsSize())
      throw new IllegalArgumentException("Erreur SBatchAggregationFunction 001 : L'agrgation n'est pas admissible, car les poids n'ont pas la bonne dimension.");
        
    // Validation de l'ordre des poids.
    if(weights.getOrder() != this.getOrder())
      throw new IllegalArgumentException("Erreur SBatchAggregationFunction 002 : L'agrgation n'est pas admissible, car les poids ne sont par du bon ordre.");
   
    // Valisation de la taille des biais.
    if(bias.getNeurons() != this.getOutputsSize())
      throw new IllegalArgumentException("Erreur SBatchAggregationFunction 003 : L'agrgation n'est pas admissible, car les biais n'ont pas la bonne dimension.");
       
    // Affectation des paramtres.
    this.W = weights;
    this.B = bias;
   
    return batchForwardPropagation(b, batch_inputs);   
  }
  
  @Override
  public List<double[]> batchForwardPropagation(int b, List<double[]> batch_inputs) throws IllegalArgumentException 
  {
    // Vrification de la taille de la batch b  oprer sur le vecteur d'entre.
    if(b > batch_inputs.size())
      throw new IllegalArgumentException("Erreur SBatchNormalizationFunction 004 : La taile de la batch a excuter ne peut pas tre plus grande et la batch du vecteur d'entre.");
    
    // Dfinir la taille de la batch pour la fonction et allocation de mmoire si ncessaire.
    this.setBatchSize(b);
    
    this.batch_forward_inputs = batch_inputs;
    
    // Excuter la fonction sur la batch et enregister la rfrence de la solution dans la batch de retour.
    for(int i = 0; i < getBatchSize(); i++)
      this.batch_forward_outputs.set(i, this.aggregation_function_list.get(i).aggregation(batch_inputs.get(i), W, B));
    
    return this.batch_forward_outputs; 
  }
  
  @Override
  public List<double[]> batchBackPropagation(int b, List<double[]> batch_delta) throws IllegalArgumentException
  {
    // Validation de la taille de la batch  calculer.
    if(b != this.getBatchSize())
      throw new IllegalArgumentException("Erreur SBatchAggregationFunction 005 : Le paramtre b n'est pas gal  la taille de la batch en calcul lors de l'activation.");
    
    // Excuter la fonction sur la batch et enregister la rfrence de la solution dans la batch de retour.
    for(int i = 0; i < getBatchSize(); i++)
      this.batch_backward_outputs.set(i, this.aggregation_function_list.get(i).backPropagation(batch_delta.get(i)));
    
    return this.batch_backward_outputs;
  }
    
  @Override
  public boolean upgradeMaximumBatchSize(int size)
  {
    if(super.upgradeMaximumBatchSize(size))
    {
      // valuer l'ampleur de la mise  jour.
      int upgrade = size - this.aggregation_function_list.size();
      
      // Augmenter l'espace mmoire uniquement si ncessaire
      if(upgrade > 0)
        for(int i = 0; i < upgrade; i++)
          this.aggregation_function_list.add(SAggregationFunctionBuilder.build(this.getFunctionName(), this.getInputsSize(), this.getOutputsSize()));
    
      return true;
    }
    else
      return false;
  }
  
  @Override
  public String toString()
  {
    return getFunctionName();
  }
  
}
