/**
 * 
 */
package sim.nn.function.error;

import java.util.ArrayList;
import java.util.List;

import sim.exception.SConstructorException;
import sim.nn.function.SMemoryBatchFunction;

/**
 * La classe <b>SBatchErrorFunction</b> reprsente une fonction d'erreur pour l'activation en mode batch-activation.
 * 
 * @author Simon Vezina
 * @since 2018-10-25
 * @version 2021-09-11
 */
public class SBatchErrorFunction extends SMemoryBatchFunction {

  /**
   * La variable <b>error_function_list</b> reprsente la liste des fonctions d'erreur.
   */
  protected final List<SErrorFunction> error_function_list;
   
  /**
   * ...
   */
  protected List<double[]> batch_expected;
  
  //----------------
  // CONSTRUCTEUR //
  //----------------
  
  /**
   * ...
   * 
   * @param function_name
   * @param neurons
   * @param depth
   * @throws SConstructorException
   */
  public SBatchErrorFunction(int neurons, String function_name) throws SConstructorException
  {
    super(neurons, neurons);
    
    this.error_function_list = new ArrayList<SErrorFunction>();
    
    for(int i = 0; i < this.getMaximumBatchSize(); i++)
      this.error_function_list.add(SErrorFunctionBuilder.build(function_name, neurons));
  }

  //------------
  // MTHODES //
  //------------
  
  @Override
  public String getFunctionName()
  {
    return this.error_function_list.get(0).getFunctionName();
  }
    
  /**
   * Mthode pour raliser le calcul de l'erreur sur une seule donne.
   * 
   * @param calculated
   * @param expected
   * @return
   * @throws IllegalArgumentException
   */
  public double[] errors(double[] calculated, double[] expected) throws IllegalArgumentException
  {
    List<double[]> batch_calculated = new ArrayList<double[]>();
    batch_calculated.add(calculated);
    
    List<double[]> batch_expected = new ArrayList<double[]>();
    batch_expected.add(expected);
    
    return batchErrors(batch_calculated, batch_expected).get(0);
  }
  
  /**
   * <p>
   * Mthode pour raliser le calcul d'erreur sur une <i>batch</i> de donne attendue.
   * </p>
   * <p>
   * Il est important de prciser que la taille de la batch sera dfinie par la taille de la liste des donnes attendues
   * et non de l'activation qui elle dpend de l'espace requis dans l'activation des rseaux (ces tailles ne sont pas ncessairement gale).
   * </p>
   * 
   * @param batch_calculated
   * @param batch_expected
   * @return
   * @throws IllegalArgumentException
   */
  public List<double[]> batchErrors(List<double[]> batch_calculated, List<double[]> batch_expected) throws IllegalArgumentException
  {
    // Dfinir la taille de la batch avec la liste des vecteurs attendus.
    // Ce choix est ncessaire, car la taille de la liste de batch_calculated 
    // contient le nombre maximal de vecteur pouvant tre dans une batch et non le nombre actuellement solicit.
    int b = batch_expected.size();
    
    this.batch_expected = batch_expected;
       
    return batchForwardPropagation(b, batch_calculated);
  }
    
  @Override
  public List<double[]> batchForwardPropagation(int b, List<double[]> batch_inputs) throws IllegalArgumentException 
  {
    // Vrification de la taille de la batch b  oprer sur le vecteur d'entre.
    if(b > batch_inputs.size())
      throw new IllegalArgumentException("Erreur SBatchErrorFunction 001 : La taile de la batch a excuter ne peut pas tre plus grande et la batch du vecteur d'entre.");
    
    // Dfinir la taille de la batch pour la fonction et allocation de mmoire si ncessaire.
    this.setBatchSize(b);
    
    this.batch_forward_inputs = batch_inputs;
   
    // Excuter la fonction sur la batch et enregister la rfrence de la solution dans la batch de retour.
    for(int i = 0; i < getBatchSize(); i++)
      this.batch_forward_outputs.set(i, this.error_function_list.get(i).errors(batch_inputs.get(i), this.batch_expected.get(i)));
    
    return this.batch_forward_outputs;
  }
  
  /**
   * Mthode pour obtenir la propagation de l'erreur de la fonction d'erreur pour une <i>batch</i>.
   * 
   * @return La liste des propagations d'erreur des fonctions d'erreur.
   */
  public List<double[]> batchBackPropagation()
  {
    // Excuter la fonction sur la batch et enregister la rfrence de la solution dans la batch de retour.
    for(int i = 0; i < this.getBatchSize(); i++)
      this.batch_backward_outputs.set(i, this.error_function_list.get(i).backPropagation());
    
    return this.batch_backward_outputs;
  }
  
  @Override
  public List<double[]> batchBackPropagation(int b, List<double[]> batch_delta) throws IllegalArgumentException
  {
    // Validation de la taille de la batch  calculer.
    if(b != this.getBatchSize())
      throw new IllegalArgumentException("Erreur SBatchErrorFunction 002 : Le paramtre b = " + b + " n'est pas gal  la taille de la batch en calcul lors de l'activation (" + this.batch_size + ").");
   
    // Excuter la fonction sur la batch et enregister la rfrence de la solution dans la batch de retour.
    for(int i = 0; i < this.getBatchSize(); i++)
      this.batch_backward_outputs.set(i, this.error_function_list.get(i).backPropagation(batch_delta.get(i)));
    
    return this.batch_backward_outputs;
  }
  
  @Override
  public boolean upgradeMaximumBatchSize(int size)
  {
    // Bien que cet espace mmoire ne sera pas exploite, nous allons l'augmenter quand mme.
    if(super.upgradeMaximumBatchSize(size))
    {
      // valuer l'ampleur de la mise  jour.
      int upgrade = size - this.error_function_list.size();
      
      // Augmenter l'espace mmoire uniquement si ncessaire
      if(upgrade > 0)
        for(int i = 0; i < upgrade; i++)
          this.error_function_list.add(SErrorFunctionBuilder.build(this.getFunctionName(), this.getOutputsSize()));
      
      return true;
    }
    else
      return false;
  }
  
  @Override
  public String toString()
  {
    return this.getFunctionName();
  }

  @Override
  public boolean asParameter() 
  {
    return this.error_function_list.get(0).asParameter();
  }
  
}
