/**
 * 
 */
package sim.nn.function.error;

import sim.exception.SNoImplementationException;
import sim.nn.function.SAbstractNeuralFunction;

/**
 * La classe <b>SErrorFunction</b> permet de gnrer la valeur de l'erreur entre une valeur calcule par un rseau de neurone 
 * et une valeur attendue dtermine par l'chantillonnage d'entranement.
 * 
 * Dans la littrature, les fonctions d'erreur portent galement de fonction de cot (<i>cost</i>).
 * 
 * @author Simon Vzina
 * @since 2018-01-25
 * @version 2022-09-14 (Version labo : Apprentissage par rseau de neurones)
 */
public abstract class SErrorFunction extends SAbstractNeuralFunction {

  /**
   * La constante <b>DEFAULT_BACKWARD_INPUT</b> correspond  l'erreur  propager par la fonction d'erreur par dfaut.
   * Traditionnellement, cette valeur est gale  1 ou elle peut prendre la valeur du taux d'apprentissage.
   * Dans cette implmentation, nous l'avons initialis  {@value}.
   */
  private final static double DEFAULT_BACKWARD_INPUT = 1.0;
  
  /**
   * La variable <b>y</b> reprsente l'espace mmoire o sera enregistre l'activation attendue du rseau de neurones pour un vecteur d'entre donne.
   */
  protected double[] y;
  
  /**
   * La variable <b>default_backward_inputs</b> reprsente l'erreur  propager par dfaut de la fonction d'erreur.
   */
  protected final double[] default_backward_inputs;
  
  //----------------
  // CONSTRUCTEUR //
  //----------------
  
  /**
   * Constructeur d'une fonction d'erreur abstraite.
   * 
   * @param neurons Le nombre de neurones.
   */
  public SErrorFunction(int neurons)
  {
    super(neurons, neurons);
    
    this.y = null;
    
    // Initialiser  1 l'ensemble des erreurs initiales  propager dans la fonction d'erreur par dfaut.
    // Ainsi, c'est l'erreur "1" qui sera propag  moins que la mthode backPropagation(double[] delta)
    // change la valeur d'erreur initiale  propager.
    this.default_backward_inputs = new double[this.getOutputsSize()];
    
    for(int i = 0; i < this.default_backward_inputs.length; i++)
      this.default_backward_inputs[i] = DEFAULT_BACKWARD_INPUT;
    
    // Nous allons affecter ce type d'entre pour la propagation de l'erreur galement par dfaut.
    this.backward_inputs = this.default_backward_inputs;
  }
  
  //------------
  // MTHODES //
  //------------
  
  /**
   * Mthode pour obtenir l'erreur de la fonction d'erreur pour l'ensemble des neurones de la fonction.
   * 
   * @param calculated La valeur calcule (l'activation du rseau).
   * @param expected La valeur attendue (celle de la donne d'entranement).
   * @return Les erreurs de la fonction d'erreur. 
   * @throws IllegalArgumentException Si les paramtres n'ont pas le bon nombre de neurones.
   */
  public double[] errors(double[] calculated, double[] expected) throws IllegalArgumentException
  {
    // Validation de la taille du vecteur attendu.
    if(expected.length != this.getInputsSize())
      throw new IllegalArgumentException("Erreur SErrorFunction 002 : Erreur dans la taille de la valeur attendue.");
     
    // Affectation de la donne attendue.
    this.y = expected;
    
    // Effectuer la propagation des donnes (l'activation de la fonction).
    return this.forwardPropagation(calculated);
  }
  
  /**
   * Mthode pour raliser la propagation de l'erreur par dfaut par la fonction d'erreur.
   * L'erreur par dfaut est "1" pour l'ensemble des neurones.
   * Ainsi, la propagation de l'erreur de la fonction d'erreur ne dpendera uniquement que de la drive de la fonction d'erreur.
   * 
   * @return La propagation de l'erreur.
   */
  public double[] backPropagation()
  {
    // Faire la propagation des erreurs avec l'erreur initiale par dfaut.
    return this.backPropagation(this.default_backward_inputs);
  }
  
  @Override
  protected double backward(int u)
  {
    throw new SNoImplementationException();
    
  }
  
  /**
   * Mthode permettant d'valuer la drive de la fonction d'erreur pour une activation donne en lien avec une activation attendue pour un neurone d'indice u.
   * 
   * Il est important d'initialiser la fonction d'erreur avant d'utiliser cette fonctionnalit.
   * 
   * @param L'indice du neurone.
   * @return La drive de la fonction d'erreur pour le neurone u.
   */
  public abstract double dCdA(int u);
  
}
