/**
 * 
 */
package sim.nn.function.filtering;

import java.util.ArrayList;
import java.util.List;

import sim.exception.SConstructorException;
import sim.nn.function.SMemoryBatchFunction;
import sim.nn.function.SNeuralMath;

/**
 * La classe <b>SBatchPoolingFunction</b> reprsente une fonction de pooling pouvant tre excut par <i>batch</i>.
 * 
 * @author Simon Vezina
 * @since 2021-11-21
 * @version 2022-01-08
 */
public class SBatchPoolingFunction extends SMemoryBatchFunction {

  /**
   * La variable <b>convolution_function_list</b> reprsente la liste des fonctions de convolution.
   */
  private final List<SPoolingFunction> pooling_function_list;
  
  /**
   * ...
   * 
   * @param inputs
   * @param inputs_depth
   * @param filter_offset
   * @param stride
   * @param pooling_function_name
   * @throws SConstructorException
   */
  public SBatchPoolingFunction(int inputs, int inputs_depth, int filter_offset, int stride, String pooling_function_name) throws SConstructorException 
  {
    super(inputs, SNeuralMath.computeOutputsSizeForSquareInputsDepth(inputs, inputs_depth, inputs_depth, filter_offset, 0, stride));
    
    this.pooling_function_list = new ArrayList<SPoolingFunction>();
    
    try {
      
      for(int i = 0; i < this.getMaximumBatchSize(); i++)
        this.pooling_function_list.add(SPoolingFunctionBuilder.build(pooling_function_name, inputs, inputs_depth, filter_offset));
    
    }catch(IllegalArgumentException e) {
      throw new SConstructorException("Erreur SBatchPoolingFunction 001 : La fonction de pooling de batch n'a pas pu tre construite", e);
    }
  }

  @Override
  public int getInputsDepth()
  {
    return this.pooling_function_list.get(0).getInputsDepth();
  }
  
  @Override
  public int getOutputsDepth()
  {
    return this.pooling_function_list.get(0).getOutputsDepth();
  }
  
  @Override
  public String getFunctionName()
  {
    return this.pooling_function_list.get(0).getFunctionName();
  }
    
  @Override 
  public String getDimension()
  {
    return this.pooling_function_list.get(0).getDimension();
  }
  
  /**
   * Mthode pour obtenir la dimension de largeur de l'entre de forme carr.
   * 
   * @return
   */
  public int getInputOffset()
  {
    return this.pooling_function_list.get(0).getInputOffset();
  }
  
  /**
   * Mthode pour obtenir la dimension de largeur de la sortie de forme carr.
   * 
   * @return
   */
  public int getOutputOffset()
  {
    return this.pooling_function_list.get(0).getOutputOffset();
  }
  
  @Override
  public int getNumberOfParameter()
  {
    return this.pooling_function_list.get(0).getNumberOfParameter();
  }
  
  /**
   * Mthode pour obtenir la taille du filtre de filtrage.
   * 
   * @return La taille du filtre.
   */
  public int getFilterOffset()
  {
    return this.pooling_function_list.get(0).getFilterOffset();
  }
  
  /**
   * ...
   * 
   * @return
   */
  public int getStride()
  {
    return this.pooling_function_list.get(0).getStride();
  }
  
  /**
   * ...
   * 
   * @param b
   * @param batch_inputs
   * @return
   * @throws IllegalArgumentException
   */
  public List<double[]> batchPooling(int b, List<double[]> batch_inputs) throws IllegalArgumentException
  {
    return batchForwardPropagation(b, batch_inputs);
  }
  
  @Override
  public List<double[]> batchForwardPropagation(int b, List<double[]> batch_inputs) throws IllegalArgumentException 
  {
    // Vrification de la taille de la batch b  oprer sur le vecteur d'entre.
    if(b > batch_inputs.size())
      throw new IllegalArgumentException("Erreur SBatchPoolingFunction 002 : La taile de la batch a excuter ne peut pas tre plus grande et la batch du vecteur d'entre.");
    
    // Dfinir la taille de la batch pour la fonction et allocation de mmoire si ncessaire.
    this.setBatchSize(b);
    
    this.batch_forward_inputs = batch_inputs;
   
    // Excuter la fonction sur la batch et enregister la rfrence de la solution dans la batch de retour.
    for(int i = 0; i < getBatchSize(); i++)
      this.batch_forward_outputs.set(i, this.pooling_function_list.get(i).pooling(batch_inputs.get(i)));
    
    return this.batch_forward_outputs; 
  }
  
  /**
   * ...
   * 
   * @param b
   * @param batch_delta
   * @return
   * @throws IllegalArgumentException
   */
  public List<double[]> batchBackPropagation(int b, List<double[]> batch_delta) throws IllegalArgumentException
  {
    // Validation de la taille de la batch  calculer.
    if(b != this.getBatchSize())
      throw new IllegalArgumentException("Erreur SBatchPooliongFunction 003 : Le paramtre b n'est pas gal  la taille de la batch en calcul lors de l'activation.");
    
    // Excuter la fonction sur la batch et enregister la rfrence de la solution dans la batch de retour.
    for(int i = 0; i < getBatchSize(); i++)
      this.batch_backward_outputs.set(i, this.pooling_function_list.get(i).backPropagation(batch_delta.get(i)));
    
    return this.batch_backward_outputs;
  }
  
  @Override
  public boolean upgradeMaximumBatchSize(int size)
  {
    if(super.upgradeMaximumBatchSize(size))
    {
      // valuer l'ampleur de la mise  jour.
      int upgrade = size - this.pooling_function_list.size();
      
      // Augmenter l'espace mmoire uniquement si ncessaire
      if(upgrade > 0)
        for(int i = 0; i < upgrade; i++)
          this.pooling_function_list.add(SPoolingFunctionBuilder.build(this.getFunctionName(), this.getInputsSize(), this.getInputsDepth(), this.getFilterOffset()));
      
      return true;
    }
    else
      return false;
  }
  
}
