/**
 * 
 */
package sim.nn.function.filtering;

import sim.exception.SConstructorException;
import sim.exception.SNoImplementationException;
import sim.math.SArrayTensor;
import sim.nn.function.SAbstractFilterFunction;
import sim.nn.function.SComputeErrorFunctionGradientBias;
import sim.nn.function.SComputeErrorFunctionGradientFiltersWeights;
import sim.nn.function.SNeuralFunction;
import sim.nn.parameter.SBias;
import sim.nn.parameter.SFiltersWeights;
import sim.util.SArrays;

/**
 * La classe <b>SConvolutionFunction</b> reprsente la fonction de convolution applicable sur une entre de neurones dcomposable en deux dimensions avec profondeur.
 * 
 * @author Simon Vezina
 * @since 2019-02-03
 * @version 2022-09-13 (Version labo : Apprentissage par rseau de neurones)
 */
public class SConvolutionFunction extends SAbstractFilterFunction implements SComputeErrorFunctionGradientFiltersWeights, SComputeErrorFunctionGradientBias {
  
  /**
   * ...
   */
  protected SFiltersWeights W;
  
  /**
   * ...
   */
  protected SBias B;
  
  /**
   * <p>
   * La variable <b>grad_W</b> contient les rsultats des gradients de la fonction d'erreur par rapport aux poids de convolution de la couche de neurones.
   * </p>
   * <p>
   * Puisque l'algorithme retenu pour faire ce calcul require une accumulation de termes raliss tout au long de la propagation de l'erreur sur l'ensemble des neurones de la couche,
   * il sera impossible de calculer un terme de gradient  la fois sans sacrifier de la performance. Ainsi, obtenir un terme de gradient ncessitera le calcul de tous les termes.
   * </p>
   */
  protected final SArrayTensor grad_W;
  
  /**
   * <p>
   * La variable <b>grad_B</b> contient les rsultats des gradients de la fonction d'erreur par rapport aux biais de convolution de la couche de neurones.
   * <p>
   * Puisque l'algorithme retenu pour faire ce calcul require une accumulation de termes raliss tout au long de la propagation de l'erreur sur l'ensemble des neurones de la couche,
   * il sera impossible de calculer un terme de gradient  la fois sans sacrifier de la performance. Ainsi, obtenir un terme de gradient ncessitera le calcul de tous les termes.
   * <p>
   */
  protected final double[] grad_B;
  
  /**
   * ...
   * 
   * @param inputs Le nombre de neurones d'entre dans la fonction.
   * @param inputs_depth La profondeur d'entre de la fonction.
   * @param outputs_depth La profondeur de sortie de la fonction correspondant aux nombres de filtres.
   * @param filter_offset La taille des filtres.
   * @param padding Le padding autour de la matrice d'entre.
   * @param stride Le dcalage du filtre lors de la convolution.
   * @throws SConstructorException Si les paramtres pour la fonction de la fonction est inadquate.
   */
  public SConvolutionFunction(int inputs, int inputs_depth, int outputs_depth, int filter_offset, int padding, int stride) throws SConstructorException
  {
    super(inputs, inputs_depth, outputs_depth, filter_offset, padding, stride);
       
    if(this.padding != 0)
      throw new SConstructorException("Erreur SConvolutionFunction 001 : Pour l'instant, c'est uniquement un PADDING de 0 qui est support.");
  
    // Dfinir l'espace mmoire pour la sauvegarde des gradients des poids.
    this.grad_W = new SArrayTensor(outputs_depth, inputs_depth, filter_offset, filter_offset);
    
    // Dfinir l'espace mmoire pour la sauvegarde des gradients de biais.
    this.grad_B = new double[outputs_depth];
    
    // Construction des paramtres.
    this.W = new SFiltersWeights(filter_offset, inputs_depth, outputs_depth);
    this.B = new SBias(outputs_depth);
  }
         
  /**
   * ...
   * 
   * @param inputs
   * @param W
   * @param B
   * @return
   * @throws IllegalArgumentException
   */
  public double[] convolution(double[] inputs, SFiltersWeights W, SBias B) throws IllegalArgumentException
  { 
    // Validation du nombre de biais.
    if(B.getNeurons() != this.getOutputsDepth())
      throw new IllegalArgumentException("Erreur SConvolutionFunction 003 : Le nombre de biais " + B.getNeurons() + " n'est pas gal  " + this.getOutputsDepth() + " tant le niveau de profondeur de sortie de la fonction.");
        
    // Validation de la taille des filtres.
    if(W.getOffset() != this.filterOffset)
      throw new IllegalArgumentException("Erreur SConvolutionFunction 004 : La dimension du filtre " + W.getOffset() + " n'est pas gal  " + this.filterOffset + " tat la valeur prdfinie.");
        
    // Affectation des paramtres ncessaires au fonctionnement de la fonction. 
    this.W = W;
    this.B = B;
    
    // Effectuer l'activation de la fonction.
    return this.forwardPropagation(inputs);
  }
  
  @Override
  public int getNumberOfParameter()
  {
    return this.W.getNumberOfParameter() + this.B.getNumberOfParameter();
  }
  
  /**
   * ...
   * 
   * Dans la notation dZ(f)_mn/dW(f)_ijd
   * 
   * @param f La profondeur de sortie (le numro du filtre).
   * @param d L'indice de profondeur d'entre du tenseur W (le numro de profondeur de l'entre).
   * @param m L'indice de ligne de Z.
   * @param n L'indice de colonne de Z.
   * @param i L'indice de la ligne du filtre W.
   * @param j L'indice de la colonne du filtre W.
   * @return
   */
  public double dZdW(int f, int d, int m, int n, int i, int j)
  {
    throw new SNoImplementationException();
    
  }
  
  /**
   * 
   * @param f 
   * @param u
   * @param v
   * @param d
   * @return
   */
  public double dZdA(int f, int d, int u, int v)
  {
    throw new SNoImplementationException();
  }
  
  /**
   * ...
   * 
   * @param f La profondeur de sortie (le numro du filtre).
   * @return La drive de la fonction de convolution par rapport  B.
   */
  public double dZdB(int f) 
  {
    throw new SNoImplementationException();
    
  }
  
  @Override
  public double dCdW(int f, int d, int i, int j)
  {
    throw new SNoImplementationException();
    
  }
  
  @Override
  public double dCdB(int f)
  {
    throw new SNoImplementationException();
    
  }
    
  @Override
  protected double filtering(int f, int m, int n)
  {
    throw new SNoImplementationException();
    
  }
  
  @Override
  protected double[] backwardPass()
  {
    throw new SNoImplementationException();
    
  }
      
  @Override
  public String getFunctionName()
  {
    return SNeuralFunction.CONVOLUTION_FUNCTION;
  }

  
}
