/**
 * 
 */
package sim.nn.function.regularization;

import java.util.ArrayList;
import java.util.List;

import sim.nn.function.SMemoryBatchFunction;

/**
 * La classe <b>SBatchDropoutFunction</b> reprsente la fonction dropout pouvant tre active par <i>batch</i>.
 * 
 * @author Simon Vezina
 * @since 2021-11-21
 * @version 2022-01-08
 */
public class SBatchDropoutFunction extends SMemoryBatchFunction {
                                           
  /**
   * La variable <b>convolution_function_list</b> reprsente la liste des fonctions de convolution.
   */
  private final List<SDropoutFunction> dropout_function_list;
  
  /**
   * La variable <b>dropping_probability</b> reprsente la probabilit de faire le drop.
   */
  protected double dropping_probability;
  
  /**
   * Constructeur d'une fonction dropout de type <i>batch</i>.
   * 
   * @param neurons Le nombre de neurones.
   */
  public SBatchDropoutFunction(int neurons) 
  {
    super(neurons, neurons);
    
    this.dropout_function_list = new ArrayList<SDropoutFunction>();
    
    for(int i = 0; i < getMaximumBatchSize(); i++)
      this.dropout_function_list.add(new SDropoutFunction(neurons));
    
    this.dropping_probability = SDropoutFunction.DEFAULT_DROPPING_PROBABILITY;
  }

  @Override
  public int getInputsSize()
  {
    return this.dropout_function_list.get(0).getInputsSize();
  }
  
  @Override
  public int getOutputsSize()
  {
    return this.dropout_function_list.get(0).getOutputsSize();
  }
     
  @Override
  public String getFunctionName()
  {
    return this.dropout_function_list.get(0).getFunctionName();
  }
    
  /**
   * 
   * @param b
   * @param batch_inputs
   * @param p
   * @return
   * @throws IllegalArgumentException
   */
  public List<double[]> batchDrop(int b, List<double[]> batch_inputs, double p) throws IllegalArgumentException
  {
    // Validation de la probabilit.
    if(p < 0.0 || p > 1.0)
      throw new IllegalArgumentException("Erreur SBatchDropoutFunction 001 : La probabilit p = " + p + " n'est pas contenu entre 0 et 1.");
    
    this.dropping_probability = p;
    
    return batchForwardPropagation(b, batch_inputs);
  }
  
  @Override
  public List<double[]> batchForwardPropagation(int b, List<double[]> batch_inputs) throws IllegalArgumentException 
  {
    // Vrification de la taille de la batch b  oprer sur le vecteur d'entre.
    if(b > batch_inputs.size())
      throw new IllegalArgumentException("Erreur SBatchDropoutFunction 002 : La taile de la batch a excuter ne peut pas tre plus grande et la batch du vecteur d'entre.");
    
    // Dfinir la taille de la batch pour la fonction et allocation de mmoire si ncessaire.
    this.setBatchSize(b);
    
    this.batch_forward_inputs = batch_inputs;
    
    // Excuter la fonction sur la batch et enregister la rfrence de la solution dans la batch de retour.
    for(int i = 0; i < this.getBatchSize(); i++)
      this.batch_forward_outputs.set(i, this.dropout_function_list.get(i).drop(batch_inputs.get(i), this.dropping_probability));
    
    return this.batch_forward_outputs;
  }
  
  /**
   * ...
   * 
   * @param b
   * @param batch_delta
   * @return
   * @throws IllegalArgumentException
   */
  public List<double[]> batchBackPropagation(int b, List<double[]> batch_delta) throws IllegalArgumentException
  {
    // Validation de la taille de la batch  calculer.
    if(b != this.getBatchSize())
      throw new IllegalArgumentException("Erreur SBatchDropoutFunction 003 : Le paramtre b = " + b + " n'est pas gal  la taille de la batch en calcul lors de l'activation (" + this.batch_size + ").");
   
    // Excuter la fonction sur la batch et enregister la rfrence de la solution dans la batch de retour.
    for(int i = 0; i < this.getBatchSize(); i++)
      this.batch_backward_outputs.set(i, this.dropout_function_list.get(i).backPropagation(batch_delta.get(i)));
    
    return this.batch_backward_outputs;
  }
  
  @Override
  public boolean upgradeMaximumBatchSize(int size)
  {
    // Mise  jour des taille de la batch pour les donnes.
    if(super.upgradeMaximumBatchSize(size))
    { 
      // valuer l'ampleur de la mise  jour.
      int upgrade = size - this.dropout_function_list.size();
      
      // Augmenter l'espace mmoire uniquement si ncessaire
      if(upgrade > 0)
        for(int i = 0; i < upgrade; i++)
          this.dropout_function_list.add(new SDropoutFunction(this.getOutputsSize()));
    
      return true;
    }
    else
      return false;
  }
  
  @Override
  public String toString()
  {
    return this.getFunctionName();
  } 

}
