/**
 * 
 */
package sim.nn.function.regularization;

import java.util.ArrayList;
import java.util.List;

import sim.exception.SNoImplementationException;
import sim.nn.function.SAbstractBatchFunction;
import sim.nn.function.SComputeErrorFunctionGradientScale;
import sim.nn.function.SComputeErrorFunctionGradientShift;
import sim.nn.function.SNeuralFunction;
import sim.nn.parameter.SScale;
import sim.nn.parameter.SShift;

/**
 * La classe <b>SBatchNormalizationFunction</b> reprsente la fonction de rgularisation portant le nom de <i>batch-normalization</i>.
 * 
 * @author Simon Vezina
 * @since 2018-09-13
 * @version 2022-09-13 (Version labo : Apprentissage par rseau de neurones)
 */
public class SBatchNormalizationFunction extends SAbstractBatchFunction implements SComputeErrorFunctionGradientScale, SComputeErrorFunctionGradientShift {

  //--------------
  // CONSTANTES //
  //--------------
  
  /**
   * La constante <b>EPSILON</b> reprsente un petit chiffre ncessaire si la variance est nulle afin d'viter une division par zro. 
   */
  private static final double EPSILON = 1e-6;
  
  //-------------
  // VARIABLES //
  //-------------
  
  /**
   * La variable <b>SC</b> reprsente le paramtre de scale. 
   */
  protected SScale SC;
  
  /**
   * La variable <b>SH</b> reprsente le paramtre de shift.
   */
  protected SShift SH;
  
  /**
   * La variable <b>running</b> reprsente la moyenne et la variance en volution pour toutes les couches de neurones en profondeur.
   */
  protected SRunningMeanVariance running;
  
  /**
   * La variable <b>mean</b> reprsente la moyenne de toutes les activations ralises par tous les neurones.
   */
  protected final double[] mean;
  
  /**
   * La variable <b>variance</b> reprsente l'cart typede toutes les activations ralises par tous les neurones. 
   */
  protected final double[] variance;
     
  /**
   * La variable <b>normalized_Z</b> reprsente l'agrgation normalise de la fonction <i>batch-normalization</i>.
   */
  protected final List<double[]> normalized_Z;
  
  //----------------
  // CONSTRUCTEUR //
  //----------------
  
  /**
   * Constructeur de la fonction de normalisation <i>batch-normalization</i>.
   * 
   * @param neurons Le nombre de neurones.
   */
  public SBatchNormalizationFunction(int neurons) 
  {
    super(neurons);
    
    this.mean = new double[neurons];
    this.variance = new double[neurons];
    
    // Construire le vecteur des agrgations normaliss avec profondeur.
    this.normalized_Z = new ArrayList<double[]>();
    
    // Faire l'allocation de la mmoire pour la batch.
    for(int b = 0; b < this.getMaximumBatchSize(); b++)
      this.normalized_Z.add(new double[neurons]);   
    
    // Construction des paramtres.
    this.SC = new SScale(neurons);
    this.SH = new SShift(neurons);
  }
  
  //------------
  // MTHODES //
  //------------
  
  @Override
  public int getNumberOfParameter()
  {
    return this.SC.getNumberOfParameter() + this.SH.getNumberOfParameter();
  }
  
  /**
   * Mthode pour obtenir la moyenne des agrgations d'un neurone.
   * 
   * @param u L'indice du neurone.
   * @return La moyenne des agrgations du neurone.
   */
  public double getMean(int u)
  {
    // Si la batch est suprieure  1, retourner la moyenne de la batch.
    if(this.getBatchSize() > 1)
      return this.mean[u];
    else
      return this.running.getMean(u);    // Si la batch est 1, retourner la moyenne volutive. 
  }
  
  /**
   * Mthode pour obtenir la variance des agrgations d'un neurone.
   * 
   * @param u L'indice du neurone.
   * @return La variance des agrgations du neurone.
   */
  public double getVariance(int u)
  {
    // Si la batch est suprieure  1, retourner la variance de la batch.
    if(this.getBatchSize() > 1)
      return this.variance[u];
    else
      return this.running.getVariance(u);  // Si la batch est 1, retourner la variance volutive.
  }
  
  /**
   * Mthode pour obtenir la racine carre de la variance + epsilon.
   * 
   * @param u L'indice du neurone.
   * @return La racine carre de la variance.
   */
  public double getSquareVariance(int u)
  {
    return Math.sqrt(this.getVariance(u) + EPSILON);
  }
  
  @Override
  public double dCdSC(int u)
  {
    throw new SNoImplementationException();
    
  }
  
  @Override
  public double dCdSH(int u)
  {
    throw new SNoImplementationException();
    
  }
  
  @Override
  protected double forward(int b, int u) 
  {
    throw new SNoImplementationException();
    
  }

  @Override
  protected double backward(int b, int v) 
  {
    throw new SNoImplementationException();
    
  }

  @Override
  public String getFunctionName() 
  {
    return SNeuralFunction.BATCH_NORMALIZATION_FUNCTION;
  }
  
  /**
   * ...
   * 
   * @param b La taille de la batch.
   * @param batch_inputs La liste des entres dans la fonction.
   * @param scale Le scale.
   * @param shift Le shift.
   * @param running Les valeurs voluante de moyenne et variance.
   * @return La batch des donnes d'entre normalises par la fonction.
   */
  public List<double[]> batchNormalization(int b, List<double[]> batch_inputs, SScale scale, SShift shift, SRunningMeanVariance running)
  {
    // Valisation des paramtres.
    if(scale.getNeurons() != this.getOutputsSize())
      throw new IllegalArgumentException("Erreur SBatchNormalizationFunction 003 : La normalisation n'est pas admissible, car les scales n'ont pas la bonne dimension.");
    
    if(shift.getNeurons() != this.getOutputsSize())
      throw new IllegalArgumentException("Erreur SBatchNormalizationFunction 004 : La normalisation n'est pas admissible, car les shift n'ont pas la bonne dimension.");
        
    if(running.getNeurons() != this.getOutputsSize())
      throw new IllegalArgumentException("Erreur SBatchNormalizationFunction 005 : La normalisation n'est pas admissible, car le running n'est pas la bonne dimension.");
    
    // Affectation des paramtres.
    this.SC = scale;
    this.SH = shift;
    this.running = running;
    
    // valuer la moyenne des agrgations.
    computeMean(b, batch_inputs);
             
    // valuer les variances des agrgations.
    computeVariance(b, batch_inputs);

    // valuer la fonction sur l'ensemble des neurones.
    return batchForwardPropagation(b, batch_inputs);
  }
    
  @Override
  public boolean upgradeMaximumBatchSize(int size)
  {
    // Faire l'upgrade de la mthode parent et continuer dans cette classe s'il y a eu mise  jour.
    if(super.upgradeMaximumBatchSize(size))
    {
      // valuer l'ampleur de la mise  jour.
      int upgrade = size - this.normalized_Z.size();
      
      // Augmenter l'espace mmoire uniquement si ncessaire
      if(upgrade > 0)
        for(int i = 0; i < upgrade; i++)
          this.normalized_Z.add(new double[this.getOutputsSize()]);
    
      return true;
    }
    else
      return false;
  }
  
  /**
   * Mthode pour valuer la moyenne des agrgation.
   * 
   * @param m Le nombre d'agrgation  considrer (n'est pas ncessairement gale  la taille de la liste).
   * @param Z La liste des agrgations.
   */
  private void computeMean(int m, List<double[]> Z)
  {
    throw new SNoImplementationException();
    
  }
  
  /**
   * ...
   * 
   * @param Z
   */
  private void computeVariance(int m, List<double[]> Z)
  {
    throw new SNoImplementationException();
    
  }
  
}
