/**
 * 
 */
package sim.nn.gradient;

import sim.nn.function.SComputeErrorFunctionGradientActivationParameters;
import sim.nn.function.SComputeErrorFunctionGradientBias;
import sim.nn.function.SComputeErrorFunctionGradientFiltersWeights;
import sim.nn.function.activation.SActivationFunctionBuilder;
import sim.nn.layer.SLayer;
import sim.nn.parameter.SActivationParameters;
import sim.nn.parameter.SBias;
import sim.nn.parameter.SContainActivationParameters;
import sim.nn.parameter.SContainBias;
import sim.nn.parameter.SContainFiltersWeights;
import sim.nn.parameter.SFiltersWeights;

/**
 * La classe <b>SConvolutionGradientLayer</b> reprsente un le gradient de la fonction d'erreur applicable  une couche de neurone ayant une fonction de convolution  plusieurs niveau de profondeur.
 * 
 * @author Simon Vezina
 * @since 2020-01-12
 * @version 2021-12-31
 */
public class SConvolutionGradientLayer implements SGradientLayer, SContainFiltersWeights, SContainBias, SContainActivationParameters {


  /**
   * ...
   */
 protected final SFiltersWeights grad_W;
  
  /**
   * ...
   */
  protected final SBias grad_B;
  
  
  /**
   * ...
   */
  protected final SActivationParameters grad_K;
  
  /**
   * ...
   * 
   * @param outputSize
   * @param filter_offset La taille du filtre en ligne et colonne.
   * @param input_depth Le nombre de matrice par filtre (la profondeur d'entre).
   * @param output_depth Le nombre de filtre dans la convolution (la profondeur de sortie).
   * @param activation_function_name
   */
  public SConvolutionGradientLayer(int outputSize, int filter_offset, int input_depth, int output_depth, String activation_function_name) 
  {
    // Dfinir le poids de convolution.
    this.grad_W = new SFiltersWeights(filter_offset, input_depth, output_depth); 
    this.grad_W.clear();
    
    // Dfinir la taille des biais de la fonction de convolution en fonction de la profondeur de sortie (le nombre de filtre).
    this.grad_B = new SBias(output_depth);
    this.grad_B.clear();
    
    // Dfinir la taille pour les paramtres d'activation.
    this.grad_K = SActivationFunctionBuilder.buildParameters(activation_function_name, outputSize);
    this.grad_K.clear();
  }

  @Override
  public SFiltersWeights getFiltersWeights() 
  {
    return this.grad_W;
  }
  
  @Override
  public SBias getBias() 
  {
    return this.grad_B;
  }
  
  @Override
  public SActivationParameters getActivationParameters() 
  {
    return this.grad_K;
  }
  
  /**
   * Mthode pour obtenir la taille du filtre dans la convolution.
   * 
   * @return La taille du filtre.
   */
  public int getFilterOffset()
  {
    return this.grad_W.getOffset();
  }
    
  @Override
  public void add(SGradientLayer gradient) throws IllegalArgumentException
  {
    this.grad_W.add(((SContainFiltersWeights)gradient).getFiltersWeights());
    this.grad_B.add(((SContainBias)gradient).getBias());
    this.grad_K.add(((SContainActivationParameters)gradient).getActivationParameters());
  }

  @Override
  public void add(SLayer layer) throws IllegalArgumentException
  {
    SComputeErrorFunctionGradientFiltersWeights dCdW = (SComputeErrorFunctionGradientFiltersWeights) layer;
    SComputeErrorFunctionGradientBias dCdB = (SComputeErrorFunctionGradientBias) layer;
    SComputeErrorFunctionGradientActivationParameters dCdK = (SComputeErrorFunctionGradientActivationParameters) layer;
    
    double value;
    
    // Itrer sur l'ensemble des profondeur de sortie.
    for(int f = 0; f < this.grad_W.getOutputDepth(); f++)
    {
      
      // Itrer sur l'ensemble des indice d, i, et j pour obtenir les gradient des poids de convolution.
      for(int d = 0; d < this.grad_W.getInputDepth(); d++)
        for(int i = 0; i < this.grad_W.getOffset(); i++)
          for(int j = 0; j < this.grad_W.getOffset(); j++)
          {
            value = this.grad_W.get(f, d, i, j) + dCdW.dCdW(f, d, i, j);
            
            this.grad_W.set(value , f, d, i, j);
          }
      
      // Dfinir le gradient du biais pour le filtre f.
      value = this.grad_B.get(f) + dCdB.dCdB(f);
      this.grad_B.set(f, value);
    }
        
    //-------------------------------------------------
    // valuer le gradient des paramtres d'activation.
    //-------------------------------------------------
    if(!this.grad_K.isEmpty())
    {
      for(int i = 0; i < this.grad_K.getNbParameterType(); i++)
        for(int u = 0; u < this.grad_K.getNbNeurons(); u++)
        {
          value = this.grad_K.get(i, u) + dCdK.dCdK(i, u);
          this.grad_K.set(i, u, value);
        }
    }
    
  }
  
  @Override
  public void clear() 
  {
    this.grad_W.clear();
    this.grad_B.clear();
    this.grad_K.clear();
  }

  @Override
  public void multiply(double v) 
  {
    this.grad_W.multiply(v);
    this.grad_B.multiply(v);
    this.grad_K.multiply(v);
  }

  @Override
  public void multiply(SGradientLayer gradient)
  {
    this.grad_W.multiply(((SContainFiltersWeights)gradient).getFiltersWeights());
    this.grad_B.multiply(((SContainBias)gradient).getBias());
    this.grad_K.multiply(((SContainActivationParameters)gradient).getActivationParameters());
  }

}
