/**
 * 
 */
package sim.nn.gradient;

import java.util.Arrays;

import sim.exception.SConstructorException;
import sim.math.SImpossibleNormalizationException;
import sim.nn.layer.SBatchConvolutionLayer;
import sim.nn.layer.SBatchDropoutLayer;
import sim.nn.layer.SBatchFullConnectedLayer;
import sim.nn.layer.SBatchPoolingLayer;
import sim.nn.layer.SConvolutionLayer;
import sim.nn.layer.SDropoutLayer;
import sim.nn.layer.SFullConnectedLayer;
import sim.nn.layer.SLayer;
import sim.nn.layer.SPoolingLayer;
import sim.nn.network.SNetwork;

/**
 * La classe <b>SErrorFunctionGradient</b> reprsente le gradient de la fonction d'erreur associ  un rseau de neurones.
 * 
 * @author Simon Vezina
 * @since 2019-08-17
 * @version 2021-12-31
 */
public class SErrorFunctionGradient {

  /**
   * La variable <b>grad_layer</b> reprsente le tableau des couches de gradients pour chaque paramtre de la fonction d'erreur.
   */
  private final SGradientLayer[] grad_layer;
  
  /**
   * La variable <b>mini_batch_size</b> reprsente le nombre de vecteurs utilis pour comptabilit le gradient de la fonction d'erreur.
   * Si la variable n'est pas gale  1, ceci reprsente un gradient en mode <i>mini-batch</i>.
   */
  private int mini_batch_size;
  
  //-------------
  // CONSTRUCTEUR
  //-------------
  
  /**
   * Constructeur du gradient d'une fonction d'erreur avec un rseau.
   * 
   * @param network Le rseau sur lequel la fonction d'erreur sera value. 
   */
  public SErrorFunctionGradient(SNetwork network) throws SConstructorException
  {
    // Construction du tableau des gradients.
    grad_layer = new SGradientLayer[network.getNbLayers()];
    
    // Remplir le tableau en fonction du type de couche que l'on retrouve dans le rseau.
    for(int i = 0; i < grad_layer.length; i++)
    {
      SLayer layer =  network.getLayers()[i];
      
      // Validation du type de couche.
      // Remarque : L'ordre est important, car SBatchFullConnectedLayer pourrait hriter de SFullConnectedLayer !!!
      
      if(layer instanceof SBatchFullConnectedLayer)
        grad_layer[i] = new SBatchFullConnectedGradientLayer(layer.getInputsSize(), layer.getOutputsSize(), ((SBatchFullConnectedLayer) layer).getAggregationFunctionOrder(), ((SBatchFullConnectedLayer) layer).getActivationFunctionName());
      else
        if(layer instanceof SFullConnectedLayer)
          grad_layer[i] = new SFullConnectedGradientLayer(layer.getInputsSize(), layer.getOutputsSize(), ((SFullConnectedLayer) layer).getAggregationFunctionOrder(), ((SFullConnectedLayer) layer).getActivationFunctionName());
        else
        {  
          // Validation du type de couche.
          // Remarque : L'ordre est important, car SBatchConvolutionLayer pourrait hriter de SConvolutionLayer !!!
          
          if(layer instanceof SBatchConvolutionLayer)
          {
            SBatchConvolutionLayer l = (SBatchConvolutionLayer) layer;
            grad_layer[i] = new SBatchConvolutionGradientLayer(l.getOutputsSize(), l.getFilterOffset(), l.getInputsDepth(), l.getOutputsDepth(), l.getActivationFunctionName());
          }
          else
            if(layer instanceof SConvolutionLayer)
            {
              SConvolutionLayer l = (SConvolutionLayer) layer;
              grad_layer[i] = new SConvolutionGradientLayer(l.getOutputsSize(), l.getFilterOffset(), l.getInputsDepth(), l.getOutputsDepth(), l.getActivationFunctionName());
            }
            else
              if(layer instanceof SBatchDropoutLayer || layer instanceof SDropoutLayer)
                grad_layer[i] = new SEmptyParameterGradientLayer();
              else
                if(layer instanceof SBatchPoolingLayer || layer instanceof SPoolingLayer)
                  grad_layer[i] = new SEmptyParameterGradientLayer();
                else
                  throw new SConstructorException("Erreur SErrorFunctionGradient 001 : La couche i = " + i + " est d'un type inconnu.");
   
        }
    }
    
    this.mini_batch_size = 0;
  }
  
  //------------
  // MTHODES //
  //------------
  
  /**
   * Mthode pour obtenir la taille de la mini-batch utilise pour valuer le gradient de la fonction d'erreur.
   * Si la valeur est gale  zro, c'est que le gradient a t rinitialis.
   * 
   * @return La taille de la batch.
   */
  public int getMiniBatchSize()
  {
    return mini_batch_size;
  }
  
  /**
   * Mthode pour obtenir le gradient de la fonction d'erreur d'une couche du rseau associ.
   * 
   * @param i L'index de la couche.
   * @return Le gradient de la fonction d'erreur d'une couche.
   */
  public SGradientLayer getGradientLayer(int i)
  {
    return this.grad_layer[i];
  }
  
  /**
   * Mthode pour ajouter au gradient de la fonction d'erreur la contribution d'un rseau.
   *  
   * @param network Le rseau.
   */
  public void increase(SNetwork network)
  {
    // Ajout de la donne du rseau  la mini-batch accumul dans le gradient de la fonction d'erreur.
    // S'il y a descente stochastique du gradient, cette valeur sera gale  1 durant l'apprentissage.
    this.mini_batch_size++;
    
    //--------------------------------------------------------------
    // VERSION LABORATOIRE : L'apprentissage par rseau de neurones.
    //--------------------------------------------------------------
    //throw new SNoImplementationException();
    
    SLayer[] layers = network.getLayers();
    
    for(int i = 0; i < network.getNbLayers(); i++)
      this.grad_layer[i].add(layers[i]);
  }
  
  /**
   * Mthode pour ajouter au gradient de la fonction d'erreur la contribution d'une autre fonction d'erreur.
   * 
   * @param gradient Le gradient  ajouter.
   */
  public void increase(SErrorFunctionGradient gradient)
  {
    //--------------------------------------------------------------
    // VERSION LABORATOIRE : L'apprentissage par rseau de neurones.
    //--------------------------------------------------------------
    //throw new SNoImplementationException();
    
    // Ajouter la contribution d'un autre gradient uniquement si n'est pas vide.
    if(!gradient.isEmpty())
    {
      for(int i = 0; i < this.grad_layer.length; i++)
        this.grad_layer[i].add(gradient.grad_layer[i]);
      
      this.mini_batch_size += gradient.mini_batch_size;
    }
  }
  
  /**
   * Mthode pour additionner deux gradients de fonction d'erreur ensemble. Il faudra que ces deux gradients soient normaliss.
   * 
   * @param gradient Le gradient  additionner.
   * @throws IllegalArgumentException Si l'un des deux gradients n'est pas normalis.
   */
  public void add(SErrorFunctionGradient gradient) throws IllegalArgumentException 
  {
    if(!this.isNormalized())
      throw new IllegalArgumentException("Erreur SErrorFunctionGradient 003 : L'addition de gradient n'est pas possible, car le gradient courant n'est pas normalis (batch_size = " + this.getMiniBatchSize() + ").");
    
    if(!gradient.isNormalized())
      throw new IllegalArgumentException("Erreur SErrorFunctionGradient 004 : L'addition de gradient n'est pas possible, car le gradient  ajouter n'est pas normalis (batch_size = " + gradient.getMiniBatchSize() + ").");
    
    increase(gradient);
    
    // Mettre la taille de la batch du rsultat  1, car la mthode increase l'aura tablie  2.
    this.mini_batch_size = 1;
  }
  
  /**
   * Mthode pour multiplier le gradient de la fonction d'erreur par une constante.
   * Cette action aura pour but de multiplier l'ensemble des paramtres du gradient par la constante.
   * 
   * @param v La constante.
   */
  public void multiply(double v)
  {
    if(!isEmpty())
      for(int i = 0; i < this.grad_layer.length; i++)
        this.grad_layer[i].multiply(v);
  }
  
  /**
   * Mthode pour multiplier le grandient de la fonction d'erreur avec un autre gradient.
   * Le produit sera effectu composante par composante.
   * 
   * @param grad Le gradient.
   */
  public void multiply(SErrorFunctionGradient grad)
  {
    for(int i = 0; i < this.grad_layer.length; i++)
      this.grad_layer[i].multiply(grad.grad_layer[i]);
  }
  
  /**
   * <p>
   * Mthode pour normaliser un gradient de fonction d'erreur.
   * </p>
   * <p>
   * Cette action consiste  valuer l'ensemble des paramtres sur une base d'un <i>batch</i> de 1.
   * Ainsi, comparer le module d'un gradient normaliser sera indpendant de la taille de la <i>batch</i>
   * utiliser pour incrmenter la valeur de celui-ci.
   * </p>
   * <p>
   * Cependant, si le gradient est vide, il sera modifi et considr comme tant construit  l'aide d'une <i>batch</i> de 1.
   * </p>
   * <p>
   * Il est important de prciser qu'aprs l'excution de cette mthode, le gradient sera <u>modifi</u>.
   * La valeur de retour sera ainsi une rfrence  l'objet lui-mme.
   * </p>
   * @return Le gradient normalis.
   * @throws SImpossibleNormalizationException Si le gradient est de taille zro.
   */
  public SErrorFunctionGradient normalize() throws SImpossibleNormalizationException
  {
    if(getMiniBatchSize() == 0)
      throw new SImpossibleNormalizationException("Erreur SErrorFunctionGradient 005 : Le gradient ne peut pas tre normalis, car il n'a pas d'lment  l'intrieur (il est de taille zro).");
    
    // Modifier les champs du gradient uniquement si la batch_size est suprieur  0.
    if(getMiniBatchSize() > 1)
    {
      double value = 1.0 / (double) getMiniBatchSize();
      
      for(int i = 0; i < this.grad_layer.length; i++)
        this.grad_layer[i].multiply(value);
      
      this.mini_batch_size = 1;
    }
     
    return this;
  }
  
  /**
   * Mthode pour dterminer si le gradient est normalise.
   * 
   * @return <b>true</b> si le gradient est normalise et <b>false</b> sinon.
   */
  public boolean isNormalized()
  {
    return this.mini_batch_size == 1;
  }
  
  /**
   * <p>
   * Mthode pour rinitialiser le gradient de la fonction d'erreur.
   * Cette action a pour but de mettre  zro l'ensemble des valeurs des gradients des poids et biais ainsi que les erreur de propagation.
   * </p>
   * <p>
   * Un gradient vid n'est pas normalis.
   * </p>
   */
  public void clear()
  {
    for(int i = 0; i < grad_layer.length; i++)
      grad_layer[i].clear();
    
    mini_batch_size = 0;
  }
  
  /**
   * <p>
   * Mthode pour dterminer si le gradient est vide. Celui-ci sera vide aprs l'appel de la mthode <b>clear()</b>.
   * </p>
   * <p>
   * Il est important de noter qu'un gradient vide n'est pas quivalent  un gradient nul. 
   * Si celui-ci est nul, toutes ses valeurs seront gales  zro, mais aprs un calcul mathmatique.
   * La taille de la batch ne sera pas gal  zro.
   * </p>
   * 
   * @return <b>true</b> si le gradient est vide et <b>false</b> sinon.
   */
  public boolean isEmpty()
  {
    return mini_batch_size == 0;
  }
  
  @Override
  public boolean equals(Object obj)
  {
    if (this == obj)
      return true;
    
    if (obj == null)
      return false;
    
    if (getClass() != obj.getClass())
      return false;
    
    SErrorFunctionGradient other = (SErrorFunctionGradient) obj;
    if (!Arrays.equals(grad_layer, other.grad_layer))
      return false;
    
    if (mini_batch_size != other.mini_batch_size)
      return false;
    
    return true;
  }
  
}
