/**
 * 
 */
package sim.nn.gradient;

import sim.nn.function.SComputeErrorFunctionGradientActivationParameters;
import sim.nn.function.SComputeErrorFunctionGradientBias;
import sim.nn.function.SComputeErrorFunctionGradientWeights;
import sim.nn.function.activation.SActivationFunctionBuilder;
import sim.nn.layer.SLayer;
import sim.nn.parameter.SActivationParameters;
import sim.nn.parameter.SBias;
import sim.nn.parameter.SContainActivationParameters;
import sim.nn.parameter.SContainBias;
import sim.nn.parameter.SContainWeights;
import sim.nn.parameter.SWeights;

/**
 * La classe <b>SFullConnectedGradientLayer</b> reprsente le gradient de la fonction d'erreur value sur une couche d'un rseau
 * de type <i>full connected</i>.
 * 
 * @author Simon Vezina
 * @since 2019-08-17
 * @version 2021-12-05
 */
public class SFullConnectedGradientLayer implements SGradientLayer, SContainWeights, SContainBias, SContainActivationParameters {

  /**
   * La variable <b>grad_matrix</b> reprsente le gradient des poids (<i>weights</i>) de la couche de neurones.
   */
  protected final SWeights grad_W;
  
  /**
   * La variable <b>grad_bias</b> reprsente le gradient des biais (<i>bias</i>) de la couche de neurones.
   */
  protected final SBias grad_B;
  
  /**
   * ...
   */
  protected final SActivationParameters grad_K;
  
  //----------------
  // CONSTRUCTEUR //
  //----------------
  
  /**
   * Constructeur du gradient d'une couche de neurones de type <i>FullConnected</i>.
   * 
   * @param inputSize La taille de l'entre de la couche.
   * @param outputSize La taille de la sortie de la couche.
   * @param activation_function Le nom de la fonction d'activation.
   */
  public SFullConnectedGradientLayer(int inputSize, int outputSize, String activation_function) 
  {
    this(inputSize, outputSize, 1, activation_function);
  }
      
    
  /**
   * ...
   * 
   * @param inputSize
   * @param outputSize
   * @param order
   * @param activation_function
   */
  public SFullConnectedGradientLayer(int inputSize, int outputSize, int order, String activation_function) 
  {
    this.grad_W = new SWeights(inputSize, outputSize, order, false);
    this.grad_W.clear();
    
    this.grad_B = new SBias(outputSize, false);
    this.grad_B.clear();
    
    this.grad_K = SActivationFunctionBuilder.buildParameters(activation_function, outputSize);
    this.grad_K.clear();
  }
  
  //------------
  // MTHODES //
  //------------
  
  @Override
  public SBias getBias() 
  {
    return this.grad_B;
  }

  @Override
  public SWeights getWeights() 
  {
    return this.grad_W;
  }
      
  @Override
  public SActivationParameters getActivationParameters() 
  {
    return this.grad_K;
  }
  
  @Override
  public void add(SGradientLayer gradient) throws IllegalArgumentException
  {
    this.grad_W.add(((SContainWeights)gradient).getWeights());
    this.grad_B.add(((SContainBias)gradient).getBias());
    this.grad_K.add(((SContainActivationParameters)gradient).getActivationParameters());
  }
  
  @Override
  public void add(SLayer layer) throws IllegalArgumentException
  {
    SComputeErrorFunctionGradientWeights dCdW = (SComputeErrorFunctionGradientWeights) layer;
    SComputeErrorFunctionGradientBias dCdB = (SComputeErrorFunctionGradientBias) layer;
    SComputeErrorFunctionGradientActivationParameters dCdK = (SComputeErrorFunctionGradientActivationParameters) layer;
    
    double value;
    
    //-----------------------------------------------
    // valuer les gradients pour des poids et biais. 
    //-----------------------------------------------
    if(grad_W.getOrder() == 1)
    {
      //------------------------------
      // POUR ORDRE 1 : Version rapide
      //------------------------------
      
      // Itrer sur l'ensemble des neurones de sortie u de la couche.
      for(int u = 0; u < layer.getOutputsSize(); u++)
      {
        // valuer l'ensemble des gradients par rapport au poids pour tous les neurones n'entre v et de sortie u.
        for(int v = 0; v < layer.getInputsSize(); v++)
        {
          value = this.grad_W.get(v, u) + dCdW.dCdW(v, u);
          this.grad_W.set(v, u, value);
        }
        
        // valuer le gradients par rapport au biais pour le neurones de sortie u.
        value = this.grad_B.get(u) + dCdB.dCdB(u);
        this.grad_B.set(u, value); 
      }
    }
    else
    {
      //-------------------------------------
      // POUR ORDRE 2 et + : Version gnrale
      //-------------------------------------
      
      // Itrer sur l'ensemble des neurones de sortie u de la couche.
      for(int u = 0; u < layer.getOutputsSize(); u++)
      {
        // valuer l'ensemble des gradients par rapport au poids pour tous les neurones n'entre v et de sortie u d'ordre o.
        for(int o = 1; o <= grad_W.getOrder(); o++)
          for(int v = 0; v < layer.getInputsSize(); v++)
          {
            value = this.grad_W.get(v, u, o) + dCdW.dCdW(v, u, o);
            this.grad_W.set(v, u, o, value);
          }
        
        // valuer le gradients par rapport au biais pour le neurones de sortie u.
        value = this.grad_B.get(u) + dCdB.dCdB(u);
        this.grad_B.set(u, value);   
      }
    }  
   
    //-------------------------------------------------
    // valuer le gradient des paramtres d'activation.
    //-------------------------------------------------
    if(!this.grad_K.isEmpty())
    {
      for(int i = 0; i < this.grad_K.getNbParameterType(); i++)
        for(int u = 0; u < this.grad_K.getNbNeurons(); u++)
        {
          value = this.grad_K.get(i, u) + dCdK.dCdK(i, u);
          this.grad_K.set(i, u, value);
        }
    }
    
  }
     
  @Override
  public void clear()
  {
    this.grad_W.clear();
    this.grad_B.clear();
    this.grad_K.clear();
  }

  @Override
  public void multiply(double v) 
  {
    this.grad_W.multiply(v);
    this.grad_B.multiply(v);
    this.grad_K.multiply(v);
  }

  @Override
  public void multiply(SGradientLayer gradient) 
  {
    this.grad_W.multiply(((SContainWeights)gradient).getWeights());
    this.grad_B.multiply(((SContainBias)gradient).getBias());
    this.grad_K.multiply(((SContainActivationParameters)gradient).getActivationParameters());
  }
  
  @Override
  public boolean equals(Object obj)
  {
    if (this == obj)
      return true;
    
    if (obj == null)
      return false;
    
    if (getClass() != obj.getClass())
      return false;
    
    SFullConnectedGradientLayer other = (SFullConnectedGradientLayer) obj;
    if (grad_B == null) {
      if (other.grad_B != null)
        return false;
    } else if (!grad_B.equals(other.grad_B))
      return false;
    
    if (grad_W == null) {
      if (other.grad_W != null)
        return false;
    } else if (!grad_W.equals(other.grad_W))
      return false;
    
    if (grad_K == null) {
      if (other.grad_K != null)
        return false;
    } else if (!grad_K.equals(other.grad_K))
      return false;
    
    return true;
  }

  @Override
  public String toString()
  {
    if(this.grad_K.isEmpty())
      return "SFullConnectedGradientLayer [grad_W=" + grad_W + ", grad_B=" + grad_B + "]";
    else
      return "SFullConnectedGradientLayer [grad_W=" + grad_W + ", grad_B=" + grad_B + ", grad_K=" + grad_K + "]";
  }
  
}
