/**
 * 
 */
package sim.nn.gradient;

import java.util.ArrayList;
import java.util.List;

import sim.exception.SNoImplementationException;
import sim.exception.SRuntimeException;
import sim.nn.SNNData;
import sim.nn.function.error.SBatchErrorFunction;
import sim.nn.function.error.SErrorFunction;
import sim.nn.network.SBatchNetwork;
import sim.nn.network.SNetwork;

/**
 * La classe <b>SGradientAlgorithm</b> est une classe utilitaire effectuant des calculs de gradient de fonction d'erreur sur des rseaux.
 * 
 * @author Simon Vezina
 * @since 2019-09-06
 * @version 2022-09-14 (Version labo : Apprentissage par rseau de neurones)
 */
public class SGradientAlgorithm {
 
  /**
   * <p>
   * Mthode pour ajouter  un gradient la contribution de la propagation du gradient provenant de l'activation d'une donne.
   * </p>
   * <p>
   * Il est important de prciser que dans l'exerice de cette mthode, le gradient pass en paramtre sera <u>modifi</u> et 
   * retourn en guise de rponse du calcul et ce afin de rduire l'allocation en mmoire.
   * </p>
   * 
   * @param network Le rseau.
   * @param error_function La fonction d'erreur.
   * @param gradient Le gradient de la fonction d'erreur.
   * @param data La donne.
   * @return Le gradient pass en paramtre, mais modifi par cette mthode.
   */
  public static SErrorFunctionGradient computeGradient(SNetwork network, SErrorFunction error_function, SErrorFunctionGradient gradient, SNNData data)
  {
    throw new SNoImplementationException();
    
  }

  /**
   * <p>
   * Mthode pour ajouter  un gradient la contribution de la propagation du gradient provenant de l'activation de plusieurs donnes en mode <i>batch</i>.
   * </p>
   * <p>
   * Il est important de prciser que dans l'exerice de cette mthode, le gradient pass en paramtre sera <u>modifi</u> et 
   * retourn en guise de rponse du calcul et ce afin de rduire l'allocation en mmoire.
   * </p>
   * @param batch_network Le rseau de type <i>batch</i>.
   * @param error_function La fonction d'erreur en <i>batch</i>.
   * @param gradient Le gradient de la fonction d'erreur.
   * @param data_list La liste des donnes.
   * @return Le gradient pass en paramtre, mais modifi par cette mthode.
   */
  public static SErrorFunctionGradient batchComputeGradient(SBatchNetwork batch_network, SBatchErrorFunction error_function, SErrorFunctionGradient gradient, List<SNNData> data_list) 
  {
    throw new SNoImplementationException();
    
  }
  
  /**
   * <p>
   * Mthode pour ajouter  un gradient la contribution de la propagation du gradient provenant d'une liste d'activation de donnes.
   * Le calcul sera ralis en parallle.
   * </p>
   * <p>
   * Il est important de prciser que dans l'exerice de cette mthode, le gradient pass en paramtre sera <u>modifi</u> et 
   * retourn en guise de rponse du calcul et ce afin de rduire l'allocation en mmoire.
   * </p>
   * 
   * @param gradient Le gradient qui contiendra le rsultat du calcul.
   * @param computer_list La liste des calculateurs de gradient.
   * @param data_list Les donnes utilises lors de calcul du gradient.
   * @return Le gradient calcul.
   */
  public static SErrorFunctionGradient parallelComputeGradient(SErrorFunctionGradient gradient, List<SGradientComputer> computer_list, List<SNNData> data_list)
  {
    // Vider les calculateurs de leur ancienne donnes.
    for(SGradientComputer computer : computer_list)
      computer.clear();
    
    // Remplir les calculateurs avec les nouvelles donnes en partagant entre le calculateur.
    for(int i = 0; i < data_list.size(); i++)
      computer_list.get(i % computer_list.size()).addData(data_list.get(i));
    
    // Construire la list de threads.
    List<Thread> thread_list = new ArrayList<Thread>();
    
    // Remplir la liste de theads.
    for(SGradientComputer c : computer_list)
      thread_list.add(new Thread(c));
    
    // Partir l'ensemble des threads
    for(Thread t : thread_list)
      t.start();
    
    // Faire l'attente de la fin de toutes les threads de la liste
    for(Thread t : thread_list)
      try{
        t.join();
      }catch(InterruptedException e){
        throw new SRuntimeException("Erreur SGradientAlgorithm 002 : Un thread fut interropue spontanment.", e);
      }
    
    // Collecter les rsultats (version non parallle).
    for(SGradientComputer computer : computer_list)
      gradient.increase(computer.getGradient());
      
    return gradient;
  }
  
  /**
   * <p>
   * Mthode valuant le gradient avec inertie. 
   * </p>
   * <p>
   * Il est important de prciser que 
   * <ul>- Les objets <b>gradient</b> et <b>momentum_gradient</b> seront <u>modifis</u> dans l'exercice de cette mthode.</ul>
   * <ul>- Les deux gradients en opration mathmatique se doivent d'tre normalis pralablement.</ul>
   * <ul>- Le gradient perturbant le gradient avec inertie sera multipli par une constante.</ul>
   * <ul>- Le gradient avec inertie contiendra le rsultat des calculs.</ul>
   * </p>
   * 
   * @param new_gradient Le gradient  ajouter au gradient avec inertie.
   * @param momentum_gradient Le gradient avec inertie. 
   * @param momentum L'inertie dans le calcul du gradient.
   * @return Le gradient avec inertie.
   */
  public static SErrorFunctionGradient updateGradientWithMomentum(SErrorFunctionGradient momentum_gradient, SErrorFunctionGradient new_gradient, double momentum)
  {
    throw new SNoImplementationException();
    
  }
  
}
