package sim.nn.layer;

import java.util.List;

import sim.exception.SConstructorException;
import sim.exception.SNoImplementationException;
import sim.nn.function.SComputeErrorFunctionGradientActivationParameters;
import sim.nn.function.SComputeErrorFunctionGradientBias;
import sim.nn.function.SComputeErrorFunctionGradientFiltersWeights;
import sim.nn.function.SComputeErrorFunctionGradientScale;
import sim.nn.function.SComputeErrorFunctionGradientShift;
import sim.nn.function.activation.SActivationFunctionBuilder;
import sim.nn.function.activation.SBatchActivationFunction;
import sim.nn.function.filtering.SBatchConvolutionFunction;
import sim.nn.function.regularization.SBatchNormalizationFunction;
import sim.nn.function.regularization.SRunningMeanVariance;
import sim.nn.gradient.SGradientLayer;
import sim.nn.parameter.SActivationParameters;
import sim.nn.parameter.SBias;
import sim.nn.parameter.SContainActivationParameters;
import sim.nn.parameter.SContainBias;
import sim.nn.parameter.SContainFiltersWeights;
import sim.nn.parameter.SContainScale;
import sim.nn.parameter.SContainShift;
import sim.nn.parameter.SFiltersWeights;
import sim.nn.parameter.SScale;
import sim.nn.parameter.SShift;
import sim.util.SStringUtil;

/**
 * La classe <b>SBatchConvolutionLayer</b> reprsente une couche de neurones de type <i>batch</i> avec la fonction de convolution.
 * 
 * @author Simon Vezina
 * @since 2021-11-02
 * @version 2022-09-13 (Version labo : Apprentissage par rseau de neurones)
 */
public class SBatchConvolutionLayer implements SBatchLayer,
  SComputeErrorFunctionGradientFiltersWeights, 
  SComputeErrorFunctionGradientBias, 
  SComputeErrorFunctionGradientActivationParameters, 
  SComputeErrorFunctionGradientScale,
  SComputeErrorFunctionGradientShift,
  SContainFiltersWeights, 
  SContainBias,
  SContainActivationParameters,
  SContainScale,
  SContainShift
{

  //--------------
  // CONSTANTES //
  //--------------
  
  /**
   * La constante <b>DEFAULT_ACTIVATION_FUNCTION</b> correspond  la fonction d'activation par dfaut tant {@value}.
   */
  final static String DEFAULT_ACTIVATION_FUNCTION = "sigmoid";
  
  
  /**
   * ...
   */
  protected final SFiltersWeights W;
  
  /**
   * ...
   */
  protected final SBias B;
  
  
  /**
   * ...
   */
  protected final SActivationParameters K;  
  
  /**
   * ...
   */
  private final SScale SC;
  
  /**
   * ....
   */
  private final SShift SH;
  
  /**
   * ...
   */
  private final SRunningMeanVariance running;
  
  /**
   * ...
   */
  private final SBatchConvolutionFunction Z;
  
  /**
   * ...
   */
  private final SBatchActivationFunction A;
  
  /**
   * ...
   */
  private final SBatchNormalizationFunction N;
  
    
  /**
   * ...
   * 
   * @param input
   * @param input_depth
   * @param nb_filter
   * @param filter_offset
   * @param padding
   * @param stride
   * @param activation_function_name
   * @throws SConstructorException
   */
  public SBatchConvolutionLayer(int input, int input_depth, int nb_filter, int filter_offset, int padding, int stride, String activation_function_name) throws SConstructorException
  {
    // Dfinition de la fonction de convolution et la taille de la couche.
    this.Z = new SBatchConvolutionFunction(input, input_depth, nb_filter, filter_offset, padding, stride);
    
    // Dfinition de la matrice des poids en fonction de l'ordre de la fonction d'agrgation.
    this.W = new SFiltersWeights(filter_offset, input_depth, nb_filter);
    
    // Dfinition du vecteur des biais.
    this.B = new SBias(nb_filter);
    
    // Dfinition des paramtres d'activation
    this.K = SActivationFunctionBuilder.buildParameters(activation_function_name, this.getOutputsSize());
    
    this.SC = new SScale(this.getOutputsSize());
    
    this.SH = new SShift(this.getOutputsSize());
    
    this.running = new SRunningMeanVariance(this.getOutputsSize());
    
    // Initialisation de la fonction d'activation.
    this.A = new SBatchActivationFunction(this.getOutputsSize(), activation_function_name);
    
    // Initialisation de la fonction de normalisation.
    this.N = new SBatchNormalizationFunction(this.getOutputsSize());
  }

  /**
   * ...
   * 
   * @param filters
   * @param bias
   * @param scale
   * @param shift
   * @param input_size
   * @param padding
   * @param stride
   * @param activation_function_name
   * @throws SConstructorException
   */
  public SBatchConvolutionLayer(SFiltersWeights filters, SBias bias, SScale scale, SShift shift, int input_size, int padding, int stride, String activation_function_name) throws SConstructorException
  {
    this(filters, bias, SActivationParameters.EMPTY_ACTIVATION_PARAMETERS, scale, shift, input_size, padding, stride, activation_function_name);
  }
  
  /**
   * ...
   * 
   * @param filters
   * @param bias
   * @param activation_parameters
   * @param scale
   * @param shift
   * @param input_size
   * @param padding
   * @param stride
   * @param activation_function_name
   * @throws SConstructorException
   */
  public SBatchConvolutionLayer(SFiltersWeights filters, SBias bias, SActivationParameters activation_parameters, SScale scale, SShift shift, int input_size, int padding, int stride, String activation_function_name) throws SConstructorException
  {
    // Dfinition de la fonction de convolution et la taille de la couche.
    this.Z = new SBatchConvolutionFunction(input_size, filters.getInputDepth(), filters.getOutputDepth(), filters.getOffset(), padding, stride);
    
    // Initialisation de la fonction d'activation.
    this.A = new SBatchActivationFunction(this.getOutputsSize(), activation_function_name);
    
    // Initialisation de la fonction de normalisation.
    this.N = new SBatchNormalizationFunction(this.getOutputsSize());
    
    this.running = new SRunningMeanVariance(this.getOutputsSize());
    
    this.W = filters;
    this.B = bias;
    this.K = activation_parameters;
    this.SC = scale;
    this.SH = shift;
  }
  
  /**
   * ...
   * 
   * @param layer
   */
  private SBatchConvolutionLayer(SBatchConvolutionLayer layer)
  {
    // Effectuer la copie de la rfrence des paramtres.
    this.W = layer.W;
    this.B = layer.B;
    this.K = layer.K;
    this.SC = layer.SC;
    this.SH = layer.SH;
    this.running = layer.running;
    
    // Construire les fonctions.
    this.Z = new SBatchConvolutionFunction(layer.getInputsSize(), layer.getInputsDepth(), layer.getOutputsDepth(), layer.getFilterOffset(), layer.getPadding(), layer.getStride());
    this.A = new SBatchActivationFunction(layer.getOutputsSize(), layer.A.getFunctionName());
    this.N = new SBatchNormalizationFunction(layer.getOutputsSize());
  }
  
  @Override
  public int getInputsSize() 
  {
    return this.Z.getInputsSize();
  }

  @Override
  public int getOutputsSize() 
  {
    return this.Z.getOutputsSize();
  }

  @Override
  public int getInputsDepth()
  {
    return this.Z.getInputsDepth();
  }
  
  @Override
  public int getOutputsDepth()
  {
    return this.Z.getOutputsDepth();
  }
  
  /**
   * Mthode pour obtenir la taille du filtre dans la convolution.
   * 
   * @return La taille du filtre.
   */
  public int getFilterOffset()
  {
    return this.W.getOffset();
  }
  
  /**
   * ...
   * 
   * @return
   */
  public int getStride()
  {
    return this.Z.getStride();
  }
  
  /**
   * ...
   * 
   * @return
   */
  public int getPadding()
  {
    return this.Z.getPadding();
  }
  
  @Override
  public SFiltersWeights getFiltersWeights()
  {
    return this.W;
  }
  
  @Override
  public SBias getBias()
  {
    return this.B;
  }
  
  @Override
  public SActivationParameters getActivationParameters() 
  {
    return this.K;
  }
 
  @Override
  public SScale getScale()
  {
    return this.SC;
  }
  
  @Override
  public SShift getShift()
  {
    return this.SH;
  }
  
  @Override
  public double dCdW(int f, int d, int i, int j)
  {
    return this.Z.dCdW(f, d, i, j);
  }
  
  @Override
  public double dCdB(int f)
  {
    return this.Z.dCdB(f);
  }
  
  @Override
  public double dCdK(int i, int u) throws IllegalArgumentException
  {
    return this.A.dCdK(i, u);
  }
  
  @Override
  public double dCdSH(int u)
  {
    return this.N.dCdSC(u);
  }

  @Override
  public double dCdSC(int u) 
  {
    return this.N.dCdSH(u);
  }
  
  @Override
  public String getLayerName()
  {
    return SLayer.BATCH_CONVOLUTION_LAYER;
  }
  
  @Override
  public String getDescription()
  {
    return this.getLayerName() + " [ Dimension = " + this.getDimension() + ", Z = " + this.Z + " , N = " + this.N + ", A = " + this.A + "]" + SStringUtil.END_LINE_CARACTER;
  }

  @Override
  public SLayer copyWithSameParameter() 
  {
    return new SBatchConvolutionLayer(this);
  }

  @Override
  public int getNumberOfParameter()
  {
    return this.W.getNumberOfParameter() + this.B.getNumberOfParameter() +  this.K.getNumberOfParameter() + this.SC.getNumberOfParameter() + this.SH.getNumberOfParameter();
  }
  
  @Override
  public List<double[]> batchForwardPropagation(int b, List<double[]> batch_inputs) throws IllegalArgumentException 
  {
    throw new SNoImplementationException();
    
  }
  
  @Override
  public List<double[]> batchBackPropagation(int b, List<double[]> batch_delta) throws IllegalArgumentException 
  {
    throw new SNoImplementationException();
    
  }

  @Override
  public int getBatchSize()
  {
    return this.N.getBatchSize();
  }

  /**
   * Mthode pour obtenir le nom de la fonction d'activation.
   * 
   * @return Le nom de la fonction d'activation.
   */
  public String getActivationFunctionName()
  {
    return this.A.getFunctionName();
  }
  
  @Override
  public String toString()
  {
    return getDescription();
  }
  
  @Override
  public String getDimension() 
  {
    return this.Z.getDimension();
  }
  
  @Override
  public void learn(SGradientLayer gradient, double learning_rate) throws IllegalArgumentException
  {
    this.W.learn(gradient, learning_rate);
    this.B.learn(gradient, learning_rate);
    this.K.learn(gradient, learning_rate);
    this.SC.learn(gradient, learning_rate);
    this.SH.learn(gradient, learning_rate);
  }

  @Override
  public void setBatchSize(int b) throws IllegalArgumentException 
  {
    this.Z.setBatchSize(b);
    this.A.setBatchSize(b);
    this.N.setBatchSize(b);
  }

  @Override
  public int getMaximumBatchSize()
  {
    return this.Z.getMaximumBatchSize();
  }

  @Override
  public boolean upgradeMaximumBatchSize(int size)
  {
    this.Z.upgradeMaximumBatchSize(size);
    this.A.upgradeMaximumBatchSize(size);
    
    return this.N.upgradeMaximumBatchSize(size);
  }
  
}
