/**
 * 
 */
package sim.nn.layer;

import java.util.List;

import sim.exception.SConstructorException;
import sim.exception.SNoImplementationException;
import sim.math.SBufferedMatrix;
import sim.nn.function.SComputeErrorFunctionGradientActivationParameters;
import sim.nn.function.SComputeErrorFunctionGradientBias;
import sim.nn.function.SComputeErrorFunctionGradientScale;
import sim.nn.function.SComputeErrorFunctionGradientShift;
import sim.nn.function.SComputeErrorFunctionGradientWeights;
import sim.nn.function.SNeuralFunction;
import sim.nn.function.activation.SActivationFunctionBuilder;
import sim.nn.function.activation.SBatchActivationFunction;
import sim.nn.function.aggregation.SAggregationFunction;
import sim.nn.function.aggregation.SAggregationFunctionBuilder;
import sim.nn.function.aggregation.SBatchAggregationFunction;
import sim.nn.function.regularization.SBatchNormalizationFunction;
import sim.nn.function.regularization.SRunningMeanVariance;
import sim.nn.gradient.SGradientLayer;
import sim.nn.parameter.SActivationParameters;
import sim.nn.parameter.SBias;
import sim.nn.parameter.SContainActivationParameters;
import sim.nn.parameter.SContainBias;
import sim.nn.parameter.SContainScale;
import sim.nn.parameter.SContainShift;
import sim.nn.parameter.SContainWeights;
import sim.nn.parameter.SScale;
import sim.nn.parameter.SShift;
import sim.nn.parameter.SWeights;
import sim.util.SStringUtil;

/**
 * La classe <b>SBatchFullConnectedLayer</b> reprsente un couche de neurones pouvant s'activ par batch et contenant une fonction de normalisation de type batch-normalization.
 * 
 * @author Simon Vezina
 * @since 2018-09-15
 * @version 2022-09-13 (Version labo : Apprentissage par rseau de neurones)
 */
public class SBatchFullConnectedLayer implements SBatchLayer,
  SComputeErrorFunctionGradientWeights, 
  SComputeErrorFunctionGradientBias, 
  SComputeErrorFunctionGradientActivationParameters,
  SComputeErrorFunctionGradientScale,
  SComputeErrorFunctionGradientShift,
  SContainWeights, 
  SContainBias,
  SContainActivationParameters,
  SContainScale,
  SContainShift
{

  //--------------
  // CONSTANTES //
  //--------------
  
  /**
   * La constante <b>DEFAULT_ACTIVATION_FUNCTION</b> correspond  la fonction d'activation par dfaut tant {@value}.
   */
  final static String DEFAULT_ACTIVATION_FUNCTION = SNeuralFunction.SIGMOID_FUNCTION;
  
  /**
   * La constante <b>DEFAULT_AGGREGATION_FUNCTION</b> correspond  la fonction d'agrgation par dfaut tant {@value}.
   */
  final static String DEFAULT_AGGREGATION_FUNCTION = SNeuralFunction.WEIGHTED_SUM_FUNCTION;
  
  //-------------
  // VARIABLES //
  //-------------
  
  /**
   * ...
   */
  protected final SWeights W;
  
  /**
   * ...
   */
  protected final SBias B;
  
  /**
   * La variable <b>K</b> reprsente les paramtres d'activation de la fonction d'activation.
   * Dans l'usage de fonction d'activation simple (ex: sigmode), ce paramtres n'est pas utilis et il sera considr comme tant vide. 
   */
  protected final SActivationParameters K;
  
  /**
   * ...
   */
  private final SScale SC;
  
  /**
   * ....
   */
  private final SShift SH;
    
  /**
   * ...
   */
  private final SRunningMeanVariance running;
  
  /**
   * ...
   */
  private final SBatchAggregationFunction Z;
  
  /**
   * ...
   */
  private final SBatchActivationFunction A;
  
  /**
   * ...
   */
  private final SBatchNormalizationFunction N;
  
  //----------------
  // CONSTRUCTEUR //
  //----------------
  
  /**
   * ...
   * @param nb_input
   * @param nb_output
   */
  public SBatchFullConnectedLayer(int nb_input, int nb_output)
  {
    this(nb_input, nb_output, DEFAULT_ACTIVATION_FUNCTION);
  }

  /**
   * ...
   * 
   * @param nb_input
   * @param nb_output
   * @param activation_function_name
   */
  public SBatchFullConnectedLayer(int nb_input, int nb_output, String activation_function_name)
  {
    this(nb_input, nb_output, DEFAULT_AGGREGATION_FUNCTION, activation_function_name);
  }
  
  /**
   * ...
   * 
   * @param nb_input
   * @param nb_output
   * @param aggregation_function_name
   * @param activation_function_name
   */
  public SBatchFullConnectedLayer(int nb_input, int nb_output, String aggregation_function_name, String activation_function_name)
  {
    this.Z = new SBatchAggregationFunction(nb_input, nb_output, aggregation_function_name);
    
    this.A = new SBatchActivationFunction(nb_output, activation_function_name);
    
    this.N = new SBatchNormalizationFunction(nb_output);
    
    this.W = new SWeights(nb_input, nb_output, this.Z.getOrder());
    
    this.B = new SBias(nb_output);
    
    // Dfinition des paramtres d'activation (mme s'il n'y en a pas pour la fonction d'activation donne).
    this.K = SActivationFunctionBuilder.buildParameters(activation_function_name, nb_output);
    
    this.SC = new SScale(nb_output);
    
    this.SH = new SShift(nb_output);
    
    this.running = new SRunningMeanVariance(nb_output);
  }
  
  /**
   * ...
   * 
   * @param weights
   * @param bias
   * @param scale
   * @param shift
   * @param aggregation_function
   * @param activation_function
   * @throws SConstructorException
   */
  public SBatchFullConnectedLayer(SBufferedMatrix weights, double[] bias, double[] scale, double[] shift, String aggregation_function, String activation_function) throws SConstructorException
  {
    this(weights, bias, SBufferedMatrix.ZERO_DIMENSION_MATRIX, scale, shift, aggregation_function, activation_function);
  }
  
  /**
   * ...
   * 
   * @param weights
   * @param bias
   * @param activation_parameters
   * @param scale
   * @param shift
   * @param aggregation_function
   * @param activation_function
   * @throws SConstructorException
   */
  public SBatchFullConnectedLayer(SBufferedMatrix weights, double[] bias, SBufferedMatrix activation_parameters, double[] scale, double[] shift, String aggregation_function, String activation_function) throws SConstructorException
  {
    // En raison d'un problme technique d'hritage, je vais construire une fonction d'agrgation
    // de type non 'batch' afin d'y extraire de l'information pour mieux construire ma fonction avec 'batch'
    try {
      
      SAggregationFunction f = SAggregationFunctionBuilder.build(aggregation_function, weights);
      this.Z = new SBatchAggregationFunction(f.getInputsSize(), f.getOutputsSize(), aggregation_function);
      
    }catch(IllegalArgumentException e) {
      throw new SConstructorException("Erreur SBatchFullConnectedLayer 001 : La fonction d'agrgation n'a pas pu tre construite.", e);
    }
    
    // Construction des autres fonctions.
    this.A = new SBatchActivationFunction(this.Z.getOutputsSize(), activation_function);
    this.N = new SBatchNormalizationFunction(this.Z.getOutputsSize());
    
    // Validation du nombre de neurones pour les poids et biais.
    if(weights.getNbLine() != bias.length)
      throw new SConstructorException("Erreur SBatchFullConnectedLayer 002 : Les poids possdent " + weights.getNbLine() + " neurones (lignes) et les biais possdent " + bias.length + " neurones ce qui n'est pas identique et donc invalide.");
    
    this.W = new SWeights(weights, this.Z.getOrder());
    this.B = new SBias(bias);
    this.K = new SActivationParameters(activation_parameters);
    
    this.SC = new SScale(scale);
    this.SH = new SShift(shift);
    
    // Nous allons dfinir les valeurs de moyenne et variance mouvante par dfaut.
    this.running = new SRunningMeanVariance(this.W.getOutputSize());
  }
  
  /**
   * ...
   * 
   * @param layer
   */
  private SBatchFullConnectedLayer(SBatchFullConnectedLayer layer)
  {
    this.W = layer.W;
    this.B = layer.B;
    this.K = layer.K;
    this.SC = layer.SC;
    this.SH = layer.SH;
    this.running = layer.running;
    
    this.Z = new SBatchAggregationFunction(layer.getInputsSize(), layer.getOutputsSize(), layer.Z.getFunctionName());
    this.A = new SBatchActivationFunction(layer.getOutputsSize(), layer.A.getFunctionName());
    this.N = new SBatchNormalizationFunction(layer.getOutputsSize());
  }
    
  //------------
  // MTHODES //
  //------------
  
  @Override
  public int getInputsSize() 
  {
    return this.Z.getInputsSize();
  }

  @Override
  public int getOutputsSize() 
  {
    return this.Z.getOutputsSize();
  }

  @Override
  public int getNumberOfParameter()
  {
    return this.W.getNumberOfParameter() + this.B.getNumberOfParameter() + this.SC.getNumberOfParameter() + this.SH.getNumberOfParameter();
  }
  
  /**
   * Mthode pour obtenir le nom de la fonction d'agrgation.
   * 
   * @return Le nom de la fonction d'agrgation.
   */
  public String getAggregationFunctionName()
  {
    return this.Z.getFunctionName();
  }
  
  /**
   * Mthode pour obtenir le nom de la fonction d'activation.
   * 
   * @return Le nom de la fonction d'activation.
   */
  public String getActivationFunctionName()
  {
    return this.A.getFunctionName();
  }
  
  @Override
  public SWeights getWeights() 
  {
    return this.W;
  }
  
  @Override
  public SBias getBias()
  {
    return this.B;
  }
  
  @Override
  public SActivationParameters getActivationParameters()
  {
    return this.K;
  }
  
  @Override
  public SScale getScale()
  {
    return this.SC;
  }
  
  @Override
  public SShift getShift()
  {
    return this.SH;
  }
  
  /**
   * <p>
   * Mthode pour obtenir l'ordre de la fonction d'agrgation.
   * </p>
   * <p>
   * Par exemple, la fonction d'agrgation de la somme pondre (<i>weighted sum</i>) est une fonction
   * d'agrgation d'ordre 1.
   * </p>
   * 
   * @return l'ordre de la fonction d'agrgation.
   */
  public int getAggregationFunctionOrder()
  {
    return this.Z.getOrder();
  }
  
  @Override
  public SLayer copyWithSameParameter() 
  {
    return new SBatchFullConnectedLayer(this);
  }

  @Override
  public List<double[]> batchForwardPropagation(int b, List<double[]> batch_inputs) throws IllegalArgumentException 
  {
    throw new SNoImplementationException();
    
  }

  @Override
  public List<double[]> batchBackPropagation(int b, List<double[]> batch_delta) throws IllegalArgumentException 
  {
    throw new SNoImplementationException();
    
  }

  @Override
  public int getBatchSize()
  {
    return this.N.getBatchSize();
  }

  @Override
  public String toString()
  {
    return getDescription();
  }
  
  @Override
  public void learn(SGradientLayer gradient, double learning_rate) throws IllegalArgumentException
  {
    this.W.learn(gradient, learning_rate);
    this.B.learn(gradient, learning_rate);
    this.K.learn(gradient, learning_rate);
    this.SC.learn(gradient, learning_rate);
    this.SH.learn(gradient, learning_rate);
  }
    
  @Override
  public double dCdW(int v, int u, int o)
  {
    return this.Z.dCdW(v, u, o);
  }
  
  @Override
  public double dCdB(int u)
  {
    return this.Z.dCdB(u);
  }
  
  @Override
  public double dCdK(int i, int u) throws IllegalArgumentException
  {
    return this.A.dCdK(i, u);
  }
  
  @Override
  public double dCdSC(int u)
  {
    return this.N.dCdSC(u);
  }
  
  @Override
  public double dCdSH(int u)
  {
    return this.N.dCdSH(u);
  }

  @Override
  public String getLayerName()
  {
    return SLayer.BATCH_FULL_CONNECTED_LAYER;
  }
  
  @Override
  public String getDescription()
  {
    return this.getLayerName() + " [ Dimension = " + this.getDimension() + ", Z = " + this.Z + " , N = " + this.N + ", A = " + this.A + "]" + SStringUtil.END_LINE_CARACTER;
  }

  @Override
  public void setBatchSize(int b) throws IllegalArgumentException
  {
    this.Z.setBatchSize(b);
    this.A.setBatchSize(b);
    this.N.setBatchSize(b);
  }

  @Override
  public int getMaximumBatchSize()
  {
    return this.Z.getMaximumBatchSize();
  }

  @Override
  public boolean upgradeMaximumBatchSize(int size)
  {
    this.Z.upgradeMaximumBatchSize(size);
    this.A.upgradeMaximumBatchSize(size);
    
    return this.N.upgradeMaximumBatchSize(size);
  }
  
}
