/**
 * 
 */
package sim.nn.layer;

import sim.exception.SConstructorException;
import sim.exception.SNoImplementationException;
import sim.nn.function.SComputeErrorFunctionGradientActivationParameters;
import sim.nn.function.SComputeErrorFunctionGradientBias;
import sim.nn.function.SComputeErrorFunctionGradientFiltersWeights;
import sim.nn.function.activation.SActivationFunction;
import sim.nn.function.activation.SActivationFunctionBuilder;
import sim.nn.function.filtering.SConvolutionFunction;
import sim.nn.gradient.SGradientLayer;
import sim.nn.parameter.SActivationParameters;
import sim.nn.parameter.SBias;
import sim.nn.parameter.SContainActivationParameters;
import sim.nn.parameter.SContainBias;
import sim.nn.parameter.SContainFiltersWeights;
import sim.nn.parameter.SFiltersWeights;
import sim.util.SStringUtil;

/**
 * La classe <b>SConvolutionLayer</b> reprsente une couche de type convolution.
 * 
 * @author Simon Vezina
 * @since 2020-01-12
 * @version 2022-09-13 (Version labo : Apprentissage par rseau de neurones)
 */
public class SConvolutionLayer implements SLayer,
  SComputeErrorFunctionGradientFiltersWeights, 
  SComputeErrorFunctionGradientBias,
  SComputeErrorFunctionGradientActivationParameters,
  SContainFiltersWeights, 
  SContainBias,
  SContainActivationParameters
{

  /**
   * ...
   */
  protected final SFiltersWeights W;
  
  /**
   * ...
   */
  protected final SBias B;
  
  /**
   * ...
   */
  protected final SActivationParameters K;
  
  /**
   * ...
   */
  protected final SConvolutionFunction Z;
  
  /**
   * ...
   */
  protected final SActivationFunction A;
  
  //----------------
  // CONSTRUCTEUR //
  //----------------
  
  /**
   * Constructeur d'une couche de neurones avec
   * 
   * @param input
   * @param input_depth
   * @param nb_filter
   * @param filter_offset
   * @param padding
   * @param stride
   * @param activation_function
   * @throws SConstructorException
   */
  public SConvolutionLayer(int input, int input_depth, int nb_filter, int filter_offset, int padding, int stride, String activation_function) throws SConstructorException
  {
    // Dfinition de la fonction de convolution et la taille de la couche.
    this.Z = new SConvolutionFunction(input, input_depth, nb_filter, filter_offset, padding, stride);
    
    // Initialisation de la fonction d'activation.
    try{
      this.A = SActivationFunctionBuilder.build(activation_function, this.getOutputsSize());
    }catch(IllegalArgumentException e){
      throw new SConstructorException("Erreur SConvolutionLayer 001 : La fonction d'activation '" + activation_function + "' n'a pas t initialise.", e);
    }
    
    // Dfinition de la matrice des poids en fonction de l'ordre de la fonction d'agrgation.
    this.W = new SFiltersWeights(filter_offset, input_depth, nb_filter);
    
    // Dfinition du vecteur des biais.
    this.B = new SBias(nb_filter); 
    
    // Dfinition des paramtres d'activation
    this.K = SActivationFunctionBuilder.buildParameters(activation_function, this.getOutputsSize());
  }

  /**
   * ...
   * 
   * @param filters
   * @param bias
   * @param input_size
   * @param padding
   * @param stride
   * @param activation_function_name
   * @throws SConstructorException
   */
  public SConvolutionLayer(SFiltersWeights filters, SBias bias, int input_size, int padding, int stride, String activation_function_name) throws SConstructorException
  {
    this(filters, bias, SActivationParameters.EMPTY_ACTIVATION_PARAMETERS, input_size, padding, stride, activation_function_name);
  }
  
  /**
   * ...
   * 
   * @param filters
   * @param bias
   * @param activation_parameters
   * @param input_size
   * @param padding
   * @param stride
   * @param activation_function_name
   * @throws SConstructorException
   */
  public SConvolutionLayer(SFiltersWeights filters, SBias bias, SActivationParameters activation_parameters, int input_size, int padding, int stride, String activation_function_name) throws SConstructorException
  {
    // Dfinition de la fonction de convolution et la taille de la couche.
    this.Z = new SConvolutionFunction(input_size, filters.getInputDepth(), filters.getOutputDepth(), filters.getOffset(), padding, stride);
    
    // Initialisation de la fonction d'activation.
    try{
      this.A = SActivationFunctionBuilder.build(activation_function_name, this.getOutputsSize());
    }catch(IllegalArgumentException e){
      throw new SConstructorException("Erreur SConvolutionLayer 002 : La fonction d'activation '" + activation_function_name + "' n'a pas t initialise.", e);
    }
    
    this.W = filters;
    this.B = bias;
    this.K = activation_parameters;
        
    // Validation que le paramtre d'activation concorde avec le format de la fonction d'activation.
    if(this.K.getNbParameterType() != this.A.getNbParameterType())
      throw new SConstructorException("Erreur SConvolutionLayer 003 : Le paramtre d'activation contient " + this.K.getNbParameterType() + " ce qui n'est pas adquat pour le choix de la fonction d'activation '" + this.A.getFunctionName() + "' qui doit avoir " + this.A.getNbParameterType() + " type de paramtre."); 
  }
  
  /**
   * ...
   * 
   * @param layer
   */
  private SConvolutionLayer(SConvolutionLayer layer)
  {
    this.W = layer.W;
    this.B = layer.B;
    this.K = layer.K;
    
    // Dfinition de la fonction de convolution et la taille de la couche.
    this.Z = new SConvolutionFunction(layer.getInputsSize(), layer.getInputsDepth(), layer.getOutputsDepth(), layer.W.getOffset(), layer.Z.getPadding(), layer.Z.getStride());
    
    // Dfinir la fonction d'activation.
    this.A = SActivationFunctionBuilder.build(layer.A.getFunctionName(), layer.getOutputsSize());
  }
  
  //------------
  // MTHODES //
  //------------
  
  /**
   * Mthode pour obtenir la taille du filtre dans la convolution.
   * 
   * @return La taille du filtre.
   */
  public int getFilterOffset()
  {
    return this.W.getOffset();
  }
  
  @Override
  public int getInputsSize() 
  {
    return this.Z.getInputsSize();
  }

  @Override
  public int getOutputsSize() 
  {
    return this.Z.getOutputsSize();
  }

  @Override
  public int getInputsDepth()
  {
    return this.Z.getInputsDepth();
  }
    
  @Override
  public int getOutputsDepth()
  {
    return this.Z.getOutputsDepth();
  }
  
  /**
   * Mthode pour obtenir le nom de la fonction d'activation.
   * 
   * @return Le nom de la fonction d'activation.
   */
  public String getActivationFunctionName()
  {
    return this.A.getFunctionName();
  }
  
  @Override
  public SFiltersWeights getFiltersWeights()
  {
    return this.W;
  }
  
  @Override
  public SBias getBias()
  {
    return this.B;
  }
  
  @Override
  public SActivationParameters getActivationParameters() 
  {
    return this.K;
  }
  
  /**
   * ...
   * 
   * @return
   */
  public int getStride()
  {
    return this.Z.getStride();
  }
  
  /**
   * ...
   * 
   * @return
   */
  public int getPadding()
  {
    return this.Z.getPadding();
  }
    
  @Override
  public double[] forwardPropagation(double[] input) throws IllegalArgumentException 
  {
    throw new SNoImplementationException();
    
  }

  @Override
  public double[] backPropagation(double[] delta) throws IllegalArgumentException
  {
    throw new SNoImplementationException();
    
  }

  @Override
  public double dCdW(int f, int d, int i, int j)
  {
    return this.Z.dCdW(f, d, i, j);
  }
  
  @Override
  public double dCdB(int f)
  {
    return this.Z.dCdB(f);
  }
  
  @Override
  public double dCdK(int i, int u) throws IllegalArgumentException 
  {
    return this.A.dAdK(i, u);
  }
  
  @Override
  public int getNumberOfParameter()
  {
    return this.W.getNumberOfParameter() + this.B.getNumberOfParameter() + this.K.getNumberOfParameter();
  }
  
  @Override
  public void learn(SGradientLayer gradient, double learning_rate) throws IllegalArgumentException
  {
    this.W.learn(gradient, learning_rate);
    this.B.learn(gradient, learning_rate);
    this.K.learn(gradient, learning_rate);
  }

  @Override
  public SLayer copyWithSameParameter() 
  {
    return new SConvolutionLayer(this);
  }

  @Override
  public String getLayerName()
  {
    return SLayer.CONVOLUTION_LAYER;
  }
  
  @Override
  public String getDescription()
  {
    StringBuffer buffer = new StringBuffer();
    
    buffer.append(this.getLayerName() + " [ Dimension = " + this.getDimension() + ", Z = " + Z + ", A = " + A +"]" + SStringUtil.END_LINE_CARACTER);
    
    return buffer.toString();
  }
  
  @Override
  public String getDimension() 
  {
    return this.getInputsSize() + " x " + this.getOutputsSize() + " --> (" + this.Z.getInputsDepth() + " x " + this.Z.getInputOffset() + " x " + this.Z.getInputOffset() + ") x (" + this.Z.getOutputsDepth() + " x " + this.Z.getOutputOffset() + " x " + this.Z.getOutputOffset() + ")";
  }

  @Override
  public String toString()
  {
    return getDescription();
  }
  
}
