/**
 * 
 */
package sim.nn.layer;

import sim.exception.SConstructorException;
import sim.exception.SNoImplementationException;
import sim.math.SBufferedMatrix;
import sim.nn.function.SComputeErrorFunctionGradientActivationParameters;
import sim.nn.function.SComputeErrorFunctionGradientBias;
import sim.nn.function.SComputeErrorFunctionGradientWeights;
import sim.nn.function.SNeuralFunction;
import sim.nn.function.activation.SActivationFunction;
import sim.nn.function.activation.SActivationFunctionBuilder;
import sim.nn.function.aggregation.SAggregationFunction;
import sim.nn.function.aggregation.SAggregationFunctionBuilder;
import sim.nn.gradient.SGradientLayer;
import sim.nn.parameter.SActivationParameters;
import sim.nn.parameter.SBias;
import sim.nn.parameter.SContainActivationParameters;
import sim.nn.parameter.SContainBias;
import sim.nn.parameter.SContainWeights;
import sim.nn.parameter.SWeights;
import sim.util.SStringUtil;

/**
 * <p>
 * La classe <b>SFullConnectedLayer</b> reprsente une couches de neurones de type <b><i>Full connected</i></b>.
 * Cela signifie que l'ensemble des entres sont connects  l'aide d'un poids  un neurone de sortie o une fonction d'activation
 * sera applique avant la sortie de la couche.
 * </p>
 * <p>
 * Par dfaut, la fonction d'agrgation utilise sera la somme pondre (<i>weighted sum</i>) et la fonction d'activation
 * sera la fonction sigmod.
 * </p>
 * 
 * @author Simon Vzina
 * @since 2018-02-17
 * @version 2022-09-14 (Version labo : Apprentissage par rseau de neurones)
 */
public class SFullConnectedLayer implements SLayer, 
  SComputeErrorFunctionGradientWeights, 
  SComputeErrorFunctionGradientBias, 
  SComputeErrorFunctionGradientActivationParameters,
  SContainWeights, 
  SContainBias,
  SContainActivationParameters
{

  //--------------
  // CONSTANTES //
  //--------------
  
  /**
   * La constante <b>DEFAULT_AGGREGATION_FUNCTION</b> correspond  la fonction d'agrgation par dfaut tant {@value}.
   */
  public final static String DEFAULT_AGGREGATION_FUNCTION = SNeuralFunction.WEIGHTED_SUM_FUNCTION;
  
  /**
   * La constante <b>DEFAULT_ACTIVATION_FUNCTION</b> correspond  la fonction d'activation par dfaut tant {@value}.
   */
  public final static String DEFAULT_ACTIVATION_FUNCTION = SNeuralFunction.SIGMOID_FUNCTION;
  
  //-------------
  // VARIABLES //
  //-------------
  
  /**
   * <p>
   * La variable <b>W</b> reprsente les poids (<i>weight</i>) de tous les neurones de la couche.
   * Chaque <u>ligne</u> de cette reprsentation matricielle correspond  <u>un neurone</u> de la couche o les poids serviront  pondr l'entre (<i>input</i>) lors des calculs.
   * </p>
   * 
   * Ainsi, une ligne <i>u</i> de la matrice est associe au neurone <i>u</i> de la couche.
   */
  protected final SWeights W;
  
  /**
   * <p>
   * La variable <b>B</b> reprsente le biais (<i>bias</i>) de tous les neurones de la couche.
   * Chaque <u>lment</u> de cette reprsentation en vecteur correspond  <u>un neurone</u> de la couche o les biais serviront  corriger l'aggrgation de l'entre avec les poids lors des calculs.
   * </p>
   * 
   * Ainsi, l'indice <i>u</i> du vecteur est associ au neurone <i>u</i> de la couche.
   */
  protected final SBias B;
   
  /**
   * La variable <b>K</b> reprsente les paramtres d'activation de la fonction d'activation.
   * Dans l'usage de fonction d'activation simple (ex: sigmode), ce paramtres n'est pas utilis et il sera considr comme tant vide. 
   */
  protected final SActivationParameters K;
  
  /**
   * La variable <b>Z<b> reprsente la fonction d'aggrgation de la couche de neurones.
   */
  protected final SAggregationFunction Z;
  
  /**
   * La variable <b>A</b> reprsente la fonction d'activation de la couche de neurones.
   */
  protected final SActivationFunction A;
    
  //----------------
  // CONSTRUCTEUR //
  //----------------
  
  /**
   * Constructeur d'une couche de neurones par dfaut. La couche aura comme fonction d'activation <b>sigmode</b>.
   * 
   * @param nb_input Le nombre de neurones en entre.
   * @param nb_output Le nombre de neurones en sortie.
   */
  public SFullConnectedLayer(int nb_input, int nb_output)
  {
    this(nb_input, nb_output, DEFAULT_ACTIVATION_FUNCTION);
  }
  
  /**
   * Constructeur d'une couche de neurones.
   * 
   * @param nb_input Le nombre de neurones en entre.
   * @param nb_output Le nombre de neurones en sortie.
   * @param activation_function La fonction d'activation.
   */
  public SFullConnectedLayer(int nb_input, int nb_output, String activation_function) throws SConstructorException
  {
    this(nb_input, nb_output, DEFAULT_AGGREGATION_FUNCTION, activation_function);
  }
  
  /**
   * Constructeur d'une couche de neurones.
   * 
   * @param nb_input Le nombre de neurones en entre.
   * @param nb_output Le nombre de neurones en sortie.
   * @param aggregation_function La fonction d'agrgation.
   * @param activation_function La fonction d'activation.
   */
  public SFullConnectedLayer(int nb_input, int nb_output, String aggregation_function, String activation_function) throws SConstructorException
  {
    // Initialisation de la fonction d'agrgation.
    try{
      this.Z = SAggregationFunctionBuilder.build(aggregation_function, nb_input, nb_output);
    }catch(IllegalArgumentException e){
      throw new SConstructorException("Erreur SFullConnectedLayer 001 : La fonction d'agrgation '" + aggregation_function + "' n'a pas t initialise.", e);
    } 
       
    // Initialisation de la fonction d'activation.
    try{
      this.A = SActivationFunctionBuilder.build(activation_function, nb_output);
    }catch(IllegalArgumentException e){
      throw new SConstructorException("Erreur SFullConnectedLayer 002 : La fonction d'activation '" + activation_function + "' n'a pas t initialise.", e);
    }
    
    // Dfinition de la matrice des poids en fonction de l'ordre de la fonction d'agrgation.
    this.W = new SWeights(nb_input, nb_output, this.Z.getOrder());
    
    // Dfinition du vecteur des biais.
    this.B = new SBias(nb_output);
    
    // Dfinition des paramtres d'activation (mme s'il n'y en a pas pour la fonction d'activation donne).
    this.K = SActivationFunctionBuilder.buildParameters(activation_function, nb_output);
  }
      
  /**
   * Constructeur d'une couche de neurones dj entran.
   * 
   * @param weights Les poids de la couche de neurones.
   * @param bias Les biais de la couche de neurones.
   * @param activation_function Le nom de la fonction d'activation.
   * @throws SConstructorException S'il y a eu une erreur lors de la construction de l'objet.
   */
  public SFullConnectedLayer(SBufferedMatrix weights, double[] bias, String activation_function) throws SConstructorException
  {
    this(weights, bias, DEFAULT_AGGREGATION_FUNCTION, activation_function);
  }
  
  /**
   * ...
   * 
   * @param weights
   * @param bias
   * @param aggregation_function
   * @param activation_function
   * @throws SConstructorException
   */
  public SFullConnectedLayer(SBufferedMatrix weights, double[] bias, String aggregation_function, String activation_function) throws SConstructorException
  {
    this(weights, bias, SBufferedMatrix.ZERO_DIMENSION_MATRIX, aggregation_function, activation_function);
  }
  
  /**
   * ...
   * 
   * @param weights
   * @param bias
   * @param activation_parameters
   * @param aggregation_function
   * @param activation_function
   * @throws SConstructorException
   */
  public SFullConnectedLayer(SBufferedMatrix weights, double[] bias, SBufferedMatrix activation_parameters, String aggregation_function, String activation_function) throws SConstructorException
  {
    // Initialisation de la fonction d'agrgation.
    try{
      this.Z = SAggregationFunctionBuilder.build(aggregation_function, weights);
    }catch(IllegalArgumentException e){
      throw new SConstructorException("Erreur SFullConnectedLayer 003 : La fonction d'agrgation '" + aggregation_function + "' n'a pas t initialise.", e);
    }
        
    // Initialisation de la fonction d'activation.
    try{
      this.A = SActivationFunctionBuilder.build(activation_function, bias.length);
    }catch(IllegalArgumentException e){
      throw new SConstructorException("Erreur SFullConnectedLayer 004 : La fonction d'activation '" + activation_function + "' n'a pas t initialise.", e);
    }
    
    this.W = new SWeights(weights, this.Z.getOrder());
    this.B = new SBias(bias);
    this.K = new SActivationParameters(activation_parameters);
    
    // Validation du nombre de neurones pour les poids et biais.
    if(this.W.getOutputSize() != this.B.getNeurons())
      throw new SConstructorException("Erreur SFullConnectedLayer 005 : Les poids possdent " + this.W.getOutputSize() + " neurones (lignes) et les biais possdent " + this.B.getNeurons() + " neurones ce qui n'est pas identique et donc invalide.");
    
    // Validation que le paramtre d'activation concorde avec le format de la fonction d'activation.
    if(this.K.getNbParameterType() != this.A.getNbParameterType())
      throw new SConstructorException("Erreur SFullConnectedLayer 006 : Le paramtre d'activation contient " + this.K.getNbParameterType() + " ce qui n'est pas adquat pour le choix de la fonction d'activation '" + this.A.getFunctionName() + "' qui doit avoir " + this.A.getNbParameterType() + " type de paramtre.");
  }
  
  /**
   * ...
   * 
   * @param W
   * @param B
   * @param K
   * @param aggregation_function
   * @param activation_function
   */
  private SFullConnectedLayer(SWeights W, SBias B, SActivationParameters K, String aggregation_function, String activation_function)
  {
    // Copier la rfrence des paramtres pour utiliser les mmes.
    this.W = W;
    this.B = B;
    this.K = K;
    
    // Initialisation de la fonction d'agrgation.
    try{
      this.Z = SAggregationFunctionBuilder.build(aggregation_function, W.getInputSize(), W.getOutputSize());
    }catch(IllegalArgumentException e){
      throw new SConstructorException("Erreur SFullConnectedLayer 007 : La fonction d'agrgation '" + aggregation_function + "' n'a pas t initialise.", e);
    } 
        
    // Initialisation de la fonction d'activation.
    try{
      this.A = SActivationFunctionBuilder.build(activation_function, W.getOutputSize());
    }catch(IllegalArgumentException e){
      throw new SConstructorException("Erreur SFullConnectedLayer 008 : La fonction d'activation '" + activation_function + "' n'a pas t initialise.", e);
    } 
  }
  
  //------------
  // MTHODES //
  //------------
      
  @Override
  public int getInputsSize()
  {
    return this.Z.getInputsSize();
  }
  
  @Override
  public int getOutputsSize() 
  {
    return this.Z.getOutputsSize();
  }
  
  /**
   * <p>
   * Mthode pour obtenir l'ordre de la fonction d'agrgation.
   * </p>
   * <p>
   * Par exemple, la fonction d'agrgation de la somme pondre (<i>weighted sum</i>) est une fonction
   * d'agrgation d'ordre 1.
   * </p>
   * 
   * @return l'ordre de la fonction d'agrgation.
   */
  public int getAggregationFunctionOrder()
  {
    return this.Z.getOrder();
  }
  
  @Override
  public SWeights getWeights() 
  {
    return this.W;
  }
  
  @Override
  public SBias getBias()
  {
    return this.B;
  }
  
  @Override
  public SActivationParameters getActivationParameters() 
  {
    return this.K;
  }
  
  @Override
  public int getNumberOfParameter()
  {
    return this.W.getNumberOfParameter() + this.B.getNumberOfParameter() + this.K.getNumberOfParameter();
  }

  /**
   * Mthode pour obtenir le nom de la fonction d'agrgation.
   * 
   * @return Le nom de la fonction d'agrgation.
   */
  public String getAggregationFunctionName()
  {
    return this.Z.getFunctionName();
  }
  
  /**
   * Mthode pour obtenir le nom de la fonction d'activation.
   * 
   * @return Le nom de la fonction d'activation.
   */
  public String getActivationFunctionName()
  {
    return this.A.getFunctionName();
  }

  /**
   * <p>
   * Mthode pour gnrer l'quation linaire correspondant au neurone i.
   * </p>
   * 
   * <p>
   * L'quation reprsent sera : w_i0 * a_0 + w_i1 * a_1 + ... + w_in * a_n + b_i = 0.
   * Le tableau contiendra alors l'ensemble des coefficients w et le biais en dernier paramtre.
   * </p>
   * 
   * @param i Le choix du neurone
   * @return L'quation linaire du neurone.
   */
  public double[] getEquation(int u)
  {
    double[] array = new double[this.getInputsSize() + 1];
    
    for(int v = 0; v < this.getInputsSize(); v++)
      array[v] = this.W.get(v, u);
    
    // Mettre le bias dans la dernire case.
    array[this.getInputsSize()] = this.B.get(u);
    
    return array;
  }
  
  @Override
  public double[] forwardPropagation(double[] input) throws IllegalArgumentException
  {
    throw new SNoImplementationException();
  
  }
  
  @Override
  public double[] backPropagation(double[] delta) throws IllegalArgumentException
  {
    throw new SNoImplementationException();
    
  }
    
  @Override
  public void learn(SGradientLayer gradient, double learning_rate) throws IllegalArgumentException
  {
    this.W.learn(gradient, learning_rate);
    this.B.learn(gradient, learning_rate);
    this.K.learn(gradient, learning_rate);
  }
  
  @Override
  public double dCdW(int v, int u)
  {
    return this.Z.dCdW(v, u);
  }
  
  @Override
  public double dCdW(int v, int u, int o)
  {
    return this.Z.dCdW(v, u, o);
  }
  
  @Override
  public double dCdB(int u)
  {
    return this.Z.dCdB(u);
  }
   
  @Override
  public double dCdK(int i, int u) throws IllegalArgumentException 
  {
    return this.A.dCdK(i, u);
  }
  
  @Override
  public String getLayerName()
  {
    return SLayer.FULL_CONNECTED_LAYER;
  }
  
  @Override
  public String getDescription()
  {
    StringBuffer buffer = new StringBuffer();
    
    buffer.append(this.getLayerName() + " [ Dimension = " + this.getDimension() + ", Z = " + Z + ", A = " + A +"]" + SStringUtil.END_LINE_CARACTER);
    
    return buffer.toString();
  }

  @Override
  public String toString()
  {
    return getDescription();
  }
  
  @Override
  public SLayer copyWithSameParameter()
  {
    return new SFullConnectedLayer(this.W, this.B, this.K, this.Z.getFunctionName(), this.A.getFunctionName());
  }
  
}
