/**
 * 
 */
package sim.nn.network;

import java.util.List;

import sim.exception.SConstructorException;
import sim.nn.layer.SBatchFullConnectedLayer;
import sim.nn.layer.SLayer;

/**
 * La classe <b>SBatchFullConnectedNetwork</b> reprsente un rseau de type <i>full connected<i/>
 * utilisant une fonction de normalisation portant le nom de <i>batch-normalization</i>
 * ncessitant une action des donnes par batch (<i>batch-activation</i>).
 * 
 * @author Simon Vezina
 * @since 2018-09-15
 * @version 2021-09-22
 */
public class SBatchFullConnectedNetwork implements SBatchNetwork {

  /**
   * La variable <b>layers</b> reprsente les couches du rseau de neurones. 
   */
  private final SBatchFullConnectedLayer[] layers;
   
  //----------------
  // CONSTRUCTEUR //
  //----------------
  
  /**
   * ...
   * 
   * @param neurons
   */
  public SBatchFullConnectedNetwork(int... neurons) 
  {
    if(neurons.length < 2)
      throw new SConstructorException("Erreur SBatchFullConnectedNetwork 001 : Le nombre de couche de neurones du rseau est infrieur  2 (gal  " + neurons.length + ").");
    
    // Identification du nombre de couche dans le rseau.
    this.layers = new SBatchFullConnectedLayer[neurons.length-1];
   
    // Construction des couches de neurones (avec fonction d'activation par dfaut).
    for(int i = 0; i < layers.length; i++)
      this.layers[i] = new SBatchFullConnectedLayer(neurons[i], neurons[i+1]);   
    
    if(!isValidLayersSequence())
      throw new SConstructorException("Erreur SBatchFullConnectedNetwork 002 : La squence des couches est invalide, car " + getDimension() + ".");  
  }

  /**
   * ...
   * 
   * @param l
   * @throws SConstructorException
   */
  public SBatchFullConnectedNetwork(SBatchFullConnectedLayer... l) throws SConstructorException
  {
    if(l.length < 1)
      throw new SConstructorException("Erreur SBatchFullConnectedNetwork 003 : Le nombre de couche L = " + l.length + " est infrieur  1.");
       
    // Identification du nombre de couche dans le rseau.
    this.layers = new SBatchFullConnectedLayer[l.length];
   
    // Affectation des couches dans le rseau.
    for(int i = 0; i < l.length; i++)
      this.layers[i] = l[i];
    
    // Vrifier si la squence de couche est valide.
    if(!isValidLayersSequence())
      throw new SConstructorException("Erreur SBatchFullConnectedNetwork 004 : La squence des couches est invalide, car " + getDimension() + ".");  
  }
  
  /**
   * 
   * @param l
   * @throws SConstructorException
   */
  public SBatchFullConnectedNetwork(List<SBatchFullConnectedLayer> layer_list) throws SConstructorException
  {
    if(layer_list.size() < 1)
      throw new SConstructorException("Erreur SBatchFullConnectedNetwork 005 : Le nombre de couche L = " + layer_list.size() + " est infrieur  1.");
    
    // Identification du nombre de couche dans le rseau.
    this.layers = new SBatchFullConnectedLayer[layer_list.size()];
   
    // Affectation des couches dans le rseau.
    for(int i = 0; i < this.layers.length; i++)
      this.layers[i] = layer_list.get(i);
    
    // Vrifier si la squence de couche est valide.
    if(!isValidLayersSequence())
      throw new SConstructorException("Erreur SBatchFullConnectedNetwork 006 : La squence des couches est invalide, car " + getDimension() + "."); 
  }
  
  @Override
  public SLayer[] getLayers() 
  {
    return this.layers;
  }

  
  @Override
  public SNetwork copyWithSameParameter()
  {
    SBatchFullConnectedLayer[] copy_layers = new SBatchFullConnectedLayer[this.getNbLayers()];
    
    for(int i = 0 ; i < copy_layers.length; i++)
      copy_layers[i] = (SBatchFullConnectedLayer) this.layers[i].copyWithSameParameter();
    
    return new SBatchFullConnectedNetwork(copy_layers);
  }

  @Override
  public String getNetworkName()
  {
    return SNetwork.FULL_CONNECTED_BATCH_NETWORK;
  }
  
  @Override
  public String toString()
  {
    return getDescription();
  }
  
}
