/**
 * 
 */
package sim.nn.network;

import java.util.List;
import java.util.StringTokenizer;

import sim.exception.SConstructorException;
import sim.nn.layer.SFullConnectedLayer;
import sim.nn.layer.SLayer;

/**
 * La classe <b>SFullConnectedNetwork</b> reprsente un rseau de neurones contenant des couches de type <b>full connected</b>.
 * 
 * @author Simon Vzina
 * @since 2018-02-18
 * @version 2022-01-08
 */
public class SFullConnectedNetwork implements SNetwork {

  /**
   * ...
   */
  private static final String DEFAULT_ACTIVATION_FUNCTION = "sigmoid";
  
  /**
   * La variable <b>layers</b> reprsente les couches du rseau de neurones. 
   */
  protected final SFullConnectedLayer[] layers;
   
  //-----------------
  // CONSTRUCTEURS //
  //-----------------
  
  /**
   * Constructeur d'un rseau de neurones de type FullConnected. Toutes les couches auront comme fonction d'activation <b>sigmode</b>.
   * 
   * @param expression L'expression dcriant le rseau (ex : 4 x 7 x 3).
   * @throws SConstructorException S'il y a eu une erreur lors de la construction du rseau.
   * @throws IllegalArgumentException Si l'expression n'est pas valide.
   */
  public SFullConnectedNetwork(String expression) throws SConstructorException, IllegalArgumentException
  {
    this(networkExpressionToArray(expression));
  }
  
  /**
   * ...
   * 
   * @param expression
   * @param activation_function_name
   * @throws SConstructorException
   * @throws IllegalArgumentException
   */
  public SFullConnectedNetwork(String expression, String activation_function_name) throws SConstructorException, IllegalArgumentException
  {
    this(activation_function_name, networkExpressionToArray(expression));
  }
  
  /**
   * Constructeur d'un rseau de neurones de type FullConnected. Toutes les couches auront comme fonction d'activation <b>sigmode</b>.
   * 
   * @param neurons La taille des couches de neurones en squence.
   * @throws SConstructorException Si le nombre de couche de neurones est infrieur  2.
   */
  public SFullConnectedNetwork(int... neurons) throws SConstructorException
  {
    this(DEFAULT_ACTIVATION_FUNCTION, neurons);
  }
  
  /**
   * Constructeur d'un rseau de neurones de type FullConnected. Toutes les couches auront comme fonction d'activation <b>sigmode</b>.
   * 
   * @param neurons La taille des couches de neurones en squence.
   * @throws SConstructorException Si le nombre de couche de neurones est infrieur  2.
   */
  public SFullConnectedNetwork(String activation_function_name, int... neurons) throws SConstructorException
  {
    if(neurons.length < 2)
      throw new SConstructorException("Erreur SFullConnectedNetwork 001 : Un rseau de taille 1 doit avoir 2 paramtres (entre x sortie). Votre dfinition de rseau contient prsentement " + neurons.length + " paramtre.");
    
    // Allocation de la mmoire pour le nombre de couches.
    this.layers = new SFullConnectedLayer[neurons.length-1];
    
    // Construction des couches de neurones (avec fonction d'activation par dfaut).
    for(int i = 0; i < layers.length; i++)
      this.layers[i] = new SFullConnectedLayer(neurons[i], neurons[i+1], activation_function_name);   
    
    if(!isValidLayersSequence())
      throw new SConstructorException("Erreur SFullConnectedNetwork 002 : La squence des couches est invalide, car " + getDimension() + ".");  
  }
  
  /**
   * Construction d'un rseau de neurones  partir d'une squence de couche de neurones.
   * 
   * @param l La couche du rseau.
   * @throws SConstructorException Si le nombre de couche de neurones est infrieur  2.
   * @throws SConstructorException Si la squence des couches de neurones est invalide.
   */
  public SFullConnectedNetwork(SFullConnectedLayer... l) throws SConstructorException
  {
    if(l.length < 1)
      throw new SConstructorException("Erreur SFullConnectedNetwork 003 : Le nombre de couche L = " + l.length + " est infrieur  1.");
       
    // Identification du nombre de couche dans le rseau.
    this.layers = new SFullConnectedLayer[l.length];
   
    // Affectation des couches dans le rseau.
    for(int i = 0; i < l.length; i++)
      this.layers[i] = l[i];
    
    // Vrifier si la squence de couche est valide.
    if(!isValidLayersSequence())
      throw new SConstructorException("Erreur SFullConnectedNetwork 004 : La squence des couches est invalide, car " + getDimension() + ".");  
  }
  
  /**
   * Constructeur d'un rseau de neurones  partir d'une liste de couches.
   * 
   * @param layer_list La liste des couches.
   */
  public SFullConnectedNetwork(List<SFullConnectedLayer> layer_list)
  {
    if(layer_list.size() < 1)
      throw new SConstructorException("Erreur SFullConnectedNetwork 005 : Le nombre de couche L = " + layer_list.size() + " est infrieur  1.");
    
    // Identification du nombre de couche dans le rseau.
    this.layers = new SFullConnectedLayer[layer_list.size()];
   
    // Affectation des couches dans le rseau.
    for(int i = 0; i < this.layers.length; i++)
      this.layers[i] = layer_list.get(i);
    
    // Vrifier si la squence de couche est valide.
    if(!isValidLayersSequence())
      throw new SConstructorException("Erreur SFullConnectedNetwork 006 : La squence des couches est invalide, car " + getDimension() + "."); 
  }
  
  //------------
  // MTHODES //
  //------------
  
  @Override
  public SLayer[] getLayers()
  {
    return this.layers;
  }
     
  @Override
  public SNetwork copyWithSameParameter()
  {
    SFullConnectedLayer[] copy_layers = new SFullConnectedLayer[this.getNbLayers()];
    
    for(int i = 0 ; i < copy_layers.length; i++)
      copy_layers[i] = (SFullConnectedLayer) this.layers[i].copyWithSameParameter();
    
    return new SFullConnectedNetwork(copy_layers);
  }
  
  @Override
  public String getNetworkName()
  {
    return SNetwork.FULL_CONNECTED_NETWORK;
  }
  
  @Override
  public String toString()
  {
    return getDescription();
  }
  
  /**
   * Mthode pour convertir une expression dcrivant un rseau en un tableau de dimension du rseau.
   * Une expression valide aura la forme suivante : "4 x 7 x 9 x 2".
   * 
   * @param expression L'expression dcrivant le rseau.
   * @return Le tableau des dimensions du rseau.
   * @throws IllegalArgumentException Si l'expression n'est pas valide.
   */
  private static int[] networkExpressionToArray(String expression) throws IllegalArgumentException
  {
    // Obtenir les tokens de l'expression du rseau (excluant les caractres, " ", "x" et "X").
    StringTokenizer tokens = new StringTokenizer(expression," xX");
    
    // Remplir le tableau des dimensions du rseau.
    int[] dimension = new int[tokens.countTokens()];
    
    // Validation de l'expression.
    try {
      
      int i = 0;
      
      while(tokens.hasMoreTokens())
      {
        dimension[i] = Integer.parseInt(tokens.nextToken());
        i++;
      }
    }catch(NumberFormatException e) {
      throw new IllegalArgumentException("Erreur SFullConnectedNetwork 007 : L'expression " + expression + " n'est pas valide pour dcrire le rseau.", e);
    }
    
    return dimension;
  }
  
}
