/**
 * 
 */
package sim.nn.network;

import java.util.ArrayList;
import java.util.List;

import sim.exception.SNoImplementationException;
import sim.nn.SBatchActivable;
import sim.nn.gradient.SErrorFunctionGradient;
import sim.nn.layer.SBatchLayer;
import sim.nn.layer.SLayer;
import sim.util.SStringUtil;

/**
 * L'interface <b>SNetwork</b> reprsente un rseau de neurones.
 * 
 * @author Simon Vzina
 * @since 2019-02-09
 * @version 2022-09-14 (Version labo : Apprentissage par rseau de neurones)
 */
public interface SNetwork extends SBatchActivable {

  public static final String NETWORK = "network";
  public static final String FULL_CONNECTED_NETWORK = "full-connected network";
  public static final String CUSTOM_NETWORK = "custom network";
  public static final String BATCH_NETWORK = "batch-network";
  public static final String FULL_CONNECTED_BATCH_NETWORK = "full-connected batch-network";
  public static final String CUSTOM_BATCH_NETWORK = "custom batch-network";
   
  /**
   * Mthode pour effectuer l'activation du rseau de neurones  partir d'un vecteur d'entre.
   * 
   * @param input Le vecteur d'entre (<i>input</i>).
   * @return L'activation du rseau de neurones  la sortir du rseau.
   * @throws IllegalArgumentException Si le paramtre n'est pas conforme.
   */
  public default double[] forwardPropagation(double[] input) throws IllegalArgumentException
  {
    throw new SNoImplementationException();
    
  }
  
  /**
   * Mthode pour effectuer la propagation de l'erreur dans le rseau  partir d'un vecteur d'erreur.
   * 
   * @param delta L'erreur  propagager dans le rseau.
   * @return L'erreur propage  l'entre du rseau.
   * @throws IllegalArgumentException Si le paramtre n'est pas conforme.
   */
  public default double[] backPropagation(double[] delta) throws IllegalArgumentException
  {
    throw new SNoImplementationException();
    
  }
    
  /**
   * <p>
   * Mthode pour raliser l'activation d'une <i>batch</i> de vecteur d'entre dans le rseau.
   * </p>
   * <p>
   * Une mise en garde sur la mmoire est ncessaire d'tre prcise :  
   * <ul>Si le rseau est de type "Network", une copie des activations sera retourne lors de l'activation.</ul>
   * <ul>Si le rseau est de type "Batch-Network", une rfrence sur les activations sera retourne lors de l'activation.</ul>
   * </p>
   * 
   * @param b La taille de la <i>batch</i>.
   * @param batch_inputs La liste des entres  activer.
   * @return La liste des activations.
   * @throws IllegalArgumentException Si un paramtre n'est pas adquat.
   */
  public default List<double[]> batchForwardPropagation(int b, List<double[]> batch_inputs) throws IllegalArgumentException
  {
    // Validation de 1er instance sur b.
    if(b < 1)
      throw new IllegalArgumentException("Erreur SNetwork 001 : La taille de la batch doit tre suprieur  0.");
    
    // Vrifier que la taille de la batch est de taille infrieur  la taille de la liste.
    if(b > batch_inputs.size())
      throw new IllegalArgumentException("Erreur SNetwork 002 : La taille de la batch ne peut pas tre suprieur  la taille de la liste d'entre.");
    
    List<double[]> list = new ArrayList<double[]>();
    
    // Construire une liste contenant les copies des activations des donnes d'entre par le rseau.
    // Puisque le rseau retourne une rfrence sur les activations, il faudra les copies autrement, elles seraient au fur est  mesure des activations.
    for(int i = 0; i < b; i++) 
      list.add(forwardPropagation(batch_inputs.get(i)).clone());
    
    return list;
  }
  
  /**
   * <p>
   * Mthode pour raliser la propagation de l'ereur d'une <i>batch</i> de donnes d'entre.
   * </p>
   * <p>
   * Il est important de prciser que cette implmentation est uniquement admissible pour des rseaux de type "batch-network".
   * </p>
   * 
   * @param b La taille de la batch.
   * @param batch_delta La liste des erreurs  propager.
   * @return La liste des erreurs propage.
   * @throws IllegalArgumentException Si un paramtre n'est pas adquat.
   */
  public default List<double[]> batchBackPropagation(int b, List<double[]> batch_delta) throws IllegalArgumentException
  {
    // Validation de la batch. 
    if(b != 1)
      throw new IllegalArgumentException("Erreur SNetwork 003 : Ce type de rseau ne support uniquement que des tailles de la batch gal  1.");
    
    // Vrifier que la taille de la batch est de taille infrieur  la taille de la liste.
    if(b > batch_delta.size())
      throw new IllegalArgumentException("Erreur SNetwork 004 : La taille de la batch ne peut pas tre suprieur  la taille de la liste d'entre.");
    
    List<double[]> list = new ArrayList<double[]>();
    
    // Puisque seulement 1 vecteur peut tre support, alors nous ajoutons qu'une seule propagation  la liste.
    list.add(this.backPropagation(batch_delta.get(0)));
    
    return list;
  }
  
  
  /**
   * Mthode permettant d'obtenir la squence des couches du rseau.
   * 
   * @return Les couches du rseau.
   */
  public SLayer[] getLayers();
  
  /**
   * Mthode pour obtenir le nombre de paramtres dcrivant le rseau.
   * 
   * @return Le nombre de paramtre dans le rseau.
   */
  public default int getNumberOfParameter()
  {
    SLayer[] layers = this.getLayers();
    
    int count = 0;
    
    for(SLayer l : layers)
      count += l.getNumberOfParameter();
      
    return count;
  }
  
  /**
   * Mthode pour raliser l'apprentissage du rseau  partir du gradient de la fonction d'erreur par rapport  l'ensemble des paramtres du rseau.
   * 
   * @param gradient Le gradient de la fonction d'erreur.
   * @param learning_rate Le taux d'apprentissage.
   * @throws IllegalArgumentException Si le gradient n'a pas d'information sur la fonction d'erreur.
   */
  public default void learn(SErrorFunctionGradient gradient, double learning_rate) throws IllegalArgumentException
  {
    // Autoriser l'apprentissage uniquement qu'avec des gradients normalise (o la batch = 1).
    if(!gradient.isNormalized())
      throw new IllegalArgumentException("Erreur SNetwork 005 : Le gradient n'est pas normalis. Il ne peut pas tre utilis dans un processus d'apprentissage.");
     
    SLayer[] layers = this.getLayers();
    
    for(int i = 0; i < layers.length; i++)
      layers[i].learn(gradient.getGradientLayer(i), learning_rate);
  }
    
  /**
   * Mthode pour obtenir le nombre de neurones en entre dans le rseau.
   * 
   * @return Le nombre de neurones en entre.
   */
  public default int getInputsSize()
  {
    return this.getLayers()[0].getInputsSize();
  }
  
  /**
   * Mthode pour obtenir le nombre de neurones en sortie dans le rseau.
   * 
   * @return Le nombre de neurones en sortie.
   */
  public default int getOutputsSize()
  {
    SLayer[] layers = this.getLayers();
    
    return layers[layers.length-1].getOutputsSize();
  }
    
  /**
   * Mthode pour obtenir le niveau de profondeur d'entre du rseau.
   * 
   * @return Le niveau de profondeur d'entre du rseau.
   */
  public default int getInputDepth()
  {
    return this.getLayers()[0].getInputsDepth();
  }
  
  /**
   * Mthode pour obtenir le niveau de profondeur de sortie du rseau.
   * 
   * @return Le niveau de profondeur de sortie du rseau.
   */
  public default int getOutputDepth()
  {
    SLayer[] layers = this.getLayers();
    
    return layers[layers.length-1].getOutputsDepth();
  }
  
  /**
   * <p>
   * Mthode qui retourne l'indice de la dernire couche accessible dans le rseau de neurones.
   * </p>
   * <p>
   * Il est important de prciser que l'indice L ne correspond pas au nombre de couche dans le rseau.
   * La numrotation des couches dbute  0 et se termine  SIZE-1.
   * </p>
   * @return L'indice <i>L</i> du rseau (la dernire couche).
   */
  public default int getL()
  {
    return this.getLayers().length - 1;
  }
  
  /**
   * <p>
   * Mthode qui donne le nombre de couches dans le rseau (excluant la couche d'entre).
   * </p>
   * <p>
   * Il est important de prciser que la variable <i>L</i> ne correspond pas au nombre de couches.
   * Puisque <i>L</i> correspond  l'indice de la <u>dernire couche</u> et que celles-ci sont enregistres dans un tableau,
   * l'indice <i>L</i> sera gal  SIZE-1.
   * </p>
   * 
   * @return Le nombre de couches dans le rseau.
   */
  public default int getNbLayers()
  {
    return this.getLayers().length;
  }
  
  /**
   * <p>
   * Mthode pour obtenir la taille de la batch utilise dans le rseau.
   * </p>
   * <p>
   * Un rseau de type "network" aura toujours une taille de <i>batch</i> gal  1 mme si l'on lui demande de s'activer sur sur <i>batch</i>.
   * Un rseau de type "batch-network" aura toujours une taille de <i>batch</i> gal  la taille de la dernire <i>batch</i> active.
   * </p>
   * @return La taille de la batch.
   */
  public default int getBatchSize()
  {
    return 1;
  }
  
  /**
   * Mthode pour valider une squence de couches de neurones.
   * La squence sera adquate si la profondeur et le nombre de neurones de sortie d'une couche 
   * est gale  la profondeur et le nombre de neurones d'entre de la couche suivante pour toute la squence de couche.
   * 
   * @return <b>true</b> si la squence est valide et <b>false</b> sinon.
   */
  public default boolean isValidLayersSequence()
  {
    SLayer[] layers = this.getLayers();
    
    // Vrifier que la couche i possde en output autant de neurones que la couche i+1 en input.
    
    // Pour la vrification de la profondeur, puisque je n'ai pas de couche de type "dense",
    // alos je ne peux pas procder  cette vrification en squence des profondeurs.
    for(int i = 0; i < layers.length-1; i++)
    {
      if(layers[i].getOutputsSize() != layers[i+1].getInputsSize())
        return false;
     
      // Puisqu'il n'y a pas de couche de type 'dense' pour rduire les profondeurs,
      // je ne peux pas faire cette validation.
      //if(layers[i].getOutputDepth() != layers[i+1].getInputDepth())
      //    return false;
    }
    
    return true;
  }
  
  /**
   * Mthode pour dterminer si le rseau est de type <i>batch-network</i> pouvant raliser des activations et propagation d'erreur par <i>batch</i>.
   * 
   * @return <b>true</b> si le rseau est de type <i>batch-network</i> et <b>false</b> sinon.
   */
  public default boolean isBatchNetwork()
  {
    return false;
  }
  
  /**
   * Mthode pour obtenir la dimension du rseau.
   * 
   * @return La dimension du rseau.
   */
  public default String getDimension()
  {
    SLayer[] layers = this.getLayers();
    StringBuffer buffer = new StringBuffer();
    
    // Itrer sur toutes les couches. 
    for(int i = 0; i < layers.length; i++)
    {
      if(i != 0)
        buffer.append(" --> ");
      
      buffer.append(layers[i].getDimension());
    }
   
    return buffer.toString();
  }
  
  /**
   * Mthode pour obtenir la description du rseau.
   * 
   * @return La description du rseau.
   */
  public default String getDescription()
  {
    SLayer[] layers = this.getLayers();
    StringBuffer buffer = new StringBuffer();
    
    // Nom du rseau.
    buffer.append("-- " + this.getNetworkName() + " (" + this.getInputsSize() + " x " + this.getOutputsSize() + ") --" + SStringUtil.END_LINE_CARACTER);
    
    // Description des couches.
    for(int i = 0 ; i < layers.length; i++)
    {
      buffer.append("Couche #" + (i+1) + SStringUtil.END_LINE_CARACTER);
      buffer.append(layers[i].getDescription() + SStringUtil.END_LINE_CARACTER);
    }
    
    buffer.append("Nombre de paramtres du rseau : " + this.getNumberOfParameter() + " paramtres");
    
    return buffer.toString();
  }
  
  /**
   * <p>
   * Mthode pour raliser une copie du rseau tout en gardant accs sur les mmes paramtres.
   * </p>
   * <p>
   * Ce type de copie n'est pas quivalente  la mthode <u>clone()</u>, car mme si la copie du rseau 
   * est sur un espace mmoire diffrent, les paramtres des deux rseaux partageront les mmes espaces mmoires.
   * Ce type de copie permet alors au deux rseaux d'effectuer leurs activations sur leur propre espace mmoire
   * tout en exploitant le mme espace mmoire pour les paramtres ce qui correspond  un conomie.
   * </p>
   * <p>
   * Ce type de copie de rseau permettra ainsi l'activation et la propagation de l'erreur en <u>parallle</u>.
   * </p>
   * 
   * @return Une copie du rseau avec usage des mmes paramtres en mmoire.
   */
  public SNetwork copyWithSameParameter();
  
  /**
   * Mthode pour obtenir le nom du rseau.
   * 
   * @return Le nom du rseau.
   */
  public String getNetworkName();
  
  @Override
  public default void setBatchSize(int b) throws IllegalArgumentException 
  {
    // Validation que le rseau est de type batch.
    if(!this.isBatchNetwork())
      throw new IllegalArgumentException("Erreur SNetwork 006 : Le rseau n'est pas de type batch, donc cette opration n'est pas valide.");
    
    SBatchLayer[] layers = (SBatchLayer[]) this.getLayers();
      
    for(SBatchLayer l : layers)
      l.setBatchSize(b);
  }

  @Override
  public default int getMaximumBatchSize() 
  {
    // Validation que le rseau est de type batch.
    if(!this.isBatchNetwork())
      throw new IllegalArgumentException("Erreur SNetwork 007 : Le rseau n'est pas de type batch, donc cette opration n'est pas valide.");
   
    SLayer[] layers = this.getLayers();
    
    return ((SBatchLayer)layers[0]).getMaximumBatchSize();
  }

  @Override
  public default boolean upgradeMaximumBatchSize(int size) 
  {
    // Validation que le rseau est de type batch.
    if(!this.isBatchNetwork())
      throw new IllegalArgumentException("Erreur SNetwork 007 : Le rseau n'est pas de type batch, donc cette opration n'est pas valide.");
    
    boolean value = false;
    
    SBatchLayer[] layers = (SBatchLayer[]) this.getLayers();
    
    for(SBatchLayer l : layers)
      value = l.upgradeMaximumBatchSize(size);
    
    return value;
  }
  
}
