/**
 * 
 */
package sim.nn.parameter;

import sim.exception.SNaNValueException;
import sim.nn.function.SNeuralMath;
import sim.util.SArrays;

/**
 * La classe abstraite <b>SArrayParameter</b> reprsente un paramtre tant dcrit  l'aide d'un tableau de valeur (un vecteur de donne). 
 * 
 * @author Simon Vezina
 * @since 2021-11-22
 * @version 2021-11-24
 */
public abstract class SArrayParameter extends SParameter {

  /**
   * La variable <b>array</b> reprsente le tableau des donne du paramtre abstrait.
   */
  protected final double[] array;
  
  /**
   * Constructeur d'un vecteur de paramtre dont les valeurs seront dtermines alatoirement.
   * 
   * @param neurons Le nombre de neurones associs au paramtre.
   */
  public SArrayParameter(int neurons)
  {
    this(neurons, true);  
  }
  
  /**
   * Constructeur d'un vecteur de paramtres associ  une couche de neurones.
   * 
   * @param neurons Le nombre de neurones en sortie de la couche.
   * @param random_value S'il faut dterminer des valeurs alatoires aux biais. Si faux, une valeur de zro sera accorde.
   */
  public SArrayParameter(int neurons, boolean random_value)
  {
    super();
    
    // Construction des ordonnes  l'origine (bias).
    this.array = new double[neurons];
    
    // Remplir le vecteur avec des donne alatoires si demand.
    if(random_value)
      for(int i = 0; i < neurons; i++)
        this.array[i] = randomParameter();    
  }
  
  /**
   * Constructeur d'un vecteur de paramtres  partir d'un tableau de donnes.
   * La rfrence  ce tableau sera copi dans le paramtre.
   * 
   * @param vector Le vecteur des paramtres.
   */
  public SArrayParameter(double[] vector)
  {
    this.array = vector;
  }
  
  /**
   * Mthode pour obtenir la valeur du paramtre associ au neurone u.
   * 
   * @param u L'indice du neurone.
   * @return Le paramtre du neurone u.
   */
  public double get(int u)
  {
    return this.array[u];
  }
  
  /**
   * Mthode pour obtenir le vecteur du paramtre.
   * 
   * @return Le vecteur.
   */
  public double[] getVector()
  {
    return this.array;
  }
  
  /**
   * Mthode pour obtenir le nombre de neurones sur lequel ce paramtre utilis.
   * 
   * @return Le nombre de neurones.
   */
  public int getNeurons()
  {
    return this.array.length;
  }
   
  @Override
  public int getNumberOfParameter()
  {
    return this.array.length;
  }
  
  /**
   * Mthode pour dfinir la valeur d'un paramtre.
   * 
   * @param u L'indice du paramtre.
   * @param value La valeur  attribuer au paramtre de neurone u.
   * @throws SNaNValueException Si un valeur NaN doit tre affecte.
   */
  public void set(int u, double value) throws SNaNValueException
  {
    this.array[u] = this.analyseValue(value);
  }

  @Override
  public void multiply(double value) 
  {
    SArrays.multiply(this.array, value, this.array);
  }
 
  @Override
  public void clear() 
  {
    SArrays.clear(this.array);
  }
    
  @Override
  public boolean equals(Object obj) 
  {
    if (this == obj) {
      return true;
    }
    
    if (obj == null) {
      return false;
    }
    
    if (!(obj instanceof SArrayParameter)) {
      return false;
    }
    
    SArrayParameter other = (SArrayParameter) obj;
    if (!SArrays.nearlyEquals(this.array, other.array)) {
      return false;
    }
    
    return true;
  }
  
  /**
   * Mthode pour raliser l'addition d'un paramtre de type vecteur avec un autre.
   * 
   * @param parameter Le paramtre  ajouter.
   * @throws IllegalArgumentException Si les paramtres n'ont pas la mme taille de vecteur.
   */
  protected void addParameter(SArrayParameter parameter) throws IllegalArgumentException
  {
    try {
      
      SArrays.add(this.array, parameter.array, this.array);
      
    }catch(ArrayIndexOutOfBoundsException e) {
      throw new IllegalArgumentException("Erreur SArrayParameter 001 : L'addition des deux paramtres n'est pas possible.", e); 
    }
  }
  
  /**
   * Mthode pour raliser la multiplication d'un paramtre de type vecteur avec un autre.
   * 
   * @param parameter Le paramtre  multiplier.
   * @throws IllegalArgumentException Si les paramtres n'ont pas la mme taille de vecteur.
   */
  protected void multiplyParameter(SArrayParameter parameter) throws IllegalArgumentException
  {
    try {
      
      SArrays.multiply(this.array, parameter.array, this.array);
    
    }catch(ArrayIndexOutOfBoundsException e) {
      throw new IllegalArgumentException("Erreur SArrayParameter 002 : La multiplication des deux paramtres n'est pas possible.", e); 
    }
  }
  
  /**
   * Mthode pour raliser l'apprentissage d'un paramtre de type vecteur  l'aide du gradient de ce paramtre.
   * 
   * @param gradient_parameter Le gradient du paramtre.
   * @param learning_rate Le taux d'apprentissage.
   */
  protected void learnParameter(SArrayParameter gradient_parameter, double learning_rate)
  {
    // Validation de la taille du gradient avec la taille du paramtre.
    if(gradient_parameter.getNeurons() != this.getNeurons())
      throw new IllegalArgumentException("Erreur SArrayParameter 003 : Le gradient possde " + gradient_parameter.getNeurons() + " neurones et le paramtre en contient " + this.getNeurons() + " neurones ce qui n'est pas compatible.");
    
    // Accs au tableau du gradient.
    double[] gradient = gradient_parameter.array;
     
    // Raliser l'apprentissage pour tous les lments du paramtre.
    for(int u = 0; u < this.getNeurons(); u++)
      this.set(u, SNeuralMath.learning(this.get(u), gradient[u], learning_rate));
  }
}
