/**
 * 
 */
package sim.nn.parameter;

import java.util.ArrayList;
import java.util.List;

import sim.exception.SConstructorException;
import sim.exception.SNaNValueException;
import sim.math.SArrayTensor;
import sim.nn.function.SNeuralMath;
import sim.nn.gradient.SGradientLayer;

/**
 * La classe <b>SFiltersWeights</b> reprsente les poids des filtres de la fonction de convolution.
 * 
 * @author Simon Vezina
 * @since 2019-02-02
 * @version 2021-12-07
 */
public class SFiltersWeights extends SParameter {

  /**
   * ...
   */
  private final List<SArrayTensor> filters;
  
  /**
   * La variable <b>offset</b> reprsente la taille de chaque matrice de poids pour l'ensemble des filtres.
   */
  private final int offset;
  
  /**
   * Constructeur de poids  filter.
   * 
   * @param offset La dimension du filter N x N.
   * @param input_depth La profondeur de l'entre.
   * @param output_depth La profondeur de sortie (le nombre de groupe de filtres).
   */
  public SFiltersWeights(int offset, int input_depth, int output_depth) 
  {
    super();
    
    if(offset < 2)
      throw new SConstructorException("Erreur SFilterWeights 001 : La taille du filtre " + offset + " doit tre suprieur  1.");
    
    if(input_depth < 1)
      throw new SConstructorException("Erreur SFilterWeights 002 : La profondeur des filtres en entre " + input_depth + " doit tre suprieur  0.");
    
    if(output_depth < 1)
      throw new SConstructorException("Erreur SFilterWeights 003 : La profondeur des filtres en sortie " + output_depth + " doit tre suprieur  0.");
    
    this.offset = offset;
    
    this.filters = new ArrayList<SArrayTensor>();
    
    // Construire l'ensemble des tenseurs correspondant  chacun des filtres.
    for(int i = 0; i < output_depth; i++)
    {
      double[] array = new double[offset*offset*input_depth];
      
      // Remplir le array de nombre alatoire.
      for(int j = 0; j < array.length; j++)
        array[j] = this.randomParameter();
      
      // Construire le tenseur avec la rfrence au tableau, car l'allocation de mmoire vient juste d'tre ralise.
      this.filters.add(SArrayTensor.buildWithReference(array, input_depth, offset, offset));
    }
  }

  /**
   * ...
   * 
   * @param filters
   * @throws SConstructorException
   */
  public SFiltersWeights(SArrayTensor... filters) throws SConstructorException
  {
    // Vrifier qu'il y a des terme dans l'affectation.
    if(filters.length == 0)
      throw new SConstructorException("Erreur SFilterWeights 004 : Il n'y a pas de filtre d'affect.");
      
    // Vrifier si l'ensemble des tenseurs est d'ordre 3.
    for(SArrayTensor t : filters)
      if(t.getOrder() != 3)
        throw new SConstructorException("Erreur SFilterWeights 005 : Le tenseur t = " + t + " n'est pas d'ordre 3.");
    
    // PLEIN DE CHOSE  VRIFIER !!!
    
    // Le offset correspondra  la 1er dimension du tenseur.
    this.offset = filters[0].getAccesDimension()[1];
    
    this.filters = new ArrayList<SArrayTensor>();
        
    for(SArrayTensor t : filters)
      this.filters.add(t);
  }
  
  @Override
  public SFiltersWeights clone()
  {
    SArrayTensor[] array = new SArrayTensor[this.filters.size()];
    
    for(int i = 0; i < array.length; i++)
      array[i] = this.filters.get(i).clone();
    
    return new SFiltersWeights(array);
  }
  
  /**
   * Mthode pour obtenir la taille des matrices carres de filtrage.
   * 
   * @return La taille du offset des matrices carres de filtrage.
   */
  public int getOffset()
  {
    return this.offset;
  }
  
  /**
   * ...
   * 
   * @param f L'indice de profondeur de sortie (le numro du filtre).
   * @return Les filtres de sortie f.
   */
  public SArrayTensor getWeights(int f)
  {
    return this.filters.get(f);
  }
  
  /**
   * ...
   * 
   * @param f profondeur de sortie. 
   * @param d profondeur d'entre.
   * @param i L'indice de la ligne.
   * @param j L'indice de la colonne.
   * @return La composante W_ij d'un filtre de sortie f et d'entre d.
   */
  public double get(int f, int d, int i, int j)
  {
    return this.filters.get(f).get(d, i, j);
  }
  
  /**
   * Mthode pour dfinir un paramtre des filtres de convolution.
   * 
   * @param value La valeur  dfinir.
   * @param f profondeur de sortie. 
   * @param d profondeur d'entre.
   * @param i L'indice de la ligne.
   * @param j L'indice de la colonne.
   * @return La composante W_ij d'un filtre de sortie f et d'entre d.
   * @throws SNaNValueException Si la valeur  affecter est gal  NaN.
   */
  public void set(double value, int f, int d, int i, int j) throws SNaNValueException
  {
    this.filters.get(f).set(this.analyseValue(value), d, i, j);
  }
    
  /**
   * Mthode pour obtenir la profondeur d'entre des filtres (le nombre de matrice par filtre).
   * 
   * @return La profondeur d'entre des filtres.
   */
  public int getInputDepth()
  {
    // Accs  la dimension de la profondeur de l'entre.
    return this.filters.get(0).getAccesDimension()[0];
  }
  
  /**
   * Mthode pour obtenir la profondeur de sortie des filtres (le nombre de filtre).
   * 
   * @return La profondeur de sortie des filtres.
   */
  public int getOutputDepth()
  {
    return this.filters.size();
  }
  
  @Override
  public int getNumberOfParameter() 
  {
    return this.filters.get(0).getAccesArray().length * this.filters.size();
  }

  /**
   * Mthode pour obtenir la liste des filtres reprsent sous la forme de tenseur.
   * Chaque lment de la liste reprsente un filtre complet o le nombre de filtres sera la profondeur de sortie de la couche.
   * 
   * @return La liste des filtres.
   */
  public List<SArrayTensor> getTensorList()
  {
    return this.filters;
  }
  
  
  @Override
  public void learn(SGradientLayer gradient, double learning_rate) throws IllegalArgumentException 
  {
    // Effectuer l'entranement si le gradient contient un accs au poids.
    if(!(gradient instanceof SContainFiltersWeights))
      throw new IllegalArgumentException("Erreur SFilterWeights 007 : Le gradient n'est pas du bon type (" + gradient.getClass().getName() + ")."); 
     
    SFiltersWeights dCdW = ((SContainFiltersWeights) gradient).getFiltersWeights();
    double value;
    
    // Validation de la taille des poids utilis pour faire l'apprentissage.
    if(dCdW.getOffset() != this.getOffset() || dCdW.getInputDepth() != this.getInputDepth() || dCdW.getOutputDepth() != this.getOutputDepth())
      throw new IllegalArgumentException("Erreur SFilterWeights 008 : Le gradient n'a pas le bon format de paramtre.");
      
    if(isTrainable())
    {
      // Apprentissage chez tous les poids.
      
      // Itrer sur l'ensemble des filtres.
      for(int f = 0; f < filters.size(); f++)
      {
        SArrayTensor tensor = filters.get(f);
        
        // Itrer sur chaque ligne/colonne et profondeur des poids du filter de numro f.
        for(int d = 0; d < tensor.getAccesDimension()[0]; d++)
          for(int i = 0; i < tensor.getAccesDimension()[1]; i++)
            for(int j = 0; j < tensor.getAccesDimension()[2]; j++)   
            {
              value = SNeuralMath.learning(tensor.get(d, i, j), dCdW.get(f, d, i, j), learning_rate);
              this.set(value, f, d, i, j);
            }
      }
    }     
    
  }
  
  @Override
  public void add(SParameter parameter) throws IllegalArgumentException 
  {
    // Vrification du type de paramtre.
    if(!(parameter instanceof SFiltersWeights))
      throw new  IllegalArgumentException("Erreur FiltersSWeights 007 : Le paramtre n'est pas du bon type (" + parameter.getClass().getName() + ")."); 
    
    SFiltersWeights p = (SFiltersWeights) parameter;
    
    if(p.getTensorList().size() != this.filters.size())
      throw new  IllegalArgumentException("Erreur FiltersSWeights 008 : Le paramtre possde " + p.getTensorList().size() + " filtres, mais le paramtre courant en possde " +  this.filters.size() + " ce qui n'est pas identique."); 
    
    // Faire l'addition tenseur par tenseur.
    for(int i = 0; i < filters.size(); i++)
      this.filters.get(i).addWith(p.getWeights(i));
  }

  @Override
  public void multiply(double value) 
  {
    // Faire la multiplication tenseur par tenseur.
    for(int i = 0; i < filters.size(); i++)
      this.filters.get(i).multiplyWith(value);
  }

  @Override
  public void multiply(SParameter parameter) throws IllegalArgumentException 
  {
    // Vrification du type de paramtre.
    if(!(parameter instanceof SFiltersWeights))
      throw new  IllegalArgumentException("Erreur FiltersSWeights 009 : Le paramtre n'est pas du bon type (" + parameter.getClass().getName() + ")."); 
    
    SFiltersWeights p = (SFiltersWeights) parameter;
    
    if(p.getTensorList().size() != this.filters.size())
      throw new  IllegalArgumentException("Erreur FiltersSWeights 010 : Le paramtre possde " + p.getTensorList().size() + " filtres, mais le paramtre courant en possde " +  this.filters.size() + " ce qui n'est pas identique."); 
    
    // Faire l'addition tenseur par tenseur.
    for(int i = 0; i < filters.size(); i++)
      this.filters.get(i).multiplyWith(p.getWeights(i));
  }

  @Override
  public void clear() 
  {
    for(SArrayTensor t : filters)
      t.clear();
  }

}
