/**
 * 
 */
package sim.nn.parameter;

import sim.exception.SConstructorException;
import sim.math.SBufferedMatrix;
import sim.nn.function.SNeuralMath;

/**
 * La classe <b>SMatrixParameter</b> reprsente un paramtre abstrait dont l'organisation des donnes sera enregistre dans une matrice.
 * 
 * @author Simon Vezina
 * @since 2021-12-03
 * @version 2021-12-31
 */
public abstract class SMatrixParameter extends SParameter {

  /**
   * La variable <b>matrix</b> reprsente la matrice contenant les paramtres organiss en ligne et colonne.
   */
  protected final SBufferedMatrix matrix;
  
  /**
   * Constructeur d'une matrice de paramtre <u>sans dimension</u>.
   */
  public SMatrixParameter()
  {
    super();
    
    this.matrix = SBufferedMatrix.ZERO_DIMENSION_MATRIX;
    
    setTrainable(false);
  }
  
  /**
   * ...
   * 
   * @param nb_line
   * @param nb_column
   * @throws SConstructor
   */
  public SMatrixParameter(int nb_line, int nb_column) throws SConstructorException
  {
    this(nb_line, nb_column, false);
  }

  /**
   * ...
   * 
   * @param nb_line
   * @param nb_colomn
   * @param random_value
   * @throws SConstructorException
   */
  public SMatrixParameter(int nb_line, int nb_colomn, boolean random_value) throws SConstructorException
  {
    super();
    
    if(nb_line < 1)
      throw new SConstructorException("Erreur SMatrixParameter 001 : Le nombre de ligne " + nb_line + " doit tre suprieur  1.");
    
    if(nb_colomn < 1)
      throw new SConstructorException("Erreur SMatrixParameter 002 : Le nombre de colonne " + nb_colomn + " doit tre suprieur  1.");
        
    // Construction de la matrice interne. 
    // Il est important de remarquer le nombre de'output correspond au nombre de ligne et le nombre d'input au nombre de colonne.
    this.matrix = new SBufferedMatrix(nb_line, nb_colomn);
    
    // Remplir la matrice avec des donne alatoires, si demand.
    if(random_value)
      for(int i = 0; i < nb_line; i++)
        for(int j = 0; j < nb_colomn; j++)
          this.matrix.set(i, j, randomParameter());
  }
  
  /**
   * Constructeur d'une matrice de paramtres  partir d'une matrice de donnes.
   * La rfrence  ce tableau sera copi dans le paramtre.
   * 
   * @param vector La matrice pris en rfrence.
   */
  public SMatrixParameter(SBufferedMatrix m)
  {
    this.matrix = m;
  }
  
  /**
   * Constructeur d'une matrice de paramtres  partir d'une liste de ligne donne.
   * 
   * @param lines La liste des lignes.
   */
  public SMatrixParameter(double[] ... lines)
  {
    this.matrix = new SBufferedMatrix(lines);
  }
  
  /**
   * Mthode pour obtenir la matrice des paramtres.
   * 
   * @return La matrice.
   */
  public SBufferedMatrix getMatrix()
  {
    return this.matrix;
  }
    
  @Override
  public int getNumberOfParameter()
  {
    return this.matrix.getNbLine() * this.matrix.getNbColumn();
  }

  @Override
  public void clear() 
  {
    this.matrix.clear();
  }
  
  @Override
  public void multiply(double value)
  {
    this.matrix.multiplyWith(value);
  }
  
  /**
   * ...
   * 
   * @param parameter
   * @throws IllegalArgumentException
   */
  protected void addParameter(SMatrixParameter parameter) throws IllegalArgumentException 
  {
    // Faire l'ajout de la matrice  la courante.
    this.matrix.addWith(((SMatrixParameter) parameter).getMatrix());    
  }

  /**
   * ...
   * 
   * @param parameter
   * @throws IllegalArgumentException
   */
  protected void multiplyParameter(SMatrixParameter parameter) throws IllegalArgumentException 
  {
    // Faire la multiplication de la matrice avec la matrice courante.
    this.matrix.hadamardProduct(parameter.getMatrix());   
  }

  /**
   * ...
   * 
   * @param gradient
   * @param learning_rate
   * @throws IllegalArgumentException
   */
  protected void learnParameter(SMatrixParameter gradient, double learning_rate) throws IllegalArgumentException 
  {
    // Validation de la taille du gradient avec la taille du paramtre.
    if(gradient.matrix.getNbLine() != this.matrix.getNbLine() || gradient.matrix.getNbColumn() != this.matrix.getNbColumn())
      throw new IllegalArgumentException("Erreur SMatrixParameter 003 : Le gradient possde n'a pas la bonne taille.");
    
    // Accs  la matrice du gradient.
    SBufferedMatrix gradient_matrix = gradient.matrix;
    
    // Raliser l'apprentissage pour tous les lments du paramtre.
    for(int u = 0; u < this.matrix.getNbLine(); u++)
      for(int v = 0; v < this.matrix.getNbColumn(); v++)
        this.matrix.set(u, v, SNeuralMath.learning(this.matrix.get(u, v), gradient_matrix.get(u, v), learning_rate));
  }

  @Override
  public int hashCode() {
    final int prime = 31;
    int result = 1;
    result = prime * result + ((matrix == null) ? 0 : matrix.hashCode());
    return result;
  }

  @Override
  public boolean equals(Object obj) {
    if (this == obj)
      return true;
    if (obj == null)
      return false;
    if (getClass() != obj.getClass())
      return false;
    SMatrixParameter other = (SMatrixParameter) obj;
    if (matrix == null) {
      if (other.matrix != null)
        return false;
    } else if (!matrix.equals(other.matrix))
      return false;
    return true;
  }

  @Override
  public String toString() 
  {
    return "SMatrixParameter [matrix=" + matrix + "]";
  }  

}
