/**
 * 
 */
package sim.nn.parameter;

import java.util.concurrent.atomic.AtomicLong;

import sim.exception.SNaNValueException;
import sim.math.SMath;
import sim.nn.gradient.SGradientLayer;

/**
 * La classe abstraite <b>SParameter</b> reprsente un paramtre dcrivant une couche de neurones.
 * Ces paramtres sont utilises dans diffrent type de fonction applicable sur des neurones d'une couche.
 * 
 * @author Simon Vzina
 * @since 2018-01-09
 * @version 2021-12-13
 */
public abstract class SParameter {

  //--------------
  // CONSTANTES //
  //--------------
    
  /**
   * ...
   */
  public final static int UNIFORM_RANDOM_DISTRIBUTION = 0;
  
  /**
   * ...
   */
  public final static int GAUSSIAN_RANDOM_DISTRIBUTION = 1;
  
  /**
   * La constante <b>DEFAULT_TRAINABLE</b> dtermine si le paramtre peut tre entrain ou non par dfaut qui est par dfaut {@value}.
   */
  private final static boolean DEFAULT_TRAINABLE = true;
  
  /**
   * La constante <b>DEFAULT_CLIPPING_NORMALIZATION</b> dtermine si le paramtre ne peut pas dpasser une certaine zone de valeur (si la coupure est active) qui est par dfaut {@value}.
   */
  private final static boolean DEFAULT_CLIPPING_NORMALIZATION = true;
    
 
  /**
   * La constante <b>MAX_RANDOM</b> reprsente la valeur maximale alatoire d'initialisation d'un paramtre tant gal  {@value}.
   */
  protected final static double ABSOLUTE_MAX_RANDOM = 1.0;
  
  /**
   * La variable <b>atomic_counteur</b> correspond  un compteur THREAD PROOF pour compter le nombre de paramtre ayant subit une coupure de normalisation. 
   */
  private static final AtomicLong atomic_counter = new AtomicLong();
  
  /**
   * ...
   */
  private static int random_distribution = UNIFORM_RANDOM_DISTRIBUTION;
  
  /**
   * La variable <b>max_clipping_normalization</b> reprsente la valeur maximale de coup d'un paramtre.
   */
  protected static double max_clipping_normalization = 4.0;
  
  /**
   * La variable <b>min_clipping_normalization</b> reprsente la valeur minimal de coup d'un paramtre.
   */
  protected static double min_clipping_normalization = -4.0;
    
  /**
   * La variable <b>global_clipping_normalization</b> reprsente l'tat globale de l'application de la normalisation de coupure (<i>clipping</i>).
   */
  private static boolean global_clipping_normalization = DEFAULT_CLIPPING_NORMALIZATION;
  //-------------
  // VARIABLES //
  //-------------
  
  /**
   * La variable <b>is_trainable</b> permet de dterminer si le paramtre peut tre entrain (raliser des apprentissage).
   */
  private boolean is_trainable;
  
  /**
   * La variable <b>is_clipping_normalization</b> dterminer si la normalisation de type coupure est applicable sur le paramtre.
   */
  private boolean is_clipping_normalization;
  
  //----------------
  // CONSTRUCTEUR //
  //----------------
  
  /**
   * Constructeur d'un paramtre.
   */
  public SParameter()
  {
    this.is_trainable = DEFAULT_TRAINABLE;
    this.is_clipping_normalization = DEFAULT_CLIPPING_NORMALIZATION;
  }
  
  //------------
  // MTHODES //
  //------------
  
  /**
   * Mthode pour faire l'analyse d'une valeur d'un paramtre.
   * 
   * @param value La valeur du paramtre  analyser.
   * @return La mme valeur  analyser si elle est adquate ou une autre valeur traite autrement.
   * @throws SNaNValueException Si la valeur  analyser est de type NaN (not un number).
   */
  protected double analyseValue(double value) throws SNaNValueException
  {
    // Validation de la NaN.
    if(Double.isNaN(value))
      throw new SNaNValueException();
    
    // Application de la normalisation de coupure si ncessaire.
    if(SParameter.global_clipping_normalization && this.is_clipping_normalization)
    {
      if(value > SParameter.max_clipping_normalization)
      {
        atomic_counter.incrementAndGet();
        return SParameter.max_clipping_normalization;
      }
      else
        if(value < SParameter.min_clipping_normalization)
        {
          atomic_counter.incrementAndGet();
          return SParameter.min_clipping_normalization;
        }
    }
    
    // Retourner la valeur s'il n'y a pas eu de coupure.
    return value;
  }
  
  /**
   * Mthode pour dterminer si le paramtre peut tre entrain.
   * 
   * @return <b>true</b> si le paramtre peut tre entrain et <b>false<b> sinon.
   */
  public boolean isTrainable()
  {
    return is_trainable;
  }
  
  /**
   * Mthode pour obtenir le nombre de variable dcrivant le paramtre.
   * 
   * @return Le nombre de variables contenus dans le paramtre.
   */
  public abstract int getNumberOfParameter();
 
  /**
   * Mthode pour dfinir si le paramtre activera sa normalisation de coupure limitant  une certaine valeur minimale et maximale les valeurs admissibles.
   * 
   * @param value L'tat de la normalisation de coupure (<i>clipping</i>.
   */
  public void setClippingNormalization(boolean value)
  {
    this.is_clipping_normalization = value;
  }
  
  /**
   * Mthode pour obtenir l'tat globale de l'application ou non de la normalisation de coupure.
   * 
   * @return <b>true</b> si la normalisation de coupure sera applique globalement (mais pas ncessairement localement) et <b>false</b> sinon.
   */
  public static boolean getGlobalClippingNormalization()
  {
    return SParameter.global_clipping_normalization;
  }
  
  /**
   * <p>
   * Mthode pour dfinir globalement l'tat de l'application de la normalisation de coupure (<i>clipping</i>).
   * </p>
   * <p>
   * Par dfinition de globalement, nous appliquons ceci  l'ensemble de tous les paramtres des rseaux de neurones.
   * Cependant, l'application globale ne sous-entend pas qu'elle sera applique localement.
   * Par contre, si elle n'est pas applique globalement, elle ne pourra pas l'tre localement.
   * </p>
   *  
   * @param value L'tat de la normalisation de coupure.
   */
  public static void setGlobalClippingNormalization(boolean value)
  {
    SParameter.global_clipping_normalization = value;
  }
  
  /**
   * Mthode pour obtenir le nombre de paramtres qui ont subit une coupe de normalisation.
   * Le dnombrement comprend les paramtres et les gradients des paramtres. 
   * 
   * @return Le nombre de normalisation de coupure.
   */
  public static long getClippingNormalizationCount()
  {
    return SParameter.atomic_counter.get();
  }
  
  
 
  
  /**
   * Mthode pour rinitialiser le comptage des paramtres ayant ncessit une normalisation de coupure.
   */
  public static void resetClippingNormalizationCount()
  {
    SParameter.atomic_counter.set(0);
  }
  
  /**
   *  Mthode pour dfinir l'tat d'entrainement du paramtre.
   *  
   *  @param value L'tat de l'entrainement du paramtre.
   */
  public void setTrainable(boolean value)
  {
    is_trainable = value;
  }
  
  /**
   * Mthode pour raliser l'apprentissage du paramtre  partir d'un gradient de la fonction d'erreur associ  une couche de neurones.
   * 
   * @param gradient Le gradient de la fonction d'erreur.
   * @param learning_rate Le taux d'apprentissage.
   * @throws IllegalArgumentException Si le gradient n'est pas du type adquatement pour l'apprentissage du paramtre.
   */
  public abstract void learn(SGradientLayer gradient, double learning_rate) throws IllegalArgumentException;
  
  /**
   * Mthode pour ajouter au paramtre courant la valeur d'un autre paramtre. L'ajout sera support uniquement si les deux paramtres sont de mme type.
   * 
   * @param parameter Le paramtre  ajouter.
   * @throws IllegalArgumentException Si le paramtre n'est pas du bon type.
   */
  public abstract void add(SParameter parameter) throws IllegalArgumentException;
  
  /**
   * Mthode pour multiplier le paramtre par une constante.
   * 
   * @param value La constante.
   */
  public abstract void multiply(double value);
  
  /**
   * Mthode pour multiplier le paramtre par un autre paramtre.
   * 
   * @param parameter Le paramtre.
   */
  public abstract void multiply(SParameter parameter) throws IllegalArgumentException;
  
  /**
   * Mthode pour mettre  zro l'ensemble des valeurs de ce paramtre.
   */
  public abstract void clear();
  
  /**
   * Mthode pour gnrer une valeur alatoire  un paramtre.
   *  
   * @return Une valeur alatoire  un paramtre.
   */
  public double randomParameter()
  {
    switch(SParameter.random_distribution)
    {
      case SParameter.UNIFORM_RANDOM_DISTRIBUTION : return SMath.random(-1.0*ABSOLUTE_MAX_RANDOM, ABSOLUTE_MAX_RANDOM);
      
      case SParameter.GAUSSIAN_RANDOM_DISTRIBUTION : return ABSOLUTE_MAX_RANDOM*SMath.randomGaussian()*SMath.randomSign();
    
      default : throw new IllegalStateException("Erreur SParameter 001 : La classe de SParameter est prsentement dans un tat invalide pour gnrer des nombres alatoires.");
    }
  }
    
}
