/**
 * 
 */
package sim.nn.parameter;

import sim.exception.SConstructorException;
import sim.exception.SNaNValueException;
import sim.math.SBufferedMatrix;
import sim.nn.gradient.SGradientLayer;

/**
 * <p>
 * La classe <b>SWeights</b> reprsente la donnes correspondant aux poids des neurones d'entre dans le calcul de l'activation d'une couche de neurones.
 * </p>
 * <p>
 * Dans la matrice des poids, chaque ligne reprsente les paramtres requis pour le calcul d'un neurone.
 * Ainsi, le nombre de ligne dans la matrice reprsente le nombre de neurones pouvant tre calcul par les poids.
 * Si les poids sont d'ordre 1, alors il y aura un nombre de colonne gal au nombre de neurone d'entre.
 * Si les poids sont d'ordre suprieur  1, alors la matrice contiendra O*Input o O est l'ordre et Input est le nombre de neurone d'entre.
 * </p>
 *  
 * @author Simon Vzina
 * @since 2018-07-29
 * @version 2021-12-31
 */
public class SWeights extends SMatrixParameter {

  /**
   * <p>
   * La variable <b>order</b> reprsente l'ordre de la matrice de poids. 
   * </p>
   * <p>
   * Si l'ordre est gal  1, alors cette matrice permet de reprsenter un hyperplan.
   * Si l'ordre est gal  2, alors cette matrice permet de reprsenter une hypersphre.
   * </p>
   */
  private final int order;
  
  /**
   * <p>
   * La variable <b>offset</b> reprsente le dcalage d'indice en colonne dans la matrice pour accder aux diffrent ordre de poids pour une mme ligne.
   * </p>
   * <p>
   * Cette valeur est toujours gal au nombre de neurone d'entre pouvant tre calcul par la matrice de poids.
   * Ainsi, ce dcalage permettra de se dplacer sur tous les paramres d'une mme ligne dans la matrice des poids et accder aux valeurs pour les diffrents ordres.
   * </p>
   */
  private final int offset;
  
  //----------------
  // CONSTRUCTEUR //
  //----------------
  
  /**
   * Constructeur des poids d'une couche de neurones  partir du nombre d'entre de la couche et du nombre de sortie de la couche.
   * Les poids sont tous initialiss avec une valeur alatoire. 
   * 
   * @param nb_input Le nombre de neurones d'entre du poids.
   * @param nb_output Le nombre de neurones de sortie du poids.
   */
  public SWeights(int nb_input, int nb_output)
  {
    this(nb_input, nb_output, 1);
  }

  /**
   * Constructeur des poids d'une couche de neurone avec prcision de l'ordre.
   * Les poids sont tous initialiss avec une valeur alatoire. 
   * 
   * @param nb_input Le nombre de neurones d'entre du poids.
   * @param nb_output Le nombre de neurones de sortie du poids.
   * @param order L'ordre du poids.
   */
  public SWeights(int nb_input, int nb_output, int order)
  {
    this(nb_input, nb_output, order, true);
  }
  
  /**
   * ...
   * 
   * @param nb_input
   * @param nb_output
   * @param random_value
   */
  public SWeights(int nb_input, int nb_output, boolean random_value)
  {
    this(nb_input, nb_output, 1, true);
  }
  
  /**
   * Constructeur des poids d'une couche de neurones.
   * 
   * @param nb_input Le nombre de neurones en entre de la couche.
   * @param nb_output Le nombre de neurones en sortie de la couche.
   * @param random_value S'il faut dterminer des valeurs alatoires aux poids. Si faux, une valeur de zro sera accorde.
   */
  public SWeights(int nb_input, int nb_output, int order, boolean random_value) throws SConstructorException
  {
    super(nb_output, nb_input*order, random_value);
    
    if(order < 1)
      throw new SConstructorException("Erreur SWeights 003 : L'ordre de la matrice order = " + order + " doit tre suprieur  1.");
       
    this.order = order;
    this.offset = nb_input;
  }
  
  /**
   * Constructeur des poids d'une couche de neurones  partir d'une matrice.
   * La matrice passe en paramtre sera prise en rfrence (ce n'est pas une copie des donnes qui sera ralise).
   * 
   * @param matrix La matrice des poids.
   */
  public SWeights(SBufferedMatrix matrix)
  {
    this(matrix, 1);
  }
  
  /**
   * Construction d'une matrice de poids  partir d'une matrice.
   * La matrice passe en paramtre sera prise en rfrence (ce n'est pas une copie des donnes qui sera ralise).
   * 
   * @param matrix La matrice des poids.
   * @param order L'ordre de la matrice.
   * @throws SConstructorException Si la matrice et l'ordre est incompatible.
   */
  public SWeights(SBufferedMatrix matrix, int order) throws SConstructorException
  {
    // Affectation de la matrice dans le paramtre this.matrix (copie de la rfrence)
    super(matrix);
    
    if(order < 1)
      throw new SConstructorException("Erreur SWeights 004 : La matrice de poids ne peut pas tre d'ordre infrieur  1.");
    
    if(matrix.getNbColumn() % order != 0)
      throw new SConstructorException("Erreur SWeights 005 : Le nombre de colonne de la matrice " + matrix.getNbColumn() + " n'est pas compatible avec l'ordre " + order +".");
        
    this.order = order;
    this.offset = matrix.getNbColumn() / order;
  }
  
  /**
   * Constructeur des poids d'une couche de neurones  partir de ligne correspondant  la matrice des poids.
   * Cette matrice sera d'ordre 1 (pour l'usage d'une fonction d'agrgation comme la somme pondre).
   * 
   * @param lines Les lignes de la matrices.
   */
  public SWeights(double[] ... lines)
  {
    super(lines);
    
    this.order = 1;
       
    this.offset = this.matrix.getNbColumn();
  }
  
  //------------
  // MTHODES //
  //------------
  
  @Override
  public SWeights clone()
  {
    SWeights w = new SWeights(this.matrix.clone(), this.getOrder());
    w.setTrainable(this.isTrainable());
       
    return w;
  }
  
  /**
   * <p>
   * Mthode pour accder au poid w_vu de la matrice (en mode matrice d'ordre 1).
   * </p>
   * <p>
   * Il est important de noter que cette accs est en reprsentation tensorielle.
   * Ainsi, le paramtre v est l'entre (mais la colonne de la matrice) et 
   * u est la sortie (mais la ligne de la matrice).
   * </p>
   * 
   * @param v L'indice du signal d'entre dans la couche de neurones (input).
   * @param u L'indice du signal de sortie dans la couche de neurones (output).
   * @return Le poids dans la matrice.
   */
  public double get(int v, int u)
  {
    return this.matrix.get(u, v);
  }
  
  /**
   * <p>
   * Mthode pour accder au poid w_vu d'ordre o de la matrice (en mode matrice d'ordre o).
   * </p>
   * <p>
   * Il est important de noter que cette accs est en reprsentation tensorielle.
   * Ainsi, le paramtre v est l'entre (mais la colonne de la matrice) et 
   * u est la sortie (mais la ligne de la matrice).
   * </p>
   * @param v L'indice du signal d'entre dans la couche de neurones (input).
   * @param u L'indice du signal de sortie dans la couche de neurones (output).
   * @param o L'ordre du poids w_vu (doit tre gal  1 ou suprieur).
   * @return Le poids dans la matrice.
   */
  public double get(int v, int u, int o)
  {
    return this.matrix.get(u, v + (o-1)*offset);
  }
  
  /**
   * Mthode pour obtenir l'ordre de la matrice.
   * 
   * @return L'ordre de la matrice.
   */
  public int getOrder()
  {
    return this.order;
  }
  
  /**
   * Mthode pour obtenir la matrice des poids.
   * 
   * @return La matrice des poids.
   */
  public SBufferedMatrix getMatrix()
  {
    return this.matrix;
  }
  
  /**
   * <p>
   * Mthode pour obtenir le nombre d'entre de neurones associ  l'usage de ce paramtre de poids (<i>weights</i>).
   * </p>
   * <p>
   * Il est important de noter que cette information n'est pas quivalente au nombre de colonnes dans la matrice
   * reprsentant ce paramtre, car l'ordre suprieur  1 permettra d'avoir ordre*input d'lment sur une mme ligne.
   * </p>
   * 
   * @return Le nombre de neurones d'entre.
   */
  public int getInputSize()
  {
    // Le paramtre offset reprsente le nombre de colonne dans la matrice si l'ordre est 1.
    return offset;
  }
  
  /**
   * Mthode pour obtenir le nombre de sortie de neurones associ  l'usage de ce paramtre de poids (<i>weights</i>).
   * 
   * @return Le nombre de neurones de sortie.
   */
  public int getOutputSize()
  {
    return this.matrix.getNbLine();
  }
  
  @Override
  public int getNumberOfParameter()
  {
    return this.matrix.getNbLine() * this.matrix.getNbColumn();
  }
  
  /**
   * <p>
   * Mthode pour dfinir la valeur d'un poids.
   * </p>
   * <p>
   * Il est important de noter que cette accs est en reprsentation tensorielle.
   * Ainsi, le paramtre v est l'entre (mais la colonne de la matrice) et 
   * u est la sortie (mais la ligne de la matrice).
   * </p>
   * 
   * @param v L'indice du signal d'entre dans la couche de neurones (input).
   * @param u L'indice du signal de sortie dans la couche de neurones (output).
   * @param value La valeur du poids.
   * @throws SNaNValueException Si un valeur NaN doit tre affecte.
   */
  public void set(int v, int u, double value) throws SNaNValueException
  {
    this.matrix.set(u, v, this.analyseValue(value));
  }

  /**
   * <p>
   * Mthode pour dfinir la valeur d'un poids en prcisant l'ordre.
   * Il est  not que la position du poids pour tous les ordres se retrouveront sur une mme ligne dans la matrice des poids.
   * </p>
   * <p>
   * Il est important de noter que cette accs est en reprsentation tensorielle.
   * Ainsi, le paramtre v est l'entre (mais la colonne de la matrice) et 
   * u est la sortie (mais la ligne de la matrice).
   * </p>
   * 
   * @param v L'indice du signal d'entre dans la couche de neurones (input).
   * @param u L'indice du signal de sortie dans la couche de neurones (output).
   * @param o L'ordre du poids w_vu (doit tre suprieur  0).
   * @param value
   * @throws SNaNValueException
   */
  public void set(int v, int u, int o, double value) throws SNaNValueException
  {
    this.matrix.set(u, v + (o-1)*offset, this.analyseValue(value));
  }
  
  @Override
  public void add(SParameter parameter) throws IllegalArgumentException 
  {
    // Vrification du type de paramtre.
    if(!(parameter instanceof SWeights))
      throw new  IllegalArgumentException("Erreur SWeights 006 : Le paramtre n'est pas du bon type (" + parameter.getClass().getName() + ")."); 
    
    this.addParameter((SMatrixParameter)parameter);
  }
  
  @Override
  public void multiply(SParameter parameter) throws IllegalArgumentException
  {
    // Vrification du type de paramtre.
    if(!(parameter instanceof SWeights))
      throw new  IllegalArgumentException("Erreur SWeights 007 : Le paramtre n'est pas du bon type (" + parameter.getClass().getName() + ")."); 
    
    this.multiplyParameter((SMatrixParameter)parameter);
  }
  
  @Override
  public void learn(SGradientLayer gradient, double learning_rate) throws IllegalArgumentException
  {
    // Effectuer l'entranement si le gradient contient un accs au poids.
    if(!(gradient instanceof SContainWeights))
      throw new IllegalArgumentException("Erreur SWeights 008 : Le gradient n'est pas du bon type (" + gradient.getClass().getName() + ")."); 
     
    this.learnParameter(((SContainWeights)gradient).getWeights(), learning_rate);
  }
    
  @Override
  public String toString()
  {
    return "SWeights [weights=" + this.matrix + "]";
  }

  @Override
  public int hashCode()
  {
    final int prime = 31;
    int result = super.hashCode();
    result = prime * result + offset;
    result = prime * result + order;
    return result;
  }

  @Override
  public boolean equals(Object obj)
  {
    if (this == obj)
      return true;
    if (!super.equals(obj))
      return false;
    if (getClass() != obj.getClass())
      return false;
    SWeights other = (SWeights) obj;
    if (offset != other.offset)
      return false;
    if (order != other.order)
      return false;
    return true;
  }

   
}
