/**
 * 
 */
package sim.nn.reader;

import java.awt.Color;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import javax.imageio.ImageIO;

import sim.exception.SConstructorException;
import sim.exception.SNoImplementationException;
import sim.graphics.SColor;
import sim.graphics.STexture;
import sim.nn.SNNData;
import sim.util.SArrays;
import sim.util.SBrowser;
import sim.util.SLog;
import sim.util.SManyFilesFoundException;

/**
 * <p>
 * La classe <b>SNeuralClassifiedImageReader</b> reprsente un lecteur de donne pour rseau de neurones
 * pour classification d'image.
 * </p>
 * 
 * <p>
 * Pour que cette classe ralise adquatement la lecteur et que la classification soit adquate, 
 * il faut respecter la structure suivante de rpertoire :
 * <ul>- Crer un rpertoire "training" et "testing". </ul>
 * <ul>- Crer des sous-rpertoire dans les 2 rpertoires prcdents avec le nom de la classification o les images s'y retrouveront.
 * </p>
 * 
 * @author Simon Vezina
 * @since 2019-05-24
 * @version 2021-10-05
 */
public class SNeuralClassifiedImageReader extends SAbstractNeuralDataReader {

  private final static String DEFAULT_TRAINING_DIRECTORY = "training";
  private final static String DEFAULT_TESTING_DIRECTORY = "testing";
  
  /**
   * La constante <b>IS_MERGE_COLOR_SEPARATED</b> dtermine le format du vecteur correspondant  une image en couleur.
   * Si le format est <b>separated</b>, le vecteur retourn aura un format de tableau de rouge, tableau de vert et tableau de bleu. 
   */
  private final static boolean IS_MERGE_COLOR_SEPARATED = true;
  
  /**
   * La constante <b>DEFAULT_BLACK_VALUE</b> reprsente la valeur numrique <u>de rfrence</u> attribuable  la couleur noire tant {@value}.
   */
  private final static double DEFAULT_BLACK_VALUE = 0.0;
  
  /**
   * La constante <b>DEFAULT_WHITE_VALUE</b> reprsente la valeur numrique <u>de rfrence</u> attribuable  la couleur blanche tant {@value}.
   */
  private final static double DEFAULT_WHITE_VALUE = 1.0;
    
  protected double black_value;
  protected double white_value;
  protected double red_value;
  protected double green_value;
  protected double blue_value;
  
  protected final boolean is_colored;
  protected final String data_directory;
  protected final String[] classifications;
  private final List<double[]> expected_vector;
  
  /**
   * Constructeur d'un lecteur de donnes pour rseau de neurones  partir d'images classifies.
   * 
   * @param data_directory Le nom du rpertoire contenant les images.
   * @param colored_image Dterminer si l'image est en couleur (<b>true</b> si en couleur et <b>false</b> si noire et blanche).
   * @throws 
   */
  public SNeuralClassifiedImageReader(String data_directory, boolean colored_image) throws SConstructorException
  {
    super();
    
    try {
      
      // Trouver toutes les classifications d'entrainement.
      this.classifications = findAllTrainingClassification(data_directory);
      
      // Exception lance s'il n'y a pas de classification.
      if(classifications.length == 0)
        throw new SConstructorException("Erreur SNeuralClassifiedImageReader 001 : Aucune classification n'a t demand pour le rpertoire '" + data_directory + "'.");

      // Valider que les classifications d'entrainement et de test son identique.
      if(!validateClassifications(data_directory, classifications))
        throw new SConstructorException("Erreur SNeuralClassifiedImageReader 002 : Les classifications exiges ne se retourve pas dans les rpertoires de " + DEFAULT_TRAINING_DIRECTORY + " et " + DEFAULT_TRAINING_DIRECTORY + ".");
    
    }catch(IllegalArgumentException e) {
      throw new SConstructorException("Erreur SNeuralClassifiedImageReader 003 : Une erreur dans la dtermination des classifications est survenue.", e);
    }
    
    this.is_colored = colored_image;
    this.data_directory = data_directory;
    this.expected_vector = buildExpectedValue(this.classifications.length);
    
    // Couleur de valeur nulle.
    this.black_value = DEFAULT_BLACK_VALUE;
    
    // Couleur de valeur gale  1.
    this.white_value = DEFAULT_WHITE_VALUE;
    this.red_value = DEFAULT_WHITE_VALUE;
    this.green_value = DEFAULT_WHITE_VALUE;
    this.blue_value = DEFAULT_WHITE_VALUE; 
  }
  
  /**
   * Constructeur d'un lecteur de donnes pour rseau de neurones  partir d'images classifies.
   * 
   * @param data_directory Le nom du rpertoire contenant les images.
   * @param classifications Les classifications  analyser.
   * @param colored_image Dterminer si l'image est en couleur (true si en couleur et false si noire et blanche).
   * @throws SConstructorException S'il y a eu une erreur lors de de la construction de l'objet.
   */
  public SNeuralClassifiedImageReader(String data_directory, String[] classifications, boolean colored_image) throws SConstructorException
  {
    super();
    
    // Exception lance s'il n'y a pas de classification.
    if(classifications.length == 0)
      throw new SConstructorException("Erreur SNeuralClassifiedImageReader 004 : Aucune classification n'a t demand pour le rpertoire '" + data_directory + "'.");

    // Validations des classifications et du rpertoire qui sera en lecture.
    try {
      
      if(!validateClassifications(data_directory, classifications))
        throw new SConstructorException("Erreur SNeuralClassifiedImageReader 005 : Les classifications exiges ne se retourve pas dans les rpertoires de " + DEFAULT_TRAINING_DIRECTORY + " et " + DEFAULT_TRAINING_DIRECTORY + ".");
    
    }catch(IllegalArgumentException e) {
      throw new SConstructorException("Erreur SNeuralClassifiedImageReader 006 : Une erreur de lecture est survenue.", e);
    }
    
    this.is_colored = colored_image;
    this.data_directory = data_directory;
    this.classifications = classifications;
    this.expected_vector = buildExpectedValue(this.classifications.length);
    
    // Couleur de valeur nulle.
    this.black_value = DEFAULT_BLACK_VALUE;
    
    // Couleur de valeur gale  1.
    this.white_value = DEFAULT_WHITE_VALUE;
    this.red_value = DEFAULT_WHITE_VALUE;
    this.green_value = DEFAULT_WHITE_VALUE;
    this.blue_value = DEFAULT_WHITE_VALUE;     
  }

  @Override
  public boolean isImageData()
  {
    return true;
  }
  
  /**
   * ...
   * 
   * @param data_type Le nom du rpertoire  lire (training/testing)
   * @return
   * @throws SManyFilesFoundException
   * @throws FileNotFoundException
   * @throws IOException
   */
  private List<SNNData> readAllData(String data_type) throws IOException
  {
    // Liste des donnes en lecture.
    List<SNNData> list = new ArrayList<SNNData>();
    
    SBrowser browser = new SBrowser();
    browser.goTo(this.data_directory);
    browser.forward(data_type);
      
    // Parcourir la classification.
    for(int c = 0; c < classifications.length; c++)
    {
      // Rentrer dans le membre de la classification  lire.
      browser.forward(classifications[c]);
        
      // *** Message console. ***
      SLog.logWriteLine("Message SNeuralClassifiedImageReader - Chargement des vecteurs : " + classifications[c]);
                
      // Parcourir les fichiers du rpertoire.
      for(File f : browser.getFileList())
        list.add(buildData(f, c));
            
      // Reculer dans le rpertoire pour accder au autres membres de la classification.
      browser.back();
    }
    
    return list;
  }
  
  /**
   * ...
   * 
   * @param file
   * @param classification_index
   * @return
   * @throws IOException
   */
  private SNNData buildData(File file, int classification_index) throws IOException
  {
    // Crer l'image.
    BufferedImage buffer = ImageIO.read(file);
      
    // Construire la texture de l'image.
    STexture texture = new STexture(file.getName(), buffer);
      
    // Retourner la donne.
    return new SNNData(texture, imageToArray(buffer), this.expected_vector.get(classification_index), this.classifications);   
  }
  
  /**
   * Mthode pour gnrer un vecteur  partir d'une image interprte en nuance de gris.
   * 
   * @param image L'image  convertir en vecteur.
   * @return Le vecteur de l'image interprte en nuance de gris.
   */
  private double[] imageToArray(BufferedImage image)
  {
    if(this.is_colored)
      return coloredImageToArray(image);
    else
      return greyImageToArray(image);
  }
  
  /**
   * Mthode pour gnrer un vecteur  partir d'une image noire et blanche.
   * 
   * @param image L'image noire et blanche.
   * @return Le vecteur associ  l'image.
   */
  private double[] greyImageToArray(BufferedImage image)
  {
    // Taille de l'image.
    final int size = image.getHeight()*image.getWidth();
    
    // La valeur delta fait l'cart entre la correspondance des "couleur noir" et "couleur blanche".
    final double delta = 2.0*(this.white_value - this.black_value);
    
    // La base de la correspondance  des "couleur noir" et "couleur blanche".
    final double max = this.white_value;
    
    double[] array = new double[size];
    int index = 0;
    
    for(int j = 0; j < image.getHeight(); j++)          // Iterer sur la ligne.
      for(int i = 0; i < image.getWidth(); i++)         // Iterer sur la colonne.
      {
        // Rfrence : https://stackoverflow.com/questions/9131678/convert-a-rgb-image-to-grayscale-image-reducing-the-memory-in-java
        int rgb = image.getRGB(j, i);
        int r = (rgb >> 16) & 0xFF;
        int g = (rgb >> 8) & 0xFF;
        int b = rgb & 0xFF;
        
        int greyLevel = (r + g + b) / 3;
        int grey = (greyLevel << 16) + (greyLevel << 8) + greyLevel;
        
        // Construire une couleur dans l'espace des couleurs (valeur entre 0.0 et 1.0).
        SColor color = new SColor(new Color(grey));
        
        // Dfinition de la valeur de la couleur entre "black_value" et "white_value".
        array[index] = color.getRed()*delta - max;
                
        // Changer le positionnement dans le vecteur.
        index++;
      }
    
    return array;
  }
  
  /**
   * Mthode pour gnrer un vecteur  partir d'une image en couleur.
   * 
   * @param image L'image en couleur.
   * @return Le vecteur associ  l'image.
   */
  private double[] coloredImageToArray(BufferedImage image)
  {
    // Taille de l'image.
    final int size = image.getHeight()*image.getWidth();
    
    double[] array_red = new double[size];
    double[] array_green = new double[size];
    double[] array_blue = new double[size];
    
    // La valeur delta fait l'cart entre la correspondance des "couleur noir" et "couleur blanche".
    final double delta_red = 2.0*(this.red_value - this.black_value);
    final double delta_green = 2.0*(this.green_value - this.black_value);
    final double delta_blue = 2.0*(this.blue_value - this.black_value);
    
    int index = 0;
    
    // Itrer sur l'ensemble des couleurs de l'image
    for(int j = 0; j < image.getHeight(); j++)          // Iterer sur la ligne.
      for(int i = 0; i < image.getWidth(); i++)         // Iterer sur la colonne.
      {
        // La couleur en (i,j)
        Color color = new Color(image.getRGB(i, j));
        
        // Extraire la couleur par canaux 0 .. 1.
        float[] compArray = new float[4];
        compArray = color.getColorComponents(compArray);
        
        // Recadrer la valeur entre les extremums 
        array_red[index] = ((double) compArray[0])*delta_red - this.red_value;
        array_green[index] = ((double) compArray[1])*delta_green - this.green_value;
        array_blue[index] = ((double) compArray[2])*delta_blue - this.blue_value;
               
        // Changer le positionnement dans le vecteur.
        index++;
      }
    
    // Retourner le vecteur correspondant  l'image dans le format dsir.
    if(IS_MERGE_COLOR_SEPARATED)
      return mergeArrayColorSeparated(array_red, array_green, array_blue);
    else
      return mergeArrayColorRGB(array_red, array_green, array_blue);
    
  }
  
  /**
   * ...
   * 
   * @param red
   * @param green
   * @param blue
   * @return
   */
  private static double[] mergeArrayColorSeparated(double[] red, double[] green, double[] blue)
  {
    return SArrays.merge(red, green, blue);
  }
  
  /**
   * ...
   * 
   * @param red
   * @param green
   * @param blue
   * @return
   */
  private static double[] mergeArrayColorRGB(double[] red, double[] green, double[] blue)
  {
    throw new SNoImplementationException();
  }
  
  /**
   * Mthode pour obtenir la liste de toutes les classifications disponible dans un rpertoire d'images classifies.
   * 
   * @param data_directory Le nom du rpertoire contenant les images.
   * @return Le tableau des classifications.
   * @throws SManyFilesFoundException S'il y a eu plusieurs rpertoire de trouv.
   */
  private static String[] findAllTrainingClassification(String data_directory) throws IllegalArgumentException
  {
    try {
      
      // Liste des noms des classifications.
      List<String> all_training = new ArrayList<String>();
            
      // Se rendre dans le rpertoire contenant l'ensemble des classifications d'entranement.
      SBrowser browser = new SBrowser();
      browser.goTo(data_directory);
      browser.forward(DEFAULT_TRAINING_DIRECTORY);
      
      List<File> directory = browser.getDirectoryList();
      
      for(File f : directory)
        all_training.add(f.getName());
      
      // Les listes sont identiques, on peut y aller.
      return all_training.toArray(new String[all_training.size()]);
      
    }catch(IOException e) {
      throw new IllegalArgumentException("Erreur SNeuralClassifiedImageReader 007 : Une erreur est survenue  la lecture", e);
    }
  }

  /**
   * Mthode pour valider les classifications. Cette mthode va vrifier que la liste exige se retrouve  la fois dans le training et dans le testing.
   * 
   * @param data_directory
   * @param classifications
   * @return
   * @throws SConstructorException
   */
  private static boolean validateClassifications(String data_directory, String[] classifications) throws IllegalArgumentException
  {
    try {
      
      // Localiser un browser pour indentifier les rpertoires contenant les classifications d'entranement.
      SBrowser browser_training = new SBrowser();
      browser_training.goTo(data_directory);
      browser_training.forward(DEFAULT_TRAINING_DIRECTORY);
      
      // Localiser un browser pour indentifier les rpertoires contenant les classifications d'entranement.
      SBrowser browser_testing = new SBrowser();
      browser_testing.goTo(data_directory);
      browser_testing.forward(DEFAULT_TESTING_DIRECTORY);
      
      List<File> directory_training = browser_training.getDirectoryList();
      List<String> string_training = new ArrayList<String>();
      
      for(File f : directory_training)
        string_training.add(f.getName());
      
      List<File> directory_testing = browser_testing.getDirectoryList();
      List<String> string_testing = new ArrayList<String>();
      
      for(File f : directory_testing)
        string_testing.add(f.getName());
      
      // Validation que l'ensemble des classifications se retrouve dans le training et le testing.
      for(String s : classifications)
      {
        if(!string_training.contains(s))
          return false;
        
        if(!string_testing.contains(s))
          return false;
      }
      
      return true;
      
    }catch(IOException e) {
      throw new IllegalArgumentException("Erreur SNeuralClassifiedImageReader 008 : Une erreur est survenue  la lecture", e);
    }
  }
  
  @Override
  public void readAllTrainingData() 
  {
    try {
      
      // Lecture des donnes.
      List<SNNData> l = readAllData(DEFAULT_TRAINING_DIRECTORY);
      
      // Copie des donnes dans la bonne liste officielle.
      this.training_list.addAll(l);
      
    } catch (IOException e) {
      throw new RuntimeException("Erreur SNeuralClassifiedImageReader 009 : Erreur de lecture.", e);
    }
    
  }

  @Override
  public void readRandomTrainingData(int nb) throws IllegalArgumentException 
  {
    throw new SNoImplementationException();  
  }

  @Override
  public void readAllTestingData()
  {
    try {
      
      // Lecture des donnes.
      List<SNNData> l = readAllData(DEFAULT_TESTING_DIRECTORY);
      
      // Copie des donnes dans la bonne liste officielle.
      this.testing_list.addAll(l);
      
    } catch (IOException e) {
      throw new RuntimeException("Erreur SNeuralClassifiedImageReader 010 : Erreur de lecture.", e);
    }
  }

  @Override
  public void readRandomTestingData() throws IllegalArgumentException
  {
    throw new SNoImplementationException();  
  }
  
}
