/**
 * 
 */
package sim.nn.reader;

import java.util.ArrayList;
import java.util.List;

import sim.nn.SNNData;

/**
 * L'interface <b>SNeuralDataReader</b> reprsente un lecteur de donne pour rseau de neurones.
 * 
 * @author Simon Vezina
 * @since 2019-12-08
 * @version 2021-09-06
 */
public interface SNeuralDataReader {

  /**
   * Mthode pour dterminer si les donnes obtenues par le lecteur est de type "image".
   * 
   * @return <b>true</b> si les donnes sont des images et <b>false</b> sinon.
   */
  public default boolean isImageData()
  {
    return false;
  }
  
  /**
   * Mthode pour obtenir le nombre de neurones des donne d'entranement/test.
   * Ce nombre correspond au nombre de neurones d'entre d'un rseau de neurones.
   * 
   * @return Le nombre de neurones des donnes.
   * @throws IllegalStateException Si aucune donne a t lue par le lecteur.
   */
  public int getDataSize() throws IllegalStateException;
  
  /**
   * Mthode pour obtenir le nombre classifications associes aux donnes d'entranement/test. 
   * Ce nombre correspond au nombre de neurones en sortie d'un rseau de neurones.
   * S'il n'y a pas de donne disponible, le nombre de classification sera gal  zro.
   * 
   * @return Le nombre de classifications associes aux donnes.
   */
  public int getClassificationSize();
  
  /**
   * Mthode pour obtenir la liste des donnes d'entranement du lecteur de donnes.
   * 
   * @return La liste des donnes d'entranement.
   */
  public List<SNNData> getAllTrainingData();
  
  /**
   * Mthode pour obtenir un nombre limits de donnes d'entranement du lecteur choisi alatoirement.
   * 
   * @param nb Le nombre de donne d'entranement.
   * @return La list des donnes d'entranement.
   */
  public default List<SNNData> getRandomTrainingData(int nb) throws IllegalArgumentException
  {
    return randomData(getAllTrainingData(), nb);
  }
  
  /**
   * Mthode pour obtenir la liste des donnes de test du lecteur de donnes.
   * 
   * @return La liste des donnes de test.
   */
  public List<SNNData> getAllTestingData();
    
  /**
   * Mthode pour obtenir un nombre limits de donnes de test du lecteur choisi alatoirement.
   * 
   * @param nb Le nombre de donne de test.
   * @return La list des donnes de test.
   */
  public default List<SNNData> getRandomTestingData(int nb) throws IllegalArgumentException
  {
    return randomData(getAllTestingData(), nb);
  }
  
  /**
   * Mthode pour faire la lecture de toutes les donnes d'entrainement.
   */
  public void readAllTrainingData();
  
  /**
   * Mthode pour faire la lecture d'un nombre limits de <u>donnes d'entrainement</u> choisies alatoirement.
   * Cette mthode peut tre utile si la base de donnes est trop grande pour tre charge compltement.
   * 
   * @param nb Le nombre de donnes.
   * @throws IllegalArgumentException Si le nombre de donne est ngatif.
   */
  public void readRandomTrainingData(int nb) throws IllegalArgumentException;
  
  /**
   * Mthode pour faire la lecture de toutes les donnes de test.
   */
  public void readAllTestingData();
  
  /**
   * Mthode pour faire la lecture d'un nombre limits de <u>donnes de test</u> choisies alatoirement.
   * Cette mthode peut tre utile si la base de donnes est trop grande pour tre charge compltement.
   * 
   * @param nb Le nombre de donnes.
   * @throws IllegalArgumentException Si le nombre de donne est ngatif.
   */
  public void readRandomTestingData() throws IllegalArgumentException;
  
  /**
   * Mthode pour effacer l'intgralit des donnes du lecteur.
   */
  public void clear();
    
  /**
   * Mthode pour obtenir un nombre limits de donnes choisi alatoirement dans une liste.
   * 
   * @param nb Le nombre de donne.
   * @return La list des donnes.
   * @throws IllegalArgumentException Si le nombre nb est ngatif.
   */
  private List<SNNData> randomData(List<SNNData> list, int nb) throws IllegalArgumentException
  {
    // Validation du nombre nb > -1.
    if(nb < 0)
      throw new IllegalArgumentException("Erreur SNeuralDataReader 001 : Le nombre nb = " + nb + " ne peut pas tre ngatif.");
    
    List<SNNData> sub_list = new ArrayList<SNNData>();
    
    for(int i = 0; i < nb; i++)
    {
      int index = (int)(Math.random() * list.size());
      
      sub_list.add(list.get(index));
    }
    
    return sub_list;
  }
  
}
