/**
 * 
 */
package sim.nn.reader;

import java.awt.Color;
import java.awt.image.BufferedImage;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import sim.exception.SConstructorException;
import sim.exception.SNoImplementationException;
import sim.nn.SNNData;
import sim.util.SBrowser;
import sim.util.SStringUtil;

/**
 * La classe <b>SNeuralImageDataReader</b> reprsente un lecteur de donne pour rseau de neurones  partir d'un <u>fichier image unique</u>.
 * L'image sera analyse et chaque couleur de l'image correspondra  un lment de la classification.
 * 
 * @author Simon Vezina
 * @since 2019-11-24
 * @version 2021-09-06
 */
public class SNeuralImageDataReader extends SAbstractNeuralDataReader {

  /**
   * ...
   */
  private final int DEFAULT_IMAGE_SIZE = 500;
  
  /**
   * ...
   */
  private final boolean DEFAULT_BALANCED_DATA = true;
  
  /**
   * ...
   */
  private final double DEFAULT_TESTING_RATIO = 0.2;
  
  /**
   * La variable <b>file_name</b> reprsente le nom du fichier de l'image  analyser.
   */
  private final String file_name;
  
  /**
   * ...
   */
  private final Map<String, List<double[]>> data_map;
  
  /**
   * ...
   */
  private final String[] classifications;
    
  /**
   * ...
   */
  private final List<double[]> expected_vector;
  
  /**
   * ...
   */
  private final double testing_ratio;
  
  /**
   * ...
   */
  private final boolean balanced_data;
  
  /**
   * La variable <b>data_image</b> reprsente l'ensemble des donnes en image.
   */
  private final BufferedImage data_image;
  
  //----------------
  // CONSTRUCTEUR //
  //----------------
  
  /**
   * Constructeur 
   * 
   * @param image_file_name
   * @throws SConstructorException
   */
  public SNeuralImageDataReader(String image_file_name) throws SConstructorException
  {
    super();
    
    this.file_name = image_file_name;
    String file_name_without_extension = SStringUtil.getFileNameWithoutExtension(image_file_name);  
    
    this.testing_ratio = DEFAULT_TESTING_RATIO;
    this.balanced_data = DEFAULT_BALANCED_DATA;
    
    try {
      
      // Lecture du fichier image.
      SBrowser browser = new SBrowser();
      this.data_image = browser.readImageFile(image_file_name);
      
      // Chargement des donnes.
      this.data_map = readImageData(this.data_image);
      
      // Dtermination des vecteurs attendue.
      this.expected_vector = buildExpectedValue(this.data_map.keySet().size());
      
      // Determination de l'ensemble des classifications.
      this.classifications = this.data_map.keySet().toArray(new String[this.data_map.keySet().size()]);
      
      // Balancer les donnes des classification si requis.
      balanceData();
      
      // Affecter les donnes  la liste d'entranement et de test.
      fillTrainingAndTestingList();
      
      // criture des donnes en image.
      browser.writeImageFile(file_name_without_extension + "_training", "png", buildImageData(this.training_list));
      browser.writeImageFile(file_name_without_extension + "_testing", "png", buildImageData(this.testing_list));
      
      // criture des donnes en text.
      browser.writeTextFile(file_name_without_extension + "_training", "txt", buildStringData(this.training_list).toString());
      browser.writeTextFile(file_name_without_extension + "_testing", "txt", buildStringData(this.testing_list).toString());
      
    }catch(IOException e) {
      throw new SConstructorException("Erreur SNeuralImageDataReader 001 : Une erreur est survenue lors de la cration de l'objet.", e);
    }
    
  }

  /**
   * ...
   * 
   * @return
   */
  public BufferedImage getDataImage()
  {
    return this.data_image;
  }
  
  /**
   * ...
   * 
   * @return
   */
  public BufferedImage getTrainingDataImage()
  {
    return buildImageData(this.training_list);
  }
  
  /**
   * ...
   * 
   * @return
   */
  public BufferedImage getTestingDataImage()
  {
    return buildImageData(this.testing_list);
  }
  
  /**
   * ...
   * 
   * @return
   */
  public BufferedImage getClassificationErrorTestingImage()
  {
    return buildClassificationErrorTestingImageData();
  }
    
  /**
   * Mthode pour obtenir le nom du fichier d'image lu (avec extension).
   * 
   * @return Le nom du fichier image.
   */
  public String getFileName()
  {
    return this.file_name;
  }
  
  /**
   * Mthode pour le tableau des couleurs utilises dans la classification de l'image.
   * 
   * @return Les couleurs de la classification.
   */
  public String[] getClassificationStringColor()
  {
    return this.classifications;
  }
  
  /**
   * Mthode pour faire la lecture de l'image et de gnrer l'ensemble des donnes.
   * 
   * @return La liste des donnes.
   */
  private Map<String,List<double[]>> readImageData(BufferedImage buffer)
  {
    // La carte des donnes tri par classification.
    Map<String, List<double[]>> map = new HashMap<String, List<double[]>>();
    
    // Dfinition de la couleur blanc/noir en int correspondant  aucune classification.
    int white = Color.white.getRGB();
    int black = Color.black.getRGB();
    
    //Iteration sur l'ensemble de l'image.
    for(int i = 0; i < buffer.getWidth(); i++)
      for(int j = 0; j < buffer.getHeight(); j++)
      {
        int color = buffer.getRGB(i, j);
        
        // Analyse de la couleur non blanche.
        if(color != white && color != black)
        {
          String string_color = Integer.toString(color);
          
          double x = (double)i / (double)buffer.getWidth();
          double y = (double)j / (double)buffer.getHeight();
          
          double[] data = { x, y };
          
          // Ajout de la donne  la carte tri par classification.
          if(map.containsKey(string_color))
            map.get(string_color).add(data);
          else
          {
            // Si la cl n'est pas dans la carte, il faudra l'ajouter.
            List<double[]> list = new ArrayList<double[]>();
            list.add(data);
            map.put(string_color, list);
          }
        }
      }
    
    return map;
  }
  
  /**
   * Mthode pour balancer les donnes afin qu'il n'y ait pas de surreprsentation de donnes pour une classification.
   * Dans ce processus, les vecteurs seront mlang dans un mme classification perdant ainsi l'ordre associ  la lecteur du fichier image.  
   */
  private void balanceData()
  {
    if(this.balanced_data)
    {
      // Identifier le nombre minimum de vecteur contenu dans un classification.
      int min = Integer.MAX_VALUE;
      
      for(String string_color : this.data_map.keySet())
        if(min > this.data_map.get(string_color).size())
          min = this.data_map.get(string_color).size();
      
      // Rduire le nombre de vecteur de chaque classification  la valeur minimale.
      for(String string_color : this.data_map.keySet())
      {
        List<double[]> list = this.data_map.get(string_color);
        
        // Mlanger les vecteurs de la liste.
        Collections.shuffle(list);
        
        // Retirer les lments (toujours le premier, car mlang.
        while(list.size() > min)
          list.remove(0);
      }
    }
  }
  
  /**
   * Mthode pour remplir les listes des donnes d'entrainement et de test.
   */
  private void fillTrainingAndTestingList()
  {
    // Itrer sur l'ensemble des classifications  partir de l'ordre du tableau contenant les noms des classifications.
    for(int i = 0; i < classifications.length; i++)
    {
      String obj = classifications[i];
      double[] expected = expected_vector.get(i);
      
      // Liste des donnes diponibles pour cette classification.
      List<double[]> data_list = this.data_map.get(obj);
      
      // Nombre de donnes  mettre dans le testing.
      int testing_number = (int) (this.testing_ratio * data_list.size());
                    
      // Remplir les listes.
      for(int d = 0; d < data_list.size(); d++)
      {
        // Construire la donne de rseau de neurones.
        SNNData data = new SNNData(obj, data_list.get(d), expected, classifications);
            
        if(d < testing_number)
          testing_list.add(data);
        else
          training_list.add(data);
      }
    }
  }
  
  /**
   * Mthode pour gnrer une image  partir d'une liste de donnes.
   * 
   * @param data_list La liste des donnes.
   * @return Une image correspondant  la liste des donnes.
   */
  private BufferedImage buildImageData(List<SNNData> data_list)
  {
    BufferedImage buffer = new BufferedImage(DEFAULT_IMAGE_SIZE, DEFAULT_IMAGE_SIZE, BufferedImage.TYPE_INT_RGB);
    
    // Itrer sur l'ensemble des donnes.
    for(SNNData data : data_list)
    {
      int x = (int) (buffer.getWidth() * data.getVector()[0]);
      int y = (int) (buffer.getHeight() * data.getVector()[1]);
      
      //int rgb = Integer.parseInt((String) data.getObj());
      int rgb = Integer.parseInt((String) data.getExpectedClassification());
      
      // crire la couleur dans le buffer.
      buffer.setRGB(x, y, rgb);
    }
    
    return buffer;
  }
  
  /**
   * Mthode pour construire l'image des erreurs de la classification des donnes de test.
   * Ce sont uniquement les pixels mal classifis qui seront affich avec leur couleur d'activation (tant errone).
   * 
   * @return L'image des erreurs de classification des donnes de test.
   */
  private BufferedImage buildClassificationErrorTestingImageData()
  {
    BufferedImage buffer = new BufferedImage(DEFAULT_IMAGE_SIZE, DEFAULT_IMAGE_SIZE, BufferedImage.TYPE_INT_RGB);
    
    for(SNNData data : this.testing_list)
    {
      // Afficher uniquement les donnes mal classifies
      if(!data.isClassified())
      {
        int x = (int) (buffer.getWidth() * data.getVector()[0]);
        int y = (int) (buffer.getHeight() * data.getVector()[1]);
        
        int rgb = Integer.parseInt((String) data.getActivationClassification());
        
        // crire la couleur dans le buffer.
        buffer.setRGB(x, y, rgb);
      }
    }
    
    return buffer;
  }
  
  /**
   * ...
   * 
   * @param data_list
   * @return
   */
  private StringBuffer buildStringData(List<SNNData> data_list)
  {
    StringBuffer buffer = new StringBuffer();
    
    // Itrer sur l'ensemble des donnes.
    for(SNNData data : data_list)
    {
      buffer.append(Double.toString(data.getVector()[0]));
      buffer.append(";");
      buffer.append(Double.toString(data.getVector()[1]));
      buffer.append(";");
      buffer.append((String) data.getObj());
      buffer.append(SStringUtil.END_LINE_CARACTER);
    }
    
    return buffer;
  }
 
  @Override
  public void readAllTrainingData() 
  {
    throw new SNoImplementationException("Erreur SNeuralImageDataReader : Prsentement, la lecture sur faire  la construction de l'objet.");      
  }

  @Override
  public void readRandomTrainingData(int nb) throws IllegalArgumentException
  {
    throw new SNoImplementationException();  
  }

  @Override
  public void readAllTestingData() 
  {
    throw new SNoImplementationException("Erreur SNeuralImageDataReader : Prsentement, la lecture sur faire  la construction de l'objet.");  
  }

  @Override
  public void readRandomTestingData() throws IllegalArgumentException 
  {
    throw new SNoImplementationException();  
  }
  
  
    /*
    try {
        
      // Pour la lecture de l'image.
      SBrowser browser = new SBrowser();
      File file = browser.findFile(image_file_name);
    
      // Pour l'criture des donnes.
      FileWriter fw = new FileWriter(data_file_name);
      BufferedWriter bw = new BufferedWriter(fw);
      
      // Crer l'image.
      BufferedImage buffer = ImageIO.read(file);
      
      // Dfinition de la couleur blanc en int correspondant  aucune classification.
      int white = Color.white.getRGB();
      
      int counter = 0;
      
      // Iteration sur l'ensemble de l'image.
      for(int i = 0; i < buffer.getWidth(); i++)
        for(int j = 0; j < buffer.getHeight(); j++)
        {
          int color = buffer.getRGB(i, j);
      
          // Analyse de la couleur non blanche.
          if(color != white)
          {
            // criture de l'information de la couleur en format : #,x,y,class
            StringBuffer s = new StringBuffer();
            s.append(counter);
            s.append(",");
            s.append((double)i / (double)buffer.getWidth());
            s.append(",");
            s.append((double)j / (double)buffer.getHeight());
            s.append(",");
            s.append(Integer.toString(color));
            
            bw.write(s.toString());
            bw.write(SStringUtil.END_LINE_CARACTER);
            bw.flush();
            
            counter++;
          }
        }
      
      bw.close();
      fw.close();
      
    }catch(IOException e) {
      e.printStackTrace();
    }
     */
    
}
