/**
 * 
 */
package sim.nn.reader;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.StringTokenizer;

import sim.exception.SConstructorException;
import sim.exception.SNoImplementationException;
import sim.exception.SParsingException;
import sim.nn.SNNData;
import sim.util.SBrowser;
import sim.util.SBufferedReader;
import sim.util.SLog;
import sim.util.SManyFilesFoundException;
import sim.util.SStringUtil;

/**
 * La classe <b>SNeuralStringDataReader</b> est un lecteur de donne dans le format suivant : "id","value1","value2",...,"classification"
 * 
 * Cette classe est un sous-produit d'un reader de type CSV.
 * 
 * @author Simon Vezina
 * @since 2019-11-24
 * @version 2022-03-04
 */
public class SNeuralStringDataReader extends SAbstractNeuralDataReader {

  protected final List<String> classifications;
  
  protected final List<double[]> expected_vector;
  
  protected final int inputs;
  
  protected final List<SNNData> training_list;
  
  protected final List<SNNData> testing_list;
  
  /**
   * ...
   * 
   * @param data_file_name
   * @param inputs
   * @throws SConstructorException
   */
  public SNeuralStringDataReader(String data_file_name, int inputs) throws SConstructorException
  {
    this.inputs = inputs;
    
    try {
      
      List<String> string_list = readAllStringData(data_file_name);
      
      this.classifications = findAllClassifications(string_list);
      this.expected_vector = buildExpectedValue(this.classifications.size());
      
      this.training_list = buildNeuralData(string_list, inputs);
      this.testing_list = this.training_list;
      
    }catch(IOException e) {
      throw new SConstructorException("Erreur SNeuralStringDataReader 001 : La lecture n'a pas t possible.", e);
    }
    
  }

    
  /**
   * Mthode pour faire la lecture de l'ensemble des donnes.
   * 
   * @param data_file
   * @return
   * @throws SManyFilesFoundException
   * @throws FileNotFoundException
   * @throws IOException
   */
  private List<String> readAllStringData(String data_file) throws SManyFilesFoundException, FileNotFoundException, IOException
  {
    SBrowser browser = new SBrowser();
    File file = browser.findFile(data_file);
    
    FileReader fr = new FileReader(file);
    SBufferedReader sbr = new SBufferedReader(fr);
      
    List<String> list = new ArrayList<String>();
    
    boolean end_reading = false;        //condition de fin de lecture
    
    while(!end_reading)
    {   
      // Lecture de la ligne
      String line = sbr.readLine();    
      
      //Si la ligne lu est "null", la fin du fichier est atteinte
      if(line == null)
        end_reading = true;
      else
        if(line.length() > 0)
          if(line.charAt(0) != '#')
            list.add(line);
     }
    
    sbr.close();
    
    return list;
  }  
   
  /**
   * Mthode pour obtenir la liste de toutes les classes.
   * 
   * @param string_list La liste des donnes en 
   */
  private List<String> findAllClassifications(List<String> string_list)
  {
    List<String> classifications = new ArrayList<String>();
    
    for(String s : string_list)
    {
      StringTokenizer tokens = new StringTokenizer(s, SStringUtil.REMOVE_CARACTER_TOKENIZER); 
      
      // Obtenir le dernier token.
      String last_token = tokens.nextToken();
      
      while(tokens.hasMoreTokens())
        last_token = tokens.nextToken();
      
      if(!classifications.contains(last_token))
        classifications.add(last_token);
    }
    
    return classifications;
  }
    
  /**
   * ...
   * 
   * @param string_list
   * @param inputs
   * @return
   * @throws SParsingException
   */
  private List<SNNData> buildNeuralData(List<String> string_list, int inputs) throws SParsingException
  {
    List<SNNData> data_list = new ArrayList<SNNData>();
    String[] classifications_array = classifications.toArray(new String[classifications.size()]);
    
    int counter = 1;
    
    for(String s: string_list)
    {
      StringTokenizer tokens = new StringTokenizer(s, SStringUtil.REMOVE_CARACTER_TOKENIZER); 
      
      // Validation de l'entre.
      if(!(tokens.countTokens() == (inputs + 2) || tokens.countTokens() == (inputs + 1)))
        throw new SParsingException("Erreur SNeuralStringDataReader 002 : La donne " + s + " n'a pas le bon nombre d'lment. Le nombre d'input requis doit tre de " + inputs + ".");
      
      // Dfinition de l'indentification du vecteur.
      String id;
      
      // Parsing avec numro/string d'identification.
      if(tokens.countTokens() == (inputs + 2))
        id = tokens.nextToken();
      else
        id = Integer.toString(counter);
      
      counter++;
      
      // Remplissage du vecteur avec le nombre d'input prvu.
      double[] vector = new double[inputs];
      
      try {
     
        for(int i = 0; i < inputs; i++)
          vector[i] = Double.parseDouble(tokens.nextToken());
        
      }catch(NumberFormatException e) {
        SLog.logWriteLine("Message SNeuralStringDataReader : La donne " + s + " n'est pas dans un format valide.");
      }
      
      // Obtenir le string de la classification.
      String classifier = tokens.nextToken();
            
      // Ajouter la donne.
      data_list.add(new SNNData(id, vector, expected_vector.get(classifications.indexOf(classifier)), classifications_array)); 
    }
    
    return data_list;
  }


  @Override
  public int getDataSize() 
  {
    // Cas particulier s'il n'y a pas de donne.
    if(this.training_list.size() == 0)
      return 0;
    
    return this.training_list.get(0).getDataSize();
  }


  @Override
  public int getClassificationSize() 
  {
    // Cas particulier s'il n'y a pas de donne.
    if(this.training_list.size() == 0)
      return 0;
    
    return this.training_list.get(0).getExpectedSize();
  }


  @Override
  public List<SNNData> getAllTrainingData()
  {
    return this.training_list;
  }


  @Override
  public List<SNNData> getAllTestingData()
  {
    return this.testing_list;
  }


  @Override
  public void readAllTrainingData() 
  {
    throw new SNoImplementationException();  
  }


  @Override
  public void readRandomTrainingData(int nb) throws IllegalArgumentException 
  {
    throw new SNoImplementationException();      
  }


  @Override
  public void readAllTestingData() 
  {
    throw new SNoImplementationException();      
  }


  @Override
  public void readRandomTestingData() throws IllegalArgumentException 
  {
    throw new SNoImplementationException();      
  }

}
