/**
 * 
 */
package sim.nn.trainer;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import sim.exception.SIllegalNegativeValueException;
import sim.nn.SNNData;
import sim.nn.network.SNetwork;
import sim.util.SStringUtil;

/**
 * La classe <b>SAbstractTrainer</b> reprsente un entraneur abstrait.
 * Le rseau de neurone ainsi que la fonction d'erreur n'y sont pas dfinis.
 * 
 * @author Simon Vezina
 * @since 2019-08-18
 * @version 2022-01-01
 */
public abstract class SAbstractTrainer extends SDataManager {

  /**
   * La constante <b>DEFAULT_LEARNING_RATE</b> reprsente le taux d'apprentissage par dfaut tant gal  {@value}.
   */
  protected final static double DEFAULT_LEARNING_RATE = 0.1;
  
  /**
   * La variable <b>learning_rate</b> reprsente le taux d'apprentissage qui sera ralis par le rseau aprs chaque entranement. 
   */
  protected double learning_rate;
  
  /**
   * La variable <b>training_count</b> dtermine le nombre d'entrainement ralis.
   */
  protected int training_count;
  
  /**
   * La variable <b>data_count</b> dtermine le nombre de donne utilis lors des entrainement.
   */
  protected int data_count;
  
  /**
   * La variable <b>is_testing_error_computed</b> dtermine si les erreurs sur les donnes de test ont t calcules. 
   */
  protected boolean is_testing_error_computed;
    
  //----------------
  // CONSTRUCTEUR //
  //----------------
  
  /**
   * ...
   * 
   * @param network
   */
  public SAbstractTrainer(SNetwork network)
  {
    super(network.getInputsSize(), network.getOutputsSize());
    
    // Dfinir le taux d'apprentissage par dfaut.
    this.learning_rate = DEFAULT_LEARNING_RATE;
    
    // Initialiser les paramtres de statistique. 
    this.training_count = 0;
    this.data_count = 0;
    
    // Dfinir les donnes de validation comme tant non values.
    this.is_testing_error_computed = false;
  }
  
  //------------
  // MTHODES //
  //------------
  
  /**
   * Mthode pour obtenir le nom de la fonction d'erreur utilise lors de l'entranement.
   * 
   * @return Le nom de la fonction d'erreur.
   */
  public abstract String getErrorFunctionName();
  
  /**
   * Mthode pour obtenir le rseau en entranement.
   * 
   * @return Le rseau en entranement.
   */
  public abstract SNetwork getNetwork();
  
  /**
   * Mthode pour obtenir le taux d'apprentissage de l'entranement.
   * 
   * @return Le taux d'apprentissage.
   */
  public double getLearningRate()
  {
    return learning_rate;
  }
  
  /**
   * Mthode pour obtenir l'erreur moyenne de l'ensemble des donnes d'une liste pralablement activ par un rseau.
   * 
   * @return L'erreur moyenne sur l'ensemble des donnes de test.
   * @throws IllegalStateException Si la liste est vide.
   */
  public double getAverageError(List<SNNData> list) throws IllegalStateException
  {
    // Validation de la taille de la liste.
    if(list.size() == 0)
      throw new IllegalStateException("Erreur SAbstractTrainer 001 : L'erreur moyenne n'a pas pu tre value, car la liste est vide.");
    
    // Calculer l'erreur moyenne des donnes de test.
    double err = 0.0;
    
    for(SNNData data : list)
      err += data.getError();
      
    return err / list.size();
  }
  
  /**
   * Mthode pour obtenir l'erreur moyenne de l'ensemble des donne de test gnre par le rseau.
   * 
   * @return L'erreur moyenne sur l'ensemble des donnes de test.
   * @throws IllegalStateException Si l'erreur n'a pas t valu sur l'ensemble des donnes de test.
   */
  public double getTestingAverageError() throws IllegalStateException
  {
    if(!is_testing_error_computed)
      throw new IllegalStateException("Erreur SAbstractTrainer 002 : Avant d'valuer l'erreur moyenne sur l'ensemble des donnes de test, vous devez obligatoirement avoir calcul l'erreur.");
    
    return getAverageError(this.testing_list);
  }
   
  /**
   * Mthode pour obtenir le taux de classification de l'ensemble des donnes de test de l'entraineur.
   * 
   * @return
   * @throws IllegalStateException Si l'erreur n'a pas t valu sur l'ensemble des donnes de test.
   */
  public double getTestingClassification() throws IllegalStateException
  {
    if(!is_testing_error_computed)
      throw new IllegalStateException("Erreur SAbstractTrainer 003 : Avant d'valuer le taux de classification des donnes de test, vous devez obligatoirement avoir calcul l'erreur.");
    
    // valuer le taux de classification des donnes de test.
    int count = 0;
    
    for(SNNData data : this.testing_list)
      if(data.isClassified())
        count++;
      
    return (double)count / (double)this.testing_list.size() * 100;
  }
  
  /**
   * Mthode pour obtenir le taux de classification, classe par classe, sur l'ensemble des donnes de test de l'entraneur.
   * 
   * @return Un expression donnant le taux de classification de chaque classe de la classification.
   * @throws IllegalStateException
   */
  public String getTestingClassificationDetails() throws IllegalStateException
  {
    if(!is_testing_error_computed)
      throw new IllegalStateException("Erreur SAbstractTrainer 004 : Avant d'valuer le taux de classification des donnes de test, vous devez obligatoirement avoir calcul l'erreur.");
   
    Map<String, Integer> nb_map = new HashMap<String, Integer>();
    Map<String, Integer> succes_map = new HashMap<String, Integer>();
    Map<String, Integer> counter_map = new HashMap<String, Integer>();
    
    // Initialiser  zro l'ensemble du comptage.
    Integer zero = Integer.valueOf(0);
    
    for(SNNData data : this.testing_list)
    {
      nb_map.put(data.getExpectedClassification(), zero);
      succes_map.put(data.getExpectedClassification(), zero);
      counter_map.put(data.getExpectedClassification(), zero);
    }
    
    // Compter le nombre d'lment par classe et les succes par classe..
    for(SNNData data : this.testing_list)
    {
      // Compter le nombre d'lment par classe.
      nb_map.put(data.getExpectedClassification(), Integer.valueOf(nb_map.get(data.getExpectedClassification()).intValue()+1));
      
      // Compter le nombre de donne bien class.
      if(data.isClassified())
        succes_map.put(data.getExpectedClassification(), Integer.valueOf(succes_map.get(data.getExpectedClassification()).intValue()+1));      
    
      // Compter quel classe est reconnue lors des activation.
      counter_map.put(data.getActivationClassification(), Integer.valueOf(counter_map.get(data.getActivationClassification()).intValue()+1));
    }
 
    // L'ensemble des classes.
    StringBuffer buffer = new StringBuffer();
    
    buffer.append("Statistique de classification :");
    buffer.append(SStringUtil.END_LINE_CARACTER);
    
    // Itrer sur l'ensemble des classes.
    for(String c : counter_map.keySet())
    {
      buffer.append("Classe " + c + " : " + Integer.toString(succes_map.get(c)) + " succs / " + Integer.toString(nb_map.get(c)) + " lments dont " + Integer.toString(counter_map.get(c)) + " ont t dtects.");
      buffer.append(SStringUtil.END_LINE_CARACTER);
    }
       
    // Retourner l'expression des statistiques.
    return buffer.toString();
  }
  
  /**
   * Mthode pour obtenir le nombre d'itration d'entrainement ralis.
   * 
   * @return Le nombre d'itration d'entrainement ralis.
   */
  public int getTrainingCount()
  {
    return this.training_count;
  }
  
  /**
   * <p>
   * Mthode pour obtenir le nombre d'poque d'entrainement ralis.
   * </p>
   * <p>
   * Une poque d'entrainement est ralise lorsqu'un nombre de donne d'entrainment gale  la taille de la liste d'entrainement a t utilis.
   * </p>
   * 
   * @return Le nombre d'poque d'entrainement ralis.
   */
  public int getEpochCount()
  {
    if(this.training_list.size() == 0)
      return 0;
    else
      return this.data_count / this.training_list.size();
  }
  
  /**
   * Mthode pour obtenir le nombre de donnes utilises lors des entrainements raliss.
   * 
   * @return Le nombre de donnes utilises.
   */
  public int getDataCount()
  {
    return this.data_count;
  }
     
  /**
   * ...
   * 
   * @return
   * @throws IllegalStateException Si l'erreur n'a pas t valu sur l'ensemble des donnes de test.
   */
  public String getStatistics() throws IllegalStateException
  {
    return "Epoch = " + this.getEpochCount() + 
        " , Training = " + this.getTrainingCount() + 
        " , Data = " + this.getDataCount() + 
        " --> Error = " + this.getTestingAverageError() +
        " , Classification = " + this.getTestingClassification() + "%";
  }
  
  /**
   * Mthode pour dfinir le taux d'apprentissage du rseau de neurones.
   * 
   * @param learning_rate Le taux d'apprentissage
   * @throws SIllegalNegativeValueException Si le taux d'apprentissage ({@code learning_rate}) est ngatif.
   */
  public void setLearningRate(double learning_rate) throws SIllegalNegativeValueException
  {
    // Vrifier que le taux d'apprentissage est positif. 
    if(learning_rate < 0.0)
      throw new SIllegalNegativeValueException("Erreur SAbstractTrainer 005 : Le taux d'apprentissage '" + learning_rate + "' ne peut pas tre ngatif");
    
    // Dfinir le taux d'apprentissage.
    this.learning_rate = learning_rate;
  }
  
  /**
   * Mthode pour raliser une mise  jour sur les statistiques de l'entranement.
   * Cette mthode ajoute <u>un entranement</u> avec l'usage de <u>une donne</u>.
   */
  protected void updateTrainingStatistic()
  {
    this.training_count++;
    this.data_count++;
    
    // Pour signaler  l'entraneur qu'une modification du rseau a t ralise et qu'un recalcul de l'erreur sur l'ensemble des donnes de test est ncessaire.
    this.is_testing_error_computed = false;
  }
  
  /**
   * Mthode pour raliser une mise  jour sur les statistiques de l'entranement.
   * Cette mthode ajoute <u>un entranement</u> avec l'usage de <u>plusieurs donne</u>.
   * @param mini_batch
   */
  protected void updateTrainingStatistic(int mini_batch)
  {
    this.training_count++;
    this.data_count += mini_batch;
    
    // Pour signaler  l'entraneur qu'une modification du rseau a t ralise et qu'un recalcul de l'erreur sur l'ensemble des donnes de test est ncessaire.
    this.is_testing_error_computed = false;
  }
  
  /**
   * Mthode pour calculer les erreurs sur une liste de donnes. Ces donnes seront actives par le rseau interne  l'entraneur 
   * et elles seront compare avec leur donne attendue  l'aide de la fonction d'erreur interne  l'entraneur.
   * 
   * @param data_list La liste des donnes.
   * @throws IllegalStateException Si la liste est vide ou elle est non conforme avec l'entraneur.
   */
  public abstract void computeErrors(List<SNNData> data_list) throws IllegalStateException;
  
  /**
   * Mthode pour calculer les erreurs sur toutes les donnes de test. 
   * L'usage de cette mthode est ncessaire pour obtenir des statistiques sur le rseau en entranement.
   * @throws IllegalStateException Si l'entraineur ne contient pas de donnes de test.
   */
  public abstract void computeErrorsTestingData() throws IllegalStateException;
    
  /**
   * Mthode pour valuer l'erreur sur une donnes.
   * 
   * @param data La donne.
   */
  public abstract void computeErrors(SNNData data);
  
  /**
   * Mthode pour obtenir le nombre de processeurs disponible pour raliser l'entranement.
   * Ce rsultat correspond au nombre de thread utilis lors de l'entranement.
   * 
   * @return Le nombre de processeurs utiliss.
   */
  public abstract int getProcessors();
    
}
