/**
 * 
 */
package sim.nn.trainer;

import java.util.ArrayList;
import java.util.List;

import sim.math.SMath;
import sim.nn.SNNData;

/**
 * Classe abstraite <b>SDataManager</b> reprsente un gestionnaire de donnes d'apprentissage et de validation pour un entraneur de rseau de neurones.
 * 
 * @author Simon Vezina
 * @since 2019-08-17
 * @version 2021-11-30
 */
public class SDataManager 
{
  /**
   * ...
   */
  protected final List<SNNData> training_list;
    
  /**
   * ...
   */
  protected final List<SNNData> testing_list;

  /**
   * ...
   */
  protected final int data_input_size;
  
  /**
   * ...
   */
  protected final int data_output_size;
  
  //----------------
  // CONSTRUCTEUR //
  //----------------
  
  /**
   * ...
   * 
   * @param input_size
   * @param output_size
   */
  public SDataManager(int input_size, int output_size)
  {
    this.training_list = new ArrayList<SNNData>();
    this.testing_list = new ArrayList<SNNData>();
    
    this.data_input_size = input_size;
    this.data_output_size = output_size;
  }
  
  //------------
  // MTHODES //
  //------------
  
  /**
   * ...
   * 
   * @return
   */
  public List<SNNData> getTrainingList()
  {
    return training_list;
  }
    
  /**
   * ...
   * 
   * @return
   */
  public List<SNNData> getTestingList()
  {
    return testing_list;
  }
    
  /**
   * ...
   * 
   * @return
   */
  public int getTrainingSize()
  {
    return training_list.size();
  }
  
  /**
   * ...
   * 
   * @return
   */
  public int getTestingSize()
  {
    return testing_list.size();
  }
  
  /**
   * Mthode pour obtenir un <u>vecteur d'entranement alatoire</u> disponible dans l'entraneur.
   * 
   * @return Un vecteur d'entranement.
   */
  public SNNData getRandomTrainingData()
  {
    return training_list.get(SMath.random(0, training_list.size()-1));
  }
  
  /**
   * Mthode pour obtenir une liste de donne d'entranement alatoire.
   * 
   * @param size La taille de la liste.
   * @return La liste de donne d'entranement.
   */
  public List<SNNData> getRandomTrainingDataList(int size)
  {
    List<SNNData> data_list = new ArrayList<SNNData>();
    
    // Obtenir une liste de donnes alatoires.
    for(int i = 0; i < size; i++)
      data_list.add(getRandomTrainingData());
    
    return data_list;
  }
  
  /**
   * Mthode pour obtenir un <u>vecteur de validation alatoire</u> disponible dans l'entraneur.
   * 
   * @return Un vecteur de validation.
   */
  public SNNData getRandomTestingData()
  {
    return testing_list.get(SMath.random(0, testing_list.size()-1));
  }
  
  /**
   * Mthode pour ajouter un nouveau vecteur d'entranement  l'entraneur.
   * 
   * @param training_object L'objet dfinissant la donne.
   * @param vector Le vecteur reprsentant la donne.
   * @param expected Le vecteur attendue lors de l'activation de cette donne par un rseau de neurones.
   * @throws IllegalArgumentException Si la donne n'est pas d'un format adquat.
   */
  public void addTraining(Object training_object, double[] vector, double[] expected) throws IllegalArgumentException
  {
    add(training_object, vector, expected, true);
  }

  /**
   * Mthode pour ajouter un nouveau vecteur d'entranement  l'entraneur.
   * 
   * @param data
   * @throws IllegalArgumentException
   */
  public void addTraining(SNNData data) throws IllegalArgumentException
  {
    add(data, true);
  }
  
  /**
   * Mthode pour ajouter des nouveaux vecteurs d'entranement  l'entraneur.
   * 
   * @param list
   * @throws IllegalArgumentException Si la donnes n'est pas du format adquat.
   */
  public void addTraining(List<SNNData> list) throws IllegalArgumentException
  {
    for(SNNData data : list)
      add(data, true);
  }
  
  /**
   * Mthode pour ajouter un nouveau vecteur de test  l'entraneur.
   * 
   * @param testing_object
   * @param vector
   * @param expected
   * @throws IllegalArgumentException
   */
  public void addTesting(Object testing_object, double[] vector, double[] expected)  throws IllegalArgumentException
  {
    add(testing_object, vector, expected, false);
  }
  
  /**
   * Mthode pour ajouter un nouveau vecteur de test d'entranement  l'entraneur.
   *
   * @param data
   * @throws IllegalArgumentException
   */
  public void addTesting(SNNData data) throws IllegalArgumentException
  {
    add(data, false);
  }
  
  /**
   * Mthode pour ajouter des nouveaux vecteurs de test  l'entraneur.
   * 
   * @param list
   * @throws IllegalArgumentException Si la donnes n'est pas du format adquat.
   */
  public void addTesting(List<SNNData> list) throws IllegalArgumentException
  {
    for(SNNData data : list)
      add(data, false);
  }
  
  /**
   * Mthode pour retirer un vecteur d'entranement alatoire de l'entraneur.
   * 
   * @return La donne correspondant au vecteur d'entranement retir.
   */
  public SNNData removeRandomTraining()
  {
    return training_list.remove(SMath.random(0, training_list.size()-1));    
  }
  
  /**
   * Mthode pour retirer l'ensemble des vecteurs d'entranement.
   */
  public void removeAllTrainingList()
  {
    training_list.clear();
  }
  
  /**
   * Mthode pour ajouter un vecteur  l'entraneur.
   * Si la variable <b>training</b> est <b>true</b>, le vecteur sera ajout  la liste d'entranement.
   * Si la variable <b>training</b> est <b>false</b>, le vecteur sera ajout  la liste de validation.
   * 
   * @param testing_object
   * @param vector
   * @param expected
   * @param training
   * @throws IllegalArgumentException Si la donnes n'est pas du format adquat.
   */
  private void add(Object object, double[] vector, double[] expected, boolean training) throws IllegalArgumentException
  {
    add(new SNNData(object, vector, expected), training);
  }
  
  /**
   * Mthode pour ajouter une donne  l'entraneur.
   * Si la variable <b>training</b> est <b>true</b>, la donne sera ajoute  la liste d'entranement.
   * Si la variable <b>training</b> est <b>false</b>, la donne sera ajoute  la liste de validation.
   *
   * @param data
   * @param training
   * @throws IllegalArgumentException Si la donnes n'est pas du format adquat.
   */
  protected void add(SNNData data, boolean training) throws IllegalArgumentException
  {
    if(data.getDataSize() != this.data_input_size)
      throw new IllegalArgumentException("Erreur SDataManager 005 : Le vecteur d'entranement possde une taille de " + data.getVector().length + ", mais l'entraneur autorise un vecteur d'entre de taille gal  " + data_input_size + ".");
    
    if(data.getExpectedSize() != this.data_output_size)
      throw new IllegalArgumentException("Erreur SDataManager 006 : Le vecteur de sortie attendu possde une taille de " + data.getExpected().length + ", mais l'entraneur autorise un vecteur de sortie de taille gal  " + this.data_output_size + ".");
    
    if(training)
      training_list.add(data);
    else
      testing_list.add(data);
  }
  
}
