/**
 * 
 */
package sim.nn.trainer;

import java.util.ArrayList;
import java.util.List;

import sim.exception.SConstructorException;
import sim.nn.SNNData;
import sim.nn.function.regularization.SDropoutFunction;
import sim.nn.gradient.SGradientComputer;
import sim.nn.network.SBatchNetwork;
import sim.nn.network.SNetwork;

/**
 * La classe <b>SFastTrainer</b> reprsente un entraneur plus rapide grce au paralllisme.
 * 
 * @author Simon Vezina
 * @since 2019-09-05
 * @version 2021-11-29
 */
public class SFastTrainer extends STrainer {

  /**
   * La variable <b>THREAD_NUMBER</b> correspond au nombre de processeurs disponible pour excuter la classe.
   * Cette valeur ne peut pas tre infrieur  1.
   */
  protected final static int THREAD_NUMBER = Runtime.getRuntime().availableProcessors();
  
  /**
   * La variable <b>computer_list</b> reprsente une liste de calculateur d'erreur sur les donnes de test.
   */
  protected final List<SErrorsComputer> computer_test_list;
  
  /**
   * ...
   */
  protected final List<SGradientComputer> computer_gradient_list;
  
  //----------------
  // CONSTRUCTEUR //
  //----------------
  
  /**
   * ...
   * 
   * @param network
   * @param error_function
   * @throws SConstructorException
   */
  public SFastTrainer(SNetwork network, String error_function) throws SConstructorException
  {
    super(network, error_function);
    
    // Construire les calculateurs d'erreur sur les donnes de test avec le nombre de thread.
    this.computer_test_list = new ArrayList<SErrorsComputer>();
    
    for(int i = 0; i < THREAD_NUMBER; i++)
      this.computer_test_list.add(new SErrorsComputer(network, error_function));
    
    // Construire les calculateur de gradient.
    this.computer_gradient_list = new ArrayList<SGradientComputer>();
    
    // Prsentement, les batchNetwork ne peuvent pas tre entran en parallle !!!
    for(int i = 0; i < THREAD_NUMBER; i++)
      this.computer_gradient_list.add(new SGradientComputer(network, error_function));
  }
  
  //------------
  // MTHODES //
  //------------
      
  @Override
  protected void computeErrorsListTestingData()
  {
    // Effectuer le calcul de l'erreur en parallle.
    STrainingAlgorithm.paralleleComputeErrors(computer_test_list);
  }
  
  @Override
  public void train(int batch)
  {
    // Pour l'usage de la fonction dropout.
    SDropoutFunction.setDropping(true);
    
    // Dterminer le type d'entrainement en fonction du type de rseau.
    if(network.isBatchNetwork())
      STrainingAlgorithm.batchTraining((SBatchNetwork)this.network, this.batch_error_function, this.gradient, getRandomTrainingDataList(batch), this.learning_rate);
    else
      STrainingAlgorithm.parallelTraining(this.network, this.gradient, this.computer_gradient_list, getRandomTrainingDataList(batch), this.learning_rate);
   
    // Faire la mise  jour avec 1 entranement utilisant X donnes.
    this.updateTrainingStatistic(batch);
    
    // Pour la dsactivation de la fonction dropout.
    SDropoutFunction.setDropping(false);
  }
  
  @Override
  protected void add(SNNData data, boolean training)
  {
    // Mettre la donne dans la bonne liste des donnes (training ou testing).
    super.add(data, training);
    
    // Mettre en alternance la donne de test dans tous les calculateurs de test.
    // De cette faon, le tout sera bien balanc.
    if(!training)
      computer_test_list.get(getTestingSize() % this.computer_test_list.size()).add(data);
  }
   
  @Override
  public int getProcessors() 
  {
    return THREAD_NUMBER;
  }
  
}
