/**
 * 
 */
package sim.nn.trainer;

import java.util.ArrayList;
import java.util.List;

import sim.exception.SConstructorException;
import sim.math.SPointFunction;
import sim.nn.SNNData;
import sim.nn.function.SNeuralFunction;
import sim.nn.function.error.SBatchErrorFunction;
import sim.nn.function.error.SErrorFunction;
import sim.nn.function.error.SErrorFunctionBuilder;
import sim.nn.function.regularization.SDropoutFunction;
import sim.nn.gradient.SErrorFunctionGradient;
import sim.nn.network.SBatchNetwork;
import sim.nn.network.SNetwork;
import sim.util.SChronometer;

/**
 * La classe <b>SBasicTrainer</b> reprsente un entraneur de rseau de neurones de base.
 * 
 * @author Simon Vezina
 * @since 2019-08-18
 * @version 2021-11-30
 */
public class STrainer extends SAbstractTrainer {
  
  /**
   * La variable <b>network</b> reprsente le rseau qui sera entran.
   */
  protected final SNetwork network;
  
  /**
   * La variable <b>error_function</b> reprsente la fonction d'erreur utilie pour valuer l'erreur du rseau.
   */
  protected final SErrorFunction error_function;
  
  /**
   * La variable <b>batch_error_function</b> reprsente la fonction d'erreur utilie pour valuer l'erreur du rseau en mode <i>batch activation</i>.
   */
  protected final SBatchErrorFunction batch_error_function;
  
  /**
   * La variable <b>gradient</b> reprsente le gradient de la fonction d'erreur associ  un rseau.
   */
  protected final SErrorFunctionGradient gradient;
  
  /**
   * La variable <b>time_compute_error_testing</b> reprsente le temps pour valuer l'erreur sur l'ensemble des donnes de test.
   */
  protected double time_compute_error_testing;
  
  /**
   * La variable <b>statistic_function</b> reprsente une fonction de point qui va enregistrer l'volution de l'erreur engendr par l'activation des donnes de test du rseau dans son processus d'apprentissage.
   */
  protected SPointFunction statistic_function;
  
  //----------------
  // CONSTRUCTEUR //
  //----------------
  
  /**
   * Constructeur d'un entraneur avec la fonction d'erreur <u>nergie</u>.
   * 
   * @param network Le rseau de neurones.
   */
  public STrainer(SNetwork network)
  {
    this(network, SNeuralFunction.ENERGY_ERROR_FUNCTION);
  }
  
  /**
   * Constructeur d'un entraneur.
   * 
   * @param network Le rseau  entraner.
   * @param error_function_name Le nom de la fonction d'erreur.
   * @throws SConstructorException Si une erreur est survenue lors de la construction de l'objet.
   */
  public STrainer(SNetwork network, String error_function_name) throws SConstructorException
  {
    super(network);
    
    // Dfinir le rseau.
    this.network = network;
    
    // Construction de la fonction d'erreur.
    try {
      this.error_function = SErrorFunctionBuilder.build(error_function_name, network.getOutputsSize());
      this.batch_error_function = new SBatchErrorFunction(network.getOutputsSize(), error_function_name);
    }catch(IllegalArgumentException e) {
      throw new SConstructorException("Erreur STrainer 001 : Le nom de fonction d'erreur '" + error_function_name + "' n'est pas reconnue.", e);
    }
    
    // Constructeur du gradient de la fonction d'erreur en fonction du type de rseau.
    this.gradient = new SErrorFunctionGradient(this.network);
    
    // S'assurer que le gradient est vide.
    this.gradient.clear();   
    
    this.time_compute_error_testing = 0.0;
    
    this.statistic_function = new SPointFunction();
  }

  //------------
  // MTHODES //
  //------------
  
  @Override
  public String getErrorFunctionName()
  {
    return this.error_function.getFunctionName();
  }

  @Override
  public SNetwork getNetwork()
  {
    return network;
  }
  
  /**
   * Mthode pour obtenir le temps moyen pour activer une donne.
   * Le calcul sera effectue lorsqu'il y aura calcul de l'erreur sur les donnes de test.
   * Ce temps moyen exclut l'efficacit donne par le calcul en multi-tche.
   * 
   * @return Le temps moyen pour activer une donne.
   */
  public double getAverageTimeActivation()
  {
    return this.time_compute_error_testing * this.getProcessors() / this.testing_list.size();
  }
  
  /**
   * Mthode pour obtenir la fonction de statistique de l'entranement.
   * 
   * @return La fonction de statistique de l'entranement.
   */
  public SPointFunction getStatisticFunction()
  {
    return this.statistic_function;
  }
  
  /**
   * Mthode qui effectue l'entranement d'un rseau de neurones avec <u>un vecteur d'entranement</u> dtermin alatoirement.
   * 
   * @throws IllegalStateException S'il n'y a pas de vecteur d'entrainement d'enregistr.
   * @throws UnsupportedOperationException Si l'entrainement est ralis avec un rseau de type <i>batch</i>.
   */
  public void train() throws IllegalStateException, UnsupportedOperationException
  {
    // Vrifier si l'entrainement est possible.
    if(this.training_list.size() == 0)
      throw new IllegalStateException("Erreur STrainer 002 : L'entrainement n'est pas possible puisqu'il n'y a pas de vecteur d'entrainement d'enregistr.");
    
    // Pour l'usage de la fonction dropout.
    SDropoutFunction.setDropping(true);
    
    // Raliser l'entranement avec une donne alatoire.
    STrainingAlgorithm.training(this.network, this.error_function, this.gradient, getRandomTrainingData(), this.learning_rate);
    
    // Faire la mise  jour avec 1 entranement utilisant 1 donne.
    updateTrainingStatistic();  
    
    // Pour retirer l'usage de la fonction dropout.
    SDropoutFunction.setDropping(false);
  }

  /**
   * Mthode qui effectue l'entranement d'un rseau de neurones sur une collection de donne ensemble 
   * en mode <i>batch</i> ou <i>mini-batch</i> selon le type de rseau en entranement.
   * 
   * @param batch La taille de la <i>mini-batch</i>(mode activation) ou taille de la batch(mode batch-activation).
   * @throws IllegalStateException S'il n'y a pas de vecteur d'entrainement d'enregistr.
   */
  public void train(int batch) throws IllegalStateException
  {
    // Vrifier si l'entrainement est possible.
    if(this.training_list.size() == 0)
      throw new IllegalStateException("Erreur STrainer 004 : L'entrainement n'est pas possible puisqu'il n'y a pas de vecteur d'entrainement d'enregistr.");
    
    // Pour l'usage de la fonction dropout.
    SDropoutFunction.setDropping(true);
    
    // Dterminer le type d'entrainement en fonction du type de rseau.
    if(network.isBatchNetwork())
      STrainingAlgorithm.batchTraining((SBatchNetwork)this.network, this.batch_error_function, this.gradient, getRandomTrainingDataList(batch), this.learning_rate);
    else
      STrainingAlgorithm.training(this.network, this.error_function, this.gradient, getRandomTrainingDataList(batch), this.learning_rate);
    
    // Faire la mise  jour avec 1 entranement utilisant X donnes.
    this.updateTrainingStatistic(batch);
    
    // Pour retirer l'usage de la fonction dropout.
    SDropoutFunction.setDropping(false);
  }
  
  
  /**
   * Mthode que effectue l'entranement d'un rseau de neurones sur l'ensemble des donnes d'entranement.
   *
   * @throws IllegalStateException S'il n'y a pas de vecteur d'entrainement d'enregistr.
   */
  public void trainAll() throws IllegalStateException
  {
    // Vrifier si l'entrainement est possible.
    if(this.training_list.size() == 0)
      throw new IllegalStateException("Erreur STrainer 005 : L'entrainement n'est pas possible puisqu'il n'y a pas de vecteur d'entrainement d'enregistr.");
    
    // Pour l'usage de la fonction dropout.
    SDropoutFunction.setDropping(true);
    
    // Mlanger les donnes d'entranement.
    java.util.Collections.shuffle(this.training_list);
    
    // Itrer sur l'ensemble des donne.
    for(SNNData data : this.training_list)
    {
      STrainingAlgorithm.training(this.network, this.error_function, this.gradient, data, this.learning_rate);
      updateTrainingStatistic();  
    }
     
    // Pour retirer l'usage de la fonction dropout.
    SDropoutFunction.setDropping(false);
  }
    
  /**
   * Mthode pour effectuer l'entranement d'un rseau de neurones sur l'ensemble des donnes d'entranement, mais en utilisant sous-groupe de vecteur (<i>mini-batch</i>)  chaque apprentissage.
   * 
   * @param mini_batch La taille de la <i>mini-batch</i>.
   * @throws IllegalStateException S'il n'y a pas de vecteur d'entrainement d'enregistr.
   */
  public void trainAll(int mini_batch) throws IllegalStateException
  {
 // Vrifier si l'entrainement est possible.
    if(this.training_list.size() == 0)
      throw new IllegalStateException("Erreur STrainer 006 : L'entrainement n'est pas possible puisqu'il n'y a pas de vecteur d'entrainement d'enregistr.");
    
    // Pour l'usage de la fonction dropout.
    SDropoutFunction.setDropping(true);
    
    // Mlanger les donnes d'entranement.
    java.util.Collections.shuffle(this.training_list);
    
    // Liste des donnes dans la mini-batch.
    List<SNNData> mini_batch_list = new ArrayList<SNNData>();
    
    // Itrer sur l'ensemble des donne.
    // P.S.
    // Dans cette version, il est possible qu'un certain nombre de vecteurs n'est pas t utilis,
    // car si la liste n'est pas entirement rempli, elle ne sera pas utilise.
    for(SNNData data : this.training_list)
      if(mini_batch_list.size() < mini_batch)
        mini_batch_list.add(data);
      else
      {
        STrainingAlgorithm.training(this.network, this.error_function, this.gradient, mini_batch_list, this.learning_rate);
        updateTrainingStatistic(mini_batch);
      }
     
    // Pour retirer l'usage de la fonction dropout.
    SDropoutFunction.setDropping(false);
  }
  
  @Override
  public void computeErrors(List<SNNData> data_list) throws IllegalArgumentException
  {
    // Vrifier qu'il y a bien des donnes de test.
    if(data_list.size() == 0)
      throw new IllegalStateException("Erreur STrainer 007 : Le calcul de l'erreur n'est pas possible puisque la liste des donnes est vide.");
   
    // Itrer sur l'ensemble des donnes de test.
    for(SNNData data : data_list)
      STrainingAlgorithm.computeErrors(this.network, this.error_function, data);
  }
  
  @Override
  public void computeErrors(SNNData data)
  {
    STrainingAlgorithm.computeErrors(this.network, this.error_function, data);
  }
  
  @Override
  public void computeErrorsTestingData() throws IllegalStateException
  {
    SChronometer chrono = new SChronometer();
    chrono.start();
    
    // Calculer l'erreur sur les donnes de test.
    computeErrorsListTestingData();
        
    // Dfinir le test comme tant calcul.
    is_testing_error_computed = true;
    
    // Ajouter l'erreur calcul dans la fonction de statistique ( nb_train , error )
    this.statistic_function.add(this.getTrainingCount(), this.getTestingAverageError());
        
    // valuation du temps pour activer une donne dans l'avantage du multi-tche.
    chrono.stop();
    this.time_compute_error_testing = chrono.getTime();
  }

  /**
   * Mthode pour effectuer le calcul de l'erreur sur la liste des donnes de test contenu dans l'entraneur.
   * 
   * @throws IllegalStateException S'il n'y a pas de donne de test.
   */
  protected void computeErrorsListTestingData() throws IllegalStateException
  {
    computeErrors(this.testing_list);
  }
    
  @Override
  public String toString() 
  {
    return "STrainer [error_function = " + error_function + ", learning_rate = " + learning_rate + ", Processors = " + this.getProcessors() + ", Training vector = " + training_list.size() + ", Testing vector = " + testing_list.size() + "]";
  }

  @Override
  public int getProcessors() 
  {
    return 1;
  }
  
}
