/**
 * 
 */
package sim.nn.wdbc;

import java.util.NoSuchElementException;
import java.util.StringTokenizer;

import sim.exception.SConstructorException;
import sim.util.SStringUtil;

/**
 * La classe <b>SPatient</b> reprsente un patient de l'examen wdbc (Diagnostic Wisconsin Breast Cancer Database)
 * 
 * @author Simon Vzina
 * @since 2018-02-09
 * @version 2018-02-14
 */
public class SPatient {

  private static final int NB_DATA_9_FIELD_FORMAT = 9;   // Le 10ime champ tant le diagnostique.
  private static final int NB_DATA_30_FIELD_FORMAT = 30;  // Le diagnostique est inscrit avant les donnes.
  
  private final int id;
  private final double[] data;
  private final boolean diagnostic;
  
  /**
   * Constructeur d'un patient.
   * 
   * @param description La description du patient.
   * @throws SConstructorException Si la description ne permet pas la construction complte du dossier du patient.
   */
  public SPatient(String description) throws SConstructorException
  {
    
    // Usage d'un tokenizer avec caractre ",".
    StringTokenizer tokens = new StringTokenizer(description, SStringUtil.REMOVE_CARACTER_TOKENIZER);     
    
    // Lecture du numro d'identification. Ce champ est le premier dans tous les formats de lecture.
    try{
      id = Integer.parseInt(tokens.nextToken());        // 1. Sample code number (id number)
    }catch(NoSuchElementException e){
      throw new SConstructorException("Le patient de description '" + description + "' n'est pas valide, car il manque d'information.");
    }catch(NumberFormatException e){
      throw new SConstructorException("Le patient de description '" + description + "' n'est pas valide, car le 1ier champ ne peut pas tre interprt comme un numro d'identification.");
    }
    
    // Lecture des autres champs selon le type de patient (version 10 ou 30 lments).
    switch(tokens.countTokens())
    {
      case (NB_DATA_9_FIELD_FORMAT+1) : 
        
        data = new double[NB_DATA_9_FIELD_FORMAT];
        diagnostic = parse10Field(tokens, data, description);
        break;
        
      case (NB_DATA_30_FIELD_FORMAT+1) :
        
        data = new double[NB_DATA_30_FIELD_FORMAT];
        diagnostic = parse30Field(tokens, data, description);
        break;
        
        default : throw new SConstructorException("Le patient de description '" + description + "' n'est pas valide, car le nombre de champs disponibles " + tokens.countTokens() + " n'est pas reconnu.");
    }
   
  }
  
  /**
   * Mthode pour obtenir le numro d'identification du patient.
   * 
   * @return Le numro d'identification.
   */
  public int getID()
  {
    return id;
  }
  
  /**
   * Mthode pour obtenir le vecteur des donnes du patient.
   * @return
   */
  public double[] getVector()
  {
    return data;
  }
  
  /**
   * Mthode pour obtenir la valeur du diagnostique du patient.
   * 
   * @return La valeur du diagnostique du patient.
   */
  public double[] getValue()
  {
    double[] output = new double[1];
    
    if(diagnostic)
      output[0] = 1.0;
    else
      output[0] = 0.0;
    
    return output;
  }
  
  /**
   * ...
   * 
   * @param tokens
   * @param array
   * @param description
   * @return
   * @throws SConstructorException
   */
  private boolean parse10Field(StringTokenizer tokens, double[] array, String description) throws SConstructorException
  {
    if(tokens.countTokens() != (NB_DATA_9_FIELD_FORMAT+1))
      throw new SConstructorException("Le nombre de champs en lecture " + tokens.countTokens() + " pour le patient " + description + " n'est pas en accord avec la version en interprtation ncessitant " + (NB_DATA_9_FIELD_FORMAT+1) + " champs.");
    
    if(array.length != NB_DATA_9_FIELD_FORMAT)
      throw new SConstructorException("Le tableau  remplir possde une taille " + array.length + " ce qui n'est pas valide avec le format en lecture qui require " + NB_DATA_9_FIELD_FORMAT + " champs.");
    
    // Une normalisation par une division par 10 sera ralis pour recadrer les donnes entre 0.0 et 1.0.
    try{
      array[0] = (double)Integer.parseInt(tokens.nextToken()) / 10.0;   // 2. Clump Thickness               1 - 10
      array[1] = (double)Integer.parseInt(tokens.nextToken()) / 10.0;   // 3. Uniformity of Cell Size       1 - 10
      array[2] = (double)Integer.parseInt(tokens.nextToken()) / 10.0;   // 4. Uniformity of Cell Shape      1 - 10
      array[3] = (double)Integer.parseInt(tokens.nextToken()) / 10.0;   // 5. Marginal Adhesion             1 - 10
      array[4] = (double)Integer.parseInt(tokens.nextToken()) / 10.0;   // 6. Single Epithelial Cell Size   1 - 10
      array[5] = (double)Integer.parseInt(tokens.nextToken()) / 10.0;   // 7. Bare Nuclei                   1 - 10
      array[6] = (double)Integer.parseInt(tokens.nextToken()) / 10.0;   // 8. Bland Chromatin               1 - 10
      array[7] = (double)Integer.parseInt(tokens.nextToken()) / 10.0;   // 9. Normal Nucleoli               1 - 10
      array[8] = (double)Integer.parseInt(tokens.nextToken()) / 10.0;   // 10. Mitoses                      1 - 10
      int d = Integer.parseInt(tokens.nextToken());                     // 11. Class:                       (2 for benign, 4 for malignant)
      
      // Analyse du diagnostique
      if(d == 2)
        return false;
      else
        if(d == 4)
          return true;
        else
          throw new SConstructorException("Le code de diagnostique '" + d + "' pour le patient " + description + " n'est pas reconnu (pas 2 ou 4).");
      
    }catch(NoSuchElementException e){
      throw new SConstructorException("Le patient de description '" + description + "' n'est pas valide, car il manque d'information.");
    }catch(NumberFormatException e){
      throw new SConstructorException("Le patient de description '" + description + "' n'est pas valide, car un paramtre numrique ne peut pas tre interprt.");
    }   
  }
  
  /**
   * ...
   * 
   * @param tokens
   * @param array
   * @param description
   * @return
   * @throws SConstructorException
   */
  private boolean parse30Field(StringTokenizer tokens, double[] array, String description) throws SConstructorException
  {
    if(tokens.countTokens() != (NB_DATA_30_FIELD_FORMAT+1))
      throw new SConstructorException("Le nombre de champs en lecture " + tokens.countTokens() + " pour le patient " + description + " n'est pas en accord avec la version en interprtation ncessitant " + (NB_DATA_30_FIELD_FORMAT+1) + " champs.");
    
    if(array.length != NB_DATA_30_FIELD_FORMAT)
      throw new SConstructorException("Le tableau  remplir possde une taille " + array.length + " ce qui n'est pas valide avec le format en lecture qui require " + NB_DATA_30_FIELD_FORMAT + " champs.");
    
    try{
      
      // Lecture du diagnostique.
      String d = tokens.nextToken();
      d = d.toUpperCase();
      
      boolean diag;
      
      if(d.equals("M"))
        diag = true;
      else
        if(d.equals("B"))
          diag = false;
        else
          throw new SConstructorException("Le code de diagnostique '" + d + "' pour le patient " + description + " n'est pas reconnu (pas M ou B).");
        
      // Lecture des 30 autres champs.
      for(int i = 0; i < NB_DATA_30_FIELD_FORMAT; i++)
        array[i] = Double.parseDouble(tokens.nextToken());
      
      // Faire une renormalisation des donnes en fonction des chantillons reu.
      // La renormalisation a pour but de transformer tous les champs du vecteur en une valeur entre 0.0 et 1.0.
      array[0] /= 100.0;
      array[1] /= 100.0;
      array[2] /= 100.0;
      array[3] /= 2000.0;
      array[11] /= 10.0;
      array[12] /= 10.0;
      array[13] /= 100.0;
      array[20] /= 100.0;
      array[21] /= 100.0;
      array[22] /= 100.0;
      array[23] /= 2000.0;
      
      // Retourner le diagnostique.
      return diag;
      
    }catch(NoSuchElementException e){
      throw new SConstructorException("Le patient de description '" + description + "' n'est pas valide, car il manque d'information.");
    }catch(NumberFormatException e){
      throw new SConstructorException("Le patient de description '" + description + "' n'est pas valide, car un paramtre numrique ne peut pas tre interprt.");
    }   
  }
  
  
  @Override
  public String toString()
  {
    return "SPatient [id=" + id + ", diagnostic=" + diagnostic + "]";
  }
  
}// fin de la clase SPatient
