/**
 * 
 */
package sim.readwrite.object;

import java.awt.Color;
import java.awt.image.BufferedImage;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;

import javax.imageio.ImageIO;

import sim.exception.SConstructorException;
import sim.graphics.SChart;
import sim.graphics.SImageUtil;
import sim.math.SRealFunction;
import sim.readwrite.SAbstractRW;
import sim.util.SArrays;
import sim.util.SBrowser;
import sim.util.SBufferedReader;
import sim.util.SInitializationException;
import sim.util.SKeyWordDecoder;
import sim.util.SLog;
import sim.util.SManyFilesFoundException;
import sim.util.SReadingException;
import sim.util.SStringUtil;

/**
 * La classe <b>SImageFunctionRW</b> reprsente une fonction pouvant tre charge  partir d'une image.
 * 
 * @author Simon Vzina
 * @since 2018-03-31
 * @version 2018-04-09
 */
public class SImageFunctionRW extends SAbstractRW {

  //--------------
  // CONSTANTES //
  //--------------
  
  /**
   * La constante <b>KEYWORD_PARAMETER</b> correspond  un tableau contenant l'ensemble des mots cls 
   *  utiliser reconnus lors de la dfinition de l'objet par une lecture en fichier.
   */
  private static final String[] KEYWORD_PARAMETER = { 
      SKeyWordDecoder.KW_X_MIN, 
      SKeyWordDecoder.KW_X_MAX,
      SKeyWordDecoder.KW_Y_MIN, 
      SKeyWordDecoder.KW_Y_MAX, 
      SKeyWordDecoder.KW_IMAGE_FILE_NAME
  };
  
  private static final String DEFAULT_FILE_NAME = "image_fonction.txt";
  
  private static final String DEFAULT_IMAGE_FILE_NAME = "image_fonction.bmp";
  
  private static final int DEFAULT_IMAGE_HEIGHT = 1000;
  
  private static final double DEFAULT_X_MIN = 0.0;
  private static final double DEFAULT_X_MAX = 1.0;
  private static final double DEFAULT_Y_MIN = 0.0;
  private static final double DEFAULT_Y_MAX = 1.0;
  
  //-------------
  // VARIABLES //
  //-------------
  private double x_min;
  private double x_max;
  private double y_min;
  private double y_max;
  
  private int image_height;
  
  private String image_file_name;
  private double[] values;
  
  /**
   * 
   */
  public SImageFunctionRW()
  {
    super();
    
    x_min = DEFAULT_X_MIN;
    x_max = DEFAULT_X_MAX;
    y_min = DEFAULT_Y_MIN;
    y_max = DEFAULT_Y_MAX;
    
    image_height = DEFAULT_IMAGE_HEIGHT;
    image_file_name = null;
    values = null;
  }

  /**
   * ...
   * 
   * @param file_name
   */
  public SImageFunctionRW(String file_name)
  {
    super(file_name);
    
    x_min = DEFAULT_X_MIN;
    x_max = DEFAULT_X_MAX;
    y_min = DEFAULT_Y_MIN;
    y_max = DEFAULT_Y_MAX;
    
    image_height = DEFAULT_IMAGE_HEIGHT;
    image_file_name = null;
    values = null;
  }

  /**
   * ...
   * 
   * @param values
   * @throws SConstructorException
   */
  public SImageFunctionRW(double[] values) throws SConstructorException
  {
    this(DEFAULT_FILE_NAME, DEFAULT_IMAGE_FILE_NAME, DEFAULT_X_MIN, DEFAULT_X_MAX, values);
  }
  
  /**
   * ...
   * 
   * @param file_name
   * @param image_file_name
   * @param x_min
   * @param x_max
   * @param values
   * @throws SConstructorException
   */
  public SImageFunctionRW(String file_name, String image_file_name, double x_min, double x_max, double[] values) throws SConstructorException
  {
    this(file_name, image_file_name, x_min, x_max, SArrays.findMin(values), SArrays.findMax(values), values);
  }
  
  /**
   * ...
   * 
   * @param file_name
   * @param image_file_name
   * @param x_min
   * @param x_max
   * @param y_min
   * @param y_max
   * @param values
   * @throws SConstructorException
   */
  public SImageFunctionRW(String file_name, String image_file_name, double x_min, double x_max, double y_min, double y_max, double[] values) throws SConstructorException
  {
    super(file_name);
    
    this.image_file_name = image_file_name;
    
    if(x_min > x_max)
      throw new SConstructorException("Erreur SImageFunctionRW 001 : La valeur de x_min = " + x_min + " est suprieure  x_max = " + x_max + ".");
    
    if(y_min > y_max)
      throw new SConstructorException("Erreur SImageFunctionRW 002 : La valeur de y_min = " + y_min + " est suprieure  y_max = " + y_max + ".");
    
    image_height = DEFAULT_IMAGE_HEIGHT;
    
    this.x_min = x_min;
    this.x_max = x_max;
    this.y_min = y_min;
    this.y_max = y_max;
        
    this.values = new double[values.length];
    SArrays.copy(values, this.values);
  }
  /**
   * @param sbr
   * @throws IOException
   * @throws SConstructorException
   */
  public SImageFunctionRW(SBufferedReader sbr) throws IOException, SConstructorException
  {
    this();
    
    try{
      read(sbr);
    }catch(SInitializationException e){
      throw new SConstructorException("Erreur SImageFunctionRW 003 : Une erreur d'initialisation est survenue.", e);
    }
    
  }

  public double getMinX()
  {
    return x_min;
  }
  
  public double getMaxX()
  {
    return x_max;
  }
  
  public double getMinY()
  {
    return y_min;
  }
  
  public double getMaxY()
  {
    return y_max;
  }
  
  public double[] getValues()
  {
    return values;
  }
  
  /**
   * Mthode pour dfinir la hauteur de l'image gnre.
   * Ce paramtre permet de contrler la prcision de la fonction lors de sa lecture.
   * 
   * @param height La hauteur de l'image gnre.
   */
  public void setImageHeight(int height)
  {
    if(height < 10)
      throw new IllegalArgumentException("Erreur SImageFunction 004 : La hauteur h = " + height + " est infrieure  10.");
    
    this.image_height = height;
  }
  
  /**
   * Mthode pour dfinir les valeurs maximales en x d'affichage de la fonction.
   * 
   * @param y_min La valeur minimale.
   * @param y_max La valeur maximale.
   * @throws IllegalArgumentException Si les paramtres de la fonction sont illgaux.
   */
  public void setX(double min, double max) throws IllegalArgumentException
  {
    if(min > max)
      throw new IllegalArgumentException("Erreur SImageFunctionRW 005 : La valeur de x_min = " + min + " est suprieure  x_max = " + max + ".");
      
    this.x_min = min;
    this.x_max = max;
  }
  /**
   * Mthode pour dfinir les valeurs maximales en y d'affichage de la fonction.
   * 
   * @param y_min La valeur minimale.
   * @param y_max La valeur maximale.
   * @throws IllegalArgumentException Si les paramtres de la fonction sont illgaux.
   */
  public void setY(double min, double max) throws IllegalArgumentException
  {
    if(min > max)
      throw new IllegalArgumentException("Erreur SImageFunctionRW 005 : La valeur de y_min = " + min + " est suprieure  y_max = " + max + ".");
      
    this.y_min = min;
    this.y_max = max;
  }
  
  @Override
  public String getRWName()
  {
    return SKeyWordDecoder.KW_IMAGE_FUNCTION;
  }

  @Override
  public void writeInformation(BufferedWriter bw) throws IOException
  {
    super.writeInformation(bw);
 
    bw.write(SKeyWordDecoder.KW_IMAGE_FILE_NAME);
    bw.write("\t");
    bw.write(image_file_name);
    bw.write(SStringUtil.END_LINE_CARACTER);
    
    bw.write(SKeyWordDecoder.KW_X_MIN);
    bw.write("\t");
    bw.write(Double.toString(x_min));
    bw.write(SStringUtil.END_LINE_CARACTER);
    
    bw.write(SKeyWordDecoder.KW_X_MAX);
    bw.write("\t");
    bw.write(Double.toString(x_max));
    bw.write(SStringUtil.END_LINE_CARACTER);
    
    bw.write(SKeyWordDecoder.KW_Y_MIN);
    bw.write("\t");
    bw.write(Double.toString(y_min));
    bw.write(SStringUtil.END_LINE_CARACTER);
    
    bw.write(SKeyWordDecoder.KW_Y_MAX);
    bw.write("\t");
    bw.write(Double.toString(y_max));
    bw.write(SStringUtil.END_LINE_CARACTER);    
    
    if(image_file_name != null)
    {
      // Construire la fonction.
      SRealFunction f = new SRealFunction(values, x_min, x_max);
      
      // Construire le graphique.
      SChart chart = new SChart(f, y_min, y_max, image_height);
      
      //Gnrer le fichier image
      SImageUtil.writeImageBMP(image_file_name, chart.buildBufferedImage());     
    }
  }
    
  @Override
  protected boolean read(SBufferedReader sbr, int code, String remaining_line) throws SReadingException
  {
    switch(code)
    {
      case SKeyWordDecoder.CODE_IMAGE_FILE_NAME : image_file_name = readStringNotEmpty(remaining_line, SKeyWordDecoder.KW_IMAGE_FILE_NAME); return true;
      case SKeyWordDecoder.CODE_X_MIN : x_min = readDouble(remaining_line, SKeyWordDecoder.KW_X_MIN); return true;  
      case SKeyWordDecoder.CODE_X_MAX : x_max = readDouble(remaining_line, SKeyWordDecoder.KW_X_MAX); return true;  
      case SKeyWordDecoder.CODE_Y_MIN : y_min = readDouble(remaining_line, SKeyWordDecoder.KW_Y_MIN); return true;  
      case SKeyWordDecoder.CODE_Y_MAX : y_max = readDouble(remaining_line, SKeyWordDecoder.KW_Y_MAX); return true;  
            
      default : return super.read(sbr, code, remaining_line);
    }
  }
  
  @Override
  protected void initialization() throws SInitializationException
  {
    super.initialization();
 
    if(x_min > x_max)
      throw new SInitializationException("Erreur SImageFunctionRW 006 : La valeur de x_min = " + x_min + " est suprieure  x_max = " + x_max + ".");
    
    if(y_min > y_max)
      throw new SInitializationException("Erreur SImageFunctionRW 007 : La valeur de y_min = " + y_min + " est suprieure  y_max = " + y_max + ".");
      
    try {
      SBrowser browser = new SBrowser();
      File f = browser.findFile(image_file_name);
      
      int color_black = Color.black.getRGB();
      
      // Obtenir les positions des pixels dans l'image.
      BufferedImage image = ImageIO.read(f);
      
      int[] tab = new int[image.getWidth()];
      
      int nb_not_found = 0;
      
      // Itrer sur la largeur de l'image.
      for(int i = 0; i < image.getWidth(); i++)
      {
        boolean found = false;
      
        // Itrer sur la hauteur de l'image  la recherche du pixel noir.
        for(int j = 0; j < image.getHeight() && !found; j++)
        {
          if(image.getRGB(i,j) == color_black)
          {
            tab[i] = j;
            found = true;
          }
        }
        
        if(!found)
        {
          nb_not_found++;
          tab[i] = (int) Double.NaN;          
        }
          
      }
      
      // Affichage du message du nombre de chiffre NaN.
      SLog.logWriteLine("Message SImageFunctionRW : Aprs la lecture de la fonction, il y a " + nb_not_found + " lments discrets non dtermins (NaN).");
      
      // Faire l'inversion des valeurs.
      for(int i = 0; i < tab.length; i++)
        tab[i] = image.getHeight()-1 - tab[i];
      
      // Convertir les positions en valeur de la fonction.
      values = SArrays.mappingIntToDouble(tab, y_min, y_max);
      
    }catch(FileNotFoundException e) {
      throw new SInitializationException("Erreur SImageFunction 008 : Le ficher '" + image_file_name + "' n'a pas t trouv.", e);
    }catch(SManyFilesFoundException e) {
      throw new SInitializationException("Erreur SImageFunction 009 : Le ficher '" + image_file_name + "' a t trouv sous plusieurs exemplaire.", e);
    }catch(IOException e) {
      throw new SInitializationException("Erreur SImageFunction 010 : Le ficher '" + image_file_name + "' n'a pas t charg adquatement en raison d'une erreur de type I/O.", e);
    }

  }
  
  @Override
  public String[] getRWParameterName()
  {
    String[] other_parameters = super.getRWParameterName();
    
    return SStringUtil.merge(other_parameters, KEYWORD_PARAMETER);
  }
  
}
