/**
 * 
 */
package sim.readwrite.object.nn;

import java.io.BufferedWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import sim.exception.SConstructorException;
import sim.math.SArrayTensor;
import sim.math.SBufferedMatrix;
import sim.nn.layer.SBatchConvolutionLayer;
import sim.nn.parameter.SActivationParameters;
import sim.nn.parameter.SBias;
import sim.nn.parameter.SFiltersWeights;
import sim.nn.parameter.SScale;
import sim.nn.parameter.SShift;
import sim.readwrite.object.SMatrixRW;
import sim.util.SBufferedReader;
import sim.util.SInitializationException;
import sim.util.SKeyWordDecoder;
import sim.util.SReadingException;
import sim.util.SStringUtil;

/**
 * La classe <b>SBatchConvolutionLayerRW</b> reprsente une couche de neurones de type batch de convolution pouvant tre lu dans un fichier.
 * 
 * @author Simon Vezina
 * @since 2021-10-31
 * @version 2022-02-28
 */
public class SBatchConvolutionLayerRW extends SConvolutionLayerRW {

  //--------------
  // CONSTANTES //
  //--------------
  
  /**
   * La constante <b>KEYWORD_PARAMETER</b> correspond  un tableau contenant l'ensemble des mots cls 
   *  utiliser reconnus lors de la dfinition de l'objet par une lecture en fichier.
   */
  private static final String[] KEYWORD_PARAMETER = {
    SKeyWordDecoder.KW_SCALE, SKeyWordDecoder.KW_SHIFT 
  };
  
  /**
   * ...
   */
  private final static double[] DEFAULT_SCALE = { };
  
  /**
   * ...
   */
  private final static double[] DEFAULT_SHIFT = { };
  
  //-------------
  // VARIABLES //
  //-------------
  
  /**
   * La variable <b>scale</b> reprsente les donnes correspondant au scale de la couche.
   */
  private double[] scale;
  
  /**
   * La variable <b>shift</b> reprsente les donnes correspondant au shift de la couche.
   */
  private double[] shift;
  
  /**
   * 
   */
  public SBatchConvolutionLayerRW() 
  {
    super();
    
    this.scale = DEFAULT_SCALE;
    this.shift = DEFAULT_SHIFT;
  }

  /**
   * @param file_name
   */
  public SBatchConvolutionLayerRW(String file_name)
  {
    super(file_name);

    this.scale = DEFAULT_SCALE;
    this.shift = DEFAULT_SHIFT;
  }

  /**
   * ...
   * 
   * @param file_name
   * @param activation_function
   * @param W
   * @param bias
   * @param activation_parameters
   * @param scale
   * @param shift
   * @param input_size
   * @param stride
   * @param padding
   * @throws SConstructorException
   */
  public SBatchConvolutionLayerRW(String file_name, String activation_function, SFiltersWeights W, double[] bias, SMatrixRW activation_parameters, double[] scale, double[] shift, int input_size, int stride, int padding) throws SConstructorException
  {
    super(file_name, activation_function, W, bias, activation_parameters, input_size, stride, padding);
    
    this.scale = scale;
    this.shift = shift;
    
    try{
      initialization();
    }catch(SInitializationException e){
      throw new SConstructorException("Erreur SBatchConvolutionLayerRW 001 : Une erreur d'initialisation est survenue.", e);
    }
  }

  /**
   * ...
   * 
   * @param sbr
   * @throws IOException
   * @throws SConstructorException
   */
  public SBatchConvolutionLayerRW(SBufferedReader sbr) throws IOException, SConstructorException 
  {
    this();   
    
    try{
      read(sbr);
    }catch(SInitializationException e){
      throw new SConstructorException("Erreur SBatchConvolutionLayerRW 002 : Une erreur d'initialisation est survenue.", e);
    }
  }
  
  /**
   * ...
   * 
   * @return
   */
  public double[] getScale()
  {
    return this.scale;
  }
  
  /**
   * ...
   * 
   * @return
   */
  public double[] getShift()
  {
    return this.shift;
  }
  
  public SBatchConvolutionLayer toBatchConvolutionLayer()
  {
    // Construire les filtres.
    List<SFiltersRW> l = this.getFilters();
    SArrayTensor[] array = new SArrayTensor[l.size()];
      
    for(int i = 0; i < l.size(); i++)
    {
      SFiltersRW filtersRW = l.get(i);
        
      List<SMatrixRW> matrixRW_list = filtersRW.getMatrix();
      List<SBufferedMatrix> matrix_list = new ArrayList<SBufferedMatrix>();
        
      for(SMatrixRW m : matrixRW_list)
        matrix_list.add(new SBufferedMatrix(m.getMatrix()));
        
      array[i] = SArrayTensor.buildWithMatrixList(matrix_list);
    }
                
    SFiltersWeights W = new SFiltersWeights(array);
     
    // Ajouter la couche  la liste.
    return new SBatchConvolutionLayer(
        W, 
        new SBias(this.getBias()), 
        new SActivationParameters(new SBufferedMatrix(this.getActivationParameters().getMatrix())),
        new SScale(this.getScale()), 
        new SShift(this.getShift()), 
        this.getInputSize(), 
        this.getPadding(), 
        this.getStride(), 
        this.getActivationFunctionName()
        );
  }
  
  @Override
  public void writeInformation(BufferedWriter bw) throws IOException
  {
    // criture des autres paramtres.
    super.writeInformation(bw);
 
    // criture du SCALE.
    bw.write(SKeyWordDecoder.KW_SCALE);
    bw.write("\t");
    
    // crire tous les composants de la ligne.
    for(int i = 0; i < this.scale.length; i++)
    {
      bw.write(Double.toString(scale[i]));
      bw.write("\t");
    }
    bw.write(SStringUtil.END_LINE_CARACTER);
    bw.write(SStringUtil.END_LINE_CARACTER);
    bw.write(SStringUtil.END_LINE_CARACTER);
    
    // criture du SHIFT.
    bw.write(SKeyWordDecoder.KW_SHIFT);
    bw.write("\t");
    
    // crire tous les composants de la ligne.
    for(int i = 0; i < this.shift.length; i++)
    {
      bw.write(Double.toString(shift[i]));
      bw.write("\t");
    }
    bw.write(SStringUtil.END_LINE_CARACTER);   
    bw.write(SStringUtil.END_LINE_CARACTER);
    bw.write(SStringUtil.END_LINE_CARACTER);
  }
  
  @Override
  public String getRWName() 
  {
    return SKeyWordDecoder.KW_BATCH_CONVOLUTION_LAYER;
  }
  
  @Override
  protected void initialization() throws SInitializationException
  {
    super.initialization();
    
    // Initialisation si la lecture/affectation a t ralise.  
  }
  
  @Override
  protected boolean read(SBufferedReader sbr, int code, String remaining_line) throws SReadingException 
  {
    switch(code)
    {
      case SKeyWordDecoder.CODE_SCALE: this.scale = readDoubleArray(remaining_line, SKeyWordDecoder.KW_SCALE); return true;
      case SKeyWordDecoder.CODE_SHIFT: this.shift = readDoubleArray(remaining_line, SKeyWordDecoder.KW_SHIFT); return true;
      
      default : return super.read(sbr, code, remaining_line);
    }
  }

  @Override
  public String[] getRWParameterName()
  {
    String[] other_parameters = super.getRWParameterName();
    
    return SStringUtil.merge(other_parameters, KEYWORD_PARAMETER);
  }
  
}
