/**
 * 
 */
package sim.readwrite.object.nn;

import java.io.BufferedWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import sim.exception.SConstructorException;
import sim.readwrite.SAbstractRW;
import sim.readwrite.object.SMatrixRW;
import sim.util.SBufferedReader;
import sim.util.SInitializationException;
import sim.util.SKeyWordDecoder;
import sim.util.SReadingException;
import sim.util.SStringUtil;

/**
 * La classe <b>SFiltersRW</b> reprsente un tenseur initialis lors d'une lecture en fichier et pouvant tre crite dans un fichier texte.
 * 
 * @author Simon Vezina
 * @since 2021-09-24
 * @version 2022-02-28
 */
public class SFiltersRW extends SAbstractRW {

  //--------------
  // CONSTANTES //
  //--------------
  
  /**
   * La constante <b>KEYWORD_PARAMETER</b> correspond  un tableau contenant l'ensemble des mots cls 
   *  utiliser reconnus lors de la dfinition de l'objet par une lecture en fichier.
   */
  private static final String[] KEYWORD_PARAMETER = { SKeyWordDecoder.KW_MATRIX };
  
  private List<SMatrixRW> matrix_list;
  
  /**
   * 
   */
  public SFiltersRW() 
  {
    this.matrix_list = new ArrayList<SMatrixRW>();
  }
  
  /**
   * ...
   * 
   * @param matrix_list
   */
  public SFiltersRW(List<SMatrixRW> matrix_list)
  {
    this.matrix_list = matrix_list;
  }
  
  /**
   * ...
   * 
   * @param sbr
   * @throws IOException
   * @throws SConstructorException
   */
  public SFiltersRW(SBufferedReader sbr) throws SConstructorException
  {
    this();   
    
    try{
      read(sbr);
    }catch(SInitializationException e){
      throw new SConstructorException("Erreur SFiltersRW 001 : Une erreur d'initialisation est survenue.", e);
    }
  }
  
  /**
   * Mthode pour obtenir le niveau de profondeur d'entre de ce filtre.
   * 
   * @return Le niveau de profondeur d'entre du filtre.
   */
  public int getInputDepth()
  {
    return this.matrix_list.size();
  }
  
  /**
   * Mthode pour obtenir la taille des filtres carr.
   * 
   * @return La taille des filtres carr.
   */
  public int getKernelOffset()
  {
    return this.matrix_list.get(0).getNbColumn();
  }
  
  /**
   * Mthode pour obtenir la liste des matrices reprsentant l'ensemble des filtres.
   * Le nombre de filtre correspond  la profondeur d'entre (input depth) de la couche de neurones.
   * 
   * @return La liste des matrices.
   */
  public List<SMatrixRW> getMatrix()
  {
    return this.matrix_list;
  }
   
  @Override
  public String getRWName() 
  {
    return SKeyWordDecoder.KW_FILTERS;
  }

  @Override
  public void writeInformation(BufferedWriter bw) throws IOException
  {
    // criture de toutes les matrices.
    for(SMatrixRW m : this.matrix_list)
      m.write(bw);
  }
  
  @Override
  protected void initialization() throws SInitializationException
  {
    super.initialization();
    
    // Vrifier qu'il y a au moins une matrice.
    if(this.matrix_list.size() < 1)
      throw new SInitializationException("Erreur SFiltersRW 003 : Le filtre ne contient pas de matrice.");
    
    
    // Vrifier que la 1re matrice est carr.
    SMatrixRW m = this.matrix_list.get(0);
      
    int nb_line = m.getNbLine();
    int nb_column = m.getNbColumn();
      
    if(nb_line != nb_column)
      throw new SInitializationException("Erreur SFiltersRW 004 : La matrice #1 est de taille " + nb_line + "x" + nb_column + " ce qui n'est pas une matrice carr.");
    
    // Vrifier que toute les autres matrices ont le bon format.
    for(int i = 1; i < this.matrix_list.size(); i++)
    {
      m = this.matrix_list.get(i);
      
      if(m.getNbLine() != nb_line || m.getNbColumn() != nb_column)
        throw new SInitializationException("Erreur SFiltersRW 005 : La matrice #" + (i+1) + " est de taille " + m.getNbLine() + "x" + m.getNbColumn() + " ce qui n'est pas " + nb_line + "x" + nb_column + " comme la 1re matrice."); 
    }
    
  }
  
  @Override
  protected boolean read(SBufferedReader sbr, int code, String remaining_line) throws SReadingException
  {
    switch(code)
    {
      case SKeyWordDecoder.CODE_MATRIX : this.matrix_list.add(new SMatrixRW(sbr)); return true; 
      
      default : return super.read(sbr, code, remaining_line);
    }
  }
  
  @Override
  public String[] getRWParameterName()
  {
    String[] other_parameters = super.getRWParameterName();
    
    return SStringUtil.merge(other_parameters, KEYWORD_PARAMETER);
  }
  
}
