/**
 * 
 */
package sim.readwrite.object.nn;

import java.io.BufferedWriter;
import java.io.IOException;
import java.util.Arrays;

import sim.exception.SConstructorException;
import sim.math.SBufferedMatrix;
import sim.nn.layer.SFullConnectedLayer;
import sim.readwrite.object.SMatrixRW;
import sim.util.SBufferedReader;
import sim.util.SInitializationException;
import sim.util.SKeyWordDecoder;
import sim.util.SReadingException;
import sim.util.SStringUtil;

/**
 * La classe <b>SFullConnectedLayerRW</b> reprsente une couche appartenant  un rseau de neurones initialise lors d'une lecture en fichier et pouvant tre crite dans un fichier texte.
 * 
 * @author Simon Vzina
 * @since 2018-03-014
 * @version 2022-03-03
 */
public class SFullConnectedLayerRW extends SLayerRW {

  //--------------
  // CONSTANTES //
  //--------------
  
  /**
   * La constante <b>KEYWORD_PARAMETER</b> correspond  un tableau contenant l'ensemble des mots cls 
   *  utiliser reconnus lors de la dfinition de l'objet par une lecture en fichier.
   */
  private static final String[] KEYWORD_PARAMETER = {
      SKeyWordDecoder.KW_WEIGHTS, 
      SKeyWordDecoder.KW_BIAS, 
      SKeyWordDecoder.KW_ACTIVATION_PARAMETERS, 
      SKeyWordDecoder.KW_ACTIVATION_FUNCTION, 
      SKeyWordDecoder.KW_AGGREGATION_FUNCTION 
  };
       
  //-------------
  // VARIABLES //
  //-------------
    
  /**
   * ...
   */
  private int aggregation_function;
  
  /**
   * La variable <b>activation_function</b> reprsente le code de rfrence du choix de la fonction d'activation.
   */
  private int activation_function;
  
  /**
   * La variable <b>weights</b> reprsente les donnes coorrespondant au poids (weight) de la couche.
   */
  private SMatrixRW weights;
  
  /**
   * La variable <b>bias</b> reprsente les donnes correspondant au bias de la couche.
   */
  private double[] bias;
  
  /**
   * La variable <b>activation_parameters</b> reprsente les donnes correspondant aux paramtres d'activation de la couche.
   */
  private SMatrixRW activation_parameters;
  
  /**
   * Constructeur d'une lecture d'une couche de neurones
   */
  public SFullConnectedLayerRW()
  {
    super();
    
    this.aggregation_function = 0;
    this.activation_function = 0;
    this.weights = DEFAULT_MATRIX;
    this.bias = DEFAULT_BIAS;
    this.activation_parameters = DEFAULT_MATRIX;
  }
  
  /**
   * ...
   * 
   * @param file_name
   */
  public SFullConnectedLayerRW(String file_name)
  {
    super(file_name);
    
    this.aggregation_function = 0;
    this.activation_function = 0;
    this.weights = DEFAULT_MATRIX;
    this.bias = DEFAULT_BIAS;
    this.activation_parameters = DEFAULT_MATRIX;
  }
  
  /**
   * ...
   * 
   * @param name
   * @param aggregation_function
   * @param activation_function
   * @param matrix
   * @param bias
   * @throws SConstructorException
   */
  public SFullConnectedLayerRW(String file_name, String aggregation_function, String activation_function, SMatrixRW matrix, double[] bias, SMatrixRW activation_paramters) throws SConstructorException
  {
    super(file_name);
    
    // Validation du nom de la fonction d'agrgation.
    int agg_index = Arrays.asList(AGGREGATION_FUNCTIONS).indexOf(aggregation_function);
    
    if(agg_index == -1)
      throw new SConstructorException("Erreur SFullConnectedLayerRW 001 : La fonction d'aggrgation '" + aggregation_function + "' n'est pas reconnue. Les fonctions admissibles sont les suivantes : " + SStringUtil.join(AGGREGATION_FUNCTIONS, ", "));
    
    // Validation du nom de la fonction d'agrgation.
    int act_index = Arrays.asList(ACTIVATION_FUNCTIONS).indexOf(activation_function);
    
    if(act_index == -1)
      throw new SConstructorException("Erreur SFullConnectedLayerRW 002 : La fonction d'activation '" + activation_function + "' n'est pas reconnue. Les fonctions admissibles sont les suivantes : " + SStringUtil.join(ACTIVATION_FUNCTIONS, ", "));
    
    
    this.aggregation_function = agg_index;
    this.activation_function = act_index;
    this.weights = matrix;
    this.bias = bias;
    this.activation_parameters = activation_paramters;
    
    try{
      initialization();
    }catch(SInitializationException e){
      throw new SConstructorException("Erreur SFullConnectedLayerRW 001 : Une erreur d'initialisation est survenue.", e);
    }
  }
  
  /**
   * Constructeur d'une lecture d'une couche de neurones  partir d'information lue dans un fichier de format txt.
   * 
   * @param sbr Le BufferedReader cherchant l'information de le fichier .txt.
   * @throws IOException Si une erreur de l'objet SBufferedWriter est lance.
   * @throws SConstructorException Si un paramtre lors de la lecture rend l'objet invalide.
   * @see SBufferedReader
   */
  public SFullConnectedLayerRW(SBufferedReader sbr) throws IOException, SConstructorException
  {
    this();   
    
    try{
      read(sbr);
    }catch(SInitializationException e){
      throw new SConstructorException("Erreur SFullConnectedLayerRW 002 : Une erreur d'initialisation est survenue.", e);
    }
  }
   
  /**
   * ...
   * 
   * @return
   */
  public SMatrixRW getWeights()
  {
    return this.weights;
  }
  
  /**
   * ....
   * 
   * @return
   */
  public double[] getBias()
  {
    return this.bias;
  }
  
  /**
   * ...
   * 
   * @return
   */
  public SMatrixRW getActivationParameters()
  {
    return this.activation_parameters;
  }
  
  /**
   * Mthode pour obtenir le nom de la fonction d'agration de la couche.
   * 
   * @return Le nom de la fonction d'agrgation.
   */
  public String getAggregationFunctionName()
  {
    return AGGREGATION_FUNCTIONS[aggregation_function];
  }
  
  /**
   * Mthode pour obtenir le nom de la fonction d'activation de la couche.
   * 
   * @return Le nom de la fonction d'activation.
   */
  public String getActivationFunctionName()
  {
    return ACTIVATION_FUNCTIONS[activation_function];
  }
  
  /**
   * Mthode pour construire un objet de la classe <b>SFullConnectedLayer</b>  partir de ces informations.
   * 
   * @return Une couche de neurones de type <i>full connected</i>
   */
  public SFullConnectedLayer toFullConnectedLayer()
  {
    return new SFullConnectedLayer(
        new SBufferedMatrix(this.getWeights().getMatrix()), 
        this.getBias().clone(), 
        new SBufferedMatrix(this.getActivationParameters().getMatrix()), 
        this.getAggregationFunctionName(), 
        this.getActivationFunctionName()
    );
  }
  
  @Override
  public void writeInformation(BufferedWriter bw) throws IOException
  {
    super.writeInformation(bw);
 
    // criture nom des fonctions.
    bw.write(SKeyWordDecoder.KW_AGGREGATION_FUNCTION);
    bw.write("\t\t");
    bw.write(AGGREGATION_FUNCTIONS[aggregation_function]);
    bw.write(SStringUtil.END_LINE_CARACTER);
    
    bw.write(SKeyWordDecoder.KW_ACTIVATION_FUNCTION);
    bw.write("\t\t");
    bw.write(ACTIVATION_FUNCTIONS[activation_function]);
    bw.write(SStringUtil.END_LINE_CARACTER);
    
    // criture des poids.
    bw.write(SKeyWordDecoder.KW_WEIGHTS);
    bw.write(SStringUtil.END_LINE_CARACTER);
    weights.writeInformation(bw);
    bw.write(SKeyWordDecoder.KW_END);     // fermeture du mot cl KW_WEIGHTS.
    bw.write(SStringUtil.END_LINE_CARACTER);
    bw.write(SStringUtil.END_LINE_CARACTER);
    
    // criture des biais.
    bw.write(SKeyWordDecoder.KW_BIAS);
    bw.write("\t");
    
    // crire tous les composants de la ligne.
    for(int i = 0; i < bias.length; i++)
    {
      bw.write(Double.toString(bias[i]));
      bw.write("\t");
    }
    bw.write(SStringUtil.END_LINE_CARACTER);
    bw.write(SStringUtil.END_LINE_CARACTER);

    // criture des parametres d'activation.
    bw.write(SKeyWordDecoder.KW_ACTIVATION_PARAMETERS);
    bw.write(SStringUtil.END_LINE_CARACTER);
    activation_parameters.writeInformation(bw);
    bw.write(SKeyWordDecoder.KW_END);     // fermeture du mot cl KW_ACTIVATION_PARAMETERS.
    bw.write(SStringUtil.END_LINE_CARACTER);
    bw.write(SStringUtil.END_LINE_CARACTER);
  }
  
  @Override
  public String getRWName() 
  {
    return SKeyWordDecoder.KW_FULL_CONNECTED_LAYER;
  }

  @Override
  protected void initialization() throws SInitializationException
  {
    super.initialization();
    
    // Vrifier que le nombre de bias correspond au nombre de ligne de la matrice.
    if(weights.getMatrix().size() != bias.length)
      throw new SInitializationException("Erreur SFullConnectedLayerRW 003 : La couche possde '" + weights.getMatrix().size() + " lignes dans la matrice et '" + bias.length + "' lments dans le bias ce qui n'est pas gaux."); 
   
    // Validation des paramtres d'activation, s'il y en a.
    if(activation_parameters.getNbLine() != 0)
      if(activation_parameters.getMatrix().get(0).length != bias.length)
        throw new SInitializationException("Erreur SFullConnectedLayerRW 004 : La couche possde " + bias.length + "' lments de biais, mais contient " + activation_parameters.getMatrix().get(0).length + " lment dans les paramtres d'activation ce qui n'est pas gaux."); 
  }

  @Override
  protected boolean read(SBufferedReader sbr, int code, String remaining_line) throws SReadingException
  {
    switch(code)
    {
      case SKeyWordDecoder.CODE_AGGREGATION_FUNCTION : aggregation_function = readIntOrExpression(remaining_line, SKeyWordDecoder.KW_AGGREGATION_FUNCTION, AGGREGATION_FUNCTIONS); return true;  
      case SKeyWordDecoder.CODE_ACTIVATION_FUNCTION : activation_function = readIntOrExpression(remaining_line, SKeyWordDecoder.KW_ACTIVATION_FUNCTION, ACTIVATION_FUNCTIONS); return true;
      case SKeyWordDecoder.CODE_WEIGHTS : weights = new SMatrixRW(sbr); return true;
      case SKeyWordDecoder.CODE_BIAS : bias = readDoubleArray(remaining_line, SKeyWordDecoder.KW_BIAS); return true;
      case SKeyWordDecoder.CODE_ACTIVATION_PARAMETERS : activation_parameters = new SMatrixRW(sbr); return true;
      
      default : return super.read(sbr, code, remaining_line);
    }
  }

  @Override
  public String[] getRWParameterName()
  {
    String[] other_parameters = super.getRWParameterName();
    
    return SStringUtil.merge(other_parameters, KEYWORD_PARAMETER);
  }
  
}
