/**
 * 
 */
package sim.readwrite.object.nn;

import java.io.BufferedWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import sim.exception.SConstructorException;
import sim.readwrite.SAbstractRW;
import sim.util.SBufferedReader;
import sim.util.SInitializationException;
import sim.util.SKeyWordDecoder;
import sim.util.SReadingException;
import sim.util.SStringUtil;

/**
 * La classe <b>SNeuralNetworkRW</b> reprsente un rseau de neurones initialise lors d'une lecture en fichier et pouvant tre crite dans un fichier texte.
 * 
 * @author Simon Vzina
 * @since 2018-03-24
 * @version 2021-10-19
 */
public class SNeuralNetworkRW extends SAbstractRW {

  //--------------
  // CONSTANTES //
  //--------------
  
  /**
   * La constante <b>KEYWORD_PARAMETER</b> correspond  un tableau contenant l'ensemble des mots cls 
   *  utiliser reconnus lors de la dfinition de l'objet par une lecture en fichier.
   */
  private static final String[] KEYWORD_PARAMETER = {
    SKeyWordDecoder.KW_BATCH_NETWORK, 
    SKeyWordDecoder.KW_FULL_CONNECTED_LAYER, 
    SKeyWordDecoder.KW_BATCH_FULL_CONNECTED_LAYER, 
    SKeyWordDecoder.KW_CONVOLUTION_LAYER, 
    SKeyWordDecoder.KW_BATCH_CONVOLUTION_LAYER,
    SKeyWordDecoder.KW_DROPOUT_LAYER, 
    SKeyWordDecoder.KW_POOLING_LAYER
  };
  
  /**
   * ...
   */
  private static final boolean DEFAULT_BATCH_NETWORK = false;
  
  //-------------
  // VARIABLES //
  //-------------
  private final List<SLayerRW> layer_list;
  
  protected boolean batch_network;
  
  /**
   * Constructeur d'une lecture d'un rseau de neurones.
   */
  public SNeuralNetworkRW()
  {
    super();
    
    batch_network = DEFAULT_BATCH_NETWORK;
    layer_list = new ArrayList<SLayerRW>();
  }
  
  /**
   * ...
   * 
   * @param file_name
   */
  public SNeuralNetworkRW(String file_name)
  {
    super(file_name);
    
    batch_network = DEFAULT_BATCH_NETWORK;
    layer_list = new ArrayList<SLayerRW>();
  }
  
  /**
   * ...
   * 
   * @param file_name
   * @param is_batch_network
   * @param list
   * @throws SConstructorException
   */
  public SNeuralNetworkRW(String file_name, boolean is_batch_network, List<SLayerRW> list) throws SConstructorException
  {
    super(file_name);
    
    this.batch_network = is_batch_network;
    this.layer_list = new ArrayList<SLayerRW>(list);
    
    try{
      initialization();
    }catch(SInitializationException e){
      throw new SConstructorException("Erreur SNeuralNetworkRW 001 : Une erreur d'initialisation est survenue.", e);
    }
    
  }
  
  /**
   * Constructeur d'une lecture d'un rseau de neurones  partir d'information lue dans un fichier de format txt.
   * 
   * @param sbr Le BufferedReader cherchant l'information de le fichier .txt.
   * @throws IOException Si une erreur de l'objet SBufferedWriter est lance.
   * @throws SConstructorException Si un paramtre lors de la lecture rend l'objet invalide.
   * @see SBufferedReader
   */
  public SNeuralNetworkRW(SBufferedReader sbr) throws IOException, SConstructorException
  {
    this();   
    
    try{
      read(sbr);
    }catch(SInitializationException e){
      throw new SConstructorException("Erreur SNeuralNetworkRW 002 : Une erreur d'initialisation est survenue.", e);
    }
  }
  
  /**
   * ...
   * 
   * @return
   */
  public boolean isBatchNetwork()
  {
    return this.batch_network;
  }
  
  /**
   * ...
   * 
   * @return
   */
  public List<SLayerRW> getLayerList()
  {
    return this.layer_list;
  }
  
  @Override
  public void writeInformation(BufferedWriter bw) throws IOException
  {
    super.writeInformation(bw);
    
    // criture du type de couche.
    bw.write(SKeyWordDecoder.KW_BATCH_NETWORK);
    bw.write("\t\t");
    bw.write(Boolean.toString(batch_network));
    bw.write(SStringUtil.END_LINE_CARACTER);
    bw.write(SStringUtil.END_LINE_CARACTER);
    
    // criture du type de la couche avec le nom du fichier o se retrouve la description de la couche.
    for(SLayerRW l : layer_list)
    {
      // criture du type de couche en jeu.
      if(l instanceof SBatchFullConnectedLayerRW)
        bw.write(SKeyWordDecoder.KW_BATCH_FULL_CONNECTED_LAYER);
      else
        if(l instanceof SFullConnectedLayerRW)
          bw.write(SKeyWordDecoder.KW_FULL_CONNECTED_LAYER);
        else
          if(l instanceof SBatchConvolutionLayerRW)
            bw.write(SKeyWordDecoder.KW_BATCH_CONVOLUTION_LAYER);
          else
            if(l instanceof SConvolutionLayerRW)
              bw.write(SKeyWordDecoder.KW_CONVOLUTION_LAYER);
            else
              if(l instanceof SDropoutLayerRW)
                bw.write(SKeyWordDecoder.KW_DROPOUT_LAYER);
              else
                if(l instanceof SPoolingLayerRW)
                  bw.write(SKeyWordDecoder.KW_POOLING_LAYER);
                else
                  throw new UnsupportedOperationException("Erreur SNeuralNetworkRW : Un type de couche doit tre crit qui n'est pas support.");
          
      // criture du nom du fichier contenant l'information de la couche. 
      bw.write("\t");
      bw.write(l.getFileName());
      bw.write(SStringUtil.END_LINE_CARACTER);
    }
    
    bw.write(SStringUtil.END_LINE_CARACTER);
    
    // Lancer l'criture des couches dans leur propre fichier.
    for(SLayerRW l : layer_list)
      l.write();  
  }
  
  @Override
  public String getRWName()
  {
    return SKeyWordDecoder.KW_NEURAL_NETWORK;
  }

  @Override
  protected void initialization() throws SInitializationException
  {
    super.initialization();
    
    // Vrifier que la squence de couches est valide.
    //   FAIRE !!!
  }
  
  @Override
  protected boolean read(SBufferedReader sbr, int code, String remaining_line) throws SReadingException
  {
    switch(code)
    {
      case SKeyWordDecoder.CODE_BATCH_NETWORK : batch_network = this.readTrueFalseExpressionOrInt(remaining_line, SKeyWordDecoder.KW_BATCH_NETWORK); return true;
    
      case SKeyWordDecoder.CODE_FULL_CONNECTED_LAYER :
      {  
        String file_name = readStringNotEmpty(remaining_line, SKeyWordDecoder.KW_FULL_CONNECTED_LAYER); 
        SFullConnectedLayerRW layer = new SFullConnectedLayerRW(file_name);
        
        try {
          layer.read();
        }catch(SConstructorException e) {
          throw new SReadingException("Erreur SNeuralNetworkRW 003 : Une erreur d'initialisation est survenue avec la couche '" + layer.getFileName() + "'.");
        }
        
        layer_list.add(layer); 
      }
      return true;
      
      case SKeyWordDecoder.CODE_BATCH_FULL_CONNECTED_LAYER :
      {  
        String file_name = readStringNotEmpty(remaining_line, SKeyWordDecoder.KW_BATCH_FULL_CONNECTED_LAYER); 
        SBatchFullConnectedLayerRW layer = new SBatchFullConnectedLayerRW(file_name);
        
        try {
          layer.read();
        }catch(SConstructorException e) {
          throw new SReadingException("Erreur SNeuralNetworkRW 004 : Une erreur d'initialisation est survenue avec la couche '" + layer.getFileName() + "'.", e);
        }
        
        layer_list.add(layer); 
      }
      return true;
        
      case SKeyWordDecoder.CODE_CONVOLUTION_LAYER :
      {  
        String file_name = readStringNotEmpty(remaining_line, SKeyWordDecoder.KW_CONVOLUTION_LAYER); 
        SConvolutionLayerRW layer = new SConvolutionLayerRW(file_name);
        
        try {
          layer.read();
        }catch(SConstructorException e) {
          throw new SReadingException("Erreur SNeuralNetworkRW 005 : Une erreur d'initialisation est survenue avec la couche '" + layer.getFileName() + "'.", e);
        }
        
        layer_list.add(layer); 
      }
      return true;
      
      case SKeyWordDecoder.CODE_BATCH_CONVOLUTION_LAYER :
      {  
        String file_name = readStringNotEmpty(remaining_line, SKeyWordDecoder.KW_BATCH_CONVOLUTION_LAYER); 
        SBatchConvolutionLayerRW layer = new SBatchConvolutionLayerRW(file_name);
        
        try {
          layer.read();
        }catch(SConstructorException e) {
          throw new SReadingException("Erreur SNeuralNetworkRW 004 : Une erreur d'initialisation est survenue avec la couche '" + layer.getFileName() + "'.", e);
        }
        
        layer_list.add(layer); 
      }
      return true;
      
      case SKeyWordDecoder.CODE_DROPOUT_LAYER :
      {  
        String file_name = readStringNotEmpty(remaining_line, SKeyWordDecoder.KW_DROPOUT_LAYER); 
        SDropoutLayerRW layer = new SDropoutLayerRW(file_name);
        
        try {
          layer.read();
        }catch(SConstructorException e) {
          throw new SReadingException("Erreur SNeuralNetworkRW 006 : Une erreur d'initialisation est survenue avec la couche '" + layer.getFileName() + "'.", e);
        }
        
        layer_list.add(layer); 
      }
      return true;
      
      case SKeyWordDecoder.CODE_POOLING_LAYER :
      {  
        String file_name = readStringNotEmpty(remaining_line, SKeyWordDecoder.KW_POOLING_LAYER); 
        SPoolingLayerRW layer = new SPoolingLayerRW(file_name);
        
        try {
          layer.read();
        }catch(SConstructorException e) {
          throw new SReadingException("Erreur SNeuralNetworkRW 007 : Une erreur d'initialisation est survenue avec la couche '" + layer.getFileName() + "'.", e);
        }
        
        layer_list.add(layer); 
      }
      return true;
      
      default : return super.read(sbr, code, remaining_line);
    }
  }

  @Override
  public String[] getRWParameterName()
  {
    String[] other_parameters = super.getRWParameterName();
    
    return SStringUtil.merge(other_parameters, KEYWORD_PARAMETER);
  }
  
}
