/**
 * 
 */
package sim.nn.function.filtering;

import java.util.ArrayList;
import java.util.List;

import sim.exception.SConstructorException;
import sim.exception.SNoImplementationException;
import sim.nn.function.SMemoryBatchFunction;
import sim.nn.function.SComputeErrorFunctionGradientBias;
import sim.nn.function.SComputeErrorFunctionGradientFiltersWeights;
import sim.nn.function.SNeuralMath;
import sim.nn.parameter.SBias;
import sim.nn.parameter.SFiltersWeights;

/**
 * La classe <b>SBatchConvolutionFunction</b> reprsente une fonction de convolution pouvant tre exploite en <i>batch</i>.
 * 
 * @author Simon Vezina
 * @since 2021-11-02
 * @version 2022-09-13 (Version labo : Apprentissage par rseau de neurones)
 */
public class SBatchConvolutionFunction extends SMemoryBatchFunction implements SComputeErrorFunctionGradientFiltersWeights, SComputeErrorFunctionGradientBias {

  /**
   * La variable <b>convolution_function_list</b> reprsente la liste des fonctions de convolution.
   */
  private final List<SConvolutionFunction> convolution_function_list;
  
  /**
   * ...
   */
  protected SFiltersWeights W;
  
  /**
   * ...
   */
  protected SBias B;
  
  /**
   * ...
   * 
   * @param inputs
   * @param inputs_depth
   * @param outputs_depth
   * @param filter_offset
   * @param padding
   * @param stride
   * @throws SConstructorException
   */
  public SBatchConvolutionFunction(int inputs, int inputs_depth, int outputs_depth, int filter_offset, int padding, int stride) throws SConstructorException 
  {
    super(inputs, SNeuralMath.computeOutputsSizeForSquareInputsDepth(inputs, inputs_depth, outputs_depth, filter_offset, padding, stride));
    
    this.convolution_function_list = new ArrayList<SConvolutionFunction>();
    
    try {
      
      for(int i = 0; i < this.getMaximumBatchSize(); i++)
        this.convolution_function_list.add(new SConvolutionFunction(inputs, inputs_depth, outputs_depth, filter_offset, padding, stride));
    
    }catch(IllegalArgumentException e) {
      throw new SConstructorException("Erreur SBatchConvolutionFunction 001 : La fonction de convolution de batch n'a pas pu tre construite", e);
    }
    
    // Construction des paramtres.
    this.W = new SFiltersWeights(filter_offset, inputs_depth, outputs_depth);
    this.B = new SBias(outputs_depth);
  }

  @Override
  public String getFunctionName()
  {
    return this.convolution_function_list.get(0).getFunctionName();
  }
  
  @Override
  public int getInputsDepth()
  {
    return this.convolution_function_list.get(0).getInputsDepth();
  }
  
  @Override
  public int getOutputsDepth()
  {
    return this.convolution_function_list.get(0).getOutputsDepth();
  }
  
  @Override
  public String getDimension()
  {
    return this.convolution_function_list.get(0).getDimension();
  }
  
  /**
   * Mthode pour obtenir la dimension de largeur de l'entre de forme carr.
   * 
   * @return
   */
  public int getInputOffset()
  {
    return this.convolution_function_list.get(0).getInputOffset();
  }
  
  /**
   * Mthode pour obtenir la dimension de largeur de la sortie de forme carr.
   * 
   * @return
   */
  public int getOutputOffset()
  {
    return this.convolution_function_list.get(0).getOutputOffset();
  }
  
  @Override
  public boolean asParameter() 
  {
    return this.convolution_function_list.get(0).asParameter();
  }
  
  @Override
  public int getNumberOfParameter()
  {
    return this.convolution_function_list.get(0).getNumberOfParameter();
  }
  
  /**
   * Mthode pour obtenir la taille du filtre de filtrage.
   * 
   * @return La taille du filtre.
   */
  public int getFilterOffset()
  {
    return this.convolution_function_list.get(0).getFilterOffset();
  }
  
  /**
   * ...
   * 
   * @return
   */
  public int getStride()
  {
    return this.convolution_function_list.get(0).getStride();
  }
  
  /**
   * ...
   * 
   * @return
   */
  public int getPadding()
  {
    return this.convolution_function_list.get(0).getPadding();
  }
  
  /**
   * ...
   * 
   * @param b
   * @param batch_inputs
   * @param W
   * @param B
   * @return
   * @throws IllegalArgumentException
   */
  public List<double[]> batchConvolution(int b, List<double[]> batch_inputs, SFiltersWeights W, SBias B) throws IllegalArgumentException
  {
    // Validation du nombre de biais.
    if(B.getNeurons() != this.getOutputsDepth())
      throw new IllegalArgumentException("Erreur SBatchConvolutionFunction 002 : Le nombre de biais " + B.getNeurons() + " n'est pas gal  " + this.getOutputsDepth() + " tant le niveau de profondeur de sortie de la fonction.");
        
    // Validation de la taille des filtres.
    if(W.getOffset() != this.getFilterOffset())
      throw new IllegalArgumentException("Erreur SBatchConvolutionFunction 003 : La dimension du filtre " + W.getOffset() + " n'est pas gal  " + this.getFilterOffset() + " tat la valeur prdfinie.");
    
    // Affectation des paramtres.
    this.W = W;
    this.B = B;
    
    return batchForwardPropagation(b, batch_inputs);
  }
  
  @Override
  public List<double[]> batchForwardPropagation(int b, List<double[]> batch_inputs) throws IllegalArgumentException 
  {
    // Vrification de la taille de la batch b  oprer sur le vecteur d'entre.
    if(b > batch_inputs.size())
      throw new IllegalArgumentException("Erreur SBatchConvolutionFunction 004 : La taile de la batch a excuter ne peut pas tre plus grande et la batch du vecteur d'entre.");
    
    // Dfinir la taille de la batch pour la fonction et allocation de mmoire si ncessaire.
    this.setBatchSize(b);
    
    this.batch_forward_inputs = batch_inputs;
   
    // Excuter la fonction sur la batch et enregister la rfrence de la solution dans la batch de retour.
    for(int i = 0; i < getBatchSize(); i++)
      this.batch_forward_outputs.set(i, this.convolution_function_list.get(i).convolution(batch_inputs.get(i), W, B));
    
    return this.batch_forward_outputs; 
  }
  
  /**
   * ...
   * 
   * @param b
   * @param batch_delta
   * @return
   * @throws IllegalArgumentException
   */
  public List<double[]> batchBackPropagation(int b, List<double[]> batch_delta) throws IllegalArgumentException
  {
    // Validation de la taille de la batch  calculer.
    if(b != this.getBatchSize())
      throw new IllegalArgumentException("Erreur SBatchConvolutionFunction 005 : Le paramtre b n'est pas gal  la taille de la batch en calcul lors de l'activation.");
    
    // Excuter la fonction sur la batch et enregister la rfrence de la solution dans la batch de retour.
    for(int i = 0; i < getBatchSize(); i++)
      this.batch_backward_outputs.set(i, this.convolution_function_list.get(i).backPropagation(batch_delta.get(i)));
    
    return this.batch_backward_outputs;
  }
  
  @Override
  public double dCdW(int f, int d, int i, int j)
  {
    throw new SNoImplementationException();
    
  }
  
  @Override
  public double dCdB(int f)
  {
    throw new SNoImplementationException();
    
  }
  
  @Override
  public boolean upgradeMaximumBatchSize(int size)
  {
    if(super.upgradeMaximumBatchSize(size))
    {
      // valuer l'ampleur de la mise  jour.
      int upgrade = size - this.convolution_function_list.size();
      
      // Augmenter l'espace mmoire uniquement si ncessaire
      if(upgrade > 0)
        for(int i = 0; i < upgrade; i++)
          this.convolution_function_list.add(new SConvolutionFunction(this.getInputsSize(), this.getInputsDepth(), this.getOutputsDepth(), this.getFilterOffset(), this.getPadding(), this.getStride()));
      
      return true;
    }
    else
      return false;
  }
  
  
}
