/**
 * 
 */
package sim.nn.gradient;

import java.util.ArrayList;
import java.util.List;

import sim.exception.SConstructorException;
import sim.nn.SNNData;
import sim.nn.function.error.SBatchErrorFunction;
import sim.nn.function.error.SErrorFunction;
import sim.nn.function.error.SErrorFunctionBuilder;
import sim.nn.network.SBatchNetwork;
import sim.nn.network.SNetwork;

/**
 * <p>
 * La classe <b>SGradientComputer</b> reprsente un calculateur de gradient de fonction d'erreur. 
 * Cette classe a pour objectif d'tre utilis en parallle (multi-tche).
 * </p>
 * <p>
 * Dans le cas d'un rseau de type batch (ralisant des activations en batch),
 * ce calculateur va sparer la batch  activer entre toute les threads
 * ce qui va rduire la taille de la batch trait par les rseaux.
 * Ainsi, il est prfrable d'augmenter la batch s'il y a un grand nombre de processeur
 * afin d'assurer les bien fait des fonction de type batch qui require beaucoup
 * de vecteur pour bien fonctionner.
 * </p>
 * 
 * @author Simon Vezina
 * @since 2019-09-06
 * @version 2021-09-12
 */
public class SGradientComputer implements Runnable {

  /**
   * La variable <b>network</b> reprsente la copie du rseau qui sera utilise dans cette classe pour effectuer le calcul du gradient.
   */
  private final SNetwork network;
  
  /**
   * La variable <b>error_function</b> reprsente la fonction d'erreur qui sera utilise si le rseau est ordinaire.
   */
  private final SErrorFunction error_function;
  
  /**
   * La variable <b>error_function</b> reprsente la fonction d'erreur qui sera utilise si le rseau est de type <i>batch</i>.
   */
  private final SBatchErrorFunction batch_error_function;
  
  /**
   * La variable <b>gradient</b> reprsente le gradient de la fonction d'erreur qui sera valu par ce calculateur.
   */
  private final SErrorFunctionGradient gradient;
  
  /**
   * La variable <b>data_list</b> reprsente la liste des donnes qui seront exploite pour valuer le gradient de la fonction d'erreur.
   */
  private final List<SNNData> data_list;
  
  /**
   * Constructeur d'un calculateur de gradient de la fonction d'erreur.
   * 
   * @param network Le rseau de neurones qui sera copi dans le calculateur (pour supporter le calcul en parallle).
   * @param error_function_name Le nom de la fonction d'erreur.
   */
  public SGradientComputer(SNetwork network, String error_function_name) throws SConstructorException
  {
    // Afin de supporter le multi-processeur, je dois faire une copie de ce rseau
    // puisque l'activation ncessite de l'accs en mmoire dans le rseau.
    this.network = network.copyWithSameParameter();
    
    // Construction de la fonction d'erreur et choix du type de calcul  effectuer.
    try {
      
      if(this.network instanceof SBatchNetwork)
      {
        this.batch_error_function = new SBatchErrorFunction(network.getOutputsSize(), error_function_name);
        this.error_function = null;
      }
      else
      {
        this.error_function = SErrorFunctionBuilder.build(error_function_name, this.network.getOutputsSize());
        this.batch_error_function = null;
      }
      
    }catch(IllegalArgumentException e) {
      throw new SConstructorException("Erreur SGradientComputer 002 : Une erreur lors de la cration de la fonction d'erreur est survenue.", e);
    }
    
    // Construction du gradient avec la configuration du rseau.
    this.gradient = new SErrorFunctionGradient(this.network);
    
    // Construire la liste des donnes qui seront exploites lors du calcul du gradient.
    this.data_list = new ArrayList<SNNData>();
  }

  @Override
  public void run() 
  {
    // Vider le gradient.
    this.gradient.clear();
    
    // Effectuer le calcul du grandient en fonction du type de rseau.
    if(batch_error_function == null)
    {
      for(SNNData data : data_list)
        SGradientAlgorithm.computeGradient(this.network, this.error_function, this.gradient, data);
    }
    else
      SGradientAlgorithm.batchComputeGradient((SBatchNetwork)this.network, this.batch_error_function, this.gradient, data_list);
    
  }

  /**
   * Mthode pour ajouter une donne dans le calculateur de gradient.
   * 
   * @param data La donne.
   */
  public void addData(SNNData data)
  {
    this.data_list.add(data);
  }
  
  /**
   * Mthode pour vide le calculateur de gradient de ses donnes.
   */
  public void clear()
  {
    this.data_list.clear();
  }
  
  /**
   * Mthode pour obtenir le gradient calcul lors de la dernire excution de run().
   * 
   * @return Le gradient.
   */
  public SErrorFunctionGradient getGradient()
  {
    return this.gradient;
  }
  
}
