/**
 * 
 */
package sim.nn.parameter;

import sim.exception.SConstructorException;
import sim.exception.SNaNValueException;
import sim.math.SBufferedMatrix;
import sim.nn.gradient.SGradientLayer;

/**
 * <p>
 * La classe <b>SActivationParameters</b> reprsente l'ensemble des paramtres utiliss pour une fonction d'activation.
 * </p>
 * <p>
 * Puisque une fonction d'activation peut exploiter 0 (ex: sigmode), 1 (ex: swish), 2 ou plusieurs paramtres diffrents mais unique pour chacun des neurones
 * permettant de spcialiser la courbe de la fonction d'activation pour chaque neurone, ce paramtre aura la structure d'une matrice
 * o chaque ligne sera assigne  un paramtre de la fonction et chaque lment de la colonne sera la valeur particulire pour chaque neurone utilis lors de l'activation de ce neurone. 
 * </p>
 * 
 * @author Simon Vezina
 * @since 2021-12-03
 * @version 2021-12-31
 */
public class SActivationParameters extends SMatrixParameter {

  /**
   * La constante <b>EMPTY_ACTIVATION_PARAMETERS</b> reprsente un paramtre d'activation vide.
   */
  public final static SActivationParameters EMPTY_ACTIVATION_PARAMETERS = new SActivationParameters();
  
  /**
   * Constructeur d'un paramtre d'activation vide.
   * Ce type de paramtre sera utilis pour les fonctions d'activation <u>sans</u> paramtre.
   * Il sera par dfaut non entranable.
   */
  private SActivationParameters() 
  {
    super();
  }

  /**
   * Constructeur d'un paramtre d'activation.
   * 
   * @param nb_parameter Le nombre de paramtres dans l'activation.
   * @param nb_neurons Le nombre de neurones par paramtre.
   * @throws SConstructorException S'il y a eu une erreur lors de la contruction de l'objet.
   */
  public SActivationParameters(int nb_parameter, int nb_neurons) throws SConstructorException 
  {
    super(nb_parameter, nb_neurons);
  }

  /**
   * ...
   * 
   * @param m La matrice contenant les paramtres d'activation.
   */
  public SActivationParameters(SBufferedMatrix m)
  {
    super(m);
  }
  
  @Override
  public SActivationParameters clone()
  {
    SActivationParameters s = new SActivationParameters(this.matrix.clone());
    s.setTrainable(this.isTrainable());
    
    return s;
  }
  
  /**
   * Mthode pour dterminer si le paramtre d'activation est vide.
   * 
   * @return <b>true</b> si le paramtre d'activation est vide et <b>false</b> sinon.
   */
  public boolean isEmpty()
  {
    return this.getNbParameterType() == 0;
  }
  
  /**
   * Mthode pour obtenir un paramtre de la fonction d'activation pour un neurone donn.
   * 
   * @param i L'indice du paramtre (le type du paramtre)
   * @param u Le neurone.
   * @return Le paramtre.
   */
  public double get(int i, int u)
  {
    return this.matrix.get(i, u);
  }
  
  /**
   * Mthode pour affecteur une valeur  un paramtre d'un certain type pour un neurone donne.
   *   
   * @param i L'indice du paramtre (le type du paramtre)
   * @param u Le neurone.
   * @param value La valeur  affecter
   * @throws SNaNValueException Si la valeur  affecter est NaN (not a number).
   */
  public void set(int i, int u, double value) throws SNaNValueException
  {
    this.matrix.set(i, u, this.analyseValue(value));
  }
  
  /**
   * Mthode pour obtenir le nombre de type de paramtre contenue le paramtre d'activation.
   * Cette information correspond  obtenir le nombre de lignes dans la matrice des paramtres.
   * 
   * @return Le nombre de type de paramtres.
   */
  public int getNbParameterType()
  {
    return this.matrix.getNbLine();
  }
    
  /**
   * Mthode pour obtenir le nombre de neurones sur lequel le paramtre d'activation pourra tre utilis.
   * Cette information correspond  obtenir le nombre de colonnes dans la matrice des paramtres.
   * 
   * @return Le nombre de neurones.
   */
  public int getNbNeurons()
  {
    return this.matrix.getNbColumn();
  }
    
  @Override
  public void add(SParameter parameter) throws IllegalArgumentException 
  {
    // Effectuer l'entranement si le gradient contient un accs au poids.
    if(!(parameter instanceof SActivationParameters))
      throw new IllegalArgumentException("Erreur SActivationParameters 003 : Le gradient n'est pas du bon type (" + parameter.getClass().getName() + ").");
    
    addParameter((SActivationParameters)parameter);
  }

  @Override
  public void multiply(SParameter parameter) throws IllegalArgumentException
  {
    // Effectuer l'entranement si le gradient contient un accs au poids.
    if(!(parameter instanceof SActivationParameters))
      throw new IllegalArgumentException("Erreur SActivationParameters 003 : Le gradient n'est pas du bon type (" + parameter.getClass().getName() + ").");
    
    multiplyParameter((SActivationParameters)parameter);
  }
  
  @Override
  public void learn(SGradientLayer gradient, double learning_rate) throws IllegalArgumentException 
  {
    // Effectuer l'entranement si le gradient contient un accs au poids.
    if(!(gradient instanceof SContainActivationParameters))
      throw new IllegalArgumentException("Erreur SActivationParameters 003 : Le gradient n'est pas du bon type (" + gradient.getClass().getName() + ").");
    
    learnParameter(((SContainActivationParameters)gradient).getActivationParameters(), learning_rate);
  }

  @Override
  public String toString()
  {
    return "SActivationParameters [K=" + this.matrix + "]";  
  }

}
